Package: ks
Version: 1.15.2
Date: 2026-05-09
Title: Kernel Smoothing
Authors@R: c(person("Tarn", "Duong", role=c("aut","cre"), email="tarn.duong@gmail.com", comment=c(ORCID="0000-0002-1198-3482")),
             person("Matt", "Wand", role="ctb", comment=c(ORCID="0000-0003-2555-896X")),
             person("Jose", "Chacon", role="ctb"),
             person("Artur", "Gramacki", role="ctb", comment=c(ORCID="0000-0002-1610-9743")))
Maintainer: Tarn Duong <tarn.duong@gmail.com>
Depends: R (>= 2.10.0)
Imports: FNN (>= 1.1), kernlab, KernSmooth (>= 2.22), Matrix, mclust,
        mgcv, multicool, mvtnorm (>= 1.0-0), pracma
Suggests: geometry, knitr, MASS, misc3d (>= 0.4-0), oz, plot3D, rgl (>=
        0.66), rmarkdown
VignetteBuilder: knitr
Encoding: UTF-8
Description: Kernel smoothers for univariate and multivariate data, with comprehensive visualisation and bandwidth selection capabilities, including for densities, density derivatives, cumulative distributions, clustering, classification, density ridges, significant modal regions, and two-sample hypothesis tests. Chacon & Duong (2018) <doi:10.1201/9780429485572>.   
License: GPL-2 | GPL-3
URL: https://mvstat.net/ks/, https://mvstat.net/mvksa/,
        https://github.com/cran/ks/
NeedsCompilation: yes
Packaged: 2026-05-09 12:34:47 UTC; tduong
Author: Tarn Duong [aut, cre] (ORCID: <https://orcid.org/0000-0002-1198-3482>),
  Matt Wand [ctb] (ORCID: <https://orcid.org/0000-0003-2555-896X>),
  Jose Chacon [ctb],
  Artur Gramacki [ctb] (ORCID: <https://orcid.org/0000-0002-1610-9743>)
Repository: CRAN
Date/Publication: 2026-05-09 14:00:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-09 16:55:55 UTC; unix
Archs: ks.so.dSYM

importFrom("grDevices", chull, col2rgb, grey, hcl.colors, rgb)
importFrom("graphics", .filled.contour, box, contour, hist, image, legend, lines, pairs, par, persp, points, polygon, rect, rug)
importFrom("KernSmooth", dpik)
importFrom("Matrix", Diagonal, Matrix, norm)
importFrom("mclust", mclustBIC)
importFrom("mvtnorm", dmvnorm, pmvnorm, rmvnorm, dmvt, pmvt, rmvt)
importFrom("pracma", meshgrid, quiver)
importFrom("stats", IQR, cov, cutree, dbeta, dist, dnorm, ecdf, fivenum, fft, hclust, kmeans, loess, median, na.omit, nlm, optim, optimise, optimize, pchisq, pnorm, predict, qnorm, quantile, rbeta, rnorm, sd, smooth.spline, var)
importFrom("utils", head, setTxtProgressBar, tail, txtProgressBar)
    
useDynLib(ks, .registration=TRUE, .fixes="C_")

export(as.kde)
export(amise.mixt, ise.mixt, mise.mixt)
export(binning)
export(block.indices)
export(col.diverging)
export(compare, compare.kda.cv, compare.kda.diag.cv)
export(contourBreaks, .contourBreaks, contourLevels, contourSizes, contourProbs)
export(dkde, pkde, qkde, rkde)
export(dnorm.mixt, rnorm.mixt)
export(dmvnorm.mixt, rmvnorm.mixt)
export(dmvt.mixt, rmvt.mixt)
export(Hamise.mixt, Hamise.mixt.diag, hamise.mixt)
export(Hmise.mixt, Hmise.mixt.diag, hmise.mixt)
export(Hnm, Hnm.diag, hnm)
export(Hns, Hns.diag, hns, Hns.kcde, hns.kcde)
export(Hbcv, Hbcv.diag)
export(Hlscv, Hlscv.diag, hlscv, Hucv, Hucv.diag, hucv)
export(Hkda, Hkda.diag, hkda)
export(Hpi, Hpi.diag, hpi)
export(Hpi.kcde, Hpi.diag.kcde, hpi.kcde)
export(Hpi.kfe, Hpi.diag.kfe, hpi.kfe)
export(histde)
export(Hscv, Hscv.diag, hscv)
export(kcde)
export(kcopula, kcopula.de)
export(kcurv)
export(kda)
export(kdcde, dckde, reg.ucv)
export(kde)
export(kde.balloon)
export(kde.boundary)
export(kde.local.test)
export(kde.sp)
export(kde.test)
export(kde.truncate, kdde.truncate)
export(kdde)
export(kdr, kdr.segment)
export(kfe)
export(kfs)
export(kms, kms.part)
export(kroc)
export(ksupp, dwsupp)
export(Lpdiff)
export(matrix.sqrt)
export(mvnorm.mixt.mode, mvnorm.mixt.part)
export(now.time, now_time)
export(plotmixt)
export(pre.scale, pre.sphere)
export(rowKpow, getRow)
export(Sdr, Sdrv, mur, nur, nurs, Qr)
export(symconv.1d, symconv.nd)
export(vec, vech, invvec, invvech)

S3method(contourLevels, histde) 
S3method(contourLevels, kcde) 
S3method(contourLevels, kcopula.de) 
S3method(contourLevels, kda)
S3method(contourLevels, kde)
S3method(contourLevels, kdde)
S3method(contourLevels, kqde)
S3method(contourLevels, kqdde)
S3method(plot, histde)
S3method(plot, kcde)
S3method(plot, kcopula)
S3method(plot, kcopula.de)
S3method(plot, kda)
S3method(plot, kde)
S3method(plot, kde.loctest)
S3method(plot, kde.part)
S3method(plot, kdde)
S3method(plot, kdr)
S3method(plot, kfs)
S3method(plot, kms)
S3method(plot, kqde)
S3method(plot, kqdde)
S3method(plot, kroc)
S3method(plot, ksupp)
S3method(predict, histde)
S3method(predict, kda)
S3method(predict, kde)
S3method(predict, kcde)
S3method(predict, kcopula)
S3method(predict, kcopula.de)
S3method(predict, kdde)
S3method(predict, kfs)
S3method(predict, kqde)
S3method(predict, kqdde)
S3method(predict, kroc)
S3method(summary, kms)
S3method(summary, kroc)

## Introduction

Kernel smoothing for data from 1- to 6-dimensions. This package forms the basis for the practical data analysis in the book [_Multivariate Kernel Smoothing and Its Applications_](https://mvstat.net/mvksa/).

There are three main types of functions in this package:

* computing kernel estimators - these function names begin with `k`
* computing bandwidth selectors - these begin with `h` (1-d) or `H` (>1-d)
* displaying kernel estimators - these begin with `plot`.
  
The kernel used throughout is the normal (Gaussian) kernel. For 1-d data, the bandwidth h is the standard deviation of the normal kernel, whereas for multivariate data, the bandwidth matrix H is the variance matrix.
 
The main function `kde()` computes a kernel density estimate. For display, its `plot` method calls `plot.kde()`. The bandwidth choice is crucial for the performance of kernel estimators. There are several varieties of bandwidth selectors available
   
* plug-in `hpi` (1-d); `Hpi`, `Hpi.diag` (2- to 6-d) 
* least squares (or unbiased) cross validation (LSCV or UCV) `hlscv` (1-d); `Hlscv`, `Hlscv.diag` (2- to 6-d) 
* biased cross validation (BCV) `Hbcv`, `Hbcv.diag` (2- to 6-d) 
* smoothed cross validation (SCV) `hscv` (1-d); `Hscv`, `Hscv.diag` (2- to 6-d) 
* normal scale `hns` (1-d); `Hns` (2- to 6-d).

For an example with bivariate data, see `vignette("ks")`. 

The other types of kernel estimators follow a similar functionality.

## Installation

Install from CRAN:

```r
install.packages("ks") 
```

## Further reading

Chacon, J.E. & Duong, T. (2018) [_Multivariate Kernel Smoothing and Its Applications_](https://mvstat.net/mvksa/). Chapman & Hall/CRC, Boca Raton. 
  
Duong, T. (2004) [_Bandwidth Matrices for Multivariate Kernel Density Estimation_](https://mvstat.net/tduong/research/publications/duong-2005-thesis.pdf) Ph.D. Thesis, University of Western Australia. 

######################################################################
## Balloon variable KDE
######################################################################
kde.balloon <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE)
{
    ## default values
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    gridsize <- ksd$gridsize
    bgridsize <- ksd$bgridsize

    ## clip data to xmin, xmax grid limits for binned estimation
    grid.clip <- binned    
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }

    ## default bandwidths
    if (d==1 & missing(h)) h <- hns(x=x, deriv.order=2)
    if (d>1 & missing(H)) H <- Hns(x=x, deriv.order=2) 
    
    if (d==2) fhat <- kde.balloon.2d(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont, verbose=verbose)
    else stop("kde.balloon only implemented for d=2")
    if (inherits(fhat$eval.points, "list")) names(fhat$eval.points) <- ksd$names

    return(fhat)
}

######################################################################
## Bivariate balloon variable KDE
######################################################################
kde.balloon.2d <- function(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE)
{
    d <- ncol(x)
    n <- nrow(x)
    fhat <- kde(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont)
    fhat.ep <- expand.grid(fhat$eval.points)
   
    fhat.pilot <- fhat
    fhat2.pilot <- kdde(x=x, deriv.order=2, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w)

    h.pi <- (d*predict(fhat.pilot, x=fhat.ep)/drop((4*pi)^(d/2)*predict(fhat2.pilot, x=fhat.ep) %*% vec(diag(d)))^2)^(1/(d+4))*n^(-1/(d+4))
    h.pi <- array(h.pi, dim=dim(fhat$estimate))
    gs <- dim(fhat$estimate)
    
    if (verbose) { pb <- txtProgressBar(max=prod(gs)); k <- 0 }

    fhat$estimate <- array(0, dim=dim(fhat$estimate))
    for (i in 1:gs[2])
        for (j in 1:gs[1])
        {
            if (!is.na(h.pi[i,j]) & !is.infinite(h.pi[i,j])) if (h.pi[i,j]>0) fhat$estimate[i,j] <- kde(x=x, w=w, H=h.pi[i,j]^2*diag(d), eval.points=fhat.ep[i+(j-1)*gs[1],])$estimate
            if (verbose) { k <- k+1; setTxtProgressBar(pb,k) }
        }
    if (verbose) close(pb)

    ## re-scale density estimate to integral 1
    delta.int <- prod(sapply(fhat$eval.points, diff)[1,]) 
    riemann.sum <- sum(fhat$estimate*delta.int)
    fhat$estimate <- fhat$estimate/riemann.sum
    
    fhat$names <- parse.name(x)  ## add variable names
    if (compute.cont)
        fhat$cont <- contourLevels(fhat, cont=1:99, approx=approx.cont)
    fhat$H <- h.pi^2
    
    return(fhat)
}

######################################################################
## Sample point variable KDE
######################################################################
kde.sp <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE)
{
    ## default values
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    gridsize <- ksd$gridsize
    bgridsize <- ksd$bgridsize

    ## clip data to xmin, xmax grid limits for binned estimation
    grid.clip <- binned    
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }

    ## default bandwidths
    if (d==1 & missing(h)) h <- hns(x=x, deriv.order=4)
    if (d>1 & missing(H)) H <- Hns(x=x, deriv.order=4) 

    if (d==2) fhat <- kde.sp.2d(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont, verbose=verbose, pre=TRUE)
    else stop("kde.sp only implemented for d=2")

    fhat$names <- parse.name(x)  ## add variable names
    if (inherits(fhat$eval.points, "list")) names(fhat$eval.points) <- ksd$names
    
    if (compute.cont)
        fhat$cont <- contourLevels(fhat, cont=1:99, approx=approx.cont)

    return(fhat)
}

######################################################################
## Bivariate sample point variable KDE
######################################################################
kde.sp.2d <- function(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE, pre=TRUE)
{
    d <- 2; n <- nrow(x)
    if (pre) 
    { 
        x.orig <- x
        S12 <- diag(apply(x.orig, 2, sd))  
        Sinv12 <- matrix.pow(S12,-1)
        x <- pre.scale(x)
        if (!missing(xmin)) xmin <- xmin %*% Sinv12
        if (!missing(xmax)) xmax <- xmax %*% Sinv12
        H <- Hns(x=x, deriv.order=4)
    }
        
    fhat <- kde(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont)
    fhat.pilot <- fhat
    fhat1.pilot <- kdde(x=x, deriv.order=1, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w)
    fhat2.pilot <- kdde(x=x, deriv.order=2, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w)
    fhat3.pilot <- kdde(x=x, deriv.order=3, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w)
    fhat4.pilot <- kdde(x=x, deriv.order=4, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w)

    fhat.pilot <- predict(fhat.pilot, x=x)
    fhat1.pilot <- predict(fhat1.pilot, x=x)
    fhat2.pilot <- predict(fhat2.pilot, x=x)
    fhat3.pilot <- predict(fhat3.pilot, x=x)
    fhat4.pilot <- predict(fhat4.pilot, x=x)

    lambda1 <- 1/fhat.pilot^5*rowKpow(fhat1.pilot, r=4)
    lambda2 <- 1/fhat.pilot^4*rowKpow(fhat2.pilot, fhat1.pilot, r=1, s=2)
    lambda3 <- 1/fhat.pilot^3*rowKpow(fhat2.pilot, r=2)
    lambda4 <- 1/fhat.pilot^3*rowKpow(fhat3.pilot, fhat1.pilot, r=1, s=1)
    lambda5 <- 1/fhat.pilot^2*fhat4.pilot
    lambda <- drop((24*lambda1 - 36*lambda2 + 6*lambda3 + 8*lambda4 - lambda5) %*% Sdr(d,r=4) %*% (vec(diag(d)) %x% vec(diag(d))))

    RK <- (4*pi)^(-d/2)
    h.Ab  <- (8*d*RK*fhat.pilot^(1+d/2)/lambda^2)^(1/(d+8))*n^(-1/(d+8))
    
    fhat$estimate <- array(0, dim=dim(fhat$estimate))
    xmin <- sapply(fhat$eval.points, min)
    xmax <- sapply(fhat$eval.points, max)
    
    if (verbose) { pb <- txtProgressBar(max=n) }
    for (i in 1:n)
    {
        if (verbose) setTxtProgressBar(pb, i)
        HAb <- h.Ab[i]^2*diag(d)
        fhat$estimate <- fhat$estimate + kde(x=matrix(x[i,], nrow=1), H=HAb, xmin=xmin, xmax=xmax, binned=binned, gridsize=dim(fhat$estimate), bgridsize=dim(fhat$estimate))$estimate
    }
    if (verbose) close(pb)
    
    fhat$estimate <- fhat$estimate/n
    if (pre)
    {
        ep <- cbind(fhat$eval.points[[1]], fhat$eval.points[[2]]) %*% S12
        fhat$eval.points[[1]] <- ep[,1]
        fhat$eval.points[[2]] <- ep[,2]
        fhat$estimate <- fhat$estimate/det(S12)
        fhat$x <- x.orig
    }
    fhat$cont <- contourLevels(fhat, cont=1:99)
    fhat$H <- h.Ab^2
    
    return(fhat)
}

#############################################################################
## Cumulative integral for KDE
#############################################################################
integral.kde <- function(q, fhat, density)
{
    gridsize <- length(fhat$eval.points)

    ## Use Simpson's rule to compute numerical integration
    simp.rule <- rep(0, gridsize-1)
    for (i in 1:(gridsize-1))
    {
        del <- fhat$eval.points[i+1] - fhat$eval.points[i]
        simp.rule[i] <- min(fhat$estimate[i], fhat$estimate[i+1])*del + 1/2*abs(fhat$estimate[i+1] - fhat$estimate[i])*del 
    }

    ## add last incomplete trapezoid
    q.ind <- findInterval(x=q, vec=fhat$eval.points)
    q.prob <- rep(0, length(q))
    i <- 0

    for (qi in q.ind)
    {
        i <- i+1

        if (qi==0)
            q.prob[i] <- 0
        else if (qi < gridsize)
        {
            ## linearly interpolate kde 
            fhat.estqi <- (fhat$est[qi+1] - fhat$est[qi])/(fhat$eval[qi+1] - fhat$eval[qi]) * (q[i] - fhat$eval[qi]) + fhat$est[qi]
            delqi <- q[i] - fhat$eval[qi] 
              
            simp.ruleqi <- min(fhat.estqi, fhat$est[qi])*delqi + 1/2*abs(fhat.estqi - fhat$est[qi])*delqi
            q.prob[i] <- sum(simp.rule[1:qi]) + simp.ruleqi
        }
        else
        {
            if (density) q.prob[i] <- 1
            else q.prob[i] <- sum(simp.rule) 
        }
    }

    if (density) q.prob[q.prob>=1] <- 1

    ## remove possible decreasing values in q.prob
    dec.ind <- which(diff(q.prob)<0)
    while (length(dec.ind)>0) 
    {
        dec.ind <- dec.ind+1 
        for (i in dec.ind) q.prob[i] <- q.prob[i-1] 
        dec.ind <- which(diff(q.prob)<0)
    }

    return(q.prob)
}

## cumulative probability P(fhat <= q)
pkde <- function(q, fhat)
{
    return(integral.kde(q=q, fhat=fhat, density=TRUE))
}

## density value of KDE at x 
## alias for predict.kde
dkde <- function(x, fhat)
{
    return(predict(fhat, x=x))
}

## p-quantile of KDE, i.e. solve for x where P(fhat < x) = p 
qkde <- function(p, fhat)
{
    if (any(p > 1) | any(p < 0)) stop("p must be <= 1 and >= 0")
  
    cumul.prob <- pkde(q=fhat$eval.points, fhat=fhat)
    ind <- findInterval(x=p, vec=cumul.prob)

    quant <- rep(0, length(ind))
    for (j in 1:length(ind))
    {
        i <- ind[j]
        if (i==0)
            quant[j] <- fhat$eval.points[1]
        else if (i>=length(fhat$eval.points))
            quant[j] <- fhat$eval.points[length(fhat$eval.points)]
        else  
        {
            quant1 <- fhat$eval.points[i]
            quant2 <- fhat$eval.points[i+1]
              
            prob1 <- cumul.prob[i]
            prob2 <- cumul.prob[i+1]
            alpha <- (p[j] - prob2)/(prob1 - prob2)
            quant[j] <- quant1*alpha + quant2*(1-alpha)
        }
    }
  
    return(quant)
}

## Silverman (1983)'s random sample from KDE
rkde <- function(n, fhat, positive=FALSE)
{
    if (positive) x <- log(fhat$x)
    else x <- fhat$x

    if (is.vector(fhat$H)) { d <- 1; nsamp <- length(x) }
    else { d <- ncol(fhat$H); nsamp <- nrow(x) } 

    x.ind <- sample(1:nsamp, size=n, replace=TRUE, prob=fhat$w)

    if (d==1)
    {
        h <- fhat$h
        rkde.val <- x[x.ind] + h*rnorm(n=n, mean=0, sd=1)
        if (positive) rkde.val <- exp(rkde.val)
    }
    else if (d>1)
    {
        H <- fhat$H
        rkde.val <- x[x.ind,] + rmvnorm(n=n, mean=rep(0,d), sigma=diag(d)) %*% matrix.sqrt(H)
    }

    return(rkde.val)
}

## plot cumulative probability as shaded region on a KDE 
plotkde.cumul <- function(fhat, q, add=FALSE, col="blue", ...)
{
    qind <- fhat$eval.points<=q
    n <- sum(qind)
    if (!add) plot(fhat)
    polygon(c(fhat$eval.points[qind],fhat$eval.points[n]), c(fhat$estimate[qind],0), col=col, ...)
}

##########################################################################
## Default values for parameter choices for ks objects
##########################################################################
ks.defaults <- function(x, w, binned, bgridsize, gridsize)
{
    ## dimensions of x
    if (is.vector(x)) { d <- 1; n <- length(x) }
    else { d <- ncol(x); n <- nrow(x) }

    ## default uniform weights
    if (missing(w)) w <- rep(1,n)
    else if (!missing(w))
    {
        if (!(identical(all.equal(sum(w), n), TRUE)))
        {
            warning("Weights don't sum to sample size - they have been scaled accordingly\n")
            w <- w*n/sum(w)
        }
    }

    ## default binning flag
    if (missing(binned)) binned <- default.bflag(d=d, n=n)

    ## default grid sizes
    if (missing(bgridsize))
    {
        if (missing(gridsize)) bgridsize <- default.bgridsize(d)
        else bgridsize <- gridsize
    }
    if (missing(gridsize)) gridsize <- default.gridsize(d)
    if (length(gridsize)==1) gridsize <- rep(gridsize, d)
    if (length(bgridsize)==1) bgridsize <- rep(bgridsize, d)
    if (inherits(x, c("matrix","data.frame"))) xnames <- colnames(x) 
    else xnames <- NULL
    ksd <- list(d=d, n=n, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize, names=xnames)

    return(ksd)
}

check.eks.version <- function()
{
    pfn <- sapply(strsplit(toString(sys.call(-sys.parent())), split=","), head,n=1)
    call.from.eks <- try(utils::getFromNamespace(pfn, "eks"), silent=TRUE)
    eks.version <- "1.1.0"
    if (!inherits(call.from.eks, "try-error"))
    {
        eks.version <- try(utils::packageVersion("eks"), silent=TRUE)
        if (inherits(eks.version, "try-error")) eks.version <- "1.1.0"
    }

    return(eks.version)
}

##########################################################################
## Parse variable name
##########################################################################
parse.name <- function(x)
{
    if (is.vector(x))
    {
        d <- 1
        x.names <- deparse(substitute(x))
    }
    else
    {  
        d <- ncol(x)
        x.names <- colnames(x)
        if (is.null(x.names))
        {
            x.names <- strsplit(deparse(substitute(x)), "\\[")[[1]][1]
            x.names <- paste(x.names, "[, ", 1:d,"]",sep="") 
        }
    }

    return(x.names)
}

##########################################################################
## Basic vectors and matrices and their operations
##########################################################################
## vec operator 
vec <- function(x, byrow=FALSE)
{
    if (is.vector(x)) return (x)
    if (byrow) x <- t(x)
    d <- ncol(x)
    vecx <- vector()
    for (j in 1:d) vecx <- c(vecx, x[,j])

    return(vecx)           
}

## vech operator
vech <- function(x)
{
    if (is.vector(x))
    {
    if (length(x)==1)
        return (x)
    else
        stop("vech undefined for vectors")
    }
    else if (is.matrix(x))
    {  
        d <- ncol(x)
        if (d!=nrow(x)) stop("vech only defined for square matrices")

        vechx <- vector()
        for (j in 1:d) vechx <- c(vechx, x[j:d,j])
        return(vechx)
    }
}

## inverse vec operator 
invvec <- function(x, ncol, nrow, byrow=FALSE)
{
    if (length(x)==1) return(x)

    d <- sqrt(length(x))
    if (missing(ncol) | missing(nrow))
    {
        ncol <- d; nrow <- d
        if (round(d) != d)
            stop("Need to specify nrow and ncol for non-square matrices")
    }

    invvecx <- matrix(0, nrow = nrow, ncol = ncol)
    if (byrow)
    for (j in 1:nrow)
        invvecx[j,] <- x[c(1:ncol) + (j-1)*ncol]
    else
        for (j in 1:ncol)
            invvecx[,j] <- x[c(1:nrow) + (j-1)*nrow]

    return(invvecx)
}

## inverse vech operator 
invvech <- function(x)
{
  if (length(x)==1) return(x)
  
  d <- (-1 + sqrt(8*length(x) + 1))/2
  if (round(d) != d)
    stop("Number of elements in x will not form a square matrix")
  invvechx <- matrix(0, nrow=d, ncol=d)

  for (j in 1:d)
        invvechx[j:d,j] <- x[1:(d-j+1)+ (j-1)*(d - 1/2*(j-2))]
  
  invvechx <- invvechx + t(invvechx) - diag(diag(invvechx))
  
  return(invvechx)
}

## trace of matrix
tr <- function(A)
{   
    count <- 0
    if (is.vector(A)) return (A[1])
    if (nrow(A)!=ncol(A)) stop("Not square matrix")
    else 
        for (i in 1:nrow(A)) count <- count + A[i,i]
  
    return(count)
}

## elementary vector 
elem <- function(i, d)
{
    elem.vec <- rep(0, d)
    elem.vec[i] <- 1

    return(elem.vec)
}      

## commutation matrix (taken from MCMCglmmm library)
comm <- function(m,n)
{
    K <- matrix(0,m*n, m*n)
    H <- matrix(0,m,n)
    for (i in 1:m)
    {
        for(j in 1:n)
        { 
            H[i,j] <- 1
            K <- K+kronecker(H,t(H))
            H[i,j] <- 0
        }
    }

    return(K)
}

##########################################################################
## Duplication matrix
## Taken from Felipe Osorio http://www.ime.usp.br/~osorio/files/dupl.q
##########################################################################
dupl <- function(order, ret.q = FALSE)
{
    ## call
    cl <- match.call()
    time1 <- proc.time()
    if (!is.integer(order))
        order <- as.integer(order)
    n <- order - 1
    
    ## initial duplication matrix
    d1 <- matrix(0, nrow = 1, ncol = 1)
    d1[1,1] <- 1
    if (!is.integer(d1))
        storage.mode(d1) <- "integer"
    
    ## recursive formula
    if (n > 0)
    {
        for (k in 1:n)
        {
            drow <- 2*k + 1 + nrow(d1)
            dcol <- k + 1 + ncol(d1)
            d2 <- matrix(0, nrow = drow, ncol=dcol)
            storage.mode(d2) <- "integer"
            d2[1,1] <- 1
            d2[2:(k+1),2:(k+1)] <- diag(k)
            d2[(k+2):(2*k+1),2:(k+1)] <- diag(k)
            d2[(2*k+2):drow,(k+2):dcol] <- d1
            ## permutation matrix
            q <- permute.mat(k)
            ## new duplication matrix
            d2 <- q %*% d2
            storage.mode(d2) <- "integer"
            d1 <- d2
        }
    }
    else {
        d2 <- q <- d1
    }
    
    ## results
    obj <- list(call=cl, order=order, d=d2)
    if (ret.q)
        obj$q <- q
    obj$time <- proc.time() - time1
    obj
}

##########################################################################
## Pre-scaling
## Parameters
## x - data points
##
## Returns
## Pre-scaled x values
##########################################################################
pre.scale <- function(x, mean.centred=FALSE)
{
    S <- diag(diag(var(x)))
    Sinv12 <- matrix.sqrt(chol2inv(chol(S)))

    if (mean.centred) x.scaled <- sweep(x, 2, apply(x, 2, mean))
    else x.scaled <- x  
    x.scaled <- x.scaled %*% Sinv12

    return (x.scaled)
}

##########################################################################
## Pre-sphering
## Parameters
## x - data points
##
## Returns
## Pre-sphered x values
##########################################################################
pre.sphere <- function(x, mean.centred=FALSE)
{
    S <- var(x)
    Sinv12 <- matrix.sqrt(chol2inv(chol(S)))

    if (mean.centred) x.sphered <- sweep(x, 2, apply(x, 2, mean))
    else x.sphered <- x  
    x.sphered <- x.sphered %*% Sinv12

  return (x.sphered)
}

##########################################################################
## Boolean functions
##########################################################################
is.even <- function(x)
{
    y <- x[x>0] %%2
    return(identical(y, rep(0, length(y))))
}

is.diagonal <- function(x)
{
    return(identical(diag(diag(x)),x))
}

##########################################################################
## Finds row index matrix
## Parameters
## x - data points
##
## Returns
## i  - if r==mat[i,]
## NA - otherwise
##########################################################################
which.mat <- function(r, mat)
{
    ind <- numeric()

    for (i in 1:nrow(mat))
    if (identical(r, mat[i,])) ind <- c(ind,i)

    return(ind)  
}

###################################################################
## Permutation functions
###################################################################

####################################################################
## Exactly the same function as combinat:::permn
####################################################################
permn.ks <- function (x, fun = NULL, ...) 
{
    if (is.numeric(x) && length(x) == 1 && x > 0 && trunc(x) == x) x <- seq(x)
    n <- length(x)
    nofun <- is.null(fun)
    out <- vector("list", gamma(n + 1))
    p <- ip <- seqn <- 1:n
    d <- rep(-1, n)
    d[1] <- 0
    m <- n + 1
    p <- c(m, p, m)
    i <- 1
    use <- -c(1, n + 2)
    while (m != 1) {
        out[[i]] <- if (nofun) 
            x[p[use]]
        else fun(x[p[use]], ...)
        i <- i + 1
        m <- n
        chk <- (p[ip + d + 1] > seqn)
        m <- max(seqn[!chk])
        if (m < n) 
            d[(m + 1):n] <- -d[(m + 1):n]
        index1 <- ip[m] + 1
        index2 <- p[index1] <- p[index1 + d[m]]
        p[index1 + d[m]] <- m
        tmp <- ip[index2]
        ip[index2] <- ip[m]
        ip[m] <- tmp
    }
    out
}

##########################################################################
## Permutations with repetitions of the first d naturals (1:d) taking 
## k elements at a time. There are d^k of them, each having length k 
## => We arrange them into a matrix of order d^k times k
## Each row represents one permutation
## Second version: filling in the matrix comlumn-wise (slightly faster)
##########################################################################
perm.rep <- function(d,r)
{
    if (r==0) { PM <- 1 }
    if (r>0)
    {
        PM <- matrix(nrow=d^r,ncol=r)
        for (pow in 0:(r-1))
        {
            t2 <- d^pow
            p1 <- 1
            while (p1<=d^r)
            {
                for (al in 1:d)
                {
                    for (p2 in 1:t2)
                    {
                        PM[p1,r-pow] <- al
                        p1 <- p1+1
                    }
                }
            }
        }
    }

    return(PM)
}

##########################################################################
## Permute a list of values
##
## Same function as EXPAND.GRID (base package), modified to take 
## list as an argument and returns a matrix 
##########################################################################
permute <- function (args) 
{
    nargs <- length(args)
    if (!nargs) 
        return(as.data.frame(list()))
    if (nargs == 1 && is.list(a1 <- args[[1]])) 
    nargs <- length(args <- a1)
    if (nargs <= 1) 
        return(as.data.frame(if (nargs == 0 || is.null(args[[1]])) list() else args, optional = TRUE))
    
    cargs <- args
    rep.fac <- 1
    orep <- prod(sapply(args, length))

    for (i in 1:nargs) 
    {
        x <- args[[i]]
        nx <- length(x)
        orep <- orep/nx
        cargs[[i]] <- rep(rep(x, rep(rep.fac, nx)), orep)
        rep.fac <- rep.fac * nx
    }
    do.call("cbind", cargs)
} 

permute.mat <- function(order)
{
    m <- as.integer(order)
    m <- m + 1
    eye <- diag(m)
    u1 <- eye[1:m,1]
    u2 <- eye[1:m,2:m]
    q1 <- kronecker(eye, u1)
    q2 <- kronecker(eye, u2)
    q <- matrix(c(q1, q2), nrow = nrow(q2), ncol = ncol(q1) + ncol(q2))
    if (!is.integer(q)) storage.mode(q) <- "integer"
    
    return(q)
}

##########################################################################
## pinv.all generates all the permutations PR_{d,r} as described in
## Appendix B of Chacon and Duong (2014)
##########################################################################
pinv.all <- function(d,r)
{
    i <- 1:d^r
    n <- i-1
    dpow <- d^(0:r)
    n.mat <- matrix(rep(n,r+1),byrow=FALSE,nrow=d^r,ncol=r+1)
    dpow.mat <- matrix(rep(dpow,d^r),byrow=TRUE,nrow=d^r,ncol=r+1)
    ndf.mat <- floor(n.mat/dpow.mat)
    ans <- ndf.mat[,r:1]-d*ndf.mat[,(r+1):2]    

    return(ans+1)
} 

##########################################################################
## Block indices for double sums
##########################################################################
block.indices <- function(nx, ny, d, r=0, diff=FALSE, block.limit=1e6, npergroup, arr.ind=FALSE)
{
    if (missing(npergroup)) 
    { 
        if (diff) npergroup <- max(c(block.limit %/% (nx*d^r), 1))
        else npergroup <- max(c(block.limit %/% nx,1))
    }
    nseq <- seq(1, ny, by=npergroup)
    if (tail(nseq,n=1) <= ny) nseq <- c(nseq, ny+1)
    if (length(nseq)==1) nseq <- c(1, ny+1)
    if (arr.ind) nseq <- cbind(head(nseq,n=-1), tail(nseq,n=-1)-1)
    
    return(nseq)
}

block.indices2 <- function(nx, ny, block.limit=1e6, npergroup)
{
    if (missing(npergroup)) npergroup <- max(c(block.limit %/% nx,1))
    nseq <- seq(1, ny, by=npergroup)
    if (tail(nseq,n=1) <= ny) nseq <- c(nseq, ny+1)
    if (length(nseq)==1) nseq <- c(1, ny+1)
    
    return(nseq)
}

####################################################################
## Differences for double sums calculations
####################################################################
differences <- function(x, y, upper=FALSE, ff=FALSE, Kpow=0)
{
    if (missing(y)) y <- x
    if (is.vector(x)) x <- t(as.matrix(x))
    if (is.vector(y)) y <- t(as.matrix(y))

    nx <- nrow(x)
    ny <- nrow(y)
    d <- ncol(x)

    if (ff) difs <- ff(init=0, dim=c(nx*ny,d))
    else difs <- matrix(ncol=d,nrow=nx*ny)

    for (j in 1:d)
    {
        difs[,j] <- rep(x[,j], times=ny) - rep(y[1:ny,j], each=nx)
        ## jth column of difs contains all the differences X_{ij}-Y_{kj}
    }

    if (upper)
    {
        ind.remove <- numeric()
        for (j in 1:(nx-1))
          ind.remove <- c(ind.remove, (j*nx+1):(j*nx+j))
          
        return(difs[-ind.remove,])
    }
    else return(difs)
}

##########################################################################
## Odd factorial
##########################################################################
OF <- function(m) { factorial(m)/(2^(m/2)*factorial(m/2)) } 

##########################################################################
## Matrix square root - taken from Stephen Lake 
## http://www5.biostat.wustl.edu/s-news/s-news-archive/200109/msg00067.html
##########################################################################
matrix.sqrt <- function(A)
{
    if (length(A)==1) return(sqrt(A))
    sva <- svd(A)
    if (min(sva$d)>=0) Asqrt <- sva$u %*% diag(sqrt(sva$d)) %*% t(sva$v)
    else stop("Matrix square root is not defined")

    return(Asqrt)
}

##########################################################################
## Matrix power
##########################################################################
matrix.pow <- function(A, n)
{
    if (nrow(A)!=ncol(A)) stop("A must be a square matrix")
    if (floor(n)!=n) stop("n must be an integer")
    if (n==0) return(diag(ncol(A)))
    if (n < 0) return(matrix.pow(A=chol2inv(chol(A)), n=-n))
        
    ## trap non-integer n and return an error
    if (n == 1) return(A)
    result <- diag(1, ncol(A))
    while (n > 0) 
    {
        if (n %% 2 != 0) 
        {
          result <- result %*% A
          n <- n - 1
        }
        A <- A %*% A
        n <- n / 2
    }
    
    return(result)
}

##########################################################################
## Kmat computes the commutation matrix of orders m,n
##########################################################################
Kmat <- function(m, n)
{ 
    K <- matrix(0,ncol=m*n,nrow=m*n)
    i <-1:m; j <- 1:n
    rows <- rowSums(expand.grid((i-1)*n,j))
    cols <- rowSums(expand.grid(i,(j-1)*m))    
    positions <- cbind(rows,cols)
    K[positions] <- 1

    return(K)
}

##########################################################################
## mat.Kprod computes row-wise Kronecker products of matrices
## Returns a matrix with rows U[i,]%x%V[i,]
##########################################################################
mat.Kprod <- function(U,V)
{ 
    n1 <- nrow(U)
    n2 <- nrow(V)
    if (n1!=n2) stop("U and V must have the same number of vectors")
    p <- ncol(U)
    q <- ncol(V)
    onep <- rep(1,p)
    oneq <- rep(1,q)
    P <- (U%x%t(oneq))*(t(onep)%x%V)

    return(P)
}

##########################################################################
## Kpow computes the Kronecker power of a matrix A
##########################################################################
Kpow <- function(A,pow)
{    
    if (floor(pow)!=pow) stop("pow must be an integer")
    Apow <- A
    if (pow==0) { Apow <- 1 }
    if (pow>1) { for(i in 2:pow) Apow <- Apow%x%A }

    return(Apow)
} 

##########################################################################
## Kronecker sum
##########################################################################
Ksum <- function(A,B)
{
    AB <- numeric()
    for (i in 1:nrow(A))
        for (j in 1:nrow(B))
            AB <- rbind(AB, A[i,] + B[j,])

    return(AB)
}

##########################################################################
## Row-wise Kronecker product
##########################################################################
rowKpow <- function(A,B,r=1,s=1)
{
    A <- as.matrix(A)
    if (missing(B))
    {
        res <- t(apply(t(A), 2, Kpow, r))
    }
    else
    {    
        B <- as.matrix(B)
        if (nrow(A)!=nrow(B)) stop("A and B must have same number of rows") 
        res <- numeric()
        
        Ar <- t(apply(t(A), 2, Kpow, r))
        Bs <- t(apply(t(B), 2, Kpow, s))
        for (i in 1:ncol(Ar)) res <- cbind(res, apply(Bs, 2, FUN="*", Ar[,i]))
    }
    
    return(drop(res))
}

getRow <- function(object, n) { return(object[n,]) }

## Returns a matrix with the pow-th Kronecker power of A[i,] in the i-th row
mat.Kpow <- function(A,pow)
{ 
    Apow <- A
    if (pow==0) { Apow <- matrix(1,nrow=nrow(A), ncol=1) }  
    if (pow>1) 
    {
        for(i in 2:pow) Apow <- mat.Kprod(Apow,A)
    }

    return(Apow)
}

## Vector of all r-th partial derivatives of the normal density at x=0, i.e., D^{\otimes r)\phi(0)
DrL0 <- function(d,r)
{
    v <- as.vector(Kpow(A=vec(diag(d)),pow=r/2))
    DL0 <- (-1)^(r/2)*(2*pi)^(-d/2)*OF(r)*matrix(Sdrv(d=d, r=r, v=v), ncol=1) 
    
    return(DL0)
}

##########################################################################
## Symmetriser matrix
##########################################################################
## Wrapper functions for Chacon & Duong (2014) 
Sdr <- function(d, r, type="recursive")
{
    type <- match.arg(type, c("recursive", "direct"))
    Sdr.mat <- do.call(paste("Sdr", type, sep="."), list(d=d, r=r))

    return(Sdr.mat)
}

Sdrv <- function(d, r, v, type="recursive")
{
    type <- match.arg(type, c("recursive", "direct"))
    v <- as.vector(v)
    Sdrvec <- do.call(paste("Sdrv", type, sep="."), list(d=d, r=r, v=v))
  
    return(Sdrvec)
}

##########################################################################
## Sdr.direct computes the symmetrizer matrix S_{d,r} based on Equation (4)
## as described in Section 3 of Chacon and Duong (2014)
##########################################################################
Sdr.direct <- function(d,r)
{ 
    S <- matrix(0,ncol=d^r,nrow=d^r)
    per <- permn.ks(r)
    per.rep <- pinv.all(d,r)
    nper <- factorial(r)
    nper.rep <- d^r
    per <- matrix(unlist(per), byrow=TRUE, ncol=r, nrow=nper)    
    pow <- 0:(r-1)
    dpow <- d^pow
    
    if (nper.rep<=nper)
    {
        dpow.mat <- matrix(rep(dpow,nper),byrow=TRUE,ncol=r,nrow=nper)
        for(i in 1:nper.rep)
        {      
            ## Loop over no. perms with reps (d^r)
            pinvi <- per.rep[i,]
            sigpinvi <- matrix(pinvi[per],byrow=FALSE,nrow=nrow(per),ncol=ncol(per))
            psigpinvi <- drop(1+rowSums((sigpinvi-1)*dpow.mat))
            S[i,] <- tabulate(psigpinvi,nbins=d^r)
        }
    }
    
    if (nper<nper.rep)
    {
        dpow.mat <- matrix(rep(dpow,nper.rep),byrow=TRUE,ncol=r,nrow=nper.rep)
        for (s in 1:nper)
        {          
            ## Loop over no. perms (r!)
            sig <- per[s,]
            sigpinv <- per.rep[,sig]
            psigpinv <- drop(1+rowSums((sigpinv-1)*dpow.mat))
            locations <- cbind(1:d^r,psigpinv)
            S[locations] <- S[locations]+1
        }
    }    
    
    return(S/nper)
}
     
##########################################################################
## Sdr.recursive computes the symmetrizer matrix S_{d,r} based on 
## the recursive approach detailed in Algorithm 1 in Section 3 of 
## Chacon and Duong (2014)
########################################################################## 
Sdr.recursive <- function(d,r)
{
    S <- diag(d)
    if (r==0) S <- 1
    if (r>=2) 
    {
        Id <- diag(d)
        T <- Id 
        A <- Kmat(d,d)        
        for (j in 2:r)
        {
            T <- ((j-1)/j)*(A%*%(T%x%Id)%*%A)+A/j
            S <- (S%x%Id)%*%T
            if (j<r) A <- Id%x%A
        }
    }
    
    return(S)
}

##########################################################################
## Symmetrizer matrix applied to a vector
##########################################################################
    
##########################################################################
## Sdrv.direct computes the result of multiplying the symmetrizer matrix
## S_{d,r} by a vector v of length d^r, based on Equation (4) as described
## in Section 4 of Chacon and Duong (2014)
##########################################################################
Sdrv.direct <- function(d,r,v)
{
    Sv <- rep(0,d^r)
    per <- permn.ks(r)
    per.rep <- pinv.all(d,r)
    nper <- factorial(r)
    nper.rep <- d^r
    per <- matrix(unlist(per), byrow=TRUE, ncol=r, nrow=nper)    
    pow <- 0:(r-1)
    dpow <- d^pow
    
    if (nper.rep<=nper)
    {
        dpow.mat <- matrix(rep(dpow,nper),byrow=TRUE,ncol=r,nrow=nper)
        for (i in 1:nper.rep)
        {      
            ## Loop over no. perms with reps (d^r)
            pinvi <- per.rep[i,]
            sigpinvi <- matrix(pinvi[per],byrow=FALSE,nrow=nrow(per),ncol=ncol(per))
            psigpinvi <- drop(1+rowSums((sigpinvi-1)*dpow.mat))
            Sv[i] <- sum(tabulate(psigpinvi,nbins=d^r)*v)
        }
    }
    
    if (nper<nper.rep)
    {
        dpow.mat <- matrix(rep(dpow,nper.rep),byrow=TRUE,ncol=r,nrow=nper.rep)
        for (s in 1:nper)
        {          
        ## Loop over no. perms (r!)
            sig <- per[s,]
            sigpinv <- per.rep[,sig]
            psigpinv <- drop(1+rowSums((sigpinv-1)*dpow.mat))
            Sv <- Sv+v[psigpinv]
        }
    }    
    
    return(Sv/nper)
}
   
##########################################################################
## Sdrv.recursive computes the result of multiplying the symmetrizer matrix
## S_{d,r} by a vector v of length d^r, based on the recursive Algorithm 2
## as described in Section 4 of Chacon and Duong (2014)
##########################################################################
Sdrv.recursive <- function(d,r,v)
{
    if ((!is.vector(v)) & (!is.matrix(v))) { stop("v must be a vector or a matrix") }
    if (is.vector(v)) { v <- matrix(v,nrow=1) }
    n <- nrow(v)   
    if (ncol(v)!=d^r) { stop("Length of the vector(s) must equal d^r") }
    
    if ((r==0)|(r==1)) { w <- v } 
    else
    {
        per.rep <- pinv.all(d,r)
        nper.rep <- d^r
        dpow <- d^((r-1):0)
        dpow.mat <- matrix(rep(dpow,d^r),ncol=r,nrow=d^r,byrow=TRUE) 

        w <- v
        for (p in 2:r)
        {
            Tv <- matrix(0,nrow=n,ncol=d^r)
            for (j in 1:p)
            {
                per.rep.tau <- per.rep
                per.rep.tau[,j] <- per.rep[,p]
                per.rep.tau[,p] <- per.rep[,j]
                positions <- 1+rowSums((per.rep.tau-1)*dpow.mat)
                Tv[,positions] <- Tv[,positions]+w
            }
            w <- Tv/p
        }
    }

    return(drop(w))
}

##########################################################################
## Lp norm between two functions (grid based)
##########################################################################
Lpdiff <- function(f1, f2, p=2, index=1)
{
    if (is.vector(f1$H)) d <- 1 else d <- ncol(f1$H)
    f.diff <- f1
    if (is.list(f1$estimate)) f.diff$estimate <- (abs(f1$estimate[[index]] - f2$estimate[[index]]))^p
    else f.diff$estimate <- (abs(f1$estimate - f2$estimate))^p
    if (d==1)
    {
        delta <- diff(f.diff$eval.points)
        riemann.sum <- sum(c(delta[1], delta)*f.diff$estimate)
    }
    else if (d>1)
    {
        delta <- sapply(f.diff$eval.points, diff)
        delta <- rbind(head(delta, n=1), delta)
        if (d==2) riemann.sum <- sum(outer(delta[,1], delta[,2]) * f.diff$estimate)
        else if (d==3) riemann.sum <- sum(outer(outer(delta[,1], delta[,2]), delta[,3]) * f.diff$estimate)
    }  

    return(riemann.sum)
}

## ISE of difference between two KDEs
ise.diff <- function(fhat1, fhat2, xmin, xmax)
{
    if(!isTRUE(all.equal(fhat1$eval.points, fhat2$eval.points)))
    stop("fhat1 and fhat2 need to de defined on the same grid")

    fhat.sq <- fhat1
    fhat.sq$estimate <- (fhat1$estimate - fhat2$estimate)^2

    if (missing(xmin) & missing(xmax))
    int <- integral.kde(fhat=fhat.sq, q=max(fhat.sq$eval.points)+0.1*abs(max(fhat.sq$eval.points)), density=FALSE)

    if (missing(xmin) & !missing(xmax))
    int <- integral.kde(fhat=fhat.sq, q=xmax, density=FALSE)

    if (!missing(xmin) & missing(xmax))
    int <- integral.kde(fhat=fhat.sq, q=max(fhat.sq$eval.points)+0.1*abs(max(fhat.sq$eval.points)), density=FALSE) - integral.kde(fhat=fhat.sq, q=xmin, density=FALSE)

    if (!missing(xmin) & !missing(xmax))
    int <- integral.kde(fhat=fhat.sq, q=xmax, density=FALSE) - integral.kde(fhat=fhat.sq, q=xmin, density=FALSE)

    return(int)   
}

##########################################################################
## Create transparent colour
## partial alias for plot3D::alpha.col
##########################################################################
## copied from plot3D::alpha.col
plot3D.alpha.col <- function (col="grey", alpha=0.5) 
{
    RGBini <- col2rgb(col)
    return(rgb(t(RGBini), maxColorValue=255, alpha=alpha*255))
}

#transparency.col <- plot3D.alpha.col 
transparency.col <- function(col, alpha)
{ 
    #trans.ind <- as.vector(na.omit(col!="transparent"))
    trans.ind <- !(col %in% "transparent")
    col[trans.ind] <- plot3D.alpha.col(col[trans.ind], alpha=alpha) 
    
    return(col)
}

round.signif <- function(x, digits=4, format=FALSE) 
{
    x <- sapply(x, function(.) ifelse(.>=10^(digits-1), round(.), signif(.,digits))) 
    if (format) x <- format(x, nsmall=digits)

    return(x)
}

closeTxtProgressBar <- function(...) 
{
    width <- getOption("width")
    cat(paste0("\r", strrep(" ", width)))
    close(...)
} 

now.time <- now_time <- function() 
{
    paste0("[", unlist(strsplit(toString(Sys.time()), "\\."))[1], "]")
}

#############################################################################
## Kernel functional estimation
#############################################################################
kfe <- function(x, G, deriv.order, inc=1, binned, bin.par, bgridsize, deriv.vec=TRUE, add.index=TRUE, verbose=FALSE)
{
    r <- deriv.order
    d <- ncol(x)
    n <- nrow(x)

    if (missing(binned)) binned <- default.bflag(d, n)
    psir <- dmvnorm.deriv.sum(x=x, Sigma=G, deriv.order=r, inc=inc, binned=binned, bin.par=bin.par, bgridsize=bgridsize, deriv.vec=deriv.vec, verbose=verbose, kfe=TRUE, add.index=FALSE)

    if (add.index)
    {
        ind.mat <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=r, only.index=TRUE)

        if (deriv.vec) return(list(psir=psir, deriv.ind=ind.mat))
        else return(list(psir=psir, deriv.ind=unique(ind.mat)))
    }
    else return(psir=psir)
}

kfe.1d <- function(x, g, deriv.order, inc=1, binned=TRUE, bin.par)
{
    r <- deriv.order
    n <- length(x)
    psir <- dnorm.deriv.sum(x=x, sigma=g, deriv.order=r, inc=1, binned=binned, bin.par=bin.par, kfe=TRUE)
    if (inc==0) psir <- (n^2*psir - n*dnorm.deriv(0, mu=0, sigma=g, deriv.order=r))/(n*(n-1))

    return(psir) 
}

kfe.scalar <- function(x, g, deriv.order, inc=1, binned=TRUE, bin.par, verbose=FALSE)
{
    r <- deriv.order
    d <- ncol(x)
    psir <- dmvnorm.deriv.scalar.sum(x=x, sigma=g, deriv.order=r, inc=inc, kfe=TRUE, binned=binned, bin.par=bin.par, verbose=verbose)
    
    return(psir)
}

###############################################################################
## Plug-in unconstrained bandwidth for KFE
##
## Returns
## Plug-in bandwidth
###############################################################################
hpi.kfe <- function(x, nstage=2, binned=FALSE, bgridsize, amise=FALSE, deriv.order=0)
{
    n <- length(x)
    d <- 1
    r <- deriv.order
    k <- 2 ## kernel order
    Kr0 <- dnorm.deriv(x=0, mu=0, sigma=1, deriv.order=r)    
    mu2K <- 1  

    if (nstage==2)
    {
        psi4.hat <- psins.1d(r=r+k+2, sigma=sd(x))
        gamse2 <- (factorial(r+2)*Kr0/(mu2K*psi4.hat*n))^(1/(r+k+3))
        psi2.hat <- kfe.1d(x=x, g=gamse2, deriv.order=r+k, inc=1, binned=binned)
    }
    else 
    {
        psi2.hat <- psins.1d(r=r+k, sigma=sd(x))
    }

    ## formula for bias annihliating bandwidths from Wand & Jones (1995, p.70)
    gamse <- (factorial(r)*Kr0/(-mu2K*psi2.hat*n))^(1/(r+k+1)) 

    return(gamse)
}

Hpi.kfe <- function(x, nstage=2, pilot, pre="sphere", Hstart, binned=FALSE, bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
{
    if (deriv.order!=0) stop("Currently only deriv.order=0 is implemented")

    n <- nrow(x)
    d <- ncol(x)
    r <- deriv.order

    if (missing(pilot)) pilot <- "dscalar"
    if(!is.matrix(x)) x <- as.matrix(x)
    pilot1 <- match.arg(pilot, c("dunconstr", "dscalar"))  
    pre1 <- match.arg(pre, c("scale", "sphere"))
    optim.fun1 <- match.arg(optim.fun, c("nlm", "optim"))

    if (pre1=="scale")
    {
        x.star <- pre.scale(x)
        S12 <- diag(sqrt(diag(var(x))))
        Sinv12 <- chol2inv(chol(S12))
    }
    else if (pre1=="sphere")
    {
        x.star <- pre.sphere(x)
        S12 <- matrix.sqrt(var(x))
        Sinv12 <- chol2inv(chol(S12))
    }

    D2K0 <- t(dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=2))
    K0 <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=0)

    if (nstage==2)
    {  
        ## stage 1
        if (pilot1=="dscalar")
        {
            psi4.ns <- psins(r=r+4, Sigma=var(x.star), deriv.vec=TRUE)
            ## gdscalar not used because it's an implicit computation without
            ## symmetriser matrices and is slower than direct computation with 
            ## symmetriser matrices. 
            A1 <- drop(t(D2K0) %*% D2K0)
            A2 <- drop(t(D2K0) %*% t(vec(diag(d)) %x% diag(d^2)) %*% psi4.ns) 
            A3 <- drop(t(psi4.ns) %*% (vec(diag(d)) %*% t(vec(diag(d))) %x% diag(d^2)) %*% psi4.ns)
            ## Special case from Chacon & Duong (2009): bias annihilation
            h2 <- (-A1/(2*A2*n))^(1/(d+4))
            H2 <- h2^2*diag(d)   
            psi2.hat <- kfe(x=x.star, G=H2, deriv.order=r+2, add.index=FALSE, binned=binned, bgridsize=bgridsize, verbose=verbose)
        }
        else if (pilot1=="dunconstr")
        {
            psi4.ns <- psins(r=r+4, Sigma=var(x), deriv.vec=TRUE)

            amse2.temp <- function(vechH)
            { 
                H <- invvech(vechH) %*% invvech(vechH)
                Hinv <- chol2inv(chol(H))
                Hinv12 <- matrix.sqrt(Hinv)
                amse2.temp <- 1/(det(H)^(1/2)*n)*((Hinv12 %x% Hinv12) %*% D2K0) + 1/2* t(vec(H) %x% diag(d^2)) %*% psi4.ns
                return(sum((amse2.temp)^2)) 
            }

            Hstart2 <- matrix.sqrt(Gns(r=r+2, n=n, Sigma=var(x)))

            if (optim.fun1=="nlm")
            {
                result <- nlm(p=vech(Hstart2), f=amse2.temp, print.level=2*as.numeric(verbose))    
                H2 <- invvech(result$estimate) %*% invvech(result$estimate)
            }
            else if (optim.fun1=="optim")
            {
                result <- optim(vech(Hstart2), amse2.temp, method="BFGS", control=list(trace=as.numeric(verbose),REPORT=1))
                H2 <- invvech(result$par) %*% invvech(result$par)
            }

            psi2.hat <- kfe(x=x, G=H2, deriv.order=r+2, add.index=FALSE, binned=binned, bgridsize=bgridsize, verbose=verbose)
        }
    }
    else
    {
        if (pilot1=="dscalar") psi2.hat <- psins(r=r+2, Sigma=var(x.star), deriv.vec=TRUE)
        else if (pilot1=="dunconstr") psi2.hat <- psins(r=r+2, Sigma=var(x), deriv.vec=TRUE)    
    }

    if (pilot1=="dscalar") { if (missing(Hstart)) Hstart <- Gns(r=r, n=n, Sigma=var(x.star)) }
    else if (pilot1=="dunconstr") { if (missing(Hstart)) Hstart <- Gns(r=r, n=n, Sigma=var(x)) }
  
  ## stage 2
  amse.temp <- function(vechH)
  { 
        H <- invvech(vechH) %*% invvech(vechH)
        amse.val <- 1/(det(H)^(1/2)*n)*K0 + 1/2* t(vec(H)) %*% psi2.hat
        return(sum((amse.val^2))) 
  }
  
    Hstart <- matrix.sqrt(Hstart)
    if (optim.fun1=="nlm")
    {
        result <- nlm(p=vech(Hstart), f=amse.temp, print.level=2*as.numeric(verbose)) 
        H <- invvech(result$estimate) %*% invvech(result$estimate)
        amise.star <- result$minimum
    }
    else if (optim.fun1=="optim")
    {
        result <- optim(vech(Hstart), amse.temp, method="BFGS", control=list(trace=as.numeric(verbose),REPORT=1))
        H <- invvech(result$par) %*% invvech(result$par)
        amise.star <- result$value
    }

    ## back-transform
    if (pilot1=="dscalar") H <- S12 %*% H %*% S12 

    if (!amise) return(H)
    else return(list(H=H, PI=amise.star))
}

###############################################################################
## Plug-in diagonal bandwidth for KFE
##
## Returns
## Plug-in bandwidth
###############################################################################
Hpi.diag.kfe <- function(x, nstage=2, pilot, pre="scale", Hstart, binned=FALSE, bgridsize, amise=FALSE,  deriv.order=0, verbose=FALSE, optim.fun="optim")
{
    if (deriv.order!=0) stop("Currently only dervi.order=0 is implemented")

    n <- nrow(x)
    d <- ncol(x)
    r <- deriv.order
    D2K0 <- t(dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=2))
    K0 <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=0)

    if(!is.matrix(x)) x <- as.matrix(x)
    if (missing(pilot)) pilot <- "dscalar"
    pilot1 <- match.arg(pilot, c("dunconstr", "dscalar"))  
    pre1 <- match.arg(pre, c("scale", "sphere"))
    optim.fun1 <- match.arg(optim.fun, c("nlm", "optim"))

    if (pre1=="sphere") stop("Using pre-sphering won't give diagonal bandwidth matrix")
    if (pilot1=="dunconstr") stop("Unconstrained pilot selectors are not suitable for Hpi.diag.kfe")

    if (pre1=="scale")
    {
        x.star <- pre.scale(x)
        S12 <- diag(sqrt(diag(var(x))))
        Sinv12 <- chol2inv(chol(S12))
    }

    if (nstage==2)
    {  
        ## stage 1
        psi4.ns <- psins(r=r+4, Sigma=var(x.star), deriv.vec=TRUE)

        if (pilot1=="dscalar")
        {
            ## h2 is on pre-transformed data scale
            A1 <- drop(t(D2K0) %*% D2K0)
            A2 <- drop(t(D2K0) %*% t(vec(diag(d)) %x% diag(d^2)) %*% psi4.ns) 
            A3 <- drop(t(psi4.ns) %*% (vec(diag(d)) %*% t(vec(diag(d))) %x% diag(d^2)) %*% psi4.ns)
            h2 <- ((4*d+8)*A1/(-d*A2 + sqrt(d^2*A2^2 + (8*d+16)*A1*A3)))^(1/(d+4))*n^(-1/(d+4))
            H2 <- h2^2*diag(d)
        }  
        psi2.hat <- kfe(x=x.star, G=H2, deriv.order=r+2, add.index=FALSE, binned=binned, bgridsize=bgridsize, verbose=verbose) 
    }
    else
        psi2.hat <- psins(r=r+2, Sigma=var(x.star), deriv.vec=TRUE)
 
    ## stage 2
    amse.temp <- function(diagH)
    { 
        H <- diag(diagH) %*% diag(diagH)
        amse.val <- 1/(det(H)^(1/2)*n)*K0 + 1/2* t(vec(H)) %*% psi2.hat
        return(sum((amse.val^2))) 
    }

    if (missing(Hstart)) Hstart <- Hns(x=x.star, deriv.order=r)
    Hstart <- matrix.sqrt(Hstart)
  
    if (optim.fun1=="nlm")
    {
        result <- nlm(p=diag(Hstart), f=amse.temp, print.level=2*as.numeric(verbose))    
        H <- diag(result$estimate) %*% diag(result$estimate)
        amise.star <- result$minimum
    }
    else if (optim.fun1=="optim")
    {
        result <- optim(diag(Hstart), amse.temp, method="BFGS", control=list(trace=as.numeric(verbose),REPORT=1))
        H <- diag(result$par) %*% diag(result$par)
    amise.star <- result$value
    }

    ## back-transform
    if (pilot1=="dscalar") H <- S12 %*% H %*% S12 
    if (!amise) return(H)
    else return(list(H=H, PI=amise.star))
}

###############################################################################
## Feature significance for ultivariate kernel density stimate 
###############################################################################
kfs <- function(x, H, h, deriv.order=2, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, positive=FALSE, adj.positive, w, verbose=FALSE, signif.level=0.05)
{
    r <- deriv.order

    ## default values 
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    bgridsize <- ksd$bgridsize
    gridsize <- ksd$gridsize
    
    ## clip data to xmin, xmax grid limits for binned estimation
    grid.clip <- binned    
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        if (positive & missing(xmin)) { xmin <- rep(0,d) } 
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }

    if (d==1)
    {
        if (missing(h)) h <- hpi(x=x, nstage=2, binned=default.bflag(d=d, n=n), deriv.order=r)

        ## KDDE for r=2
        fhatr <- kdde(x=x, h=h, deriv.order=r, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, deriv.vec=FALSE, verbose=verbose)

        ## KDE
        fhat <- kde(x=x, h=h, gridsize=gridsize, gridtype=gridtype, xmin=min(fhatr$eval.points), xmax=max(fhatr$eval.points), binned=binned, bgridsize=bgridsize, positive=positive, adj.positive=adj.positive, w=w)

        fhat.est <- as.vector(fhat$estimate)
        fhatr.est <- as.vector(fhatr$estimate)

        RDrK <- (-1)^r*psins.1d(r=2*r, sigma=1)
        fhatr.Sigma <- n^(-1)* h^(-2*r-1)*RDrK
        fhatr.Sigma12 <- sqrt(fhatr.Sigma)
        fhatr.est <- fhatr.est/fhatr.Sigma12

        local.mode <- fhatr.est <= 0
        fhatr.wald <- fhatr.est^2/abs(fhat.est)
        fhatr.wald[is.infinite(fhatr.wald)] <- max(fhatr.wald[!is.infinite(fhatr.wald)])
        gs <- length(fhat$estimate)        
    }
    else if (d>1)
    {
        if (missing(H)) H <- Hpi(x=x, nstage=2-(d>2), binned=default.bflag(d=d, n=n), deriv.order=r, verbose=verbose)
    
        ## KDDE for r=2
        fhatr <- kdde(x=x, H=H, deriv.order=r, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, deriv.vec=FALSE, verbose=verbose)

        ## KDE
        fhat <- kde(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, verbose=verbose)
                
        fhat.est <- as.vector(fhat$estimate)
        fhatr.est <- sapply(fhatr$estimate, as.vector)
    
        ## convert from vec to vech because vec'ed derivative
        ## contains repeated columns so its variance isn't invertible 
        Hinv <- chol2inv(chol(H))
        Hinv12 <- matrix.sqrt(Hinv)
        RDrK <- (-1)^r*invvec(psins(r=2*r, Sigma=diag(d)))
        dupld <- dupl(d)$d
        dupld.MPinv <- chol2inv(chol(t(dupld)%*% dupld)) %*% t(dupld)
        fhatr.Sigma.const <- n^(-1)*det(H)^(-1/2)* Kpow(Hinv12,r) %*% RDrK %*% Kpow(Hinv12,2)
        fhatr.Sigma.const <- dupld.MPinv %*% fhatr.Sigma.const %*% t(dupld.MPinv)      
        fhatr.Sigma.const12inv <- chol2inv(chol(matrix.sqrt(fhatr.Sigma.const)))
        fhatr.est <- fhatr.est %*% fhatr.Sigma.const12inv  

        ## all eigenvalues < 0 => local mode 
        fhatr.eigen <- lapply(lapply(seq(1,nrow(fhatr.est)), function(i) { invvech(fhatr.est[i,]) }), eigen, only.values=TRUE)
        fhatr.eigen <- t(sapply(fhatr.eigen, getElement, "values"))
        local.mode <- apply(fhatr.eigen <= 0, 1, all)
        fhatr.wald <- apply(fhatr.est^2, 1, sum)/fhat.est
        
        gs <- dim(fhat$estimate)
    }

    ## Hochberg adjustment for sequential tests
    pval.wald <- 1 - pchisq(fhatr.wald, d*(d+1)/2)
    pval.wald[fhat.est<=contourLevels(fhat, cont=99) | !local.mode] <- NA    
    pval.wald.ord.index <- order(pval.wald)
    pval.wald.ord <- pval.wald[pval.wald.ord.index]
    num.test <- sum(!is.na(pval.wald.ord))
    
    if (num.test>=1) num.test.seq <- c(1:num.test, rep(NA, prod(gs) - num.test))
    else num.test.seq <- rep(NA, prod(gs))

    reject.nonzero <- (pval.wald.ord <= signif.level/(num.test + 1 - num.test.seq))
    reject.nonzero.ind <- which(reject.nonzero)
    
    ## reject null hypotheses indicated in reject.nonzero.ind
    if (d==1)
    {
        signif.wald <- array(0L, dim=gs)
        signif.wald[pval.wald.ord.index[reject.nonzero.ind]] <- 1L
    }
    else
    {
        signif.wald <- array(0L, dim=gs)
        signif.wald.index <- expand.grid(lapply(gs, seq_len))
        signif.wald[as.matrix(signif.wald.index[pval.wald.ord.index[reject.nonzero.ind],])] <- 1L
    }
    
    ## ESS = effective sample size
    ## ess <- n*fhat$estimate*dmvnorm.mixt(x=rep(0,d), mu=rep(0,d), Sigma=H, props=1)
    ## signif.ess <- ess >= 5
    fhatr$estimate <- signif.wald 
    fhatr$type <- "kfs"
    class(fhatr) <- "kfs"
    
    return(fhatr)
}

#############################################################################
## S3 methodfor KFS objects
#############################################################################
## plot method
plot.kfs <- function(x, display="filled.contour", col=7, colors, abs.cont, alpha=1, alphavec=0.4, add=FALSE, ...)
{
    fhatr <- x
    fhatr$deriv.order <- NULL
    class(fhatr) <- "kde"

    if (is.vector(fhatr$H)) d <- 1 else d <- ncol(fhatr$H)
    if (d==1)
    {
        fhat <- kde(x=fhatr$x, xmin=min(fhatr$eval.points), xmax=max(fhatr$eval.points), gridsize=length(fhatr$eval.points))
        plot(fhat, col="grey", add=add, ...)
        gridsize <- length(fhatr$estimate)

        estimate.rle <- rle(as.vector(fhatr$estimate))
        estimate.rle.cumsum <- rbind(cumsum(estimate.rle$lengths), estimate.rle$lengths, estimate.rle$values)
        col <- transparency.col(col, alpha=alpha)
        for (i in which(estimate.rle$values==1))
        {
            seg.ind <- 1:estimate.rle.cumsum[2,i]   
            if (i>1)
                seg.ind <- seg.ind + estimate.rle.cumsum[1,i-1] 
            lines(fhat$eval.points[seg.ind], fhat$estimate[seg.ind], col=col, ...)
        }
    }
    else if (d==2)
    {
        if (missing(abs.cont)) abs.cont <- 0.5
        disp1 <- match.arg(display, c("slice", "persp", "image", "filled.contour", "filled.contour2"))
        if (disp1=="filled.contour2") disp1 <- "filled.contour"
        col <- c("transparent",col)
        
        plot(fhatr, abs.cont=abs.cont, drawlabels=FALSE, col=col, add=add, display=display, alpha=alpha, ...)
    }
    else if (d==3)
    {
        if (missing(abs.cont)) abs.cont <- 0.25
        e1 <- try(match.arg(display, c("plot3D", "rgl")), silent=TRUE)
        if (inherits(e1, "try-error")) display <- "plot3D"
        if (!missing(colors)) col <- colors     
        plot(fhatr, abs.cont=abs.cont, col=col, colors=colors, alphavec=alphavec, add=add, display=display, ...)
    }
    invisible()
}

## predict method
predict.kfs <- function(object, ..., x)
{
    fhat <- predict.kde(object=object, ..., x=x)
    fhat <- as.integer(fhat>=0.5)
   
    return(fhat)
}

######################################################################
## Boundary KDE
######################################################################
kde.boundary <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned=FALSE, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, boundary.supp, boundary.kernel="beta", verbose=FALSE)
{
    bk <- match.arg(boundary.kernel, c("beta", "linear")) 
    if (bk=="beta")
    {
        if (missing(boundary.supp)) boundary.supp <- 10
        fhat <- kde.beta.boundary(x=x, H=H, h=h, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont, boundary.supp=boundary.supp, verbose=verbose) 
      
    }
    else if (bk=="linear")
    {
        if (missing(boundary.supp)) boundary.supp <- 2
        fhat <- kde.linear.boundary(x=x, H=H, h=h, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, binned=binned, bgridsize=bgridsize, w=w, compute.cont=compute.cont, approx.cont=approx.cont, boundary.supp=boundary.supp, verbose=verbose)
    }

    return(fhat)
}

######################################################################
## Linear boundary KDE
######################################################################
kde.linear.boundary <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned=FALSE, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, boundary.supp=2, verbose=FALSE)
{
    ## default values 
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    if (missing(binned)) binned <- ksd$binned
    if (missing(gridsize)) gridsize <- ksd$gridsize
    if (missing(bgridsize)) bgridsize <- gridsize
  
    ## clip data to xmin,xmax grid 
    grid.clip <- binned
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }
  
    ## default bandwidths
    if (d==1 & missing(h)) 
        h <- hpi(x=x, nstage=2, binned=default.bflag(d=d, n=n))
    if (d>1 & missing(H))
        H <- Hpi(x=x, nstage=2, binned=default.bflag(d=d, n=n))
    
    ## compute exact (non-binned) estimator
    ## 1-dimensional    
    if (d==1)
    {
        stop("Not yet implemented.")  
    }
    ## multi-dimensional
    else
    {  
        if (is.data.frame(x)) x <- as.matrix(x)
        
        if (missing(eval.points))
        {
            if (d==2)
                fhat <- kde.LB.grid.2d(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, boundary.supp=boundary.supp, binned=binned, verbose=verbose)
            else 
                stop("Not yet implemented.")
        }
        else
            stop("Not yet implemented.")
    }
    
    fhat$binned <- binned
    fhat$names <- parse.name(x)  ## add variable names
    if (inherits(fhat$eval.points, "list")) names(fhat$eval.points) <- ksd$names
    fhat$w <- w
    class(fhat) <- "kde"
    
    ## compute prob contour levels
    if (compute.cont & missing(eval.points))
        fhat$cont <- contourLevels(fhat, cont=1:99, approx=approx.cont)
    
    return(fhat)
}

######################################################################
## Bivariate linear boundary KDE
######################################################################
kde.LB.grid.2d <- function(x, H, gridsize, bgridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, boundary.supp=10, binned=FALSE, verbose=FALSE)
{
   n <- nrow(x)
   d <- ncol(x)
   
   if (missing(xmin)) xmin <- apply(x, 2, min) 
   if (missing(xmax)) xmax <- apply(x, 2, max) 
   if (missing(gridtype)) gridtype <- rep("linear", d)
   
   h <- sqrt(diag(H))

   ## initialise grid 
   if (is.null(gridx))
       gridx <- make.grid.ks(x, matrix.sqrt(H), tol=supp, gridsize=gridsize, xmin=xmin, xmax=xmax, gridtype=gridtype) 
   suppx <- make.supp(x, matrix.sqrt(H), tol=supp)
   
   if (is.null(grid.pts)) grid.pts <- find.gridpts(gridx, suppx)
   fhat.grid <- matrix(0, nrow=length(gridx[[1]]), ncol=length(gridx[[2]]))
   
   ## indicator for closeness to boundary 
   bound.ind <- boundary.ind(x=x, h=h, boundary.supp=boundary.supp)
   n1 <- sum(!bound.ind)

   if (verbose) pb <- txtProgressBar()
   if (binned)
   {
      ## interior points - use normal kernel
      fhat.grid <- n1*kde(x=x[!bound.ind,], H=H, xmin=xmin, xmax=xmax, binned=binned, bgridsize=bgridsize)$estimate
       
        for (i in 1:n)
        {
            if (verbose) setTxtProgressBar(pb, i/n)
            if (bound.ind[i])
            {
                ## compute evaluation points 
                eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
                eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
                eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
                eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
                eval.x.len <- length(eval.x)
                 
                ## use linear boundary kernel for boundary points 
                fhat <- dmvnorm.LB(x=expand.grid(eval.x, eval.y), mu=x[i,], Sigma=H)
             
                ## place vector of density estimate values `fhat' onto grid 'fhat.grid' 
                for (j in 1:length(eval.y))
                    fhat.grid[eval.x.ind, eval.y.ind[j]] <- 
                        fhat.grid[eval.x.ind, eval.y.ind[j]] + 
                        w[i]*fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
            }
        }
    }
    else
    {  
        for (i in 1:n)
        {
            if (verbose) setTxtProgressBar(pb, i/n)

            ## compute evaluation points 
            eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
            eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
            eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
            eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
            eval.x.len <- length(eval.x)

            ## interior points - use normal kernel
            if (!bound.ind[i])
            {
                eval.pts <- expand.grid(list(eval.x, eval.y))
                fhat <- dmvnorm.mixt(x=eval.pts, mus=x[i,], Sigmas=H, props=1)
            }
            else
            {
                ## use linear boundary kernel for boundary points 
                 fhat <- dmvnorm.LB(x=expand.grid(eval.x, eval.y), mu=x[i,], Sigma=H)               
            }
             
            ## place vector of density estimate values fhat onto grid fhat.grid
            for (j in 1:length(eval.y))
                fhat.grid[eval.x.ind, eval.y.ind[j]] <- 
                    fhat.grid[eval.x.ind, eval.y.ind[j]] + 
                    w[i]*fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
        }
    }
    fhat.grid <- fhat.grid/n
    gridx1 <- list(gridx[[1]], gridx[[2]])
    fhat.grid <- fhat.grid/sum(fhat.grid*prod(sapply(gridx1,diff)[1,]))

    if (verbose) close(pb)
    fhat.list <- list(x=x, eval.points=gridx1, estimate=fhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE, boundary=bound.ind)
   
    return(fhat.list)
}

## bivariate linear boundary normal kernel 
dmvnorm.LB.kernel.2d <- function(x, H, xmin=c(0,0), xmax=c(1,1), ...)
{
    x1 <- seq(xmin[1], xmax[1], length=151)
    x2 <- seq(xmin[2], xmax[2], length=151)
    eval.points <- list(x1, x2)

    fhat <- list()
    fhat$eval.points <- eval.points
    fhat$estimate <- array(dmvnorm.LB(x=expand.grid(eval.points), mu=x, Sigma=H), dim=c(length(x1), length(x2)))

    x <- rmvnorm.mixt(n=1000, mus=x, Sigmas=H, props=1) 
    fhat$x <- x
    fhat$H <- H
    fhat$gridtype <- "linear"
    fhat$gridded <- TRUE
    fhat$binned <- FALSE
    fhat$names <- parse.name(x)
    fhat$w <- rep(1, nrow(x))
    class(fhat) <- "kde"
                
    return(fhat)
}

dmvnorm.LB <- function(x, mu, Sigma, a0, a1)
{
    if (!is.matrix(x)) x <- as.matrix(x)
    if (missing(a0) | missing(a1))
    {
        ev <- -sweep(x, 2, mu, FUN="-") %*% diag(sqrt(1/diag(Sigma)))
        ev.list <- list(unique(ev[,1]), unique(ev[,2]))

        delta <- prod(unlist(lapply(lapply(ev.list, diff), getElement, 1)))
        d <- ncol(Sigma)
        evalK <- dmvnorm.mixt(x=ev, mus=rep(0,d), Sigmas=diag(d), props=1)
        m0 <- sum(evalK*delta)
        m1 <- apply(evalK*delta*ev, 2, sum)
        m2 <- apply(evalK*delta*rowKpow(ev, ev), 2, sum)
        M2 <- invvec(m2)
        M2inv <- chol2inv(chol(M2))
        a0 <- 1/drop(m0 - t(m1) %*% M2inv%*%m1)
        a1 <- -a0*M2inv%*%m1
    }    
    evalK.LB <- (a0 - drop(ev %*% a1))*dmvnorm.mixt(x=x, mus=mu, Sigmas=Sigma, props=1)

    return(evalK.LB)
}

######################################################################
## Boundary kernel estimator using beta bounday kernels (2nd form)
######################################################################
kde.beta.boundary <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned=FALSE, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, boundary.supp=1, verbose=FALSE)
{
    ## default values
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    if (missing(binned)) binned <- ksd$binned
    if (missing(gridsize)) gridsize <- ksd$gridsize
    if (missing(bgridsize)) bgridsize <- gridsize
    ## if (missing(gridsize)) gridsize <- default.gridsize(d)

    ## clip data to xmin,xmax grid 
    grid.clip <- binned
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }

    ## default bandwidths
    if (d==1 & missing(h)) 
        h <- hpi(x=x, nstage=2, binned=default.bflag(d=d, n=n))
    if (d>1 & missing(H))
        H <- Hpi(x=x, nstage=2, binned=default.bflag(d=d, n=n))
  
    ## compute exact (non-binned) estimator
    ## 1-dimensional    
    if (d==1)
    {
        if (missing(eval.points))
        {
            fhat <- kde.boundary.grid.1d(x=x, h=h, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, boundary.supp=boundary.supp, binned=binned)
        }
        else
            stop("Not yet implemented.")  # fhat <- kde.points.1d(x=x, h=h, eval.points=eval.points, positive=positive, adj.positive=adj.positive, w=w)
    }
    ## multi-dimensional
    else
    {  
        if (is.data.frame(x)) x <- as.matrix(x)

        if (missing(eval.points))
        {
            if (d==2)
                fhat <- kde.boundary.grid.2d(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, boundary.supp=boundary.supp, binned=binned, verbose=verbose)
            else if (d == 3)
                fhat <- kde.boundary.grid.3d(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, boundary.supp=boundary.supp, binned=binned, verbose=verbose) 
            else 
                stop("Need to specify eval.points for more than 3 dimensions")
        }
        else
            stop("Not yet implemented.")     
    }

    fhat$binned <- binned
    fhat$names <- parse.name(x)  ## add variable names
    if (inherits(fhat$eval.points, "list")) names(fhat$eval.points) <- ksd$names
    fhat$w <- w
    class(fhat) <- "kde"
    
    ## compute prob contour levels
    if (compute.cont & missing(eval.points))
    fhat$cont <- contourLevels(fhat, cont=1:99, approx=approx.cont)

    return(fhat)
 }

kde.boundary.grid.1d <- function(x, h,gridsize, supp=3.7, xmin, xmax, gridtype, w, boundary.supp=0.5, binned=FALSE)
{
    if (missing(xmin)) xmin <- min(x)
    if (missing(xmax)) xmax <- max(x)
    if (missing(gridtype)) gridtype <- "linear"

    ## transform x into [0,1]
    x.star <- (x-xmin)/(xmax-xmin) 
    h.star <- h/(xmax-xmin)
    n <- length(x)

    gridtype1 <- match.arg(gridtype, c("linear", "sqrt")) 
    if (gridtype1=="linear")
        eval.x <- seq(0, 1, length=gridsize)
    else if (gridtype1=="sqrt")
    {
        eval.x.temp <- seq(0, 1, length=gridsize)
        eval.x <- sign(eval.x.temp) * eval.x.temp^2
    }
    gridtype.vec <- gridtype1

    ## indicator for closeness to boundary of [0,1]
    bound.ind <- boundary.ind(x=x.star, h=h.star, boundary.supp=boundary.supp) 

    n1 <- sum(!bound.ind)
    ## interior points - use normal kernel
    ## binned estimation only in the interior
    fhat.grid <- rep(0,length=gridsize)
    if (n1>0)
    {
        if (binned)
        {
            fhat.grid <- n1*kde(x=x.star[!bound.ind], h=h.star, xmin=0, xmax=1, binned=TRUE, bgridsize=gridsize)$estimate
        }
        else
        {
            fhat.grid <- n1*dnorm.mixt(x=eval.x, mus=x.star[!bound.ind], sigmas=rep(h.star, n1), props=w[!bound.ind]/n1)
        }
    }

    ## boundary points - use adjusted beta kernel
    hb.star <- 2*h.star
    for (i in 1:(n-n1))
        fhat.grid <- fhat.grid + dbeta.kernel2(x=x.star[bound.ind][i], eval.x=eval.x, h=hb.star)*w[bound.ind][i]
    fhat.grid <- fhat.grid/n

    ## backtransform
    eval.points <- (xmax-xmin)*eval.x + xmin 
    fhat.grid <- fhat.grid/(xmax-xmin)

    fhat <- list(x=x, eval.points=eval.points, estimate=fhat.grid, h=h, H=h^2, gridtype=gridtype.vec, gridded=TRUE)
    class(fhat) <- "kde"

    return(fhat)
}

kde.boundary.grid.2d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, boundary.supp=1, binned=FALSE, verbose=FALSE)
{
    n <- nrow(x)
    d <- ncol(x)

    if (missing(xmin)) xmin <- apply(x, 2, min) ##- h*supp
    if (missing(xmax)) xmax <- apply(x, 2, max) ## + h*supp
    if (missing(gridtype)) gridtype <- rep("linear", d)

    ## transform x into [0,1]^d
    x.star <- x
    for (j in 1:d) x.star[,j] <- (x[,j]-xmin[j])/(xmax[j]-xmin[j])
    H.star <- diag(1/(xmax-xmin)) %*% H %*% diag(1/(xmax-xmin))
    h.star <- sqrt(diag(H.star))

    ## initialise grid 
    if (is.null(gridx))
        gridx <- make.grid.ks(x.star, matrix.sqrt(H.star), tol=supp, gridsize=gridsize, xmin=rep(0,d), xmax=rep(1,d), gridtype=gridtype) 
    suppx <- make.supp(x.star, matrix.sqrt(H.star), tol=supp)

    if (is.null(grid.pts)) grid.pts <- find.gridpts(gridx, suppx)    
    fhat.grid <- matrix(0, nrow=length(gridx[[1]]), ncol=length(gridx[[2]]))

    ## indicator for closeness to boundary of [0,1]^d
    bound.ind <- boundary.ind(x=x.star, h=h.star, boundary.supp=boundary.supp)
    n1 <- sum(!bound.ind)

    if (verbose) pb <- txtProgressBar()
    if (binned)
    {
        ## interior points - use normal kernel
        fhat.grid <- n1*kde(x=x.star[!bound.ind,], H=H.star, xmin=rep(0,d), xmax=rep(1,d), binned=TRUE, bgridsize=gridsize)$estimate

        for (i in 1:n)
        {
            if (verbose) setTxtProgressBar(pb, i/n)
            if (bound.ind[i])
            {
                ## compute evaluation points 
                eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
                eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
                eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
                eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
                eval.x.len <- length(eval.x)

                ## convert bandwidth from normal kernel to beta kernel scale
                ## for boundary points 
                hb.star <- 2*h.star
                fhat <- dmvbeta.prod.kernel2(x=x.star[i,], eval.x=list(eval.x, eval.y), hs=hb.star)

                ## place vector of density estimate values `fhat' onto grid 'fhat.grid' 
                for (j in 1:length(eval.y))
                    fhat.grid[eval.x.ind, eval.y.ind[j]] <- 
                        fhat.grid[eval.x.ind, eval.y.ind[j]] + 
                        w[i]*fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
            }
        }
    }
    else
    {  
        for (i in 1:n)
        {
            if (verbose) setTxtProgressBar(pb, i/n)

            ## compute evaluation points 
            eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
            eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
            eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
            eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
            eval.x.len <- length(eval.x)

            ## interior points - use normal kernel
            if (!bound.ind[i])
            {
            eval.pts <- expand.grid(eval.x, eval.y)
            fhat <- dmvnorm(eval.pts, x.star[i,], H.star)
            }
            else
            {
            ## convert bandwidth from normal kernel to beta kernel scale
            ## for boundary points 
            hb.star <- 2*h.star
            fhat <- dmvbeta.prod.kernel2(x=x.star[i,], eval.x=list(eval.x, eval.y), hs=hb.star)
        }

        ## place vector of density estimate values `fhat' onto grid 'fhat.grid' 
        for (j in 1:length(eval.y))
            fhat.grid[eval.x.ind, eval.y.ind[j]] <- 
                fhat.grid[eval.x.ind, eval.y.ind[j]] + 
                w[i]*fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
        }
    }
    fhat.grid <- fhat.grid/n
    if (verbose) close(pb)
  
    ## back-transform
    gridx1 <- list((xmax[1]-xmin[1])*gridx[[1]] + xmin[1], (xmax[2]-xmin[2])*gridx[[2]] + xmin[2]) 
    fhat.grid <- fhat.grid/prod(xmax-xmin)

    fhat.list <- list(x=x, eval.points=gridx1, estimate=fhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE, boundary=bound.ind)

    return(fhat.list)
}

kde.boundary.grid.3d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, boundary.supp=0.5, verbose=FALSE, binned=FALSE)
{
    n <- nrow(x)
    d <- ncol(x)

    if (missing(xmin)) xmin <- apply(x, 2, min) 
    if (missing(xmax)) xmax <- apply(x, 2, max) 
    if (missing(gridtype)) gridtype <- rep("linear", d)

    ## transform x into [0,1]^d
    x.star <- x
    for (j in 1:d) x.star[,j] <- (x[,j]-xmin[j])/(xmax[j]-xmin[j])
    H.star <- diag(1/(xmax-xmin)) %*% H %*% diag(1/(xmax-xmin))
    h.star <- sqrt(diag(H.star)) 

    ## initialise grid 
    if (is.null(gridx))
        gridx <- make.grid.ks(x.star, matrix.sqrt(H.star), tol=supp, gridsize=gridsize, xmin=rep(0,d), xmax=rep(1,d), gridtype=gridtype) 
    suppx <- make.supp(x.star, matrix.sqrt(H.star), tol=supp)
    if (is.null(grid.pts)) grid.pts <- find.gridpts(gridx, suppx)
    fhat.grid <- array(0, dim=c(length(gridx[[1]]), length(gridx[[2]]), length(gridx[[3]])))

    ## indicator for closeness to boundary of [0,1]^d
    bound.ind <- boundary.ind(x=x.star, h=h.star, boundary.supp=boundary.supp)
    n1 <- sum(!bound.ind)

    if (verbose) pb <- txtProgressBar() 
    if (binned)
    {
        ## interior points - use normal kernel
        fhat.grid <- n1*kde(x=x.star[!bound.ind,], H=H.star, xmin=rep(0,d), xmax=rep(1,d), binned=TRUE, bgridsize=gridsize)$estimate

        for (i in 1:n)
        {
            if (verbose) setTxtProgressBar(pb, i/n)

            if (bound.ind[i])
            {
                ## compute evaluation points
                eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
                eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
                eval.z <- gridx[[3]][grid.pts$xmin[i,3]:grid.pts$xmax[i,3]]
                eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
                eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
                eval.z.ind <- c(grid.pts$xmin[i,3]:grid.pts$xmax[i,3])
                eval.x.len <- length(eval.x)
                eval.pts <- expand.grid(eval.x, eval.y)

                ## convert bandwidth from normal kernel to beta kernel scale
                ## for boundary points 
                hb.star <- 2*h.star
                fhat.xy <- dmvbeta.prod.kernel2(x=x.star[i,], eval.x=list(eval.x, eval.y), hs=hb.star[1:2])

                ## place vector of density estimate values `fhat' onto grid 'fhat.grid' 
                for (k in 1:length(eval.z))
                {
                    fhat <- w[i]*cbind(fhat.xy, dbeta.kernel2(x=x.star[i,3], eval.x=eval.z[k], h=hb.star[3]))
                    for (j in 1:length(eval.y))
                    fhat.grid[eval.x.ind,eval.y.ind[j], eval.z.ind[k]] <- 
                      fhat.grid[eval.x.ind, eval.y.ind[j], eval.z.ind[k]] + 
                    fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
                }
            }
        }
    }
    else
    {  
        for (i in 1:n)
        {
            if (verbose) setTxtProgressBar(pb, i/n)

            ## compute evaluation points
            eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
            eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
            eval.z <- gridx[[3]][grid.pts$xmin[i,3]:grid.pts$xmax[i,3]]
            eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
            eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
            eval.z.ind <- c(grid.pts$xmin[i,3]:grid.pts$xmax[i,3])
            eval.x.len <- length(eval.x)
            eval.pts <- expand.grid(eval.x, eval.y)

            ## interior points - use normal kernel
            if (!bound.ind[i])
            {
                ## place vector of density estimate values `fhat' onto grid 'fhat.grid' 
        
                for (k in 1:length(eval.z))
                {
                    fhat <- w[i]*dmvnorm(cbind(eval.pts, eval.z[k]), x[i,], H)
                    for (j in 1:length(eval.y))
                        fhat.grid[eval.x.ind,eval.y.ind[j], eval.z.ind[k]] <- 
                            fhat.grid[eval.x.ind, eval.y.ind[j], eval.z.ind[k]] + 
                            fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
                }
            }
            else
            {
                ## convert bandwidth from normal kernel to beta kernel scale
                hb.star <- 2*h.star
                fhat.xy <- dmvbeta.prod.kernel2(x=x.star[i,], eval.x=list(eval.x, eval.y), hs=hb.star[1:2])
                    
                for (k in 1:length(eval.z))
                {
                    fhat <- w[i]*cbind(fhat.xy, dbeta.kernel2(x=x.star[i,3], eval.x=eval.z[k], h=hb.star[3]))
                    for (j in 1:length(eval.y))
                        fhat.grid[eval.x.ind,eval.y.ind[j], eval.z.ind[k]] <- 
                            fhat.grid[eval.x.ind, eval.y.ind[j], eval.z.ind[k]] + 
                            fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
                }
            }
        }
    }
    fhat.grid <- fhat.grid/n
    if (verbose) close(pb)
  
    ## back-transform
    gridx1 <- list((xmax[1]-xmin[1])*gridx[[1]] + xmin[1], (xmax[2]-xmin[2])*gridx[[2]] + xmin[2], (xmax[3]-xmin[3])*gridx[[3]] + xmin[3])
    fhat.grid <- fhat.grid/prod(xmax-xmin)

    fhat.list <- list(x=x, eval.points=gridx1, estimate=fhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE, boundary=bound.ind)
    
    return(fhat.list)
}

## indicator function for boundary region of [0,1] i.e. [0,h] + [1-h, h]
boundary.ind <- function(x, h, xmin, xmax, boundary.supp=1)
{
    if (is.vector(x)) { x <- matrix(x, ncol=1) }
    n <- nrow(x)
    d <- ncol(x)

    ## indicator for closeness to boundary of [0,1]^d
    bound.ind <- matrix(NA, nrow=n, ncol=d)
    for (j in 1:d) bound.ind[,j] <- (abs(x[,j]) <= boundary.supp*h[j]) | (abs(1-x[,j]) <= boundary.supp*h[j])
    bound.ind <- apply(bound.ind, 1, any)

    return(bound.ind)
}

######################################################################
## Bivariate beta boundary KDE
######################################################################
## modified boundary beta kernel - first form (Chen, 1999)
dbeta.kernel <- function(x, eval.x, h)
{
    return (dbeta(x=eval.x, shape1=x/h^2+1, shape2=(1-x)/h^2+1))
}

## modified boundary beta kernel - second form (Chen, 1999)
dbeta.kernel2 <- function(x, eval.x, h)
{
    rhox <- function(y, hy) { if (y==0) return (1) else return(2*hy^4 + 5/2 - sqrt(4*hy^8 + 6*hy^4 + 9/4 - y^2 -y/hy^2)) }
    ind <- cut(x, c(0, 2*h^2, 1-2*h^2, 1), labels=FALSE, include.lowest=TRUE)
    dbk <- rep(0, length(eval.x))

    if (ind==1) { shape1 <- rhox(x,hy=h); shape2 <- (1-x)/h^2 }
    else if (ind==2) { shape1 <- x/h^2;   shape2 <- (1-x)/h^2 }
    else if (ind==3) { shape1 <- x/h^2;   shape2 <- rhox(1-x, hy=h) }

return(dbeta(eval.x, shape1=shape1, shape2=shape2))
}

## modified multivariate boundary beta product kernel
dmvbeta.prod.kernel2 <- function(x, eval.x, hs)
{
    d <- length(hs)
    db <- vector("list", d)
    for (i in 1:d)
    {   
        db[[i]] <- 0
        db[[i]] <- dbeta.kernel2(x=x[i], eval.x=eval.x[[i]], h=hs[i])
    }

    db <- expand.grid(db)
    db <- apply(db, 1, prod)
    
    return(db)
}

## modified multivariate boundary beta spherically symmetric kernel
dmvbeta.symm.kernel2 <- function(x, eval.x, H)
{
    d <- ncol(H)
    eval.y <- sqrt(apply(eval.x^2, 1, sum))/sqrt(d)
    y <- sqrt(sum(x^2))/sqrt(d)
  
    return(dbeta.kernel2(x=y, eval.x=eval.y, h=sqrt(tr(H)))/d)
}

rbeta.kernel2 <- function(x, n, h)
{
    rhox <- function(y, hy) { if (y==0) return (1) else return(2*hy^4 + 5/2 - sqrt(4*hy^8 + 6*hy^4 + 9/4 - y^2 -y/hy^2)) }
    
    ind <- cut(x, c(0, 2*h^2, 1-2*h^2, 1), labels=FALSE, include.lowest=TRUE)

    if (ind==1) { shape1 <- rhox(x,hy=h); shape2 <- (1-x)/h^2 }
    else if (ind==2) { shape1 <- x/h^2;   shape2 <- (1-x)/h^2 }
    else if (ind==3) { shape1 <- x/h^2;   shape2 <- rhox(1-x, hy=h) }
    
    return(rbeta(n=n, shape1=shape1, shape2=shape2))
}

dmvbeta.prod.kernel2.2d <- function(x, hs, xmin=c(0,0), xmax=c(1,1), ...)
{
    x.star <- (x - xmin)/(xmax - xmin)
    hs.star <- hs/(xmax - xmin)
    x1 <- seq(0,1, length=151)
    x1[1] <- 1e-9; x1[151] <- 1-1e-9
    eval.points <- list(x1, x1)
    fhat <- list()
    x <- cbind(rbeta.kernel2(x=x.star[1], n=1000, h=hs.star[1]), rbeta.kernel2(x=x.star[2], n=1000, h=hs.star[2]))
   
    fhat$eval.points <- list(seq(0,1, length=151), seq(0,1, length=151))
    
    fhat$estimate <- matrix(dmvbeta.prod.kernel2(x=x.star, eval.x=eval.points, hs=hs.star, ...), nrow=length(eval.points[[1]]))

    fhat$x <- sweep(sweep(x, 2, xmin, FUN="+"), 2, xmax-xmin, FUN="*") 
    fhat$eval.points[[1]] <- xmin[1] + fhat$eval.points[[1]]*(xmax[1]-xmin[1])
    fhat$eval.points[[2]] <- xmin[2] + fhat$eval.points[[2]]*(xmax[2]-xmin[2])
    fhat$H <- diag(2)
    fhat$gridtype <- "linear"
    fhat$gridded <- TRUE
    fhat$binned <- FALSE
    fhat$names <- parse.name(x)
    fhat$w <- rep(1, nrow(x))
    class(fhat) <- "kde"
                
    return(fhat)
}

##########################################################################
## Truncate unbounded KDE to polygon boundary
##########################################################################
kde.truncate <- function(fhat, boundary)
{
    ## reallocate any probability mass outside of map boundary regions
    ## to interior regions
    truncate.ind <- array(mgcv::in.out(boundary, as.matrix(expand.grid(fhat$eval.points[[1]], fhat$eval.points[[2]]))), dim=dim(fhat$estimate))
    fhat.trunc <- fhat
    fhat.trunc$estimate <- fhat.trunc$estimate*truncate.ind
    fhat.sum <- contourProbs(fhat, abs.cont=0)
    fhat.trunc.sum <- contourProbs(fhat.trunc, abs.cont=0)
    fhat.trunc$estimate <- fhat.trunc$estimate*fhat.sum/fhat.trunc.sum
    
    fhat.trunc.temp <- fhat.trunc
    fhat.trunc.temp$x <- fhat.trunc.temp$x[mgcv::in.out(boundary, fhat.trunc$x),]
    fhat.trunc$cont <- contourLevels(fhat.trunc.temp, cont=1:99, approx=TRUE)
    
    return(fhat.trunc)
}

##########################################################################
## Truncate unbounded KDDE to polygon boundary
##########################################################################
kdde.truncate <- function(fhat, boundary)
{
    ## reallocate any probability mass outside of map boundary regions
    ## to interior regions
    fhat.trunc <- fhat
    fhat0 <- kde(x=fhat$x, H=fhat$H)
    fhat0.sum <- contourProbs(fhat0, abs.cont=0)
    fhat0.trunc <- kde.truncate(fhat=fhat0, boundary=boundary)
    fhat0.trunc.sum <- contourProbs(fhat0.trunc, abs.cont=0)
    
    for (i in 1:length(fhat$estimate))
    {    
        truncate.ind <- array(mgcv::in.out(boundary, as.matrix(expand.grid(fhat$eval.points[[1]], fhat$eval.points[[2]]))), dim=dim(fhat$estimate[[i]]))
        fhat.trunc$estimate[[i]] <- fhat.trunc$estimate[[i]]*truncate.ind
        fhat.trunc$estimate[[i]] <- fhat.trunc$estimate[[i]]*fhat0.sum/fhat0.trunc.sum
    }
    fhat.trunc.temp <- fhat.trunc
    fhat.trunc.temp$x <- fhat.trunc.temp$x[mgcv::in.out(boundary, fhat.trunc$x),]
    fhat.trunc$cont <- contourLevels(fhat.trunc.temp, cont=1:99, approx=TRUE)
    
    return(fhat.trunc)
}

#############################################################################
## Estimate g_AMSE pilot bandwidths for even orders - 2-dim
##
## Parameters
## r - (r1, r2) partial derivative
## n - sample size
## psi1 - psi_(r + (2,0))
## psi2 - psi_(r + (0,2))
##
## Returns
## g_AMSE pilot bandwidths for even orders
#############################################################################
gamse.even.2d <- function(r, n, psi1, psi2)
{
    d <- 2
    num <- -2 * dmvnorm.deriv(x=c(0,0), deriv.order=r, Sigma=diag(2), deriv.vec=FALSE)
    den <- (psi1 + psi2) * n   
    g.amse <- (num/den)^(1/(2 + d + sum(r)))

    return(g.amse)
}

#############################################################################
## Estimate g_AMSE pilot bandwidths for odd orders - 2-dim
##
## Parameters
##
## r - (r1, r2) partial derivative
## n - sample size
## psi1 - psi_(r + (2,0))
## psi2 - psi_(r + (0,2))
## psi00 - psi_(0,0)
## RK - R(K^(r))
## 
## Returns
## g_AMSE pilot bandwidths for odd orders
#############################################################################
gamse.odd.2d <- function(r, n, psi1, psi2, psi00, RK)
{  
    d <- 2
    num <- 2 * psi00 * (2 * sum(r) + d) * RK
    den <- (psi1 + psi2)^2 * n^2
    g.amse <- (num/den)^(1/(2*sum(r) + d + 4))

    return(g.amse)
}

#############################################################################
## Estimate g_SAMSE pilot bandwidth - 2- to 6-dim 
##
## Parameters
## Sigma.star - scaled variance matrix
## n - sample size
##
## Returns
## g_SAMSE pilot bandwidth
#############################################################################
gsamse <- function(Sigma.star, n, modr, nstage=1, psihat=NULL)
{
    d <- ncol(Sigma.star)
    K <- numeric(); psi <- numeric()

    ## 4th order g_SAMSE
    K <- dmvnorm.deriv(x=rep(0,d), deriv.order=modr, Sigma=diag(d), add.index=TRUE, deriv.vec=FALSE)
    K <- K$deriv[apply(K$deriv.ind, 1, is.even)]

    if (modr==4)
    {
        derivt4 <- dmvnorm.deriv(x=rep(0,d), deriv.order=4, add.index=TRUE, deriv.vec=FALSE, only.index=TRUE)
        derivt6 <- dmvnorm.deriv(x=rep(0,d), deriv.order=6, add.index=TRUE, deriv.vec=FALSE, only.index=TRUE)

        for (i in 1:nrow(derivt4))
        {
            r <- derivt4[i,]
            if (is.even(r))
            {
                A3psi <- 0
                for (j in 1:d)
                {
                    if (nstage==1) A3psi <- A3psi + psins(r=r+2*elem(j,d), Sigma=Sigma.star)
                    else if (nstage==2) A3psi <- A3psi + psihat[which.mat(r=r+2*elem(j,d), mat=derivt6)]
                }
                psi <- c(psi, A3psi)    
            }
        }
    }
    ## 6th order g_SAMSE
    else if (modr==6)
    {
        derivt6 <- dmvnorm.deriv(x=rep(0,d), deriv.order=6, add.index=TRUE, deriv.vec=FALSE, only.index=TRUE)

        for (i in 1:nrow(derivt6))
        {
            r <- derivt6[i,]        
            if (is.even(r))
            {
                A3psi <- 0
                for (j in 1:d)
                A3psi <- A3psi + psins(r=r+2*elem(j,d), Sigma=Sigma.star)
                psi <- c(psi, A3psi)
            }
        }
    }

    ## see thesis for formula
    A1 <- sum(K^2)
    A2 <- sum(K * psi)  
    A3 <- sum(psi^2)
    B1 <- (2*modr + 2*d)*A1
    B2 <- (modr + d - 2)*A2
    B3 <- A3
    gamma1 <- (-B2 + sqrt(B2^2 + 4*B1*B3)) / (2*B1)
    g.samse <- (gamma1 * n)^(-1/(modr + d + 2))

    return (g.samse)      
}

############################################################################
## Scalar pilot selector for derivatives r>0 from Chacon & Duong (2011)
## Generalisation of gsamse for r>0
############################################################################
gdscalar <- function(x, d, r, n, verbose, binned=FALSE, nstage=1, scv=FALSE)
{
    if (scv) cf <- c(2^(-d), 2^(-d/2+1), 4)
    else cf <- c(1,1,1)
    S <- var(x)
    Sinv <- chol2inv(chol(S))

    if (nstage==1)
    {
        psi2r6.ns <- psins(r=2*r+6, Sigma=S)
        B3 <- norm(Matrix(psi2r6.ns, nrow=1) %*% (Matrix(vec(diag(d)), ncol=1) %*% Matrix(vec(diag(d)), nrow=1) %x% Diagonal(d^(2*r+4))) %*% Matrix(psi2r6.ns, ncol=1), type="1")
        if (!scv)
        {
            B1 <- 2*(2*pi)^(-d)*OF(2*r+4)*prod(d+2*(0:(r+2)))  
            B2 <- -(d+2*r+2)*2^(-d/2-r)*(2*pi)^(-d)*det(S)^(-1/2)*OF(2*r+4)*nu(r=r+2, A=Sinv)
        }
        else
        {
            B1 <- 2^(-2*r-3)*(4*pi)^(-d)*OF(2*r+6)*prod(d+2*(0:(r+2)))
            B2 <- -(d+2*r+2)*2^(-2*r-4)*(4*pi)^(-d)*det(S)^(-1/2)*OF(2*r+4)*nu(r=r+2, A=Sinv)
            B3 <- 4*B3
        }
        g2r4 <- (2*B1/(-B2 + sqrt(B2^2 + 4*B1*B3)))^(1/(d+2*r+8))*n^(-1/(d+2*r+8))
    }
    else if (nstage==2)
    {
        psi2r8.ns <- psins(r=2*r+8, Sigma=S)
        B3 <- norm(Matrix(psi2r8.ns, nrow=1) %*% (Matrix(vec(diag(d)), ncol=1) %*% Matrix(vec(diag(d)), nrow=1) %x% Diagonal(d^(2*r+6))) %*% Matrix(psi2r8.ns, ncol=1), type="1")
        if (!scv)
        {
            B1 <- 2*(2*pi)^(-d)*OF(2*r+6)*prod(d+2*(0:(r+3)))  
            B2 <- -(d+2*r+4)*2^(-d/2-r+2)*(2*pi)^(-d)*det(S)^(-1/2)*OF(2*r+6)*nu(r=r+4, A=Sinv)
        }
        else
        {
            B1 <- 2^(-2*r-5)*(4*pi)^(-d)*OF(2*r+6)*prod(d+2*(0:(r+3)))
            B2 <- -(d+2*r+4)*2^(-2*r-6)*(4*pi)^(-d)*det(S)^(-1/2)*OF(2*r+6)*nu(r=r+4, A=Sinv)
            B3 <- 4*B3
        }
        g2r6.ns <- (2*B1/(-B2 + sqrt(B2^2 + 4*B1*B3)))^(1/(d+2*r+8))*n^(-1/(d+2*r+8))
        L0 <- dmvnorm.mixt(x=rep(0,d), mus=rep(0,d), Sigmas=diag(d), props=1)
        if (binned)
        eta2r6 <- drop(kfe(x=x, G=g2r6.ns^2*diag(d), inc=1, binned=binned, deriv.order=2*r+6, add.index=FALSE, verbose=verbose) %*% vec(diag(d^(r+3))))
        else
        eta2r6 <- Qr(x=x, deriv.order=2*r+6, Sigma=g2r6.ns^2*diag(d), inc=1)
        
        A1 <- cf[1]*(2*d+4*r+8)*L0^2*OF(2*r+4)*nu(r=r+2, A=diag(d))
        A2 <- cf[2]*(d+2*r+2)*L0*OF(2*r+4)*eta2r6
        A3 <- cf[3]*eta2r6^2
        
        g2r4 <- (2*A1/((-A2+ sqrt(A2^2 +4*A1*A3))*n))^(1/(d+2*r+6))
    }
    return(g2r4)
}

############################################################################
## Unconstrained pilot selector for derivatives r>0 from Chacon & Duong (2011)
## Generalisation of Gunconstr for r>0
############################################################################
Gdunconstr <- function(x, d, r, n, nstage=1, verbose, binned=FALSE, scv=FALSE, optim.fun="optim")
{
    if (scv) cf <- c(2^(-(d/2+r+2)), 2)
    else cf <- c(1,1)
    S <- var(x) 
    S12 <- matrix.sqrt(S)

    optim.fun <- match.arg(optim.fun, c("nlm", "optim")) 
    if (nstage==2)
    {
        # G2r6.NR <- Gns(r=2*r+6,n=n,Sigma=S, scv=scv)
        # vecPsi2r6 <- kfe(x=x, G=G2r6.NR, binned=binned, deriv.order=2*r+6, deriv.vec=TRUE, add.index=FALSE, verbose=verbose)   
        x.star <- pre.sphere(x)
        G2r6.NR <- Gns(r=2*r+6,n=n,Sigma=var(x.star), scv=scv)
        vecPsi2r6 <- kfe(x=x.star, G=G2r6.NR, binned=binned, deriv.order=2*r+6, deriv.vec=TRUE, add.index=FALSE, verbose=verbose)   

        dls <- (0:(d^2-1))*d^(2*r+4) 

        AB2 <- function(vechG)
        {
            G <- invvech(vechG) %*% invvech(vechG)
            Ginv <- chol2inv(chol(G))
            ## direct computation
            v1 <- n^(-1)*det(G)^(-1/2)*(2*pi)^(-d/2)*(-1)^(r+2)*OF(2*r+4)* Sdrv(d=d,r=2*r+4,v=Kpow(vec(Ginv),r+2))

            v2 <- numeric(d^(2*r+4))
            for(k in 1:d^(2*r+4)) { v2[k] <- sum(vec(G)*vecPsi2r6[dls+k]) }

            AB <- cf[1]*v1 + cf[2]*(1/2)*v2
            AB2.val <- sum(AB^2)
            return(AB2.val)
        }

        Gstart <- Gns(r=2*r+4, n=n, Sigma=var(x.star), scv=scv)
        Gstart <- matrix.sqrt(Gstart)
        
        if (optim.fun=="nlm")
        {
            result <- nlm(p=vech(Gstart), f=AB2, print.level=2*as.logical(verbose))
            G2r4 <- result$estimate
        } 
        else if (optim.fun=="optim")
        { 
            result <- optim(par=vech(Gstart), fn=AB2, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
            G2r4 <- result$par
        }
        G2r4 <- invvech(G2r4) %*% invvech(G2r4)
        G2r4 <- S12 %*% G2r4 %*% S12
    }
    else
    {
        G2r4 <- Gns(r=2*r+4, n=n, Sigma=S, scv=scv)
    }
    
    return(G2r4)
}

############################################################################
## Estimate psi functionals using 1-stage plug-in 
##
## Parameters
## x.star - pre-transformed data points
## pilot - "amse" = different AMSE pilot bandwidths
##       - "samse" = optimal SAMSE pilot bandwidth
##
## Returns
## estimated psi functionals
#############################################################################
psifun1 <- function(x.star, pilot="samse", binned, bin.par, deriv.order=0, verbose=FALSE)
{
    d <- ncol(x.star)
    r <- deriv.order
    S.star <- var(x.star)
    n <- nrow(x.star)

    ## pilots are based on (2r+4)-th order derivatives
    ## compute 1 pilot for SAMSE
    if (pilot=="samse")
    {
        g.star <- gsamse(S.star, n, 4)
        psihat.star <- kfe(x=x.star, G=g.star^2*diag(d), deriv.order=4, deriv.vec=TRUE, binned=binned, add.index=TRUE, verbose=verbose)
    }
    ## compute 5 different pilots for AMSE
    else if ((pilot=="amse") & (d==2))
    {
        derivt4 <- dmvnorm.deriv(x=rep(0,d), deriv.order=4, add.index=TRUE, deriv.vec=FALSE, only.index=TRUE)
        derivt4.vec <- dmvnorm.deriv(x=rep(0,d), deriv.order=4, add.index=TRUE, deriv.vec=TRUE, only.index=TRUE)

        RK31 <- 15/(64*pi)
        psi00 <- psins(r=c(0,0), Sigma=S.star) 
        psihat.star <- vector()
        g.star <- vector()

        for (k in 1:nrow(derivt4))
        { 
            r <- derivt4[k,]
            psi1 <- psins(r=r + 2*elem(1, 2), Sigma=S.star)
            psi2 <- psins(r=r + 2*elem(2, 2), Sigma=S.star)

            ## odd order
            if (prod(r) == 3) g.star[k] <- gamse.odd.2d(r=4, n, psi1, psi2, psi00, RK31)
            ## even order
            else g.star[k] <- gamse.even.2d(r=4, n, psi1, psi2)[k]
            psihat.star[k] <- kfe.scalar(x=x.star, deriv.order=r, g=g.star[k], binned=binned, bin.par=bin.par)
        }

        ## create replicated form of psihat
        psihat.star.vec <- rep(0, nrow(derivt4.vec))
        for (k in 1:nrow(derivt4.vec))
        psihat.star.vec[k] <- psihat.star[which.mat(r=derivt4.vec[k,], mat=derivt4)]

        psihat.star <- list(psir=psihat.star.vec, deriv.ind=derivt4.vec)
    }

    return(psihat.star)
}

#############################################################################
## Estimate psi functionals using 2-stage plug-in 
##
## Parameters
## x - pre-transformed data points
## pilot - "amse" - different AMSE pilot
##       - "samse" - SAMSE pilot
## Returns
## estimated psi functionals
#############################################################################
psifun2 <- function(x.star, pilot="samse", binned, bin.par, deriv.order=0, verbose=FALSE)
{ 
    d <- ncol(x.star)
    r <- deriv.order
    S.star <- var(x.star)
    n <- nrow(x.star)

    ## pilots are based on (2r+4)-th order derivatives
    ## compute 1 pilot for SAMSE    
    if (pilot=="samse")
    {
        g6.star <- gsamse(S.star, n=n, modr=6)
        psihat6.star <- kfe(x=x.star, G=g6.star^2*diag(d), deriv.order=6, deriv.vec=TRUE, binned=binned, add.index=FALSE, verbose=verbose)
        g.star <- gsamse(S.star, n=n, modr=4, nstage=2, psihat=psihat6.star)
        psihat.star <- kfe(x=x.star, G=g.star^2*diag(d), deriv.order=4, deriv.vec=TRUE, binned=binned, add.index=TRUE, verbose=verbose)
    }
    ## compute different pilots for AMSE
    else if ((pilot=="amse") & (d==2))
    {
        derivt4 <- dmvnorm.deriv(x=rep(0,d), deriv.order=4, add.index=TRUE, deriv.vec=FALSE, only.index=TRUE)
        derivt4.vec <- dmvnorm.deriv(x=rep(0,d), deriv.order=4, add.index=TRUE, deriv.vec=TRUE, only.index=TRUE)
        derivt6 <- dmvnorm.deriv(x=rep(0,d), deriv.order=6, add.index=TRUE, deriv.vec=FALSE, only.index=TRUE)

        RK31 <- 15/(64*pi)
        RK51 <- 945/(256*pi)
        RK33 <- 225/(256*pi)
        psi00 <- psins(r=rep(0,d), Sigma=S.star) 

        psihat6.star <- vector()
        g6.star <- vector()
        psihat.star <- vector()
        g.star <- vector()

        for (k in 1:nrow(derivt6))
        {
            r <- derivt6[k,]
            psi1 <- psins(r=r + 2*elem(1, 2), Sigma=S.star)
            psi2 <- psins(r=r + 2*elem(2, 2), Sigma=S.star)
            if (prod(r) == 5)
            g6.star[k] <- gamse.odd.2d(r=6, n, psi1, psi2, psi00, RK51)
            else if (prod(r) == 9)
            g6.star[k] <- gamse.odd.2d(r=6, n, psi1, psi2, psi00, RK33) 
            else  
            g6.star[k] <- gamse.even.2d(r=6, n, psi1, psi2)[k]

            psihat6.star[k] <- kfe.scalar(x=x.star, deriv.order=r, g=g6.star[k], binned=binned, bin.par=bin.par) 
        }

        ## pilots are based on 4th order derivatives using 6th order psi functionals
        ## computed above 'psihat6.star'
        for (k in 1:nrow(derivt4))
        {
            r <- derivt4[k,]
            psi1 <- psihat6.star[7 - (r + 2*elem(1,2))[1]]
            psi2 <- psihat6.star[7 - (r + 2*elem(2,2))[1]]

            if (prod(r) == 3)
            g.star[k] <- gamse.odd.2d(r=4, n, psi1, psi2, psi00, RK31)
            else
            g.star[k] <- gamse.even.2d(r=4, n, psi1, psi2)[k]

            psihat.star[k] <- kfe.scalar(x=x.star, deriv.order=r, g=g.star[k],  binned=binned, bin.par=bin.par) 
        }

        ## create replicated form of psihat
        psihat.star.vec <- rep(0, nrow(derivt4.vec))
        for (k in 1:nrow(derivt4.vec))
        psihat.star.vec[k] <- psihat.star[which.mat(r=derivt4.vec[k,], mat=derivt4)]

        psihat.star <- list(psir=psihat.star.vec, deriv.ind=derivt4.vec)
    }

    return(psihat.star)
}

#############################################################################
## Estimate psi functionals for 6-variate data using 1-stage plug-in 
## with unconstrained pilot
##
## Parameters
## x - data points
## Sd4, Sd6 - symmetrizer matrices of order 4 and 6
##
## Returns
## estimated psi functionals
#############################################################################

psifun1.unconstr <- function(x, binned, bgridsize, deriv.order=0, verbose=FALSE)
{
    n <- nrow(x)
    r <- deriv.order
    S <- var(x)

    ## stage 1 of plug-in
    G2r4 <- Gns(r=2*r+4,n=n,Sigma=S) 
    vecPsi2r4 <- kfe(x=x, G=G2r4, deriv.order=2*r+4, binned=binned, bgridsize=bgridsize, deriv.vec=TRUE, add.index=FALSE, verbose=verbose) 
    return (vecPsi2r4)
}

#############################################################################
## Estimate psi functionals for 6-variate data using 2-stage plug-in 
## with unconstrained pilot
##
## Parameters
## x - data points
## Sd4, Sd6 - symmetrizer matrices of order 4 and 6
##
## Returns
## estimated psi functionals
############################################################################

psifun2.unconstr <- function(x, binned, bgridsize, deriv.order=0, verbose=FALSE, optim.fun="optim")
{
    d <- ncol(x)
    n <- nrow(x)
    S <- var(x)
    S12 <- matrix.sqrt(S)
    r <- deriv.order
    optim.fun <- match.arg(optim.fun, c("nlm", "optim"))

    ## stage 1 of plug-in
    x.star <- pre.sphere(x)
    G2r6 <- Gns(r=2*r+6,n=n,Sigma=var(x.star)) 
    vecPsi2r6 <- kfe(x=x.star, G=G2r6, binned=binned, bgridsize=bgridsize, deriv.order=2*r+6, deriv.vec=TRUE, add.index=FALSE, verbose=verbose)

    ## asymptotic squared bias for r = 4 for MSE-optimal G
    D2r4phi0 <- DrL0(d=d, r=2*r+4)
    Id2r4 <- diag(d^(2*r+4))

    AB2 <- function(vechG)
    {
        rr <- 2*r+4
        G <- invvech(vechG)%*%invvech(vechG)
        G12 <- matrix.sqrt(G)
        Ginv12 <- chol2inv(chol(G12))
        AB <- n^(-1)*det(Ginv12)*(Kpow(A=Ginv12,pow=rr)%*%D2r4phi0)+(1/2)*(t(vec(G))%x%Id2r4) %*% vecPsi2r6
        AB2.val <- sum(AB^2)
        return (AB2.val)
    }

    Gstart <- Gns(r=2*r+4,n=n,Sigma=var(x.star))
    Gstart <- matrix.sqrt(Gstart)

    if (optim.fun=="nlm")
    {
        res <- nlm(p=vech(Gstart), f=AB2, print.level=2*as.logical(verbose))    
        G2r4 <- res$estimate
    }
    else if (optim.fun=="optim")
    {
        res <- optim(vech(Gstart), AB2, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
        G2r4 <- res$par
    }
    G2r4 <- invvech(G2r4)%*%invvech(G2r4)
    G2r4 <- S12 %*% G2r4 %*% S12

    ## stage 2 of plug-in
    vecPsi2r4 <- kfe(x=x, G=G2r4, binned=binned, bgridsize=bgridsize, deriv.order=2*r+4, deriv.vec=TRUE, add.index=FALSE, verbose=verbose)

    return (vecPsi2r4)
}

#############################################################################
## Plug-in bandwidth selectors
#############################################################################

############################################################################
## Computes plug-in full bandwidth matrix - 2 to 6 dim
##
## Parameters
## x - data points
## Hstart - initial value for minimisation
## nstage - number of plug-in stages (1 or 2)
## pilot - "amse" - different AMSE pilot
##       - "samse" - SAMSE pilot
##       - "unconstr" - unconstrained pilot
## pre - "scale" - pre-scaled data
##     - "sphere"- pre-sphered data 
##
## Returns
## Plug-in full bandwidth matrix
#############################################################################
hpi <- function(x, nstage=2, binned=TRUE, bgridsize, deriv.order=0)
{
    ## 1-d selector is taken from KernSmooth's dpik
    d <- 1 
    if (missing(bgridsize)) bgridsize <- default.gridsize(d)
    if (deriv.order==0) h <- dpik(x=x, level=nstage, gridsize=bgridsize)
    else
    {
        n <- length(x)
        d <- 1
        r <- deriv.order  
        K2r4 <- dnorm.deriv(x=0, mu=0, sigma=1, deriv.order=2*r+4)
        K2r6 <- dnorm.deriv(x=0, mu=0, sigma=1, deriv.order=2*r+6) 
        m2 <- 1  
        mr <- psins.1d(r=2*r, sigma=1)

        ## formula for bias annihilating bandwidths from Wand & Jones (1995, p.70)
        if (nstage==2)
        {
            psi2r8.hat <- psins.1d(r=2*r+8, sigma=sd(x))
            gamse2r6 <- (2*K2r6/(-m2*psi2r8.hat*n))^(1/(2*r+9)) 
            psi2r6.hat <- kfe.1d(x=x, g=gamse2r6, deriv.order=2*r+6, inc=1, binned=binned)
            gamse2r4 <- (2*K2r4/(-m2*psi2r6.hat*n))^(1/(2*r+7))
            psi2r4.hat <- kfe.1d(x=x, g=gamse2r4, deriv.order=2*r+4, inc=1, binned=binned)
        }
        else 
        {
            psi2r6.hat <- psins.1d(r=2*r+6, sigma=sd(x))
            gamse2r4 <- (2*K2r4/(-m2*psi2r6.hat*n))^(1/(2*r+7))
            psi2r4.hat <- kfe.1d(x=x, g=gamse2r4, deriv.order=2*r+4, inc=1, binned=binned)
        }

        ## formula form Wand & Jones (1995, p.49)
        h <- ((2*r+1)*mr/(m2^2*psi2r4.hat*n))^(1/(2*r+5))
    }

    return(h)
}

Hpi <- function(x, nstage=2, pilot, pre="sphere", Hstart, binned, bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
{
    n <- nrow(x)
    d <- ncol(x)
    r <- deriv.order
    if (missing(binned)) binned <- default.bflag(d=d,n=n)
    if (d > 4) binned <- FALSE
    if (missing(bgridsize)) bgridsize <- default.bgridsize(d)
    if (!is.matrix(x)) x <- as.matrix(x)

    if (missing(pilot)) { if (d==2 & r==0) pilot <- "samse" else pilot <- "dscalar" }
    pilot1 <- match.arg(pilot, c("amse", "samse", "unconstr", "dunconstr", "dscalar"))
    pre1 <- match.arg(pre, c("scale", "sphere"))
    optim.fun <- match.arg(optim.fun, c("nlm", "optim"))

    if (pilot1=="amse" & (d>2 | r>0)) stop("amse pilot selectors not defined for d>2 and/or r>0")
    if ((pilot1=="samse" | pilot1=="unconstr") & r>0) stop("dscalar or dunconstr pilot selectors are better for derivatives r>0")
    if (pilot1=="unconstr" & d>=6) stop("Unconstrained pilots are not implemented for d>6")

    if (pre1=="scale")
    {
        x.star <- pre.scale(x)
        S12 <- diag(sqrt(diag(var(x))))
        Sinv12 <- chol2inv(chol(S12))
    }
    else if (pre1=="sphere")
    {
        x.star <- pre.sphere(x)
        S12 <- matrix.sqrt(var(x))
        Sinv12 <- chol2inv(chol(S12))
    }

    Idr <- diag(d^r)
    RKr <- nu(r=r, diag(d))*2^(-d-r)*pi^(-d/2)

    if (binned)
    {
        H.max <- (((d+8)^((d+6)/2)*pi^(d/2)*RKr)/(16*(d+2)*n*gamma(d/2+4)))^(2/(d+4))* var(x)
        bin.par <- binning(x=x, bgridsize=bgridsize, H=H.max)
        H.max.star <- (((d+8)^((d+6)/2)*pi^(d/2)*RKr)/(16*(d+2)*n*gamma(d/2+4)))^(2/(d+4))* var(x.star)
        bin.par.star <- binning(x=x.star, bgridsize=bgridsize, H=H.max.star)    
    }

    if (pilot1=="unconstr")
    {
        ## psi4.mat is on data scale
        if (nstage==1)
        psi.fun <- (-1)^r*psifun1.unconstr(x=x, binned=binned, bgridsize=bgridsize, deriv.order=r, verbose=verbose)
        else if (nstage==2)
        psi.fun <- psifun2.unconstr(x=x, binned=binned, bgridsize=bgridsize, deriv.order=r, verbose=verbose)
        psi2r4.mat <- (-1)^r*invvec(psi.fun)   

        ## use normal reference bandwidth as initial condition 
        if (missing(Hstart)) Hstart <- Hns(x=x, deriv.order=r)
    }
    else if (pilot1=="dunconstr")
    {
        ## G2r4 is on data scale
        G2r4 <- Gdunconstr(x=x, d=d, r=r, n=n, nstage=nstage, verbose=verbose, binned=binned, optim.fun=optim.fun)
        vecPsi2r4 <- kfe(x=x, G=G2r4, binned=binned, deriv.order=2*r+4, deriv.vec=TRUE, add.index=FALSE, verbose=verbose, bin.par=bin.par)
        if (missing(Hstart)) Hstart <-  Hns(x=x, deriv.order=r)
    }
    else if (pilot1=="dscalar")
    {
        ## g2r4 is on pre-transformed data scale
        g2r4 <- gdscalar(x=x.star, r=r, n=n, d=d, verbose=verbose, nstage=nstage, binned=binned)

        G2r4 <- g2r4^2 * diag(d)
        vecPsi2r4 <- kfe(x=x.star, G=G2r4, binned=binned, deriv.order=2*r+4, deriv.vec=TRUE, add.index=FALSE, verbose=verbose, bin.par=bin.par.star)
        if (missing(Hstart)) Hstart <-  Hns(x=x.star, deriv.order=r)
    }
    else
    {      
        ## psi4.mat is on pre-transformed data scale
        if (nstage==1)
        psi.fun <- psifun1(x.star, pilot=pilot, binned=binned, bin.par=bin.par.star, deriv.order=r, verbose=verbose)$psir
        else if (nstage==2)
        psi.fun <- psifun2(x.star, pilot=pilot, binned=binned, bin.par=bin.par.star, deriv.order=r, verbose=verbose)$psir
        psi2r4.mat <- invvec(psi.fun)

        ## use normal reference bandwidth as initial condition 
        if (missing(Hstart)) Hstart <- Hns(x=x.star, deriv.order=r)
        else Hstart <- Sinv12 %*% Hstart %*% Sinv12
    }

    ## PI is estimate of AMISE
    pi.temp <- function(vechH)
    { 
        H <- invvech(vechH) %*% invvech(vechH)
        Hinv <- chol2inv(chol(H))
        IdrvH <- Idr%x%vec(H)
        int.var <- 1/(det(H)^(1/2)*n)*nur(r=r, A=Hinv, mu=rep(0,d), Sigma=diag(d))*2^(-d-r)*pi^(-d/2)

        if (pilot1=="dunconstr" | pilot1=="dscalar") 
        pi.val <- int.var + (-1)^r*1/4*vecPsi2r4 %*% (vec(diag(d^r) %x% vec(H) %x% vec(H)))
        else
        pi.val <- int.var + (-1)^r*1/4* sum(diag(t(IdrvH) %*% psi2r4.mat %*% IdrvH))
        pi.val <- drop(pi.val)
        return(pi.val)
    }

    Hstart <- matrix.sqrt(Hstart)
    if (optim.fun=="nlm")
    {
        result <- nlm(p=vech(Hstart), f=pi.temp, print.level=2*as.numeric(verbose))    
        H <- invvech(result$estimate) %*% invvech(result$estimate)
        amise.star <- result$minimum
    }
    else if (optim.fun=="optim")
    {
        result <- optim(vech(Hstart), pi.temp, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
        H <- invvech(result$par) %*% invvech(result$par)
        amise.star <- result$value
    }
    if (!(pilot1 %in% c("dunconstr","unconstr")))  H <- S12 %*% H %*% S12   ## back-transform

    if (!amise) return(H)
    else return(list(H = H, PI.star=amise.star))
}     

#############################################################################
## Computes plug-in diagonal bandwidth matrix for 2 to 6-dim
##
## Parameters
## x - data points
## nstage - number of plug-in stages (1 or 2)
## pre - "scale" - pre-scaled data
##     - "sphere"- pre-sphered data 
##
## Returns
## Plug-in diagonal bandwidth matrix
#############################################################################
Hpi.diag <- function(x, nstage=2, pilot, pre="scale", Hstart, binned, bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
{
    n <- nrow(x)
    d <- ncol(x)
    r <- deriv.order
    RK <- (4*pi)^(-d/2)
    if (missing(binned)) binned <- default.bflag(d=d,n=n)
    if (d > 4) binned <- FALSE
    if (missing(bgridsize)) bgridsize <- default.bgridsize(d)
    if (!is.matrix(x)) x <- as.matrix(x)

    if (missing(pilot)) { if (d==2 & r==0) pilot <- "samse" else pilot <- "dscalar" }
    pilot1 <- match.arg(pilot, c("amse", "samse", "unconstr", "dunconstr", "dscalar"))
    pre1 <- match.arg(pre, c("scale", "sphere"))
    optim.fun <- match.arg(optim.fun, c("nlm", "optim"))

    if (pre1=="sphere") stop("Using pre-sphering won't give diagonal bandwidth matrix")
    if (pilot1=="amse" & (d>2 | r>0)) stop("samse pilot selectors are better for higher dimensions and/or deriv.order>0")
    if (pilot1=="samse" & r>0) stop("dscalar or dunconstr pilot selectors are better for derivatives r>0")
    if (pilot1=="unconstr" | pilot1=="dunconstr") stop("Unconstrained pilot selectors are not suitable for Hpi.diag")

    if (pre1=="scale")
    {
        x.star <- pre.scale(x)
        S12 <- diag(sqrt(diag(var(x))))
        Sinv12 <- chol2inv(chol(S12))
    }
    else if (pre1=="sphere")
    {
        x.star <- pre.sphere(x)
        S12 <- matrix.sqrt(var(x))
        Sinv12 <- chol2inv(chol(S12))
    }

    if (binned)
    {
        H.max <- (((d+8)^((d+6)/2)*pi^(d/2)*RK)/(16*(d+2)*n*gamma(d/2+4)))^(2/(d+4))* var(x.star)
        bin.par <- binning(x=x.star, bgridsize=bgridsize, H=H.max) 
    }

    Idr <- diag(d^r)  
    if (pilot1=="amse" | pilot1=="samse")
    {
        if (nstage==1)
        psi.fun <- psifun1(x.star, pilot=pilot, binned=binned, bin.par=bin.par, deriv.order=r, verbose=verbose)$psir
        else if (nstage==2)
        psi.fun <- psifun2(x.star, pilot=pilot, binned=binned, bin.par=bin.par, deriv.order=r, verbose=verbose)$psir
        psi2r4.mat <- invvec(psi.fun)
    }
    else if (pilot1=="dscalar")
    {
        g2r4 <- gdscalar(x=x.star, r=r, n=n, d=d, verbose=verbose, nstage=nstage, binned=binned)
        G2r4 <- g2r4^2 * diag(d)
        vecPsi2r4 <- kfe(x=x.star, G=G2r4, binned=binned, deriv.order=2*r+4, deriv.vec=TRUE, add.index=FALSE, verbose=verbose)
    }

    if (d==2 & r==0 & (pilot1=="amse" | pilot1=="samse"))
    {
        ## diagonal bandwidth matrix for 2-dim has exact formula 
        psi40 <- psi.fun[1]
        psi22 <- psi.fun[6]
        psi04 <- psi.fun[16]
        s1 <- sd(x[,1])
        s2 <- sd(x[,2])
        h1 <- (psi04^(3/4)*RK/(psi40^(3/4)*(sqrt(psi40*psi04)+psi22)*n))^(1/6)
        h2 <- (psi40/psi04)^(1/4) * h1
        H <- diag(c(s1^2*h1^2, s2^2*h2^2))
        psimat4.D <- invvech(c(psi40, psi22, psi04))
        amise.star <- drop(n^(-1)*RK*(h1*h2)^(-1) + 1/4*c(h1,h2)^2 %*% psimat4.D %*% c(h1,h2)^2)
    }
    else
    {  
        ## PI is estimate of AMISE
        pi.temp <- function(diagH)
        { 
            H <- diag(diagH) %*% diag(diagH)
            Hinv <- chol2inv(chol(H))
            IdrvH <- Idr%x%vec(H)
            int.var <- 1/(det(H)^(1/2)*n)*nu(r=r, Hinv)*2^(-d-r)*pi^(-d/2)

            if (pilot1=="dscalar")
            pi.val <- int.var + (-1)^r*1/4*vecPsi2r4 %*% (vec(diag(d^r) %x% vec(H) %x% vec(H)))
            else
            pi.val <- int.var + (-1)^r*1/4* sum(diag(t(IdrvH) %*% psi2r4.mat %*% IdrvH))

            return(drop(pi.val))
        }

        ## use normal reference bandwidth as initial condition
        if (missing(Hstart)) Hstart <- Hns(x=x.star, deriv.order=r)
        else Hstart <- Sinv12 %*% Hstart %*% Sinv12
        Hstart <- matrix.sqrt(Hstart)
        
        if (optim.fun=="nlm")
        {
            result <- nlm(p=diag(Hstart), f=pi.temp, print.level=2*as.numeric(verbose))    
            H <- diag(result$estimate^2)
            amise.star <- result$minimum
        }
        else if (optim.fun=="optim")
        {  
            result <- optim(diag(Hstart), pi.temp, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
            H <- diag(result$par) %*% diag(result$par)
            amise.star <- result$value
        }
        H <- S12 %*% H %*% S12  ## back-transform
    }

    if (!amise) return(H)
    else return(list(H = H, PI.star=amise.star))
}

#############################################################################
## Cross-validation bandwidth selectors
#############################################################################

#############################################################################
## Computes the least squares cross validation LSCV function for 2 to 6 dim
## 
## Parameters
## x - data values
## H - bandwidth matrix
##
## Returns
## LSCV(H)
#############################################################################
lscv.1d <- function(x, h, binned, bin.par, deriv.order=0)
{
    r <- deriv.order
    lscv1 <- kfe.1d(x=x, g=sqrt(2)*h, inc=1, binned=binned, bin.par=bin.par, deriv.order=2*r) 
    lscv2 <- kfe.1d(x=x, g=h, inc=0, binned=binned, bin.par=bin.par, deriv.order=2*r) 
    return((-1)^r*(lscv1 - 2*lscv2))     
}

lscv.mat <- function(x, H, binned=FALSE, bin.par, bgridsize, deriv.order=0)
{
    r <- deriv.order
    d <- ncol(x)
    n <- nrow(x)

    if (!binned)
    {
        lscv1 <- Qr(x=x, deriv.order=2*r, Sigma=2*H, inc=1)
        lscv2 <- Qr(x=x, deriv.order=2*r, Sigma=H, inc=0)
        lscv <- drop(lscv1 - 2*lscv2)
    }
    else
    {
        lscv1 <- kfe(x=x, G=2*H, inc=1, binned=binned, bin.par=bin.par, bgridsize=bgridsize, deriv.order=2*r, add.index=FALSE)
        lscv2 <- kfe(x=x, G=H, inc=0, binned=binned, bin.par=bin.par, bgridsize=bgridsize, deriv.order=2*r, add.index=FALSE)
        lscv <- (-1)^2*drop((lscv1 - 2*lscv2) %*% vec(diag(d^r)))
    }

    return(lscv)  
}

#############################################################################
## Finds the bandwidth matrix that minimises LSCV for 2 to 6 dim
## 
## Parameters
## x - data values
## Hstart - initial bandwidth matrix
##
## Returns
## H_LSCV
#############################################################################
hlscv <- function(x, binned=TRUE, bgridsize, amise=FALSE, deriv.order=0, bw.ucv=TRUE)
{
    ## adapted from versions supplied by J.E. Chacon 19/02/2021
    if (any(duplicated(x)))
    warning("Data contain duplicated values: LSCV is not well-behaved in this case")
    n <- length(x)
    d <- 1
    r <- deriv.order
    hnorm <- sqrt((4/(n*(d + 2)))^(2/(d + 4)) * var(x))
    if (missing(binned)) binned <- default.bflag(d=d,n=n)
    if (missing(bgridsize)) if (bw.ucv) bgridsize <- 10000 else bgridsize <- 10001 

    ## use stats::bw.ucv function
    if (bw.ucv)
    {
        difs <- dist(x)
        lower <- min(difs[difs>0])
        opt <- list()
        opt$minimum <- stats::bw.ucv(x=x, nb=bgridsize, lower=lower)
        opt$objective <- NA
    }
    else 
    {
        difs <- x%*%t(rep(1,n))-rep(1,n)%*%t(x)
        difs <- difs[lower.tri(difs)]  

        if (binned)
        {
            bin.par <- binning(x, bgridsize=bgridsize, h=hnorm)
            lscv.1d.temp <- function(h) { return(lscv.1d(x=x, h=h, binned=binned, bin.par=bin.par, deriv.order=r)) }
        }
        else
        {
            if (r>0) stop("Unbinned hlscv not yet implemented for deriv.order>0") 
            edifs <- exp(-difs^2/2)
            RK <- 1/(2*sqrt(pi))

            lscv.1d.temp <- function(h)
            {
                lscv1 <- (1-1/n)*sum(edifs^(1/(2*h^2)))/(h*sqrt(2)*sqrt(2*pi))
                lscv2 <- 2*sum(edifs^(1/h^2))/(h*sqrt(2*pi))
                return(RK/(n*h)+2*(lscv1-lscv2)/(n^2-n))
            }   
        }
        lower <- min(difs[difs>0]) 
        opt <- optimise(lscv.1d.temp, interval=c(lower, 2*hnorm), tol=.Machine$double.eps)
    }

    if (!amise) return(opt$minimum)
    else return(list(h=opt$minimum, LSCV=opt$objective))
}

Hlscv <- function(x, Hstart, binned=FALSE, bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim", trunc)
{
    if (any(duplicated(x))) warning("Data contain duplicated values: LSCV is not well-behaved in this case")
    if (!is.matrix(x)) x <- as.matrix(x)
    n <- nrow(x)
    d <- ncol(x)
    r <- deriv.order 
    if (missing(binned)) binned <- default.bflag(d=d,n=n)
    if (d>4) binned <- FALSE
    if (missing(bgridsize)) bgridsize <- default.bgridsize(d)

    optim.fun <- match.arg(optim.fun, c("nlm", "optim"))

    ## use normal reference selector as initial condn
    Hnorm <- Hns(x=x, deriv.order=r)
    if (missing(Hstart)) Hstart <- Hnorm
    Hstart <- matrix.sqrt(Hstart)
    if (binned) bin.par <- binning(x=x, H=Hnorm, bgridsize=bgridsize)
    if (missing(trunc)) { if (deriv.order==0) trunc <- 1e10 else trunc <- 4 }

    lscv.init <- lscv.mat(x=x, H=Hnorm, binned=binned, bin.par=bin.par, deriv.order=r)
    lscv.mat.temp <- function(vechH)
    {
        H <- invvech(vechH) %*% invvech(vechH)
        lscv <- lscv.mat(x=x, H=H, binned=binned, bin.par=bin.par, deriv.order=r)
        if (det(H) < 1/trunc*det(Hnorm) | det(H) > trunc*det(Hnorm) | abs(lscv) > trunc*abs(lscv.init)) lscv <- lscv.init 
        return(lscv)  
    }

    if (optim.fun=="nlm")
    {
        result <- nlm(p=vech(Hstart), f=lscv.mat.temp, print.level=2*as.numeric(verbose))   
        H <- invvech(result$estimate) %*% invvech(result$estimate)
        amise.opt <- result$minimum
    }
    else if (optim.fun=="optim")
    {
        result <- optim(vech(Hstart), lscv.mat.temp, method="Nelder-Mead", control=list(trace=as.numeric(verbose), REPORT=1))     
        H <- invvech(result$par) %*% invvech(result$par)
        amise.opt <- result$value
    }

    if (!amise) return(H)
    else return(list(H=H, LSCV=amise.opt))
}

#############################################################################
## Finds the diagonal bandwidth matrix that minimises LSCV for 2 to 6 dim
## 
## Parameters
## x - data values
## Hstart - initial bandwidth matrix
##
## Returns
## H_LSCV,diag
#############################################################################
Hlscv.diag <- function(x, Hstart, binned=FALSE, bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim", trunc)
{
    if (any(duplicated(x))) warning("Data contain duplicated values: LSCV is not well-behaved in this case")
    if (!is.matrix(x)) x <- as.matrix(x)
    n <- nrow(x)
    d <- ncol(x)
    r <- deriv.order
    if (missing(binned)) binned <- default.bflag(d=d,n=n)
    if (d>4) binned <- FALSE
    if (missing(bgridsize)) bgridsize <- default.bgridsize(d)
    optim.fun <- match.arg(optim.fun, c("nlm", "optim"))

    Hnorm <- Hns(x=x, deriv.order=r)
    if (missing(Hstart)) Hstart <- Hnorm

    ## don't truncate optimisation for deriv.order==0
    if (missing(trunc)) { if (deriv.order==0) trunc <- 1e10 else trunc <- 4 }

    ## linear binning
    if (binned)
    {
        bin.par <- binning(x=x, bgridsize=bgridsize, H=Hnorm)
    }

    lscv.init <- lscv.mat(x=x, H=Hnorm, binned=binned, bin.par=bin.par, deriv.order=r)

    lscv.mat.temp <- function(diagH)
    {
        H <- diag(diagH^2)
        lscv <- lscv.mat(x=x, H=H, binned=binned, bin.par=bin.par, deriv.order=r)
        if (det(H) < 1/trunc*det(Hnorm) | det(H) > trunc*det(Hnorm) | abs(lscv) > trunc*abs(lscv.init)) lscv <- lscv.init 
        return(lscv)  
    }

    Hstart <- matrix.sqrt(Hstart)
    if (optim.fun=="nlm")
    {
        result <- nlm(p=diag(Hstart), f=lscv.mat.temp, print.level=2*as.numeric(verbose))
        H <- diag(result$estimate^2)
        amise.opt <- result$minimum
    }
    else if (optim.fun=="optim")
    {
        result <- optim(diag(Hstart), lscv.mat.temp, method="Nelder-Mead", control=list(trace=as.numeric(verbose), REPORT=1))
        H <- diag(result$par^2)
        amise.opt <- result$value
    }

    if (!amise) return(H)
    else return(list(H=H, LSCV=amise.opt))
}

## aliases using "UCV" instead of "LSCV"

hucv <- function(...) { hlscv(...) }
Hucv <- function(...) { Hlscv(...) }
Hucv.diag <- function(...) { Hlscv.diag(...) }

#############################################################################
## Computes the biased cross validation BCV function for 2-dim
## 
## Parameters
## x - data values
## H1, H2 - bandwidth matrices
##
## Returns
## BCV(H)
#############################################################################
bcv.mat <- function(x, H1, H2, binned=FALSE)
{
    n <- nrow(x)
    d <- 2

    psi <- kfe(x, G=H2, deriv.order=4, add.index=TRUE, deriv.vec=TRUE, inc=0, binned=binned)
    psi40 <- psi$psir[1]
    psi31 <- psi$psir[2]
    psi22 <- psi$psir[4]
    psi13 <- psi$psir[8]
    psi04 <- psi$psir[16]

    coeff <- c(1, 2, 1, 2, 4, 2, 1, 2, 1)
    psi.fun <- c(psi40, psi31, psi22, psi31, psi22, psi13, psi22, psi13,psi04)/(n*(n-1))
    psi4.mat <- matrix(coeff * psi.fun, ncol=3, nrow=3)

    RK <- (4*pi)^(-d/2) 
    bcv <- drop(n^(-1)*det(H1)^(-1/2)*RK + 1/4*t(vech(H1)) %*% psi4.mat %*% vech(H1))

    return(list(bcv=bcv, psimat=psi4.mat))
}

#############################################################################
## Find the bandwidth matrix that minimises the BCV for 2-dim
## 
## Parameters
## x - data values
## whichbcv - 1 = BCV1
##          - 2 = BCV2 
## Hstart - initial bandwidth matrix
##
## Returns
## H_BCV
#############################################################################
Hbcv <- function(x, whichbcv=1, Hstart, binned=FALSE, amise=FALSE, verbose=FALSE)
{
    n <- nrow(x)
    d <- ncol(x)
    RK <- (4*pi)^(-d/2)
    if(!is.matrix(x)) x <- as.matrix(x)

    ## use normal reference b/w matrix for bounds
    k <- (((d+8)^((d+6)/2)*pi^(d/2)*RK)/(16*n*gamma((d+8)/2)*(d+2)))^(2/(d+4))
    Hmax <- k * abs(var(x))
    up.bound <- Hmax
    if (missing(Hstart)) Hstart <- matrix.sqrt(0.9*Hmax)

    bcv.mat.temp <- function(vechH)
    {
        H <- invvech(vechH) %*% invvech(vechH)
        bcv <- bcv.mat(x=x, H1=H, H2=whichbcv*H, binned=binned)$bcv
        return(bcv)  
    }

    result <- optim(vech(Hstart), bcv.mat.temp, method="L-BFGS-B", upper=vech(matrix.sqrt(up.bound)), lower=-vech(matrix.sqrt(up.bound)), control=list(trace=as.numeric(verbose), REPORT=1))

    H <- invvech(result$par) %*% invvech(result$par)
    amise.opt <- result$value

    if (!amise) return(H)
    else return(list(H = H, BCV=amise.opt))
}

#############################################################################
## Find the diagonal bandwidth matrix that minimises the BCV for 2-dim
## 
## Parameters
## x - data values
## whichbcv - 1 = BCV1
##          - 2 = BCV2
## Hstart - initial bandwidth matrix
##
## Returns
## H_BCV, diag
#############################################################################
Hbcv.diag <- function(x, whichbcv=1, Hstart, binned=FALSE, amise=FALSE, verbose=FALSE)
{
    n <- nrow(x)
    d <- ncol(x)
    RK <- (4*pi)^(-d/2)
    if(!is.matrix(x)) x <- as.matrix(x)

    ## use maximally smoothed b/w matrix for bounds
    k <- (((d+8)^((d+6)/2)*pi^(d/2)*RK)/(16*n*gamma((d+8)/2)*(d+2)))^(2/(d+4))
    Hmax <- k * abs(var(x))
    up.bound <- diag(Hmax)

    if (missing(Hstart)) Hstart <- 0.9*matrix.sqrt(Hmax)

    bcv.mat.temp <- function(diagH)
    {
        H <- diag(diagH) %*% diag(diagH) 
        return(bcv.mat(x, H, whichbcv*H, binned=binned)$bcv)
    } 

    result <- optim(diag(Hstart), bcv.mat.temp, method="L-BFGS-B", upper=sqrt(up.bound), control=list(trace=as.numeric(verbose), REPORT=1))
    H <- diag(result$par) %*% diag(result$par)
    amise.opt <- result$value
    
    if (!amise) return(H)
    else return(list(H = H, BCV=amise.opt))
}

#############################################################################
## Estimate scalar g_AMSE pilot bandwidth for SCV for 2 to 6 dim
##
## Parameters
## Sigma.star - scaled/ sphered variance matrix
## Hamise - (estimate) of H_AMISE 
## n - sample size
##
## Returns
## g_AMSE pilot bandwidth
#############################################################################
Theta6.elem <- function(d)
{
    Theta6.mat <- list()
    Theta6.mat[[d]] <- list()
    for (i in 1:d)
    Theta6.mat[[i]] <- list()

    for (i in 1:d)
    for (j in 1:d)
    {  
        temp <- numeric()
        for (k in 1:d)     
        for (ell in 1:d)    
        temp <- rbind(temp, elem(i,d)+2*elem(k,d)+2*elem(ell,d)+elem(j,d))

        Theta6.mat[[i]][[j]] <- temp
    }

    return(Theta6.mat)
}

gamse.scv <- function(x.star, d, Sigma.star, Hamise, n, binned=FALSE, bin.par, bgridsize, verbose=FALSE, nstage=1, Theta6=FALSE)
{
    if (nstage==0)
    {
        psihat6.star <- psins(r=6, Sigma=Sigma.star, deriv.vec=TRUE) 
    }
    else if (nstage==1)
    {  
        g6.star <- gsamse(Sigma.star, n, 6) 
        G6.star <- g6.star^2 * diag(d)
        if (Theta6) psihat6.star <- kfe(x=x.star, bin.par=bin.par, deriv.order=6, G=G6.star, deriv.vec=FALSE, add.index=FALSE, binned=binned, bgridsize=bgridsize, verbose=verbose)
        else psihat6.star <- kfe(x=x.star, bin.par=bin.par, deriv.order=6, G=G6.star, deriv.vec=TRUE, add.index=FALSE, binned=binned, bgridsize=bgridsize, verbose=verbose)
    }

    if (Theta6)
    {
        derivt6 <- dmvnorm.deriv(x=rep(0,d), deriv.order=6, add.index=TRUE, deriv.vec=FALSE, only.index=TRUE)
        Theta6.mat <- matrix(0, ncol=d, nrow=d)
        Theta6.mat.ind <- Theta6.elem(d)
        for (i in 1:d)
        for (j in 1:d)
        {
            temp <- Theta6.mat.ind[[i]][[j]]
            temp.sum <- 0
            for (k in 1:nrow(temp))
            temp.sum <- temp.sum + psihat6.star[which.mat(temp[k,], derivt6)]
            Theta6.mat[i,j] <- temp.sum 
        }

        eye3 <- diag(d)
        D4 <- dupl(d)$d
        trHamise <- tr(Hamise) 

        ## required constants - see thesis
        Cmu1 <- 1/2*t(D4) %*% vec(Theta6.mat %*% Hamise)
        Cmu2 <- 1/8*(4*pi)^(-d/2) * (2*t(D4)%*% vec(Hamise) + trHamise * t(D4) %*% vec(eye3))
        num <- 2 * (d+4) * sum(Cmu2*Cmu2)
        den <- -(d+2) * sum(Cmu1*Cmu2) + sqrt((d+2)^2 * sum(Cmu1*Cmu2)^2 + 8*(d+4)*sum(Cmu1*Cmu1) * sum(Cmu2*Cmu2))
        gamse <- (num/(den*n))^(1/(d+6)) 
    }
    else
    {  
        ## updated constants using Chacon & Duong (2010) notation
        Cmu1Cmu1 <- drop(1/4*psihat6.star %*% (Hamise %x% diag(d^4) %x% Hamise) %*% psihat6.star)
        Cmu1Cmu2 <- 3/4*(4*pi)^(-d/2)*drop(vec(Hamise %x% diag(d) %x% Hamise) %*% psihat6.star)
        Cmu2Cmu2 <- 1/64*(4*pi)^(-d)*(4*tr(Hamise%*%Hamise) + (d+8)*tr(Hamise)^2)
        num <- 2 * (d+4) * Cmu2Cmu2
        den <- -(d+2) * Cmu1Cmu2 + sqrt((d+2)^2 * Cmu1Cmu2^2 + 8*(d+4)*Cmu1Cmu1 * Cmu2Cmu2)
        gamse <- (num/(den*n))^(1/(d+6)) 
    }
    
    return(gamse)
}

#############################################################################
## Estimate unconstrained G_AMSE pilot bandwidth for SCV for 2 to 6 dim
## Code by J.E. Chacon
##
## Returns
## G_AMSE pilot bandwidth
#############################################################################
Gunconstr.scv <- function(x, binned=FALSE, bin.par, bgridsize, verbose=FALSE, nstage=1, optim.fun="optim")
{
    d <- ncol(x)
    n <- nrow(x)
    S <- var(x)
    S12 <- matrix.sqrt(S)
    optim.fun <- match.arg(optim.fun, c("nlm", "optim"))

    ## stage 1 of plug-in
    x.star <- pre.sphere(x)
    if (nstage==1)
    {  
        G6 <- Gns(r=6,n=n,Sigma=var(x.star), scv=TRUE) 
        psihat6 <- kfe(x=x.star, deriv.order=6, G=G6, deriv.vec=TRUE, add.index=FALSE, binned=binned, bgridsize=bgridsize, verbose=verbose)
    }
    else if (nstage==0)
    {
        psihat6 <- psins(r=6, Sigma=var(x.star), deriv.vec=TRUE) 
    }

    ## constants for normal reference
    D4phi0 <- DrL0(d=d,r=4)
    Id4 <- diag(d^4)

    ## asymptotic squared bias for r = 4
    AB2 <- function(vechG)
    {
        G <- invvech(vechG)%*%invvech(vechG)
        G12 <- matrix.sqrt(G)
        Ginv12 <- chol2inv(chol(G12))
        AB <- n^(-1)*det(Ginv12)*(Kpow(A=Ginv12,pow=4)%*%D4phi0)*2^(-(d+4)/2) + (t(vec(G))%x%Id4)%*%psihat6
        AB2.val <- sum(AB^2)
        return (AB2.val)
    }

    Gstart <- Gns(r=4,n=n,Sigma=S,scv=TRUE)
    Gstart <- matrix.sqrt(Gstart)
    if (optim.fun=="nlm")
    {
        result <- nlm(p=vech(Gstart), f=AB2, print.level=2*as.logical(verbose))    
        G4 <- result$estimate
    }
    else if (optim.fun=="optim")    
    { 
        result <- optim(vech(Gstart), AB2, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
        G4 <- result$par
    }   
    G4 <- invvech(G4)%*%invvech(G4)
    G4 <- S12 %*% G4 %*% S12

    return(G4) 
}

#############################################################################
## Computes the smoothed cross validation function for 2 to 6 dim
## 
## Parameters
## x - data values
## H - bandwidth matrix
## G - pilot bandwidth matrix
##
## Returns
## SCV(H)
#############################################################################
scv.1d <- function(x, h, g, binned=TRUE, bin.par, inc=1, deriv.order=0)
{
    r <- deriv.order
    if (!missing(x)) n <- length(x)
    if (!missing(bin.par)) n <- sum(bin.par$counts)
    scv1 <- kfe.1d(x=x, deriv.order=2*r, bin.par=bin.par, g=sqrt(2*h^2+2*g^2), binned=binned, inc=inc)
    scv2 <- kfe.1d(x=x, deriv.order=2*r, bin.par=bin.par, g=sqrt(h^2+2*g^2), binned=binned, inc=inc)
    scv3 <- kfe.1d(x=x, deriv.order=2*r, bin.par=bin.par, g=sqrt(2*g^2), binned=binned, inc=inc)

    bias2 <- (-1)^r*(scv1 - 2*scv2 + scv3)
    if (bias2 < 0) bias2 <- 0
    scv <- (n*h)^(-1)*(4*pi)^(-1/2)*2^(-r)*OF(2*r) + bias2

    return(scv)
}

scv.mat <- function(x, H, G, binned=FALSE, bin.par, bgridsize, verbose=FALSE, deriv.order=0)
{
    d <- ncol(x)
    n <- nrow(x)
    r <- deriv.order
    vId <- vec(diag(d))
    Hinv <- chol2inv(chol(H))

    if (!binned)
    {
        scv1 <- Qr(x=x, deriv.order=2*r, Sigma=2*H+2*G, inc=1)
        scv2 <- Qr(x=x, deriv.order=2*r, Sigma=H+2*G, inc=1)
        scv3 <- Qr(x=x, deriv.order=2*r, Sigma=2*G, inc=1)
        bias2 <- (-1)^r*(scv1 - 2*scv2 + scv3)
        if (bias2 < 0) bias2 <- 0
    }
    else
    {
        scv1 <- kfe(x=x, G=2*H + 2*G, deriv.order=2*r, inc=1, binned=binned, bin.par=bin.par, bgridsize=bgridsize, verbose=verbose, add.index=FALSE)
        scv2 <- kfe(x=x, G=H + 2*G, deriv.order=2*r, inc=1, binned=binned, bin.par=bin.par, bgridsize=bgridsize, verbose=verbose, add.index=FALSE)
        scv3 <- kfe(x=x, G=2*G, deriv.order=2*r, inc=1, binned=binned, bin.par=bin.par, bgridsize=bgridsize, verbose=verbose, add.index=FALSE)

        bias2 <- drop((-1)^r*Kpow(vId,r) %*% (scv1 - 2*scv2 + scv3))
        if (bias2 < 0) bias2 <- 0
    }
    scvmat <- 1/(det(H)^(1/2)*n)*nur(r=r, A=Hinv, mu=rep(0,d), Sigma=diag(d), type="direct")*2^(-d-r)*pi^(-d/2) + bias2
    
    return (scvmat)
}

#############################################################################
## Find the bandwidth that minimises the SCV for 1 to 6 dim
## 
## Parameters
## x - data values
## pre - "scale" - pre-scaled data
##     - "sphere"- pre-sphered data
## Hstart - initial bandwidth matrix
##
## Returns
## H_SCV
#############################################################################
hscv <- function(x, nstage=2, binned=TRUE, bgridsize, plot=FALSE)
{
    sigma <- sd(x)
    n <- length(x)
    d <- 1
    hnorm <- sqrt((4/(n*(d + 2)))^(2/(d + 4)) * var(x))
    if (missing(binned)) binned <- default.bflag(d=d,n=n)
    if (missing(bgridsize)) bgridsize <- default.bgridsize(d)
    hmin <- 0.1*hnorm
    hmax <- 2*hnorm

    bin.par <- binning(x=x, bgridsize=bgridsize, h=hnorm)
    if (nstage==1)
    {
        psihat6 <- psins.1d(r=6, sigma=sigma)
        psihat10 <- psins.1d(r=10, sigma=sigma)
    }
    else if (nstage==2)
    {
        g1 <- (2/(7*n))^(1/9)*2^(1/2)*sigma
        g2 <- (2/(11*n))^(1/13)*2^(1/2)*sigma

        psihat6 <- kfe.1d(x=x, bin.par=bin.par, binned=binned, deriv.order=6, g=g1, inc=1)
        psihat10 <- kfe.1d(x=x, bin.par=bin.par, binned=binned, deriv.order=10, g=g2, inc=1)
    }

    g3 <- (-6/((2*pi)^(1/2)*psihat6*n))^(1/7) 
    g4 <- (-210/((2*pi)^(1/2)*psihat10*n))^(1/11)
    psihat4 <- kfe.1d(x=x, bin.par=bin.par, binned=binned, deriv.order=4, g=g3, inc=1)
    psihat8 <- kfe.1d(x=x, bin.par=bin.par, binned=binned, deriv.order=8, g=g4, inc=1)

    C <- (441/(64*pi))^(1/18) * (4*pi)^(-1/5) * psihat4^(-2/5) * psihat8^(-1/9)

    scv.1d.temp <- function(h)
    {
        return(scv.1d(x=x, bin.par=bin.par, h=h, g=C*n^(-23/45)*h^(-2), binned=binned, inc=1))
    }

    if (plot)
    {  
        hseq <- seq(hmin,hmax, length=400)
        hscv.seq <- rep(0, length=length(hseq))
        for (i in 1:length(hseq))
        hscv.seq[i] <- scv.1d.temp(hseq[i])
        plot(hseq, hscv.seq, type="l", xlab="h", ylab="SCV(h)")
    }

    opt <- optimise(f=scv.1d.temp, interval=c(hmin, hmax))$minimum
    if (n >= 1e5) warning("hscv is not always stable for large samples")

    return(opt)
}

Hscv <- function(x, nstage=2, pre="sphere", pilot, Hstart, binned, bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
{
    n <- nrow(x)
    d <- ncol(x)
    r <- deriv.order
    if (missing(binned)) binned <- default.bflag(d=d,n=n)
    if (d > 4) binned <- FALSE
    if (missing(bgridsize)) bgridsize <- default.bgridsize(d)
 
    if(!is.matrix(x)) x <- as.matrix(x)
    if (missing(pilot)) { if (d==2 & r==0) pilot <- "samse" else pilot <- "dscalar" }
    pilot1 <- match.arg(pilot, c("amse", "samse", "unconstr", "dunconstr", "dscalar"))
    pre1 <- match.arg(pre, c("scale", "sphere"))
    optim.fun <- match.arg(optim.fun, c("nlm", "optim"))

    if (pilot1=="amse" & (d>2 | r>0)) stop("amse pilot selectors not defined for d>2 and/or r>0")
    if ((pilot1=="samse" | pilot1=="unconstr") & r>0) stop("dscalar or dunconstr pilot selectors are better for deriv.order>0")

    if (pre1=="scale")
    {
        x.star <- pre.scale(x)
        S12 <- diag(sqrt(diag(var(x))))
        Sinv12 <- chol2inv(chol(S12))
    }
    else if (pre1=="sphere")
    {
        x.star <- pre.sphere(x)
        S12 <- matrix.sqrt(var(x))
        Sinv12 <- chol2inv(chol(S12))
    }

    RK <- (4*pi)^(-d/2)
    if (binned)
    {
        if (pilot1=="unconstr" | pilot1=="dunconstr")
        H.max <- (((d+8)^((d+6)/2)*pi^(d/2)*RK)/(16*(d+2)*n*gamma(d/2+4)))^(2/(d+4))* var(x)
        else
        H.max <- (((d+8)^((d+6)/2)*pi^(d/2)*RK)/(16*(d+2)*n*gamma(d/2+4)))^(2/(d+4))* var(x.star)
        if (default.bflag(d=d, n=n))
        bin.par <- binning(x=x.star, bgridsize=bgridsize, H=matrix.sqrt(H.max))
    }
    if (pilot1=="unconstr")
    {
        ## Gu pilot matrix is on data scale
        Gu <- Gunconstr.scv(x=x, binned=binned, bgridsize=bgridsize, verbose=verbose, nstage=nstage-1, optim.fun=optim.fun)
        if (missing(Hstart)) Hstart <- Hns(x=x, deriv.order=r)
    }
    else if (pilot1=="dunconstr")
    {
        ## Gu pilot matrix is on data scale
        Gu <- Gdunconstr(x=x, d=d, r=r, n=n, nstage=nstage, verbose=verbose, binned=binned, scv=TRUE, optim.fun=optim.fun)
        if (missing(Hstart)) Hstart <- Hns(x=x, deriv.order=r)
    }
    else if (pilot1=="dscalar")
    {
        ## Gs is on pre-transformed data scale
        g2r4 <- gdscalar(x=x.star, d=d, r=r, n=n, nstage=nstage, verbose=verbose, scv=TRUE, binned=binned)
        Gs <- g2r4^2*diag(d)
        if (missing(Hstart)) Hstart <- Hns(x=x.star, deriv.order=r)
    }
    else
    {
        ## Gs is on transformed data scale    
        Hamise <- Hpi(x=x.star, nstage=1, deriv.order=r, pilot=pilot, pre="sphere", binned=TRUE, bgridsize=bgridsize, verbose=verbose, optim.fun=optim.fun) 
        if (any(is.na(Hamise)))
        {
            warning("Pilot bandwidth matrix is NA - replaced with maximally smoothed")
            Hamise <- (((d+8)^((d+6)/2)*pi^(d/2)*RK)/(16*(d+2)*n*gamma(d/2+4)))^(2/(d+4))* var(x.star)
        }
        gs <- gamse.scv(x.star=x.star, d=d, Sigma.star=var(x.star), Hamise=Hamise, n=n, binned=binned, bgridsize=bgridsize, verbose=verbose, nstage=nstage-1)
        Gs <- gs^2*diag(d)

        ## use normal reference bandwidth as initial condition 
        if (missing(Hstart)) Hstart <- Hns(x=x.star, deriv.order=r) 
        else Hstart <- Sinv12 %*% Hstart %*% Sinv12
    }

    ## SCV is estimate of AMISE
    scv.mat.temp <- function(vechH)
    {
        H <- invvech(vechH) %*% invvech(vechH)
        if (pilot1=="samse" | pilot1=="amse" | pilot1=="dscalar") { Gpilot <- Gs; xx <- x.star }
        else if (pilot1=="unconstr" | pilot1=="dunconstr") { Gpilot <- Gu; xx <- x }

        if (default.bflag(d=d, n=n))
        scvm <- scv.mat(x=xx, H=H, G=Gpilot, binned=binned, bin.par=bin.par, verbose=FALSE, deriv.order=r)
        else
        scvm <- scv.mat(x=xx, H=H, G=Gpilot, binned=binned, verbose=FALSE, deriv.order=r)
        return(scvm)
    }

    Hstart <- matrix.sqrt(Hstart)
    if (optim.fun=="nlm")
    {
        result <- nlm(p=vech(Hstart), f=scv.mat.temp, print.level=2*as.numeric(verbose))    
        H <- invvech(result$estimate) %*% invvech(result$estimate)
        amise.star <- result$minimum
    }
    else if (optim.fun=="optim")
    {
        result <- optim(vech(Hstart), scv.mat.temp, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
        H <- invvech(result$par) %*% invvech(result$par)
        amise.star <- result$value
    }
    if (!(pilot1 %in% c("dunconstr","unconstr")))  H <- S12 %*% H %*% S12   ## back-transform

    if (!amise) return(H)
    else return(list(H = H, SCV.star=amise.star))
}

Hscv.diag <- function(x, nstage=2, pre="scale", pilot, Hstart, binned, bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
{
    n <- nrow(x)
    d <- ncol(x)
    r <- deriv.order
    RK <- (4*pi)^(-d/2)
    if (missing(binned)) binned <- default.bflag(d=d,n=n)
    if (d > 4) binned <- FALSE
    if (missing(bgridsize)) bgridsize <- default.bgridsize(d)
    if(!is.matrix(x)) x <- as.matrix(x)

    if (missing(pilot)) { if (d==2 & r==0) pilot <- "samse" else pilot <- "dscalar" }
    pilot1 <- match.arg(pilot, c("amse", "samse", "unconstr", "dunconstr", "dscalar"))
    pre1 <- match.arg(pre, c("scale", "sphere"))
    optim.fun <- match.arg(optim.fun, c("nlm", "optim"))

    if (pilot1=="amse" & (d>2 | r>0)) stop("samse pilot selectors are better for higher dimensions and/or deriv.order>0")
    if (pilot1=="samse" & r>0) stop("dscalar pilot selectors are better for deriv.order>0")
    if (pilot1=="unconstr" | pilot1=="dunconstr") stop("Unconstrained pilot selectors are not suitable for Hscv.diag")

    if (pre1=="sphere") stop("Using pre-sphering doesn't give a diagonal bandwidth matrix")

    if (pre1=="scale")
    {
        x.star <- pre.scale(x)
        S12 <- diag(sqrt(diag(var(x))))
        Sinv12 <- chol2inv(chol(S12))
    }
    else if (pre1=="sphere")
    {
        x.star <- pre.sphere(x)
        S12 <- matrix.sqrt(var(x))
        Sinv12 <- chol2inv(chol(S12))
    }

    if (binned)
    {
        H.max <- (((d+8)^((d+6)/2)*pi^(d/2)*RK)/(16*(d+2)*n*gamma(d/2+4)))^(2/(d+4))* var(x.star)
        bin.par.star <- binning(x=x.star, bgridsize=bgridsize, H=H.max) 
    }

    if (pilot1=="dscalar")
    {
        ## Gs is on pre-transformed data scale
        g2r4 <- gdscalar(x=x.star, r=r, n=n, d=d, verbose=verbose, nstage=nstage, scv=TRUE, binned=binned)
        Gs <- g2r4^2*diag(d)
        if (missing(Hstart)) Hstart <- Hns(x=x.star, deriv.order=r)
    }
    else
    {
          ## Gs is on transformed data scale
          Hamise <- Hpi(x=x.star, nstage=1, pilot=pilot, pre="sphere", binned=binned, bgridsize=bgridsize, verbose=verbose, optim.fun=optim.fun) 
          if (any(is.na(Hamise)))
          {
            warning("Pilot bandwidth matrix is NA - replaced with maximally smoothed")
            Hamise <- (((d+8)^((d+6)/2)*pi^(d/2)*RK)/(16*(d+2)*n*gamma(d/2+4)))^(2/(d+4))* var(x.star)
          }
          gs <- gamse.scv(x.star=x.star, d=d, Sigma.star=var(x.star), Hamise=Hamise, n=n, binned=binned, bgridsize=bgridsize, verbose=verbose, nstage=nstage-1)

          Gs <- gs^2*diag(d)

        ## use normal reference bandwidth as initial condition 
        if (missing(Hstart)) Hstart <- Hns(x=x.star, deriv.order=r)
        else Hstart <- Sinv12 %*% Hstart %*% Sinv12
    }

    scv.mat.temp <- function(diagH)
    {
        H <- diag(diagH) %*% diag(diagH)
        if (default.bflag(d=d, n=n)) scvm <- scv.mat(x.star, H, Gs, binned=binned, verbose=FALSE, bin.par=bin.par.star, deriv.order=r)
        else scvm <- scv.mat(x.star, H, Gs, binned=binned, verbose=FALSE, deriv.order=r)
        return(scvm)
    }

    Hstart <- matrix.sqrt(Hstart)
    if (optim.fun=="nlm")
    {
        result <- nlm(p=diag(Hstart), f=scv.mat.temp, print.level=2*as.numeric(verbose))    
        H <- diag(result$estimate) %*% diag(result$estimate)
        amise.star <- result$minimum
    }
    else if (optim.fun=="optim")
    {  
        result <- optim(diag(Hstart), scv.mat.temp, method="Nelder-Mead", control=list(trace=as.numeric(verbose), REPORT=1))
        H <- diag(result$par) %*% diag(result$par)
        amise.star <- result$value
    }
    ## back-transform
    H <- S12 %*% H %*% S12

    if (!amise) return(H)
    else return(list(H = H, SCV.star=amise.star))
}

############################################################################
## Normal scale selector H_ns for kernel density derivate estimators
############################################################################

Hns <- function(x, deriv.order=0)
{
    if (is.vector(x)) { n <- 1; d <- length(x) } 
    else { n <- nrow(x); d <- ncol(x) }
    r <- deriv.order
    H <- (4/(n*(d+2*r+2)))^(2/(d+2*r+4)) * var(x)
    return(H)
}

Hns.diag <- function(x)
{
    if (is.vector(x)) { n <- 1; d <- length(x) } 
    else { n <- nrow(x); d <- ncol(x) }
    S <- var(x)
    Delta <- chol2inv(chol(diag(diag(S))))%*%S
    svD <- svd(Delta)
    Deltainv <- svD$v %*% diag(1/svD$d) %*% t(svD$u)
    H <- ((4*d*det(Delta)^(1/2))/(2*tr(Deltainv%*%Deltainv) + (tr(Deltainv))^2))^(2/(d+4))*diag(diag(S))*n^(-2/(d+4))
    
    return(H)
}

hns <- function(x, deriv.order=0)
{
    n <- length(x)
    d <- 1
    r <- deriv.order
    h <- (4/(n*(d+2*r+2)))^(1/(d+2*r+4))*sd(x)

    return(h)
}

#######################################################################
## Normal scale G_ns for kernel functional estimators
## scv = flag for SCV bias annihiliation constant
#######################################################################

Gns <- function(r,n,Sigma,scv=FALSE)
{
    d <- ncol(Sigma)
    const <- ifelse(scv,1,2)
    G <- (2/((n*(d+r))))^(2/(d+r+2))*const*Sigma

    return(G)
}

Gns.search <- function(G, f, n=10)
{
    Gstart.vec <- numeric(n)
    for (i in 1:length(Gstart.vec)) 
    { 
        Gstart <- matrix.sqrt(i/length(Gstart.vec)*G)
        Gstart.vec[i] <- f(vech(Gstart))
    }
    i <- which.min(Gstart.vec)/length(Gstart.vec)
    Gstart <- i*G

    return(Gstart)
}

############################################################################
## Normal mixture selector 
############################################################################

Hnm <- function(x, deriv.order=0, G=1:9, subset.ind, mise.flag=FALSE, verbose=FALSE, ...)
{
    if (!requireNamespace("mclust", quietly=TRUE)) stop("Install the mclust package as it is required.", call.=FALSE)
    modelNames <- "VVV" 
    ## set local seed
    current.seed <- .Random.seed
    on.exit( { .Random.seed <<- current.seed } )
    set.seed(8192)

    if (!missing(subset.ind)) nmixt.fit <- mclust::Mclust(x[subset.ind,], G=G, verbose=verbose, modelNames=modelNames, ...)
    else nmixt.fit <- mclust::Mclust(x, G=G, verbose=verbose, modelNames=modelNames, ...)

    if (is.vector(x)) { d <- length(x); n <- 1 } else { d <- ncol(x); n <- nrow(x) } 
    mus <- t(nmixt.fit$parameters$mean)
    Sigmas <- matrix(nmixt.fit$parameters$variance$sigma, byrow=TRUE, ncol=d)
    props <- nmixt.fit$parameters$pro
    if (mise.flag) H.nm <- Hmise.mixt(samp=n, mus=mus, Sigmas=Sigmas, props=props, deriv.order=deriv.order)
    else H.nm <- Hamise.mixt(samp=n, mus=mus, Sigmas=Sigmas, props=props, deriv.order=deriv.order)

    return(H.nm)
}

Hnm.diag <- function(x, deriv.order=0, G=1:9, subset.ind, mise.flag=FALSE, verbose=FALSE, ...)
{
    if (!requireNamespace("mclust", quietly=TRUE)) stop("Install the mclust package as it is required.", call.=FALSE)
    modelNames <- "VVI" 
    if (!missing(subset.ind)) nmixt.fit <- mclust::Mclust(x[subset.ind,], G=G, verbose=verbose, modelNames=modelNames, ...)
    else nmixt.fit <- mclust::Mclust(x, G=G, verbose=verbose, modelNames=modelNames, ...)

    if (is.vector(x)) { d <- length(x); n <- 1 } else { d <- ncol(x); n <- nrow(x) } 
    mus <- t(nmixt.fit$parameters$mean)
    Sigmas <- matrix(nmixt.fit$parameters$variance$sigma, byrow=TRUE, ncol=d)
    props <- nmixt.fit$parameters$pro
    if (mise.flag) H.nm <- Hmise.mixt.diag(samp=n, mus=mus, Sigmas=Sigmas, props=props, deriv.order=deriv.order)
    else H.nm <- Hamise.mixt.diag(samp=n, mus=mus, Sigmas=Sigmas, props=props, deriv.order=deriv.order)

    return(H.nm)
}

hnm <- function(x, deriv.order=0, G=1:9, subset.ind, mise.flag=FALSE, verbose=FALSE, ...)
{
    if (!missing(subset.ind)) nmixt.fit <- mclust::Mclust(x[subset.ind], G=G, verbose=verbose, ...)
    else nmixt.fit <- mclust::Mclust(x, G=G, verbose=verbose, ...)

    mus <- nmixt.fit$parameters$mean
    sigmas <- sqrt(nmixt.fit$parameters$variance$sigma)
    props <- nmixt.fit$parameters$pro
    n <- length(x)
    if (mise.flag) h.nm <- hmise.mixt(samp=n, mus=mus, sigmas=sigmas, props=props, deriv.order=deriv.order)
    else h.nm <- hamise.mixt(samp=n, mus=mus, sigmas=sigmas, props=props, deriv.order=deriv.order)

    return(h.nm)
}

#############################################################################
## Histogram density estimators
#############################################################################
histde <- function(x, binw, xmin, xmax, adj=0) 
{
    if (!is.null(names(x))) xnames <- names(x) else xnames <- colnames(x)
    if (is.vector(x))
    {
        d <- 1; n <- length(x)
        if (missing(binw))
        binw <- 2*3^(1/(d+2))*pi^(2/(2*d+4))*sd(x)*length(x)^(-1/(d+2))
        nbin <- round(diff(range(x))*1.2/binw, 0)
    }
    else
    {
        d <- ncol(x); n <- nrow(x)
        if (missing(binw))
            binw <- 2*3^(1/(d+2))*pi^(2/(2*d+4))*apply(x, 2, sd)*nrow(x)^(-1/(d+2))
        nbin <- round(apply(apply(x, 2, range), 2, diff)*1.2/binw,0)     
    }
    
    if (d==1) fhat <- hist.1d(x, binw=binw, xmin=xmin, xmax=xmax, adj=adj)
    else if (d==2) fhat <- hist.2d(x=x, binw=binw, xmin=xmin, xmax=xmax, adj=adj)

    fhat$names <- parse.name(x)
    if (inherits(fhat$eval.points, "list")) names(fhat$eval.points) <- xnames 
    class(fhat) <- "histde"

    return(fhat)
 }

## 1D histogram
hist.1d <- function(x, nbin, binw, x.cut, xmin, xmax, adj=0, ...)
{
    if (missing(xmin)) xmin <- min(x)
    if (missing(xmax)) xmax <- max(x)
    if (missing(nbin)) nbin <- round(diff(range(x))*1.2/binw, 0)
    if (missing(x.cut)) x.cut <- seq(from=xmin-0.1*(xmax-xmin), to=xmax+0.1*(xmax-xmin), length=nbin+1)
    if (missing(binw)) binw <- diff(x.cut)[1]
    x.cut <- x.cut + adj*binw

    hs <- hist(x=x, breaks=x.cut, plot=FALSE) 
    hs <- list(x=x, estimate=hs$density, eval.points=hs$breaks, binw=binw, nbin=nbin)
    class(hs) <- "histde"

    return(hs)
}

## 2D histogram
hist.2d <- function(x, nbin, binw, x.cut, xmin, xmax, adj=0, ...)
{ 
    if (missing(nbin)) nbin <- round(apply(apply(x, 2, range), 2, diff)*1.2/binw,0)
    if (length(nbin)==1) nbin <- rep(nbin,2)
    if (missing(xmin)) xmin <- apply(x, 2, min)
    if (missing(xmax)) xmax <- apply(x, 2, max)
    if (missing(x.cut)) x.cut <- list(seq(from=xmin[1]-0.1*(xmax[1]-xmin[1]), to=xmax[1]+0.1*(xmax[1]-xmin[1]), length=nbin[1]+1), seq(from=xmin[2]-0.1*(xmax[2]-xmin[2]), to=xmax[2]+0.1*(xmax[2]-xmin[2]), length=nbin[2]+1))
    if (missing(binw)) binw <- c(diff(x.cut[[1]])[1], diff(x.cut[[2]])[1])
    x.cut[[1]] <- x.cut[[1]] + adj*binw[1]
    x.cut[[2]] <- x.cut[[2]] + adj*binw[2] 
    index.x <- cut(x[,1], x.cut[[1]], include.lowest=TRUE)
    index.y <- cut(x[,2], x.cut[[2]], include.lowest=TRUE)

    m <- matrix(0, nrow=nbin[1], ncol=nbin[2], dimnames=list(levels(index.x), levels(index.y)))
    for (i in 1:length(index.x))
        m[index.x[i], index.y[i]] <- m[index.x[i], index.y[i]] + 1

    hs2d <- list(x=x, estimate=m/(nrow(x)*prod(binw)), eval.points=x.cut, binw=binw, nbin=nbin)
    class(hs2d) <- "histde"

    return(hs2d)
}

#############################################################################
## plot histograms
#############################################################################
plot.histde <- function(x, ...)
{
    if (is.vector(x$x)) plothistde.1d(fhat=x, ...)
    else plothistde.2d(fhat=x, ...)
    invisible()
}

plothistde.1d <- function(fhat, xlab, ylab="Density function", add=FALSE, drawpoints=FALSE, col="transparent", col.pt=4, jitter=FALSE, border=1, alpha=1, ...)
{
    if (missing(xlab)) xlab <- fhat$names
    col <- transparency.col(col, alpha=alpha)
    if (!add) plot(fhat$eval.points, c(fhat$estimate,0), type="n",  xlab=xlab, ylab=ylab, ...)
    rect(fhat$eval.points[-length(fhat$eval.points)], 0, fhat$eval.points[-1], fhat$estimate, border=border, col=col, ...)
    if (drawpoints)
        if (jitter) rug(jitter(fhat$x), col=col.pt)
        else rug(fhat$x, col=col.pt)

}

plothistde.2d <- function(fhat, breaks, nbreaks=11, xlab, ylab, zlab="Density function", cex=1, pch=1, add=FALSE, drawpoints=FALSE, col, col.fun, alpha=1, col.pt=4, lty.rect=2, cex.text=1, border, lwd.rect=1, col.rect="transparent", add.grid=TRUE, ...)
{
    if (missing(xlab)) xlab <- fhat$names[1]
    if (missing(ylab)) ylab <- fhat$names[2]
    if (missing(border)) border <- grey(0.5)
    if (!add) plot(fhat$x, col=col.pt, type="n", xlab=xlab, ylab=ylab, ...)
    if (missing(breaks)) breaks <- seq(min(fhat$estimate,0), max(fhat$estimate)+0.1*diff(range(fhat$estimate)), length=nbreaks)
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n, palette="heat", rev=TRUE, alpha=alpha) } 
    if (missing(col)) col <- col.fun(n=length(breaks))
    col <- transparency.col(col, alpha=alpha)
    
    for (i in 1:(nrow(fhat$estimate)))
        for (j in 1:(ncol(fhat$estimate)))
        {
            if (fhat$estimate[i,j]>=breaks[2])
            {
                rect(fhat$eval.points[[1]][i], fhat$eval.points[[2]][j], fhat$eval.points[[1]][i+1], fhat$eval.points[[2]][j+1], col=col[findInterval(fhat$estimate[i,j], breaks)], lty=lty.rect, border=border, lwd=lwd.rect)
            }
        }
    if (add.grid)
    {
        for (i in 1:length(fhat$eval.points[[1]]))
            lines(rep(fhat$eval.points[[1]][i],2), range(fhat$eval.points[[2]]), col=border, ...)
        for (j in 1:length(fhat$eval.points[[2]]))
            lines(range(fhat$eval.points[[1]]), rep(fhat$eval.points[[2]][j],2), col=border, ...)
    }
    
    if (drawpoints) 
        points(fhat$x[,1], fhat$x[,2], col=col.pt, cex=cex, pch=pch)
}

## predict method
predict.histde <- function(object, ..., x)
{
    fhat <- object$estimate
    if (!is.list(object$eval.points)) 
        d <- 1
    else d <- length(object$eval.points)
    if (d==1) 
    {
        gs <- length(object$eval.points)
        x.ind <- findInterval(x, object$eval.points, all.inside = FALSE)
        fhat[x.ind==0] <- object$estimate[1]
        fhat[x.ind==gs] <- object$estimate[gs]
    }
    else 
    {
        x <- matrix(x, ncol = d)
        gs <- sapply(object$eval.points, length)
        x.ind <- matrix(0, nrow=nrow(x), ncol=d)
        for (i in 1:d) x.ind[, i] <- findInterval(x[, i], object$eval.points[[i]], all.inside = FALSE)
        x.ind[x.ind==0] <- 1
        x.ind.flag <- x.ind==1
        for (i in 1:d) x.ind.flag[, i] <- x.ind.flag[, i] | x.ind[, i] ==gs[i]
    }

    return(fhat[x.ind])
}

## contourLevels method
contourLevels.histde <- function(...) { return(contourLevels.kde(...)) }

######################################################################
## Deconvolution KDE
######################################################################
dckde <- function(...) { return (kdcde(...)) }

kdcde <- function(x, H, h, Sigma, sigma, reg, bgridsize, gridsize, binned, verbose=FALSE, ...)
{
    ## default values 
    ksd <- ks.defaults(x=x, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n 
    binned <- ksd$binned
    gridsize <- ksd$gridsize
    bgridsize <- ksd$bgridsize

    x <- as.matrix(x)
    if (d==1 & missing(h)) h <- hpi(x=x, nstage=2, binned=binned, deriv.order=0)
    if (d>1 & missing(H)) H <- Hpi(x=x, nstage=2, binned=binned, deriv.order=0)

    if (d==1) stop("dckde not yet implemented for d=1")
    if (missing(reg)) reg <- reg.ucv(x=x, H=H, h=h, Sigma=Sigma, sigma=sigma, k=5, d=d, binned=binned, verbose=verbose)

    ## Deconvolution KDE is weighted KDE with non-uniform weights
    w <- dckde.weights(x=x, H=H, Sigma=Sigma, reg=reg)
    fhat <- kde(x=x, H=H, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize, verbose=verbose, ...)
    fhat$reg <- reg

    return(fhat)
}

## weights for deconvolution KDE
## code adapted from DeconWK 0.6-5, author B. Turlach
## R-forge website: https://r-forge.r-project.org/R/?group_id=630
dckde.weights <- function(x, Sigma, H, reg)
{
    n <- nrow(x)
    d <- ncol(x)
    Qmat <- matrix(0, ncol=n, nrow=n)
    bvec <- rep(0, n)
    
    for (j in 1:n)
    {  
        Qmat[j,] <- dmvnorm.mixt(x, mus=x[j,], Sigmas=2*H + 2*Sigma, props=1)
        bvec[j] <- sum(dmvnorm.mixt(x, mus=x[j,], Sigmas=2*H + Sigma, props=1))
    }
    
    if(!missing(reg)) diag(Qmat) <- diag(Qmat) + reg/n
    bvec <- bvec/n
    
    val <- kernlab::ipop(c=-bvec, H=Qmat, A=rep(1,n), b=1, r=0, l=rep(0,n), u=rep(1, n)) 
    w <- kernlab::primal(val)
    w <- w/sum(w)*n
    return(w)  
}

## unbiased k-fold cross validation choice of regularisation penalty (gamma)
reg.ucv <- function(x, H, h, Sigma, sigma, k=5, d, binned=FALSE, verbose=FALSE)
{
    gamma.ucv.temp <- function(gamma) { return(-reg.ucv.val(x=x, H=H, Sigma=Sigma, k=k, reg=gamma^2, binned=binned)) }
    gamma.val <- nlm(f=gamma.ucv.temp, p=0.1, print.level=2*as.numeric(verbose))$estimate^2
    
    return(gamma.val)
}

## k-fold UCV value for regularisation penalty
reg.ucv.val <- function(x, Sigma, H, reg, k=5, binned=FALSE)
{
    n <- nrow(x)
    n.seq <- block.indices(n, n, npergroup=round(n/k))
    cv.val <- 0
    for (j in 1:(length(n.seq)-1))
    {
        iind <- n.seq[j]:(n.seq[j+1]-1)
        w <- dckde.weights(x=x[-iind,], H=H, Sigma=Sigma, reg=reg)  
        fhat <- kde(x=x[-iind,], H=H+Sigma, w=w, binned=binned)
        cv.val <- cv.val + sum(predict(fhat, x=x[iind,]))
    }
    return(cv.val)
}

##############################################################################
## Exact MISE for normal mixtures
##############################################################################
## nu, gamma_r, gamma_r2 written by Jose Chacon 10/2008
nu <- function(r, A)
{ 
    ## using the recursive formula provided in Kan (2008)
    ei <- eigen(A)$values
    tr.vec <- numeric(r)
    for(p in 1:r)
    tr.vec[p] <- sum(ei^p)

    nu.val <- 1
    if (r>=1)
    {
        for(p in 1:r)
        {
            a <- sum(tr.vec[1:p]*rev(nu.val))/(2*p)
            nu.val <- c(nu.val,a)
        }
    } 
    return(factorial(r)*2^r*nu.val[r+1])   
}   

nu.rs <- function(r, s, A, B)
{
    if (s==0) nu.val <- nu(r=r, A=A)
    else if (s>=1)
    {
        nu.val <- 0
        for (i in 0:r)
            for (j in 0:(s-1))
                nu.val <- nu.val + choose(r,i)*choose(s-1,j) *factorial(r+s-i-j-1)*2^(r+s-i-j-1)*tr(matrix.pow(A,r-i)%*%matrix.pow(B,s-j))*nu.rs(r=i,s=j, A=A, B=B)
    }

    return(nu.val)
}

## gamma functional for normal mixture MISE
gamma_r <- function(mu, Sigma, r)
{
    Sigmainv <- chol2inv(chol(Sigma))
    d <- ncol(Sigma)
    v <- 0
    for (j in 0:r)
    v <- v + (-1)^j*choose(2*r, 2*j)*OF(2*j)*nu.rs(r=r-j, s=j, A=Sigmainv%*%mu%*%t(mu)%*%Sigmainv, B=Sigmainv)
    v <- (-1)^r*v*drop(dmvnorm.deriv(x=rep(0,d), mu=mu, Sigma=Sigma,deriv.order=0)/OF(2*r))
  
    return(v)
}

## gamma functional for normal mixture AMISE 
gamma_r2 <- function(mu, Sigma, d, r, H)
{
    Sigmainv <- chol2inv(chol(Sigma))

    if (d==1)
    w <- vec(Kpow(Sigmainv %*% Sigmainv, r)) %x% vec(Kpow(Sigmainv %*% H %*% Sigmainv, 2))
    else
    w <- matrix(Sdrv(d=d,r=2*r+4, v=vec(Kpow(Sigmainv %*% Sigmainv, r)) %x% vec(Kpow(Sigmainv %*% H %*% Sigmainv, 2))), nrow=1)

    v <- rep(0,length=d^(2*r+4))
    for(j in 0:(r+2))
    v <- v+((-1)^j*OF(2*j)*choose(2*r+4, 2*j))*(Kpow(mu,2*r-2*j+4)%x%Kpow(vec(Sigma),j))

    gamr <- (-1)^r*mvtnorm::dmvnorm(mu,mean=rep(0,d),sigma=Sigma)*sum(w %*% v)

    return(gamr)
}

##############################################################################
## Omega matrices (for exact MISE for normal mixtures)
##
## Parameters 
## mus - means
## Sigmas - variances
## k - number of mixture components
## a - subscript of Omega matrix
## H - bandwidth matrix
##
## Returns 
## Omega matrix
##############################################################################
omega <- function(mus, Sigmas, k, a, H, r)
{
    ## the (i,j) element of Omega matrix is
    ## dmvnorm(0, mu_i - mu_j, a*H + Sigma_i + Sigma_j)
    if (is.matrix(mus)) d <- ncol(mus)
    else d <- length(mus)

    if (k == 1)
        omega.mat <- gamma_r(mu=rep(0,d),Sigma=a*H + 2*Sigmas, r=r)  
    else
    {
        omega.mat <- matrix(0, nrow=k, ncol=k)
        for (i in 1:k)
        {
            Sigmai <- Sigmas[((i-1)*d+1):(i*d),]
            mui <- mus[i,]
            for (j in 1:k)
            {
                Sigmaj <- Sigmas[((j-1)*d+1):(j*d),]
                muj <- mus[j,]    
                omega.mat[i,j] <- gamma_r(mu=mui-muj, Sigma=a*H + Sigmai + Sigmaj, r=r) 
            }
        }
    }

    return(omega.mat)
}

omega.1d <- function(mus, sigmas, k, a, h, r)
{
    H <- h^2
    Sigmas <- sigmas^2

    if (k == 1)
        omega.mat <- gamma_r(mu=0, Sigma=as.matrix(a*H + 2*Sigmas), r=r)  
    else
    {   
        omega.mat <- matrix(0, nrow=k, ncol=k)
        for (i in 1:k)
        {
            Sigmai <- Sigmas[i]
            mui <- mus[i]
            for (j in 1:k)
            {
                Sigmaj <- Sigmas[j]
                muj <- mus[j]    
                omega.mat[i,j] <- gamma_r(mu=mui-muj, Sigma=as.matrix(a*H + Sigmai + Sigmaj),  r=r) 
            }
        }
    }
  
    return(omega.mat)
}

#############################################################################
## Exact MISE for normal mixtures
##
## Parameters
## mus - means
## Sigmas - variances
## Props - vector of proportions of each mixture component 
## H - bandwidth matrix
## samp - sample size
##
## Returns
## Exact MISE for normal mixtures
##############################################################################
mise.mixt <- function(H, mus, Sigmas, props, samp, h, sigmas, deriv.order=0)
{
    if (!(missing(h)))
    return(mise.mixt.1d(h=h, mus=mus, sigmas=sigmas, props=props, samp=samp, deriv.order=deriv.order)) 

    if (is.vector(mus)) d <- length(mus)
    else d <- ncol(mus)
    k <- length(props)
    r <- deriv.order

    Hinv <- chol2inv(chol(H))
    ## formula is found in Wand & Jones (1993) and Chacon, Duong & Wand (2008)
    if (k == 1) 
    {
        mise <- 2^(-r)*nu(r,Hinv)/(samp * (4 * pi)^(d/2) * sqrt(det(H))) + 
                (1-1/samp)*omega(mus, Sigmas, 1, 2, H, r) - 2*omega(mus, Sigmas, 1, 1, H, r) + omega(mus, Sigmas, 1, 0, H, r)
    }
    else
    {
        mise <- 2^(-r)*nu(r,Hinv)/(samp * (4 * pi)^(d/2) * sqrt(det(H))) +
            props %*% ((1-1/samp)*omega(mus, Sigmas, k, 2, H, r) - 2*omega(mus, Sigmas, k, 1, H, r) +  omega(mus, Sigmas, k, 0, H, r)) %*% props
    }

    return(drop(mise)) 
}

mise.mixt.1d <- function(h, mus, sigmas, props, samp, deriv.order=0)
{  
    d <- 1
    k <- length(props)
    r <- deriv.order
    H <- as.matrix(h^2)
    Hinv <- chol2inv(chol(H))
    ## formula is found in Wand & Jones (1993) and Chacon, Duong & Wand (2008)
    if (k == 1) 
    {
        mise <- 2^(-r)*nu(r,Hinv)/(samp * (4 * pi)^(d/2) * sqrt(det(H))) + 
            (1-1/samp)*omega.1d(mus, sigmas, 1, 2, h, r) -
            2*omega.1d(mus, sigmas, 1, 1, h, r) +
            omega.1d(mus, sigmas, 1, 0, h, r)
    }
    else
    {
        mise <- 2^(-r)*nu(r,Hinv)/(samp * (4 * pi)^(d/2) * sqrt(det(H))) +
            props %*% ((1-1/samp)*omega.1d(mus, sigmas, k, 2, h, r) - 
            2*omega.1d(mus, sigmas, k, 1, h, r) + 
            omega.1d(mus, sigmas, k, 0, h, r)) %*% props
    }

    return(drop(mise)) 
}

##############################################################################
## Exact AMISE for bivariate normal mixtures
##
## Parameters
## mus - means
## Sigmas - variances
## props - mixing proportions 
## H - bandwidth matrix
## samp - sample size
##
## Returns   
## Exact AMISE for normal mixtures
##############################################################################
amise.mixt <- function(H, mus, Sigmas, props, samp, h, sigmas, deriv.order=0)
{
    if (!(missing(h)))
    return(amise.mixt.1d(h=h, mus=mus, sigmas=sigmas, props=props, samp=samp, deriv.order=deriv.order))

    r <- deriv.order
    if (is.vector(mus)) { d <- length(mus); mus <- t(matrix(mus)) }
    else d <- ncol(mus)
    k <- length(props)

    if (k == 1)
    {
        Sigmasinv <- chol2inv(chol(Sigmas))
        omega.mat <- 2^(-d-r-2)*pi^(-d/2)*det(Sigmas)^(-1/2)*nu.rs(r=r, s=2, Sigmasinv, matrix.sqrt(Sigmasinv) %*% H %*% matrix.sqrt(Sigmasinv))
    }
    else
    {
        omega.mat <- matrix(0, nrow=k, ncol=k)
        for (i in 1:k)
        {
            Sigmai <- Sigmas[((i-1)*d+1):(i*d),]
            mui <- mus[i,]
            for (j in 1:k)
            {
            Sigmaj <- Sigmas[((j-1)*d+1):(j*d),]
            muj <- mus[j,]    
            omega.mat[i,j] <- gamma_r2(mu=mui-muj, Sigma=Sigmai + Sigmaj, d=d, r=r, H=H)
            }
        }
    }

    Hinv <- chol2inv(chol(H)) 
    if (k == 1) amise <- 2^(-r)*nu(r,Hinv)/(samp * (4 * pi)^(d/2) * sqrt(det(H))) + omega.mat/4
    else amise <- 2^(-r)*nu(r,Hinv)/(samp * (4 * pi)^(d/2) * sqrt(det(H))) + (props %*% omega.mat %*% props)/4

    return(drop(amise))
}

amise.mixt.1d <- function(h, mus, sigmas, props, samp, deriv.order=0)
{
    d <- 1
    r <- deriv.order
    k <- length(props)
    H <- as.matrix(h^2)

    if (k == 1)
        omega.mat <- gamma_r2(mu=rep(0,d),Sigma=as.matrix(2*sigmas^2), d=d, r=r, H=H)
    else
    {  
        omega.mat <- matrix(0, nrow=k, ncol=k)
        for (i in 1:k)
        {
            Sigmai <- as.matrix(sigmas[i]^2)
            mui <- mus[i]
            for (j in 1:k)
            {
                Sigmaj <- as.matrix(sigmas[j]^2)
                muj <- mus[j]    
                omega.mat[i,j] <- gamma_r2(mu=mui-muj, Sigma= Sigmai + Sigmaj, d=d, r=r, H=H)
            }
        }
    }
    Hinv <- h^(-2)
  
    if (k == 1) amise <- 2^(-r)*nu(r,Hinv)/(samp * (4 * pi)^(d/2) * sqrt(det(H))) + omega.mat/4
    else amise <- 2^(-r)*nu(r,Hinv)/(samp * (4 * pi)^(d/2) * sqrt(det(H))) + (props %*% omega.mat %*% props)/4

    return(drop(amise))
}

##############################################################################
## Lambda matrices (for exact AMISE for normal mixtures)
##
## Parameters 
## mus - means
## Sigmas - variances
## k - number of mixture components
## r - derivative (r1, r2)
##
## Returns 
## Lambda matrix
##############################################################################
lambda <- function(mus, Sigmas, k, r)
{
    ## the (i,j) element of Lambda matrix is d^r/ dx^r  dmvnorm(0, mu_i - mu_j,
    ## a*H + Sigma_i + Sigma_j)

    if (is.vector(mus)) d <- length(mus)
    else d <- ncol(mus)

    if (k == 1) 
    lambda.mat <- dmvnorm.deriv(deriv.order=r, x=rep(0, length(mus)), Sigma=2*Sigmas)
    else
    {   
        if (is.matrix(mus)) d <- ncol(mus)
        else d <- length(mus)
        lambda.mat <- matrix(0, nrow=k, ncol=k)
        for (i in 1:k)
        {
            Sigmai <- Sigmas[((i-1)*d+1) : (i*d),]
            mui <- mus[i,]
            for (j in 1:k)
            {
                Sigmaj <- Sigmas[((j-1)*d+1) : (j*d),]
                muj <- mus[j,]    
                lambda.mat[i,j] <- dmvnorm.deriv(deriv.order=r, x=mui-muj,Sigma=Sigmai+Sigmaj)
            }
        }
    }

    return(lambda.mat)
}

amise.mixt.2d <- function(H, mus, Sigmas, props, samp)
{  
    d <- ncol(Sigmas)
    k <- length(props)

    ## formula is found in Wand & Jones (1993)
    if (k == 1) 
    {
        amise <- 1/(samp * (4*pi)^(d/2) * sqrt(det(H))) +
            1/4 *(lambda(mus, Sigmas, k, r=c(4,0))*H[1,1]^2 +
            4*lambda(mus, Sigmas, k, r=c(3,1))*H[1,1]*H[1,2] +  
            2*lambda(mus, Sigmas, k, r=c(2,2))*(H[1,1]*H[2,2] + 2*H[1,2]^2) + 
            4*lambda(mus, Sigmas, k, r=c(1,3))*H[2,2]*H[1,2]+    
             lambda(mus, Sigmas, k, r=c(0,4))*H[2,2]^2) 
    }
    else
    {
        amise <- 1/(samp * (4*pi)^(d/2) * sqrt(det(H))) +
            1/4 * props %*% 
            (lambda(mus, Sigmas, k, r=c(4,0))*H[1,1]^2 +
            4*lambda(mus, Sigmas, k, r=c(3,1))*H[1,1]*H[1,2] +  
            2*lambda(mus, Sigmas, k, r=c(2,2))*(H[1,1]*H[2,2] + 2*H[1,2]^2) + 
            4*lambda(mus, Sigmas, k, r=c(1,3))*H[2,2]*H[1,2]+    
             lambda(mus, Sigmas, k, r=c(0,4))*H[2,2]^2) %*% props
    }

    return(drop(amise)) 
}

##############################################################################
## Finds the bandwidth matrix that minimises the MISE for normal mixtures
##
## Parameters
## mus - means
## Sigmas - variances
## props - vector of proportions of each mixture component 
## Hstart - initial bandwidth matrix
## samp - sample size
## full - 1 minimise over full bandwidth matrices
##      - 0 minimise over diagonal bandwidth matrices
## 
## Returns
## H_MISE
##############################################################################
hmise.mixt <- function(mus, sigmas, props, samp, hstart, deriv.order=0)
{
    r <- deriv.order
    d <- 1

    if (missing(hstart))
    {
        x <- rnorm.mixt(n=1000, mus=mus, sigmas=sigmas)  
        hstart <- sqrt((4/(samp*(d+2*r+2)))^(2/(d+2*r+4)) * var(x))
    }
    mise.mixt.temp <- function(h)
    {  
        return(mise.mixt.1d(h=h, mus=mus, sigmas=sigmas, props=props, samp=samp, deriv.order=deriv.order))
    }
    result <- optimize(f=mise.mixt.temp, interval=c(0, 10*hstart))
    hmise <- result$minimum
    
    return(hmise)
}

Hmise.mixt <- function(mus, Sigmas, props, samp, Hstart, deriv.order=0)
{
    r <- deriv.order
    if (is.vector(mus)) d <- length(mus)
    else d <- ncol(mus) 

    ## use normal reference estimate as initial condition
    if (missing(Hstart))
    {
        x <- rmvnorm.mixt(10000, mus, Sigmas, props)
        Hstart <- (4/(samp*(d+2*r+2)))^(2/(d+2*r+4)) * var(x)
    }

    mise.mixt.temp <- function(vechH)
    {  
        H <- invvech(vechH) %*% invvech(vechH)
        return(mise.mixt(H=H, mus=mus, Sigmas=Sigmas, props=props, samp=samp, deriv.order=deriv.order))
    }

    Hstart <- vech(matrix.sqrt(Hstart))
    result <- nlm(p=Hstart, f=mise.mixt.temp)
    Hmise <- invvech(result$estimate) %*% invvech(result$estimate)
  
    return(Hmise)
}   

Hmise.mixt.diag <- function(mus, Sigmas, props, samp, Hstart, deriv.order=0)
{   
    if (is.vector(mus)) d <- length(mus)
    else d <- ncol(mus) 

    if (missing(Hstart))
    {
        x <- rmvnorm.mixt(10000, mus, Sigmas, props)
        Hstart <- (4/(samp*(d + 2)))^(2/(d + 4)) * var(x)
    }  

    mise.mixt.temp <- function(diagH)
    {  
        H <- diag(diagH) %*% diag(diagH)
        return(mise.mixt(H=H, mus=mus, Sigmas=Sigmas, props=props, samp=samp, deriv.order=deriv.order))
    }
    Hstart <- diag(matrix.sqrt(Hstart))
    result <- nlm(p=Hstart, f=mise.mixt.temp)
    Hmise <- diag(result$estimate) %*% diag(result$estimate)
  
    return(Hmise)
}   

##############################################################################
## Finds bandwidth matrix that minimises the AMISE for normal mixtures - 2-dim
##
## Parameters
## mus - means
## Sigmas - variances
## props - vector of proportions of each mixture component 
## Hstart - initial bandwidth matrix
## samp - sample size
## 
## Returns
## Bandwidth matrix that minimises AMISE
##############################################################################
hamise.mixt <- function(mus, sigmas, props, samp, hstart, deriv.order=0)
{
    r <- deriv.order
    d <- 1

    if (missing(hstart))
    {
        x <- rnorm.mixt(n=1000, mus=mus, sigmas=sigmas)  
        hstart <- sqrt((4/(samp*(d+2*r+2)))^(2/(d+2*r+4)) * var(x))
    }

    amise.mixt.temp <- function(h)
    {  
        return(amise.mixt.1d(h=h, mus=mus, sigmas=sigmas, props=props, samp=samp, deriv.order=deriv.order))
    }
    result <- optimize(f=amise.mixt.temp, interval=c(0, 10*hstart))
    hamise <- result$minimum

    return(hamise)
}

Hamise.mixt <- function(mus, Sigmas, props, samp, Hstart, deriv.order=0)
{
    r <- deriv.order
    if (is.vector(mus)) d <- length(mus)
    else d <- ncol(mus) 

    ## use explicit formula for single normal
    if (length(props)==1)
    {
        Hamise <- (4/ (samp*(d+2*r+2)))^(2/(d+2*r+4)) * Sigmas
    }
    else
    {  
        ## use normal reference estimate as initial condition
        if (missing(Hstart)) 
        {
          x <- rmvnorm.mixt(10000, mus=mus, Sigmas=Sigmas, props=props)
          Hstart <- matrix.sqrt((4/ (samp*(d+2*r+2)))^(2/(d+2*r+4)) * var(x))
        }

        amise.mixt.temp <- function(vechH)
        {
            H <- invvech(vechH) %*% invvech(vechH)
            return(amise.mixt(H=H, mus=mus, Sigmas=Sigmas, props=props, samp=samp, deriv.order=deriv.order))
        }

        result <- nlm(p=vech(Hstart), f=amise.mixt.temp)
        Hamise <- invvech(result$estimate) %*% invvech(result$estimate)
    }

    return(Hamise)
}   

Hamise.mixt.diag <- function(mus, Sigmas, props, samp, Hstart, deriv.order=0)
{
    r <- deriv.order
    if (is.vector(mus)) d <- length(mus)
    else d <- ncol(mus) 

    ## use normal reference estimate as initial condition
    if (missing(Hstart)) 
    {
        x <- rmvnorm.mixt(10000, mus, Sigmas, props)
        Hstart <- matrix.sqrt((4/ (samp*(d+2*r+2)))^(2/(d+2*r+4)) * var(x))
    }

    amise.mixt.temp <- function(diagH)
    {
        H <- diag(diagH) %*% diag(diagH)
        return(amise.mixt(H=H, mus=mus, Sigmas=Sigmas, props=props, samp=samp, deriv.order=deriv.order))
    }
    result <- nlm(p=diag(Hstart), f=amise.mixt.temp)
    Hamise <- diag(result$estimate) %*% diag(result$estimate)

    return(Hamise)
}   
  
##############################################################################
## ISE for normal mixtures (fixed KDE)
## 
## Parameters
## x - data values
## H - bandwidth matrix
## mus - matrix of means (each row is a vector of means from each component
##       density)
## Sigmas - matrix of covariance matrices (every d rows is a covariance matrix 
##          from each component density) 
## props - mixing proportions
## lower - vector of lower end points of rectangle
## upper - vector of upper end points of rectangle
## gridsize - vector of number of grid points
## stepsize - vector of step sizes
##
## Returns
## ISE 
##############################################################################
ise.mixt <- function(x, H, mus, Sigmas, props, h, sigmas, deriv.order=0, binned=FALSE, bgridsize)
{
    if (!(missing(h)))
        return(ise.mixt.1d(x=x, h=h, mus=mus, sigmas=sigmas, props=props, deriv.order=deriv.order, binned=binned))

    if (is.vector(x)) x <- matrix(x, nrow=1)
    if (is.vector(mus)) mus <- matrix(mus, nrow=length(props))

    d <- ncol(x)
    n <- nrow(x)
    M <- length(props)
    r <- deriv.order

    ## formula is found in thesis

    vIdr <- vec(diag(d^r))
    ise1 <- 0
    ise2 <- 0
    ise3 <- 0

    if (binned)
        {
        ise1 <- dmvnorm.deriv.sum(x=x, Sigma=2*H, inc=1, deriv.order=2*r, binned=binned, bgridsize=bgridsize)
        for (j in 1:M)
        {
            Sigmaj <- Sigmas[((j - 1) * d + 1):(j * d), ]
            ise2 <- ise2 + props[j] * colSums(dmvnorm.deriv(x, mu=mus[j,],Sigma=H + Sigmaj, deriv.order=2*r))
            for (i in 1:M)
            {
                Sigmai <- Sigmas[((i - 1) * d + 1):(i * d), ]
                ise3 <- ise3 + props[i] * props[j] * dmvnorm.deriv(x=mus[i,],mu=mus[j,], Sigma = Sigmai + Sigmaj, deriv.order = 2*r)
            }
        }
        ise <- (-1)^r * sum(vIdr*(ise1/n^2 - 2 * ise2/n + ise3))
    }
    else
    {
        ise1 <- Qr(x=x, Sigma=2*H, inc=1, deriv.order=2*r)
        for (j in 1:M)
        {
            Sigmaj <- Sigmas[((j-1)*d + 1):(j*d), ]
            ise2 <- ise2 + props[j] * Qr(x=x, y=mus[j,], Sigma=H + Sigmaj, deriv.order=2*r, inc=1)
            for (i in 1:M)
            {
                Sigmai <- Sigmas[((i-1)*d + 1):(i*d), ]
                ise3 <- ise3 + props[i] * props[j] * Qr(x=mus[i,], y=mus[j,], Sigma=Sigmai + Sigmaj, deriv.order=2*r, inc=1)
            }
        }
        ise <- (-1)^r*(ise1 - 2*ise2 + ise3)
    }
 
    return(ise)
}

ise.mixt.1d <- function(x, h, mus, sigmas, props, deriv.order=0, binned=FALSE)
{  
    n <- length(x)
    M <- length(props)
    r <- deriv.order 
    ise1 <- 0
    ise2 <- 0
    ise3 <- 0

    ise1 <- dnorm.deriv.sum(x=x, sigma=sqrt(2)*h, inc=1, deriv.order=2*r, binned=binned)

    for (j in 1:M)
    {
        sigmaj <- sigmas[j]
        ise2 <- ise2 + sum(props[j]*dnorm.deriv(x=x, mu=mus[j], sigma=sqrt(h^2 + sigmaj^2), deriv.order=2*r))

        for (i in 1:M)
        {
            sigmai <- sigmas[i]
            ise3 <- ise3 + sum(props[i]*props[j]*dnorm.deriv(x=mus[i], mu=mus[j], sigma=sqrt(sigmai^2+sigmaj^2), deriv.order=2*r))
        }
    }  

    return ((-1)^r*(ise1/n^2 - 2*ise2/n + ise3))
}

Hise.mixt <- function(x, mus, Sigmas, props, Hstart, deriv.order=0)
{
    r <- deriv.order
    if (is.vector(mus)) d <- length(mus)
    else d <- ncol(mus) 
    samp <- nrow(x)

    ## use normal reference estimate as initial condition
    if (missing(Hstart))
    {
        xstart <- rmvnorm.mixt(10000, mus, Sigmas, props)
        Hstart <- (4/(samp*(d+2*r+2)))^(2/(d+2*r+4)) * var(xstart)
    }

    ise.mixt.temp <- function(vechH)
    {  
        H <- invvech(vechH) %*% invvech(vechH)
        return(ise.mixt(x=x, H=H, mus=mus, Sigmas=Sigmas, props=props, deriv.order=deriv.order))
    }
    Hstart <- vech(matrix.sqrt(Hstart))
    result <- nlm(p=Hstart, f=ise.mixt.temp)
    Hise <- invvech(result$estimate) %*% invvech(result$estimate)

    return(Hise)
}   

########################################################################
## Default grid sizes
########################################################################
default.gridsize <- function(d)
{
    if (d==1)      gridsize <- 401
    else if (d==2) gridsize <- rep(151,d)
    else if (d==3) gridsize <- rep(51, d)
    else if (d>=4) gridsize <- rep(21, d)
    
    return(gridsize)
}

default.bgridsize <- function(d)
{
    if (d==1)      gridsize <- 401
    else if (d==2) gridsize <- rep(151,d)
    else if (d==3) gridsize <- rep(31, d)
    else if (d==4) gridsize <- rep(15, d)
    else gridsize <- NA
    
    return(gridsize)
}

default.bflag <- function(d, n)
{
    if (d==1) thr <- 1
    else if (d==2) thr <- 500
    else if (d>2) thr <- 1000
    bf <- n>thr
    
    return(bf)
}

## truncate x to xmin, xmax grid
truncate.grid <- function(x, y, xmin, xmax)
{
    if (is.vector(x)) { d <- 1; n <- length(x); xvec <- TRUE; x <- matrix(x, ncol=1) }
    else { d <- ncol(x); n <- nrow(x); xvec <- FALSE }
    
    xind <- rep(TRUE, n)
    if (!(missing(xmax)))
    {
        if (d==1) msgmax <- xmax else msgmax <- paste0("c(", paste0(xmax, collapse=","), ")")
        if (any(sweep(x, 2, FUN=">", xmax))) warning(paste0("Points in x greater than xmax=", msgmax, " have been excluded."))
        for (i in 1:d) xind <- xind & x[,i]<=xmax[i]
    }
    if (!(missing(xmin)))
    {
        if (d==1) msgmin <- xmin else msgmin <- paste0("c(", paste0(xmin, collapse=","), ")")
        if (any(sweep(x, 2, FUN="<", xmin))) warning(paste0("Points in x less than xmin=", msgmin, " have been excluded."))
        for (i in 1:d) xind <- xind & x[,i]>=xmin[i]
    }

    if (d==1) 
    { 
        x <- x[xind,, drop=xvec]
        ## special case for x is 1-col matrix to force x to be 
        ## 1-col matrix, as required for eks <= 1.0.1
    }
    else 
    { 
        x <- x[xind,, drop=FALSE]
    }
    if (!missing(y)) { y <- y[xind]; x <- list(x=x, y=y) }
     
    return(x)
}

########################################################################
## Linear binning
## Courtesy of M Wand 2005
## Extended by T Duong to 3- and 4-dim 2006
## Extended by Gramack & Gramacki to include unconstrained b/w 2015 
########################################################################
binning <- function(x, H, h, bgridsize, xmin, xmax, supp=3.7, w, gridtype="linear")
{
    ## default values
    x <- as.matrix(x)
    d <- ncol(x)
    n <- nrow(x)
    if (missing(bgridsize)) bgridsize <- default.gridsize(d)
    if (missing(w)) w <- rep(1,n)
    if (missing(h)) h <- rep(0,d)
    if (!missing(H)) h <- sqrt(diag(H))
    
    range.x <- list()
    if (!missing(xmin) & !missing(xmax))
        for (i in 1:d) range.x[[i]] <- c(xmin[i], xmax[i])
    else if (!missing(xmin) & missing(xmax))
        for (i in 1:d) range.x[[i]] <- c(xmin[i], max(x[,i]) + supp*h[i])
    else if (missing(xmin) & !missing(xmax))
        for (i in 1:d) range.x[[i]] <- c(min(x[,i]) - supp*h[i], xmax[i])
    else 
        for (i in 1:d) range.x[[i]] <- c(min(x[,i]) - supp*h[i], max(x[,i]) + supp*h[i])

    a <- sapply(range.x,min) 
    b <- sapply(range.x,max) 
    
    if (missing(gridtype)) gridtype <- rep("linear", d)
    gridtype.vec <- rep("", d)
    
    gpoints <- list()
    for (id in 1:d)
    {
        gridtype1 <- match.arg(gridtype[i], c("linear", "sqrt", "quantile", "log"))
        if (gridtype1=="linear")
          gpoints[[id]] <- seq(a[id],b[id],length=bgridsize[id])  
        else if (gridtype1=="log")
          gpoints[[id]] <- seq(exp(a[id]),exp(b[id]),length=bgridsize[id])
    }
    
    if (d==1) counts <- linbin.ks(x,gpoints[[1]], w=w) 
    if (d==2) counts <- linbin2D.ks(x,gpoints[[1]],gpoints[[2]], w=w)
    if (d==3) counts <- linbin3D.ks(x,gpoints[[1]],gpoints[[2]],gpoints[[3]], w=w)
    if (d==4) counts <- linbin4D.ks(x,gpoints[[1]],gpoints[[2]],gpoints[[3]],gpoints[[4]], w=w)

    bin.counts <- list(counts=counts, eval.points=gpoints, w=w)
    if (d==1) bin.counts <- lapply(bin.counts, unlist)
    
    return(bin.counts)
}

########################################################################
## Linear binning
########################################################################
linbin.ks <- function(x, gpoints, w)
{
   n <- length(x)
   M <- length(gpoints)
   if (missing(w)) w <- rep(1, n)
   a <- gpoints[1]
   b <- gpoints[M]
   xi <- .C(C_massdist1d, x1=as.double(x[,1]), n=as.integer(n), a1=as.double(a), b1=as.double(b), M1=as.integer(M), weight=as.double(w), est=double(M))$est

   return(xi)
}

linbin2D.ks <- function(x, gpoints1, gpoints2, w)
{
    n <- nrow(x)
    M1 <- length(gpoints1)
    M2 <- length(gpoints2)
    a1 <- gpoints1[1]
    a2 <- gpoints2[1]
    b1 <- gpoints1[M1]
    b2 <- gpoints2[M2]
    if (missing(w)) w <- rep(1, n)

    ## binning for interior points
    out <- .C(C_massdist2d, x1=as.double(x[,1]), x2=as.double(x[,2]), n=as.integer(n), a1=as.double(a1), a2=as.double(a2), b1=as.double(b1), b2=as.double(b2), M1=as.integer(M1), M2=as.integer(M2), weight=as.double(w), est=double(M1*M2))
    xi <- matrix(out$est, nrow=M1, ncol=M2)

    return(xi)
}

linbin3D.ks <- function(x, gpoints1, gpoints2, gpoints3, w)
{
   n <- nrow(x)
   M1 <- length(gpoints1)
   M2 <- length(gpoints2)
   M3 <- length(gpoints3)
   a1 <- gpoints1[1]
   a2 <- gpoints2[1]
   a3 <- gpoints3[1]
   b1 <- gpoints1[M1]
   b2 <- gpoints2[M2]
   b3 <- gpoints3[M3]
   if (missing(w)) w <- rep(1, n)

   ## binning for interior points
   out <- .C(C_massdist3d, x1=as.double(x[,1]), x2=as.double(x[,2]), x3=as.double(x[,3]), n=as.integer(n), a1=as.double(a1), a2=as.double(a2), a3=as.double(a3), b1=as.double(b1), b2=as.double(b2), b3=as.double(b3), M1=as.integer(M1), M2=as.integer(M2), M3=as.integer(M3), weight=as.double(w), est=double(M1*M2*M3))
   xi <- array(out$est, dim=c(M1,M2,M3))

   return(xi)
}

linbin4D.ks <- function(x, gpoints1, gpoints2, gpoints3, gpoints4, w)
{
   n <- nrow(x)
   M1 <- length(gpoints1)
   M2 <- length(gpoints2)
   M3 <- length(gpoints3)
   M4 <- length(gpoints4)
   a1 <- gpoints1[1]
   a2 <- gpoints2[1]
   a3 <- gpoints3[1]
   a4 <- gpoints4[1]
   b1 <- gpoints1[M1]
   b2 <- gpoints2[M2]
   b3 <- gpoints3[M3]
   b4 <- gpoints4[M4]
   if (missing(w)) w <- rep(1, n)

   ## binning for interior points
   out <- .C(C_massdist4d, x1=as.double(x[,1]), x2=as.double(x[,2]), x3=as.double(x[,3]), x4=as.double(x[,4]), n=as.integer(n), a1=as.double(a1), a2=as.double(a2), a3=as.double(a3), a4=as.double(a4), b1=as.double(b1), b2=as.double(b2), b3=as.double(b3), b4=as.double(b4), M1=as.integer(M1), M2=as.integer(M2), M3=as.integer(M3), M4=as.integer(M4), weight=as.double(w), est=double(M1*M2*M3*M4))
   xi <- array(out$est, dim=c(M1,M2,M3,M4))

   return(xi)
}

########################################################################
## Discrete convolution
########################################################################
symconv.1d <- function(keval, gcounts)
{
    M <- length(gcounts)
    N <- length(keval) 
    L <- round(length(keval)/2)+1
  
    ## Smallest powers of 2 >= M + N
    P <- 2^(ceiling(log2(M + N)))
    
    ## Zero-padded version of keval and gcounts
    keval.zeropad <- rep(0, P)
    gcounts.zeropad <- rep(0, P)
    keval.zeropad[1:N] <- keval
    gcounts.zeropad[L:(L+M-1)] <- gcounts

    ## FFTs
    K <- fft(keval.zeropad)
    C <- fft(gcounts.zeropad)

    ## Invert element-wise product of FFTs and truncate and normalise it
    symconv.val <- Re(fft(K*C, inverse=TRUE)/P)[N:(N+M-1)]
    
    return(symconv.val)
}

symconv.nd <- function(keval, gcounts, d)
{
    M <- dim(gcounts)
    N <- dim(keval) 
    L <- round(dim(keval)/2)+1
    
    ## Smallest powers of 2 > M + N
    P <- 2^(ceiling(log2(M + N)))
    
    ## Zero-padded version of keval and gcounts
    keval.zeropad <- array(0, dim=P)
    gcounts.zeropad <- array(0, dim=P)
    if (d==2)
    {
        keval.zeropad[1:N[1], 1:N[2]] <- keval
        gcounts.zeropad[L[1]:(L[1]+M[1]-1), L[2]:(L[2]+M[2]-1)] <- gcounts
    }
    else if (d==3)
    {
        keval.zeropad[1:N[1], 1:N[2], 1:N[3]] <- keval
        gcounts.zeropad[L[1]:(L[1]+M[1]-1), L[2]:(L[2]+M[2]-1), L[3]:(L[3]+M[3]-1)] <- gcounts
    }
    else if (d==4)
    {
        keval.zeropad[1:N[1], 1:N[2], 1:N[3], 1:N[4]] <- keval
        gcounts.zeropad[L[1]:(L[1]+M[1]-1), L[2]:(L[2]+M[2]-1), L[3]:(L[3]+M[3]-1), L[4]:(L[4]+M[4]-1)] <- gcounts
    }
    
    ## FFTs
    K <- fft(keval.zeropad)
    C <- fft(gcounts.zeropad)

    ## Invert element-wise product of FFTs and truncate and normalise it
    symconv.val <- Re(fft(K*C, inverse=TRUE)/prod(P))
    if (d==2) symconv.val <- symconv.val[N[1]:(N[1]+M[1]-1), N[2]:(N[2]+M[2]-1)] 
    else if (d==3) symconv.val <- symconv.val[N[1]:(N[1]+M[1]-1), N[2]:(N[2]+M[2]-1), N[3]:(N[3]+M[3]-1)]
    else if (d==4) symconv.val <- symconv.val[N[1]:(N[1]+M[1]-1), N[2]:(N[2]+M[2]-1), N[3]:(N[3]+M[3]-1), N[4]:(N[4]+M[4]-1)]
    
    return(symconv.val)
}

###############################################################################
## Kernel mean shift
###############################################################################
kms <- function(x, y, H, max.iter=400, tol.iter, tol.clust, min.clust.size, merge=TRUE, keep.path=FALSE, verbose=FALSE)
{
    n <- nrow(x)
    d <- ncol(x)
    if (missing(tol.iter)) tol.iter <- 1e-3*min(apply(x, 2, IQR)) ## mean(apply(apply(x, 2, range), 2, diff))
    if (missing(tol.clust)) tol.clust <- 1e-2*max(apply(x, 2, IQR)) ## mean(apply(apply(x, 2, range), 2, diff))  
    if (missing(y)) y <- x
    if (missing(min.clust.size)) min.clust.size <- round(1e-2*nrow(y),0) 
    if (missing(H)) H <- Hpi(x, deriv.order=1, binned=default.bflag(d=d, n=n), nstage=2-(d>2))
    Hinv <- chol2inv(chol(H))
    if (is.vector(y)) y <- matrix(y, nrow=1)
   
    ## mean shift iterations
    n.seq <- block.indices(n, nrow(y), d=d, r=0, diff=FALSE, block.limit=3e6)
    if (verbose) pb <- txtProgressBar() 
    ms <- list()
    i <- 1
    if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
    ms <- kms.base(x=x, y=y[n.seq[i]:(n.seq[i+1]-1),], H=H, tol.iter=tol.iter, tol.clust=tol.clust, Hinv=Hinv, verbose=verbose, max.iter=max.iter)
    
    if (length(n.seq)>2)
    {
        for (i in 2:(length(n.seq)-1))
        {  
            if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
            ms.temp <- kms.base(x=x, y=y[n.seq[i]:(n.seq[i+1]-1),], H=H, tol.iter=tol.iter, tol.clust=tol.clust, Hinv=Hinv, verbose=verbose, max.iter=max.iter)

            ms$y <- rbind(ms$y, ms.temp$y)
            ms$end.points <- rbind(ms$end.points, ms.temp$end.points)
            ms$label <- c(ms$label, ms.temp$label + max(ms$label))
            ms$mode <- rbind(ms$mode, ms.temp$mode)
            ms$nclust <- ms$nclust + ms.temp$nclust
            ms$nclust.table <- table(ms$label)
            ms$path <- c(ms$path, ms.temp$path)

            ## merge clusters which are closer than tol.clust distance
            ms <- ms.merge.dist(ms=ms, tol=tol.clust, verbose=FALSE)
        }     
    }
    if (verbose) close(pb)
    path.temp <- ms$path
    ms$path <- NULL
    ms$tol.iter <- tol.iter
    ms$tol.clust <- tol.clust
    ms$min.clust.size <- min.clust.size
    ms$names <- parse.name(x)
    if (keep.path) ms$path <- path.temp
    
    ## merge clusters which are smaller than min.clust.size
    if (merge) ms <- ms.merge.num(ms, min.clust.size=min.clust.size, verbose=verbose)

    return(ms)
}

kms.base <- function(x, H, Hinv, y, max.iter, tol.iter, tol.clust, verbose=FALSE)
{
    ## mean shift iterations
    ## original implementation J.E. Chacon (2013)
    ## modifications T.D. (2014)
    
    if (!is.matrix(x)) x <- as.matrix(x)
    if (!is.matrix(y)) y <- as.matrix(y)
    if (missing(Hinv)) Hinv <- chol2inv(chol(H))
    nx <- nrow(x)
    ny <- nrow(y)
    d <- ncol(y)
    y.path <- split(y, row(y), drop=FALSE)
    names(y.path) <- NULL
 
    xHinv <- x %*% Hinv
    xHinvx <- rowSums(xHinv*x)
    y.update <- y
    i <- 1
    eps <- max(sqrt(rowSums(y.update^2)))

    disp.ind <- head(sample(1:nrow(y)), n=min(100,nrow(y)))
    while (eps > tol.iter & i< max.iter)
    {
        y.curr <- y.update
        yHinvy <- t(rowSums(y.curr%*%Hinv *y.curr))
        Mah <- apply(yHinvy, 2, "+", xHinvx) - 2*xHinv %*% t(y.curr)
        w <- exp(-Mah/2)
        denom <- colSums(w)
        num <- t(w)%*%x
        denom[denom<=1e-3*tol.iter] <- 1e-3*tol.iter
        mean.shift.H <- num/denom
        y.update <- mean.shift.H                             
        y.update.list <- split(y.update, row(y.update), drop=FALSE)
        y.path <- mapply(rbind, y.path, y.update.list, SIMPLIFY=FALSE)
        eps <- max(sqrt(rowSums((y.curr-y.update)^2)))
    
        if (verbose>1)
        {
            y.range <- apply(y, 2, range)
            if (d==2) plot(y.update[disp.ind,], col=1, xlim=y.range[,1], ylim=y.range[,2], xlab="x", ylab="y")
            else pairs(y.update[disp.ind,], col=1)
        }
        i <- i+1
    }
    ms.endpt <- t(sapply(y.path, tail, n=1, SIMPLIFY=FALSE))

    ## extract cluster centres
    mode.tree <- hclust(dist(ms.endpt))    
    clust.label <- cutree(mode.tree, h=tol.clust)
    nclust <- length(unique(clust.label))
    mode.val <- by(ms.endpt, INDICES=clust.label, FUN=colMeans)
    mode.val  <- t(sapply(mode.val, FUN=identity))
    colnames(mode.val) <- colnames(x)
    rownames(mode.val) <- NULL
    nclust.table <- table(clust.label, dnn="")
    
    ms <- list(x=x, y=y, end.points=ms.endpt, H=H, label=clust.label, nclust=nclust, nclust.table=nclust.table, mode=mode.val, path=y.path)
    class(ms) <- "kms"

    return(ms)
}

## merge classes in 'label' into a single class
## label is a list of vector of class labels
## i.e. all classes in label[[j]] merged into new class j
ms.merge.label <- function(ms, label, verbose=FALSE)
{
    ms.merge <- ms
    for (i in 1:length(label))
    {
        labeli <- label[[i]]
       
        if (length(labeli)>1)
        {    
            merge.label <- min(labeli)
            ms.merge$label[ms$label %in% labeli] <- merge.label
            mode.label <- ms$mode[labeli[round(length(labeli)/2,0)],]
            for (j in labeli) ms.merge$mode[j,] <- mode.label
        }
    }

    ms.merge$mode <- unique(ms.merge$mode)
    ms.merge$nclust <- nrow(ms.merge$mode)
    ms.merge$label <- as.factor(ms.merge$label)
    levels(ms.merge$label) <- 1:ms.merge$nclust
    ms.merge$label <- as.numeric(ms.merge$label)
    ms.merge$nclust.table <- table(ms.merge$label)

    if (verbose) cat("Current clusters:", ms.merge$nclust.table, "\n")
    
    return(ms.merge)  
}
   
## merge mean shift clusters based on distance threshold
ms.merge.dist <- function(ms, tol, verbose)
{   
    if (missing(tol)) tol <- 1e-1*min(apply(ms$x, 2, IQR))
    mode.tree <- hclust(dist(ms$mode))    
    merge.label <- cutree(mode.tree, h=tol)
    merge.label <- split(1:ms$nclust, merge.label, drop=FALSE)

    ## create list where each element is a vector of cluster labels
    ## to be merged into a single cluster
    ms.temp <- ms.merge.label(ms=ms, label=merge.label, verbose=verbose)
    
    return(ms.temp)
}

## merge mean shift clusters based on min cluster size
ms.merge.num <- function(ms, min.clust.size, verbose=FALSE)
{
    if (missing(min.clust.size)) min.clust.size <- round(1e-2*nrow(ms$y),0)
    min.clust.size <- round(min.clust.size, 0)

    if (any(ms$nclust.table<=min.clust.size))
    {
        if (verbose) cat("Min cluster size merging begins. Min size = ", min.clust.size, "\n")

        ms.temp <- ms
        while(any(ms.temp$nclust.table<=min.clust.size) & ms.temp$nclust>1)
        {
            nclust.table <- table(ms.temp$label)
            small.clust.ind <- which.min(nclust.table)
            
            if (nclust.table[small.clust.ind] <= min.clust.size)
            {
                nearest.clust.ind <- FNN::get.knnx(ms.temp$mode, ms.temp$mode, k=2)$nn.index[small.clust.ind,2]

                merge.label <- ms$label
                merge.label[merge.label==small.clust.ind] <- nearest.clust.ind 
                merge.label <- split(ms$label, merge.label, drop=FALSE)
                merge.label <- lapply(merge.label, unique)
                ms.temp <- ms.merge.label(ms=ms.temp, label=merge.label, verbose=verbose)
            }
        }
        ms <- ms.temp
        ms$min.clust.size <- min.clust.size
        
        if (verbose) cat("Min cluster size merging ends.\n\n")
    }
    if (verbose) { cat("Final clusters:\n"); summary(ms) }

    return(ms)
}

######################################################################
## Cluster partition for 2D kernel mean shift
#####################################################################
kms.part <- function(x, H, xmin, xmax, gridsize, verbose=FALSE, ...)
{
    if (missing(H)) H <- Hpi(x, deriv.order=1, binned=TRUE)
    tol <- 5
    tol.H <-  tol * diag(H)
    if (missing(xmin)) xmin <- apply(x, 2, min) - tol.H
    if (missing(xmax)) xmax <- apply(x, 2, max) + tol.H
    if (missing(gridsize)) gridsize <- default.gridsize(2) 
    xx <- seq(xmin[1], xmax[1], length = gridsize[1])
    yy <- seq(xmin[2], xmax[2], length = gridsize[2])
    xy <- expand.grid(xx, yy)

    xy.kms <- kms(x=x, y=xy, H=H, verbose=verbose, ...)
    xy.lab <- array(xy.kms$label, dim=gridsize)
    
    fhat <- kde(x=x, binned=TRUE, xmin=xmin, xmax=xmax, bgridsize=gridsize)
    fhat$estimate <- xy.lab

    fhat <- c(fhat, xy.kms[c("end.points", "label", "mode", "nclust", "nclust.table", "min.clust.size", "tol.iter", "tol.clust")])
    class(fhat) <- "kde.part"
    
    return(fhat)
}

plot.kde.part <- function(x, display="filled.contour", col, col.fun, alpha=1, add=FALSE, ...)
{
    clev <- sort(unique(as.vector(x$estimate)))
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n, palette="Dark2", alpha=alpha) }
    if (missing(col)) col <- col.fun(length(clev))
    
    for (i in 1:length(clev))
    {
        xtemp <- x
        xtemp$estimate <- x$estimate==clev[i]
        plot.kde(xtemp, display=display, col=c("transparent", col[i]), add=add | i>1, abs.cont=0.5, drawlabels=FALSE, alpha=alpha, ...)
    }
}

#############################################################################
## S3 methods for KMS objects
#############################################################################
## summary method
summary.kms <- function(object, ...)
{
    cat("Number of clusters =", object$nclust, "\n")
    cat("Cluster label table =", object$nclust.table, "\n")
    cat("Cluster modes =\n")
    print(as.data.frame(object$mode), ...)
}

## plot method
plot.kms <- function(x, display="splom", col, col.fun, alpha=1, xlab, ylab, zlab, theta=-30, phi=40, add=FALSE, ...)
{
    disp <- match.arg(display, c("splom", "plot3D", "rgl"))
    if (is.vector(x$H)) d <- 1 else d <- ncol(x$H)
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n, palette="Set2", alpha=alpha) }
    if (missing(col)) col <- col.fun(length(unique(x$label)))
    col <- transparency.col(col, alpha=alpha)
    if (d==1) stop("kms plot not yet implemented")
    else if (d==2)
    {
        if (missing(xlab)) xlab <- x$names[1]
        if (missing(ylab)) ylab <- x$names[2]
        if (!add) plot(x$x, col=col[x$label], xlab=xlab, ylab=ylab, ...)
        else points(x$x, col=col[x$label], ...)
    }
    else if (d==3 & disp %in% c("plot3D", "rgl"))
    {
        if (missing(xlab)) xlab <- x$names[1]
        if (missing(ylab)) ylab <- x$names[2]
        if (missing(zlab)) zlab <- x$names[3]
    
        if (disp=="plot3D")
        {
            if (!requireNamespace("plot3D", quietly=TRUE)) stop("Install the plot3D package as it is required.", call.=FALSE)
            if (!add) plot3D::points3D(x$x[,1], x$x[,2], x$x[,3], col=1, cex=0, add=add, theta=theta, phi=phi, d=4, colkey=FALSE, xlab=xlab, ylab=ylab, zlab=zlab, ticktype="detailed", bty="f", ...)
        
            for (i in 1:length(col))
                plot3D::points3D(x$x[x$label==i,1], x$x[x$label==i,2], x$x[x$label==i,3], col=col[i], add=TRUE, ...)
        }
        else if (disp=="rgl")
        {
            ## suggestions from Viktor Petukhov 08/03/2018
            if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
            if (!add) rgl::plot3d(x$x, col=col[x$label], alpha=alpha, ...)
            else rgl::points3d(x$x, col=col[x$label], alpha=alpha, ...) 
        }
    }
    else if (d>=3)
    {
        pairs(x$x, col=col[x$label], ...)
    }    
}

#############################################################################
## Kernel copula and copula density estimators
#############################################################################
## empirical pseudo-uniform transformation
## taken from pobs() function in copula package
pseudo.unif.empirical <- function (x, y) 
{
    if (missing(y)) y <- x
    if (is.vector(y)) y <- matrix(y, nrow=1)
    d <- ncol(x)
    u <- matrix(0, ncol=d, nrow=nrow(x))
    for (i in 1:d)
    {
        ecdf.fun <- ecdf(x=x[,i])
        u[,i] <- ecdf.fun(y[,i])
    }

    return(u)
}

## kernel pseudo-uniform transformation
pseudo.unif.kernel <- function(x, y, hs, binned=TRUE)
{
    if (missing(y)) y <- x
    if (is.vector(y)) y <- matrix(y, nrow=1)
    d <- ncol(x)
    u <- list()
    for (i in 1:d)
    {
        u[[i]] <- kcde(x=x[,i], h=hs[i], eval.points=y[,i], binned=binned)
    }
    u2 <- sapply(u, getElement, "estimate")

    return(u2)
}

#############################################################################
## Kernel copula estimator
#############################################################################
kcopula <- function(x, H, hs, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, marginal="kernel", verbose=FALSE)
{
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    gridsize <- ksd$gridsize
    bgridsize <- ksd$bgridsize

    ## clip data to xmin,xmax grid 
    grid.clip <- binned
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }
  
    ## default bandwidths
    if (missing(H)) H <- Hpi.kcde(x=x, binned=default.bflag(d=d,n=n))
    if (missing(hs))
    {
        hs <- rep(0, d)
        for (i in 1:d) hs[i] <- hpi.kcde(x=x[,i], binned=TRUE)
    }

    Fhat <- kcde(x=x, H=H, gridsize=gridsize, binned=binned, bgridsize=bgridsize, xmin=xmin, xmax=xmax, supp=supp, eval.points=eval.points, w=w, tail.flag="lower.tail", verbose=verbose)
    xlims <- sapply(Fhat$eval.points, range)
    xlims[1,] <- xlims[1,] - 0.1*abs(apply(xlims, 2, diff))
    xlims[2,] <- xlims[2,] + 0.1*abs(apply(xlims, 2, diff))

    ## generate pseudo-uniform values
    marginal <- match.arg(marginal, c("kernel", "empirical")) 
    if (marginal=="kernel")
    {  
        ## kernel pseudo-uniform
        u <- list()
        u.eval.points <- list()
        for (i in 1:d)
        {
            u.eval.points[[i]] <- kcde(x=x[,i], h=hs[i], eval.points=Fhat$eval.points[[i]], xmin=xlims[1,i], xmax=xlims[2,i], binned=TRUE)
        }
        y <- pseudo.unif.kernel(x=x, y=x, hs=hs, binned=TRUE)
        ep <- lapply(u.eval.points, getElement, "estimate")
    }
    else if (marginal=="empirical")
    {
        ## empirical pseudo-uniform
        y <- pseudo.unif.empirical(x=x, y=x)
        ep <- lapply(1:d, function(i) { f <- ecdf(x=x[,i]); f(Fhat$eval.points[[i]]) })
        ep <- numeric()
        for (i in 1:d)
        {
            f <- ecdf(x=x[,i])
            ep <- c(ep, list(f(Fhat$eval.points[[i]])))
        }
    }

    Chat <- Fhat
    Chat$x <- y 
    Chat$x.orig <- x
    Chat$eval.points <- ep
    if (inherits(Chat$eval.points, "list")) names(Chat$eval.points) <- ksd$names
    Chat$hs <- hs

    ## loess smoothing on a uniform grid
    if (d==2)
    {
        ## select smaller grid for d==2 for memory usage reasons
        subselect <- round(cbind(seq(1,length(Chat$eval.points[[1]]), length=101), seq(1,length(Chat$eval.points[[2]]), length=101)),0)
        eval.points.df <- data.frame(expand.grid(Chat$eval.points[[1]][subselect[,1]], Chat$eval.points[[2]][subselect[,2]]))
        names(eval.points.df) <- paste("x", 1:ncol(eval.points.df), sep="")
        eval.points.df <- data.frame(estimate=as.numeric(Chat$estimate[subselect[,1], subselect[,2]]), eval.points.df)
    }
    else if (d==3)
    {
        ## select smaller grid for d==3 for memory usage reasons
        subselect <- round(cbind(seq(1,length(Chat$eval.points[[1]]), length=21), seq(1,length(Chat$eval.points[[2]]), length=21), seq(1,length(Chat$eval.points[[3]]), length=21)),0)
        eval.points.df <- data.frame(expand.grid(Chat$eval.points[[1]][subselect[,1]], Chat$eval.points[[2]][subselect[,2]], Chat$eval.points[[3]][subselect[,3]]))
        names(eval.points.df) <- paste("x", 1:ncol(eval.points.df), sep="")
        eval.points.df <- data.frame(estimate=as.numeric(Chat$estimate[subselect[,1], subselect[,2], subselect[,3]]), eval.points.df)
    }

    if (d==2) Chat.loess <- loess(estimate ~ x1+x2, data=eval.points.df, span=0.1)
    else if (d==3)  Chat.loess <- loess(estimate ~ x1+x2+x3, data=eval.points.df, span=0.1)

    u.eval.points.regular <- list()
    for (i in 1:d)
    u.eval.points.regular[[i]] <- seq(0,1,length=length(Chat$eval.points[[i]]))
    u.eval.points.regular.df <- data.frame(expand.grid(u.eval.points.regular))
    names(u.eval.points.regular.df) <- paste("x", 1:ncol(u.eval.points.regular.df), sep="")

    Chat.smoothed <- Chat
    Chat.smoothed$eval.points <- u.eval.points.regular
    Chat.smoothed$estimate <- array(predict(Chat.loess, newdata=u.eval.points.regular.df), dim=dim(Chat$estimate))

    ## interpolate NA boundary values from loess smoothing
    gsdim <- dim(Chat$estimate)
    if (d==2)
    {
        Chat.smoothed$estimate[1,] <- 0
        Chat.smoothed$estimate[gsdim[1],] <- Chat.smoothed$estimate[gsdim[1]-1,]*1.001
        Chat.smoothed$estimate[,1] <-0
        Chat.smoothed$estimate[,gsdim[2]] <- Chat.smoothed$estimate[,gsdim[2]-1]*1.001
        Chat.smoothed$estimate[gsdim[1],gsdim[2]] <- 1
        Chat.smoothed$estimate[Chat.smoothed$estimate>1] <- 1
    }
    else if (d==3)
    {
        Chat.smoothed$estimate[,,1] <- 0
        for (k in 2:(gsdim[3]-1))
        {
           Chat.smoothed$estimate[,,k][1,] <- 0
           Chat.smoothed$estimate[,,k][gsdim[1],] <- Chat.smoothed$estimate[,,k][gsdim[1]-1,]*1.001
           Chat.smoothed$estimate[,,k][,1] <-0
           Chat.smoothed$estimate[,,k][,gsdim[2]] <- Chat.smoothed$estimate[,,k][,gsdim[2]-1]*1.001
         }
        Chat.smoothed$estimate[,,gsdim[3]] <- Chat.smoothed$estimate[,,gsdim[3]-1]*1.001
        Chat.smoothed$estimate[gsdim[1],gsdim[2],gsdim[3]] <- 1
        Chat.smoothed$estimate[Chat.smoothed$estimate>1] <- 1
    }

    Chat <- Chat.smoothed
    Chat$marginal <- marginal
    if (inherits(Chat$eval.points, "list")) names(Chat$eval.points) <- ksd$names
    class(Chat) <- "kcopula"

    return(Chat)
}

#############################################################################
## Kernel copula density estimator
#############################################################################
kcopula.de <- function(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, marginal="kernel", boundary.supp, boundary.kernel="beta", verbose=FALSE)
{
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    if (missing(binned)) binned <- ksd$binned
    if (missing(bgridsize)) bgridsize <- ksd$bgridsize
    if (missing(gridsize)) gridsize <- ksd$gridsize

    ## clip data to xmin,xmax grid for binned estimation
    grid.clip <- binned
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }

    ## default bandwidths
    hs <- rep(0, d)
    for (i in 1:d) hs[i] <- hpi.kcde(x=x[,i], binned=TRUE)

    ## generate pseudo-uniform values
    marginal <- match.arg(marginal, c("kernel", "empirical")) 
    if (marginal=="kernel") 
        y <- pseudo.unif.kernel(x=x, y=x, hs=hs, binned=TRUE)
    else if (marginal=="empirical") 
        y <- pseudo.unif.empirical(x=x, y=x)  

    colnames(y) <- colnames(x)
    if (missing(H)) H <- Hns(y)

    ## kernel copula density is boundary kernel estimator
    if (d==2 | d==3) 
        chat <- kde.boundary(x=y, H=H, gridsize=gridsize, supp=supp, xmin=rep(0,d), xmax=rep(1,d), gridtype=gridtype, w=w, boundary.supp=boundary.supp, binned=FALSE, boundary.kernel=boundary.kernel, verbose=verbose)
    else
        stop("kcopula.de requires 2-d or 3-d data.")

    ## normalise KDE to integrate to 1 
    chat$estimate <- chat$estimate/sum(chat$estimate*apply(sapply(chat$eval.points, diff), 1, prod)[1])
    chat$names <- parse.name(x)
    if (inherits(Chat$eval.points, "list")) names(Chat$eval.points) <- ksd$names
    chat$x.orig <- x
    chat$hs <- hs

    ## compute prob contour levels
    if (compute.cont & missing(eval.points))
    chat$cont <- contourLevels(chat, cont=1:99, approx=approx.cont)
    chat$marginal <- marginal
    class(chat) <- "kcopula.de"

    return(chat)
}

#############################################################################
## S3 methods
#############################################################################
## plot methods
plot.kcopula <- function(x, ...)
{
    plot.kcde(x, ...)
}

plot.kcopula.de <- function(x, ...)
{
    plot.kde(x, ...)
}

## predict methods
predict.kcopula <- function(object, ..., x, u)
{
    if (missing(u)) { if (object$marginal=="kernel") u <- pseudo.unif.kernel(x=object$x.orig, y=x, hs=object$hs) }
    return(predict.kde(object, ..., x=u))
}

predict.kcopula.de <- function(object, ..., x, u)
{
    if (missing(u)) { if (object$marginal=="kernel") u <- pseudo.unif.kernel(x=object$x.orig, y=x, hs=object$hs) }
    return(predict.kde(object, ..., x=u))
}

## contourLevel method
contourLevels.kcopula.de <- function(x, ...)
{
    x1 <- x; class(x1) <- "kde"  
    return(contourLevels(x=x1, ...))
}

##############################################################################
## Test statistic for multivariate 2-sample test
##############################################################################
kde.test <- function(x1, x2, H1, H2, h1, h2, psi1, psi2, var.fhat1, var.fhat2, binned=FALSE, bgridsize, verbose=FALSE)
{
    ## default values 
    ksd <- ks.defaults(x=x1, binned=binned, bgridsize=bgridsize)
    d <- ksd$d
    binned <- ksd$binned
    bgridsize <- ksd$bgridsize
    gridsize <- ksd$gridsize

    if (is.vector(x1) & is.vector(x2))
        return (kde.test.1d(x1=x1, x2=x2, h1=h1, h2=h2, psi1=psi1, psi2=psi2, var.fhat1=var.fhat1, var.fhat2=var.fhat2, binned=binned, bgridsize=bgridsize, verbose=verbose))

    if (!is.matrix(x1)) x1 <- as.matrix(x1)
    if (!is.matrix(x2)) x2 <- as.matrix(x2)
    n1 <- nrow(x1)
    n2 <- nrow(x2)
    d <- ncol(x1)
    K0 <- drop(dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=0))
    S1 <- var(x1)
    S2 <- var(x2)

    ## default bandwidths 
    if (missing(H1)) H1 <- Hpi.kfe(x1, deriv.order=0, binned=default.bflag(d=d,n=n1), bgridsize=bgridsize)
    if (missing(H2)) H2 <- Hpi.kfe(x2, deriv.order=0, binned=default.bflag(d=d,n=n2), bgridsize=bgridsize)
    
    ## kernel estimation for components of test statistic
    if (missing(psi1)) psi1 <- Qr(x=x1, y=x1, Sigma=H1, verbose=verbose) 
    if (missing(psi2)) psi2 <- Qr(x=x2, y=x2, Sigma=H2, verbose=verbose)

    if (missing(var.fhat1))
    {
        H1.r1 <- Hns(x=x1, deriv.order=1)
        fhat1.r1 <- predict(kdde(x=x1, H=H1.r1, deriv.order=1), x=apply(x1, 2, mean))
        var.fhat1 <- drop(fhat1.r1 %*% S1 %*% fhat1.r1)
    }
    psi12 <- Qr(x=x1, y=x2, Sigma=H1, verbose=verbose) 

    if (missing(var.fhat2))
    {
        H2.r1 <- Hns(x=x2, deriv.order=1)
        fhat2.r1 <- predict(kdde(x=x2, H=H2.r1, deriv.order=1), x=apply(x2, 2, mean))
        var.fhat2 <- drop(fhat2.r1 %*% S2 %*% fhat2.r1)
    }
    psi21 <- Qr(x=x2, y=x1, Sigma=H2, verbose=verbose) 

    ## test statistic + its parameters
    T.hat <- drop(psi1 + psi2 - (psi12 + psi21))
    muT.hat <- (n1^(-1)*det(H1)^(-1/2) + n2^(-1)*det(H2)^(-1/2))*K0
    varT.hat <- 3*(n1*var.fhat1 + n2*var.fhat2)/(n1+n2) *(1/n1+1/n2) 

    zstat <- (T.hat-muT.hat)/sqrt(varT.hat)
    pval <- 1-pnorm(zstat)
    if (length(pval==0)>0) pval[pval==0] <- pnorm(-abs(zstat[pval==0]))

    val <- list(Tstat=T.hat, zstat=zstat, pvalue=pval, mean=muT.hat, var=varT.hat, var.fhat1=var.fhat1, var.fhat2=var.fhat2, n1=n1, n2=n2, H1=H1, H2=H2, psi1=psi1, psi12=psi12, psi21=psi21, psi2=psi2)

    return(val)
}     

kde.test.1d <- function(x1, x2, h1, h2, psi1, psi2, var.fhat1, var.fhat2, binned=FALSE, bgridsize, verbose=FALSE)
{
    n1 <- length(x1)
    n2 <- length(x2)
    d <- 1
    K0 <- dnorm.deriv(x=0, mu=0, sigma=1, deriv.order=0)

    s1 <- sd(x1)
    s2 <- sd(x2)

    ## kernel estimation for components of test statistic
    if (missing(h1)) h1 <- hpi.kfe(x1, nstage=2, deriv.order=0, binned=binned, bgridsize=bgridsize)
    if (missing(h2)) h2 <- hpi.kfe(x2, nstage=2, deriv.order=0, binned=binned, bgridsize=bgridsize)

    if (missing(psi1)) psi1 <- Qr.1d(x=x1, y=x1, sigma=h1, verbose=verbose)
    if (missing(psi2)) psi2 <- Qr.1d(x=x2, y=x2, sigma=h2, verbose=verbose)

    if (missing(var.fhat1))
    {
        h1.r1 <- hns(x=x1, deriv.order=1) 
        fhat1.r1 <- predict(kdde(x=x1, h=h1.r1, deriv.order=1), x=mean(x1))
        var.fhat1 <- fhat1.r1^2*s1^2
    }
    psi12 <- Qr.1d(x=x1, sigma=h1, y=x2, verbose=verbose)

    if (missing(var.fhat2))
    {
        h2.r1 <- hns(x=x2, deriv.order=1) 
        fhat2.r1 <- predict(kdde(x=x2, h=h2.r1, deriv.order=1), x=mean(x2))
        var.fhat2 <- fhat2.r1^2*s2^2
    }
    psi21 <- Qr.1d(x=x2, sigma=h2, y=x1, verbose=verbose)

    ## test statistic + its parameters
    T.hat <- drop(psi1 + psi2 - (psi12 + psi21))
    muT.hat <- ((n1*h1)^(-1) + (n2*h2)^(-1))*K0
    varT.hat <- 3*(n1*var.fhat1 + n2*var.fhat2)/(n1+n2) *(1/n1+1/n2) 
    zstat <- (T.hat-muT.hat)/sqrt(varT.hat)
    pval <- 1-pnorm(zstat)
    if (length(pval==0)>0) pval[pval==0] <- pnorm(-abs(zstat[pval==0])) 

    val <- list(Tstat=T.hat, zstat=zstat, pvalue=pval, mean=muT.hat, var=varT.hat, var.fhat1=var.fhat1, var.fhat2=var.fhat2, n1=n1, n2=n2, h1=h1, h2=h2, psi1=psi1, psi12=psi12, psi21=psi21, psi2=psi2)

    return(val)
}

##############################################################################
## Local kde test
##############################################################################

## multivariate local test
kde.local.test <- function(x1, x2, H1, H2, h1, h2, fhat1, fhat2, gridsize, binned, bgridsize, verbose=FALSE, supp=3.7, mean.adj=FALSE, signif.level=0.05, min.ESS, xmin, xmax)
{
    ## default values 
    ksd <- ks.defaults(x=x1, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d
    if (missing(binned)) binned <- ksd$binned
    if (missing(bgridsize)) bgridsize <- ksd$bgridsize
    if (missing(gridsize)) gridsize <- ksd$gridsize

    ## clip data to xmin, xmax grid limits for binned estimation
    grid.clip <- binned    
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        x1 <- truncate.grid(x=x1, xmin=xmin, xmax=xmax)
        x2 <- truncate.grid(x=x2, xmin=xmin, xmax=xmax)
    }

    if (is.vector(x1) & is.vector(x2)) { return(kde.local.test.1d(x1=x1, x2=x2, h1=h1, h2=h2, fhat1=fhat1, fhat2=fhat2, gridsize=gridsize, binned=binned, bgridsize=bgridsize, verbose=verbose, supp=supp, mean.adj=mean.adj, xmin=xmin, xmax=xmax)) }

    if (missing(H1) & !missing(x1)) H1 <- Hpi(x=x1, deriv.order=0, binned=default.bflag(d=ncol(x1), n=nrow(x1)), bgridsize=bgridsize, verbose=verbose)
    if (missing(H2) & !missing(x2)) H2 <- Hpi(x=x2, deriv.order=0, binned=default.bflag(d=ncol(x2), n=nrow(x2)), bgridsize=bgridsize, verbose=verbose)

    if (!missing(x1) & !missing(x2))
    {
        n1 <- nrow(x1)
        n2 <- nrow(x2)
        d <- ncol(x1)
        RK <- (4*pi)^(-d/2)

        xrange <- apply(rbind(x1,x2), 2, range)
        tol.H <- sqrt(prod(diag(matrix.sqrt(H1) %*% matrix.sqrt(H2)))^(1/d))
        if (missing(xmin)) xmin <- xrange[1,] - supp*tol.H 
        if (missing(xmax)) xmax <- xrange[2,] + supp*tol.H 
    }
    else
    {
        n1 <- nrow(fhat1$x)
        n2 <- nrow(fhat2$x)
        d <- ncol(fhat1$x)
        RK <- (4*pi)^(-d/2)
        H1 <- fhat1$H
        H2 <- fhat2$H 
    }

    ## kernel estimation for components of test statistic
    if (missing(fhat1)) fhat1 <- kde(x=x1, H=H1, binned=binned, gridsize=gridsize, bgridsize=bgridsize, supp=supp, xmin=xmin, xmax=xmax, verbose=verbose)
    if (missing(fhat2)) fhat2 <- kde(x=x2, H=H2, binned=binned, gridsize=gridsize, bgridsize=bgridsize, supp=supp, xmin=xmin, xmax=xmax, verbose=verbose)

    HD2fhat <- 0
    if (mean.adj)
    { 
        D2fhat1 <- kdde(x=x1, H=H1, binned=binned, bgridsize=bgridsize, supp=supp, xmin=xmin, xmax=xmax, deriv.order=2, verbose=verbose)
        D2fhat2 <- kdde(x=x2, H=H2, binned=binned, bgridsize=bgridsize, supp=supp, xmin=xmin, xmax=xmax, deriv.order=2, verbose=verbose)
        D2fhat <- list()
        vH1 <- vec(H1)
        vH2 <- vec(H2)
        for (j in 1:length(D2fhat1$estimate)) D2fhat$estimate[[j]] <- vH1[j]*D2fhat1$estimate[[j]] - vH2[j]*D2fhat2$estimate[[j]]
        for (j in 1:length(D2fhat)) HD2fhat <- HD2fhat + D2fhat$estimate[[j]]
    }

    fhat.diff <- fhat1$estimate - fhat2$estimate - 1/2*HD2fhat  
    var.fhat.diff <- (n1^(-1)*det(H1)^(-1/2)*fhat1$estimate + n2^(-1)*det(H2)^(-1/2)*fhat2$estimate)*RK
    X2 <- fhat.diff^2/var.fhat.diff
    pvalue <- 1 - pchisq(X2, 1)
    pvalue[is.na(pvalue)] <- 0

    ## apply Hochberg multiple test adjustment
    gridsize <- sapply(fhat1$eval.points, length)
    fhat.diff.signif <- hochberg.mult.test(pvalue=pvalue, gridsize=gridsize, signif.level=signif.level)    

    fhat.diff.pos <- fhat1
    fhat.diff.neg <- fhat2
    fhat.diff.pos$estimate <- fhat.diff.signif*(fhat.diff>0)
    fhat.diff.neg$estimate <- fhat.diff.signif*(fhat.diff<0)

    if (!missing(min.ESS))
    {
        ESS1 <- n1*fhat1$estimate/dmvnorm(rep(0,d), rep(0,d), H1)
        ESS2 <- n2*fhat2$estimate/dmvnorm(rep(0,d), rep(0,d), H2)
        ESS <- pmin(ESS1, ESS2) >= min.ESS
        fhat.diff.pos$estimate <- fhat.diff.pos$estimate*ESS
        fhat.diff.neg$estimate <- fhat.diff.neg$estimate*ESS
    }
    mc <- as.character(match.call())[2:3]
    result <- list(fhat1=fhat1, fhat2=fhat2, X2=X2, pvalue=pvalue, fhat.diff=fhat.diff, mean.fhat.diff=HD2fhat, var.fhat.diff=var.fhat.diff, fhat.diff.pos=fhat.diff.pos, fhat.diff.neg=fhat.diff.neg, n1=n1, n2=n2, H1=H1, H2=H2, names=parse.name(x1), names.test=mc)
    class(result) <- "kde.loctest"

    return(result)
}

## 1-d local test
kde.local.test.1d <- function(x1, x2, h1, h2, fhat1, fhat2, gridsize=gridsize, binned=FALSE, bgridsize, verbose=FALSE, supp=3.7, mean.adj=FALSE, signif.level=0.05, min.ESS, xmin, xmax)
{
    if (missing(h1) & !missing(x1)) h1 <- hpi(x=x1, nstage=2, binned=binned, bgridsize=bgridsize)
    if (missing(h2) & !missing(x2)) h2 <- hpi(x=x2, nstage=2, binned=binned, bgridsize=bgridsize)

    if (!missing(x1) & !missing(x2))
    {
        n1 <- length(x1)
        n2 <- length(x2)
        d <- 1 
        RK <- (4*pi)^(-d/2)
        xrange <- range(c(x1,x2))
        if (missing(xmin)) xmin <- xrange[1] - supp*sqrt(h1*h2)
        if (missing(xmax)) xmax <- xrange[2] + supp*sqrt(h1*h2)
    }
    else
    {
        n1 <- length(fhat1$x)
        n2 <- length(fhat2$x)
        d <- 1
        RK <- (4*pi)^(-d/2)
        h1 <- fhat1$h
        h2 <- fhat2$h 
    }

    ## kernel estimation for components of test statistic
    if (missing(fhat1)) fhat1 <- kde(x=x1, h=h1, gridsize=gridsize, binned=binned, bgridsize=bgridsize, supp=supp, xmin=xmin, xmax=xmax)
    if (missing(fhat2)) fhat2 <- kde(x=x2, h=h2, gridsize=gridsize, binned=binned, bgridsize=bgridsize, supp=supp, xmin=xmin, xmax=xmax)

    h2D2fhat <- 0
    if (mean.adj)
    { 
        D2fhat1 <- kdde(x=x1, h=h1, gridsize=gridsize, binned=binned, bgridsize=bgridsize, supp=supp, xmin=xmin, xmax=xmax, deriv.order=2, verbose=verbose)
        D2fhat2 <- kdde(x=x2, h=h2, gridsize=gridsize, binned=binned, bgridsize=bgridsize, supp=supp, xmin=xmin, xmax=xmax, deriv.order=2, verbose=verbose)
        h2D2fhat <- (h1^2*D2fhat1$estimate - h2^2*D2fhat2$estimate)
    }

    fhat.diff <- fhat1$estimate - fhat2$estimate - 1/2*h2D2fhat  
    var.fhat.diff <- ((n1*h1)^(-1)*fhat1$estimate + (n2*h2)^(-1)*fhat2$estimate)*RK

    X2 <- fhat.diff^2/var.fhat.diff
    pvalue <- 1 - pchisq(X2, 1)

    gridsize <- length(fhat1$eval.points)
    fhat.diff.signif <- hochberg.mult.test(pvalue=pvalue, gridsize=gridsize, signif.level=signif.level)    

    fhat.diff.pos <- fhat1
    fhat.diff.neg <- fhat2
    fhat.diff.pos$estimate <- fhat.diff.signif*(fhat.diff>0)
    fhat.diff.neg$estimate <- fhat.diff.signif*(fhat.diff<0)

    if (!missing(min.ESS))
    {
        ESS1 <- n1*fhat1$estimate/dnorm(0, 0, h1)
        ESS2 <- n2*fhat2$estimate/dnorm(0, 0, h2)
        ESS <- pmin(ESS1, ESS2) >= min.ESS
        fhat.diff.pos$estimate <- fhat.diff.pos$estimate*ESS
        fhat.diff.neg$estimate <- fhat.diff.neg$estimate*ESS
    }
    mc <- as.character(match.call())[2:3]
    result <- list(fhat1=fhat1, fhat2=fhat2, chisq=X2, pvalue=pvalue, fhat.diff=fhat.diff, mean.fhat.diff=h2D2fhat, var.fhat.diff=var.fhat.diff, n1=n1, n2=n2, h1=h1, h2=h2, H1=h1^2, H2=h2^2, fhat.diff=fhat.diff, fhat.diff.pos=fhat.diff.pos, fhat.diff.neg=fhat.diff.neg, names=parse.name(x1), names.test=mc)
    class(result) <- "kde.loctest"

    return(result)
}

## Hochberg (1988) adjustment for multiple correlated tests
hochberg.mult.test <- function(pvalue, gridsize, signif.level)
{
    pvalue.ord <- pvalue[order(pvalue)]
    num.test <- sum(!is.na(pvalue.ord))

    if (num.test>=1)
        num.test.seq <- c(1:num.test, rep(NA, prod(gridsize) - num.test))
    else
        num.test.seq <- rep(NA, prod(gridsize))

    reject.nonzero <- ((pvalue.ord <= signif.level/(num.test + 1 - num.test.seq)) & (pvalue.ord > 0))  
    reject.nonzero.ind <- which(reject.nonzero)

    ## p-value == 0 => reject null hypotheses automatically
    fhat.diff.signif <- array(FALSE, dim=gridsize)
    fhat.diff.signif[which(pvalue==0, arr.ind=TRUE)] <- TRUE

    ## p-value > 0 then reject null hypotheses indicated in reject.nonzero.ind 
    for (i in reject.nonzero.ind)
        fhat.diff.signif[which(pvalue==pvalue.ord[i], arr.ind=TRUE)] <- TRUE 

    return(fhat.diff.signif)    
}

#############################################################################
## S3 method for kde.loctest objects
#############################################################################
## plot methods
plot.kde.loctest <- function(x, ...)
{
    fhat <- x
    names.test <- fhat$names.test
    if (is.vector(fhat$fhat1$x)) plotkde.loctest.1d(fhat, labels=names.test, ...)
    else
    {
        d <- ncol(fhat$fhat1$x)

        if (d==2) plotkde.loctest.2d(x, labels=names.test, ...)
        else if (d==3) plotkde.loctest.3d(x, labels=names.test, ...)
        else stop("Plot function only available for 1, 2 or 3-d data")   
    }
}

plotkde.loctest.1d <- function(x, lcol, col, add=FALSE, xlab, ylab, rugsize, add.legend=TRUE, pos.legend="topright", alpha=1, labels, ...)
{
    if (missing(xlab)) xlab <- x$fhat.diff.pos$names[1]
    if (missing(ylab)) 
    {
        if (is.null(labels)) ylab <- expression("Density difference "*f[1]-f[2])
        else ylab <- paste("Density difference", labels[1], "-", labels[2])
    }  
    if (is.null(labels)) labels <- c(expression(f[1]>f[2]), expression(f[1]<f[2]))
    else labels <- c(paste(labels[1], ">", labels[2]), paste(labels[1], "<" ,labels[2]))
    if (missing(col)) col <- hcl.colors(palette="Dark2",2)  
    col <- transparency.col(col, alpha=alpha)
    if (missing(lcol)) lcol <- 1
    if (!add) plot(x$fhat1$eval.points, x$fhat.diff, type="l", ylab=ylab, xlab=xlab, col=lcol, ...)
    else lines(x$fhat1$eval.points, x$fhat.diff, col=lcol, ...)

    plot.lim <- par()$usr
    if (missing(rugsize)) rugsize <- abs(plot.lim[4]-plot.lim[3])/50

    try(image(x$fhat.diff.pos$eval, c(plot.lim[3], plot.lim[3]+rugsize), cbind(x$fhat.diff.pos$estimate==1, x$fhat.diff.pos$estimate==1), level=0.5, add=TRUE, col=c("transparent", col[1]), ...))
    try(image(x$fhat.diff.neg$eval, c(plot.lim[3], plot.lim[3]+rugsize), cbind(x$fhat.diff.neg$estimate==1, x$fhat.diff.neg$estimate==1), level=0.5, add=TRUE, col=c("transparent", col[2]), ...))

    if (add.legend) legend(pos.legend, legend=labels, fill=col, bty="n") 
}

plotkde.loctest.2d <- function(x, col, add=FALSE, add.legend=TRUE, pos.legend="topright", alpha=1, labels, ...)
{ 
    if (missing(col)) col <- hcl.colors(palette="Dark2",2)
    col <- transparency.col(col, alpha=alpha)
    try(plot(x$fhat.diff.pos, col=c("transparent", col[1]), abs.cont=0.5, drawlabel=FALSE, disp="filled.contour", alpha=alpha, add=add, ...))
    try(plot(x$fhat.diff.neg, col=c("transparent", col[2]), abs.cont=0.5, drawlabel=FALSE, disp="filled.contour", alpha=alpha, add=TRUE, ...))

    if (is.null(labels)) labels <- c(expression(f[1]>f[2]), expression(f[1]<f[2]))
    else labels <- c(paste(labels[1], ">", labels[2]), paste(labels[1], "<" ,labels[2]))
    if (add.legend) legend(pos.legend, legend=labels, fill=col, bty="n")  
}

plotkde.loctest.3d <- function(x, col, color, add=FALSE, box=TRUE, axes=TRUE, alphavec=c(0.5, 0.5), add.legend=TRUE, labels, ...)
{
    if (length(alphavec)==1) alphavec <- rep(alphavec,2)
    if (missing(col)) col <- hcl.colors(palette="Dark2",2) 
    try(plot(x$fhat.diff.pos, color=col[1], col=col[1], abs.cont=0.5, add=add, box=FALSE, axes=FALSE, alphavec=alphavec[1],  ...))
    try(plot(x$fhat.diff.neg, color=col[2], col=col[2], abs.cont=0.5, add=TRUE, box=box, axes=axes, alphavec=alphavec[2], ...)) 

    if (add.legend) 
    {
        if (is.null(labels)) labels <- c(expression(f[1]>f[2]), expression(f[1]<f[2]))
        else labels <- c(paste(labels[1], ">", labels[2]), paste(labels[1], "<" ,labels[2]))

        if (!requireNamespace("plot3D", quietly=TRUE)) stop("Install the plot3D package as it is required.", call.=FALSE)
        plot3D::colkey(clim=c(1,2), col=col, at=c(1.25, 1.75), add=TRUE, side=1, addlines=TRUE, length=0.5, dist=-0.1, shift=0.1, labels=labels)
    }
}

###############################################################################
## Multivariate kernel regression
###############################################################################

kr <- function(x, y, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE, xsupp.truncate=TRUE, reg.order=1)
{
    ## default values 
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    bgridsize <- ksd$bgridsize
    gridsize <- ksd$gridsize

    if (is.data.frame(x)) x <- as.matrix(x)
    p <- reg.order
    if (d==1 & missing(h)) h <- hpi(x=x, binned=default.bflag(d=d, n=n))  
    if (d>1 & missing(H) & d>1) H <- Hpi(x=x, binned=default.bflag(d=d, n=n))
 
    ## compute binned estimator
    if (binned)
    {
        stop("Binning not yet available.")
    }
    else
    {
        ## compute exact (non-binned) estimator
       
        ## 1-dimensional    
        if (d==1)
        {
            ##if (missing(eval.points))
            ##{
            ##    if (unit.interval)
            ##    {
            ##        mhat <- kr.unit.interval.1d(x=x, y=y, h=h, binned=FALSE, reg.order=p)
            ##    }
            ##    else
            ##        mhat <- kr.grid.1d(x=x, y=y, h=h, gridsize=gridsize, supp=supp, positive=positive, xmin=xmin, xmax=xmax, adj.positive=adj.positive, gridtype=gridtype, w=w, reg.order=p)
           ## }
           ## else
           ##     mhat <- kr.points.1d(x=x, y=y, h=h, eval.points=eval.points, positive=positive, adj.positive=adj.positive, w=w, reg.order=p)
        }
        ## multi-dimensional
        else
        {  
            if (missing(eval.points))
            {
                if (d==2)
                    mhat <- kr.grid.2d(x=x, y=y, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, verbose=verbose, xsupp.truncate=xsupp.truncate, reg.order=p)
                ##else if (d == 3)
                ##    mhat <- kr.grid.3d(x=x, y=y, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, verbose=verbose, reg.order=p) 
                else 
                    stop("Need to specify eval.points for more than 3 dimensions")
            }
            ##else
            ##    mhat <- kr.points(x=x, y=y, H=H, eval.points=eval.points, w=w, reg.order=p)     
        }
    }

    mhat$binned <- binned
    mhat$names <- parse.name(x)  ## add variable names
    mhat$w <- w
    class(mhat) <- "kr"
    
    ## compute prob contour levels
    if (compute.cont & missing(eval.points))
    {
        mhat.temp <- mhat
        mhat.temp$estimate[is.na(mhat.temp$estimate)] <- -1e10
        mhat$cont <- contourLevels(mhat.temp, cont=1:99, approx=approx.cont)
    }
    
    return(mhat)
}


kr.1d <- function(x, y, h, eval.points, reg.order=1)
{
    n <- length(x)
    nsumy <- sum(y)/n
    wy <- y/nsumy
    if (missing(eval.points))
        s0 <- kde(x=x, h=h)
    else
        s0 <- kde(x=x, h=h, eval.points=eval.points)
    
    fhat.wy <- kde(x=x, h=h, w=wy, eval.points=s0$eval.points)
    fhat.wy$estimate <- fhat.wy$estimate * nsumy
    mhat <- fhat.wy
    if (reg.order==0)
    {  
        mhat$estimate <- mhat$estimate/s0$estimate
    }
    
    if (reg.order==1)
    {
        fhat1.wy <- kdde(x=x, h=h, w=wy, eval.points=s0$eval.points, deriv.order=1)
        fhat1.wy$estimate <- fhat1.wy$estimate * nsumy*h
        s1 <- -kdde(x=x, h=h, deriv.order=1, eval.points=s0$eval.points)$estimate*h
        s2 <- kdde(x=x, h=h, deriv.order=2, eval.points=s0$eval.points)$estimate*h^2 - s0$estimate
        
        mhat$estimate <- s2/(s2*s0$estimate-s1^2) * fhat.wy$estimate + s1/(s2*s0$estimate-s1^2) * fhat1.wy$estimate
    }
    
    return(mhat)
}

kr.grid.2d <- function(x, y, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, verbose=FALSE, xsupp.truncate=FALSE, tol.zero=1e-3, reg.order=1)
{
   p <- reg.order

   ## initialise grid 
   n <- nrow(x)
   if (is.null(gridx))
       gridx <- make.grid.ks(x, matrix.sqrt(H), tol=supp, gridsize=gridsize, xmin=xmin, xmax=xmax, gridtype=gridtype) 

   mhat.grid <- matrix(0, nrow=length(gridx[[1]]), ncol=length(gridx[[2]]))
   if (verbose) pb <- txtProgressBar()

   eval.pts <- expand.grid(gridx[[1]], gridx[[2]])
   eval.len <- nrow(eval.pts)

   if (xsupp.truncate)
   {
       fhat <- kde(x=x, binned=TRUE)
       fhat.cont <- contourLevels(fhat, cont=99, approx=TRUE)
       fhat.eval <- predict(fhat, x=eval.pts)
   }
   else
   {
       fhat.cont <- 0
       fhat.eval <- rep(1, eval.len)
   }
   
   beta.hat <- rep(NA, length=eval.len)
   
   for (i in 1:eval.len)
   {
       if (fhat.eval[i]>fhat.cont)
       {
           eval.ptsi <- unlist(eval.pts[i,])
           if (p==0) design.x <- rep(1, n)
           else if (p==1) design.x <- cbind(Inter=rep(1,n), sweep(x, 2, eval.ptsi))
           W.x <- dmvnorm(x=x, mean=eval.ptsi, sigma=H)
           A <- t(design.x * W.x)
           b <- A %*% y
           A <- A %*% design.x
           detA <- det(A)
           if (!(is.infinite(detA) | is.na(detA)))
               if (detA > tol.zero) beta.hat[i] <- solve(A,b)[1]
       }
       if (verbose) setTxtProgressBar(pb, i/eval.len)
   }
   
   mhat.grid <- array(beta.hat, dim=gridsize)
   if (verbose) close(pb)
   gridx1 <- list(gridx[[1]], gridx[[2]]) 
   
   mhat.list <- list(x=x, y=y, eval.points=gridx1, estimate=mhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE, reg.order=p)
   
   return(mhat.list)
}


plot.kr <- function(x, display="rgl", col, col.fun, xlab, ylab, ...)
{
    if (display!="rgl") plot.kdde(x=x, col=col, col.fun=col.fun, display=display, xlab=xlab, ylab=ylab, ...)
    else
    {
        ## suggestions from Viktor Petukhov 08/03/2018
        if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
        
        if (missing(col.fun)) col.fun <- terrain.colors
        if (missing(col)) col <- terrain.colors(5)[2]
        col.table <- col.fun(round(diff(range(x$estimate, na.rm=TRUE)) + 1,0))
        col <- col.table[x$estimate - min(x$estimate, na.rm=TRUE) + 1]
        if (missing(xlab)) xlab <- x$names[1]
        if (missing(ylab)) ylab <- x$names[2]
        rgl::persp3d(x$eval.points[[1]], x$eval.points[[2]], x$estimate, col=col, xlab=xlab, ylab=ylab, ...)
    }
}
   
contourLevels.kr <- function(x, ...)
{
    mhat <- x
    mhat$estimate[is.na(mhat$estimate)] <- 0
    mhat$deriv.order <- 1
    mhat$deriv.ind <- matrix(1, ncol=1, nrow=1)
    mhat$estimate <- list(mhat$estimate) 
    return(contourLevels.kdde(x=mhat,...))
}     

predict.kr <- function(object, ..., x)
{
    mhat <- object
    mhat$estimate[is.na(mhat$estimate)] <- 0
    return(predict.kde(object=mhat, ..., x=x))
}

#############################################################################
## Multivariate kernel density estimators
#############################################################################

#############################################################################
## Multivariate kernel density estimate using normal kernels
##
## Parameters
## x - points
## H - bandwidth matrix
## gridsize - number of interval points in grid
## supp - effective support of kernel
## eval.points - compute density estimate at these points (if missing
##            and dim(x) = 2, 3 compute density estimate over grid)  
## eval.levels - compute 3-D in 2-D slices taken at these level curves   
##
## Returns
## list with first d components with the points that the density
## estimate is evaluated at, and values of the density estimate 
#############################################################################
kde <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, positive=FALSE, adj.positive, w, compute.cont=TRUE, approx.cont=TRUE, unit.interval=FALSE, density=FALSE, verbose=FALSE)
{
    ## default values
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    gridsize <- ksd$gridsize
    bgridsize <- ksd$bgridsize
    if (binned & d>4) stop("Binned estimation for d>4 not implemented. Set binned=FALSE for exact estimation.")

    ## clip data to xmin, xmax grid limits for binned estimation
    grid.clip <- binned    
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        if (positive & missing(xmin)) { xmin <- rep(0,d) } 
        if (unit.interval) { if (missing(xmin)) xmin <- rep(0,d); if (missing(xmax)) xmax <- rep(1,d) }
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }
    
    ## default bandwidths
    if (d==1 & missing(h)) 
    {
        if (positive) x1 <- log(x) else x1 <- x
        if (unit.interval) x1 <- qnorm(x)
        h <- hpi(x=x1, nstage=2, binned=default.bflag(d=d, n=n), deriv.order=0)
    }
    if (d>1 & missing(H))
    {
        if (positive) x1 <- log(x) else x1 <- x
        H <- Hpi(x=x1, nstage=2, binned=default.bflag(d=d, n=n), deriv.order=0)
    }
    
    ## compute binned estimator
    if (binned)
    {
        if (positive)
        {
            if (d==1)
            {
                fhat <- kde.positive.1d(x=x, h=h, bgridsize=bgridsize, xmin=xmin, xmax=xmax, w=w, binned=binned, adj.positive=adj.positive)
            }
            else if (d==2)
            {
                fhat <- kde.positive.2d(x=x, H=H, bgridsize=bgridsize, xmin=xmin, xmax=xmax, w=w, binned=binned, adj.positive=adj.positive)
            }
        }
        else if (unit.interval)
        {
             fhat <- kde.unit.interval.1d(x=x, w=w, binned=binned, h=h)
        }
        else 
        {
            fhat <- kdde.binned(x=x, H=H, h=h, bgridsize=bgridsize, xmin=xmin, xmax=xmax, w=w, deriv.order=0, verbose=verbose)
        }
        
        if (!missing(eval.points))
        {
            fhat$estimate <- predict(fhat, x=eval.points)
            fhat$eval.points <- eval.points
        }
    }
    else
    {
        ## compute exact (non-binned) estimator   
        ## 1-dimensional    
        if (d==1)
        {
            if (missing(eval.points))
            {
                if (unit.interval)
                    fhat <- kde.unit.interval.1d(x=x, w=w, h=h, binned=FALSE)
                else if (positive)
                    fhat <- kde.positive.1d(x=x, h=h, xmin=xmin, xmax=xmax, w=w, binned=FALSE, adj.positive=adj.positive)
                else
                    fhat <- kde.grid.1d(x=x, h=h, gridsize=gridsize, supp=supp, positive=positive, xmin=xmin, xmax=xmax, adj.positive=adj.positive, gridtype=gridtype, w=w)
            }
            else
                fhat <- kde.points.1d(x=x, h=h, eval.points=eval.points, positive=positive, adj.positive=adj.positive, w=w)
        }
        ## multi-dimensional
        else
        {  
            if (is.data.frame(x)) x <- as.matrix(x)
            
            if (missing(eval.points))
            {
                if (d==2)
                {
                    if (positive)
                        fhat <- kde.positive.2d(x=x, H=H, gridsize=gridsize, xmin=xmin, xmax=xmax, w=w, binned=binned, adj.positive=adj.positive)
                    else
                        fhat <- kde.grid.2d(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, verbose=verbose)
                }
                else if (d==3)
                    fhat <- kde.grid.3d(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, verbose=verbose) 
                else 
                    fhat <- kde.grid.nd(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, verbose=verbose)
            }
            else 
                fhat <- kde.points(x=x, H=H, eval.points=eval.points, w=w, verbose=verbose)     
        }
    }
    
    if (density) fhat$estimate[fhat$estimate<0] <- 0
    fhat$binned <- binned
    fhat$names <- parse.name(x)  ## add variable names
    if (inherits(fhat$eval.points, "list")) names(fhat$eval.points) <- ksd$names
    fhat$w <- w
    fhat$type <- "kde" 
    class(fhat) <- "kde"
  
    ## compute prob contour levels
    if (compute.cont & missing(eval.points))
        fhat$cont <- contourLevels(fhat, cont=1:99, approx=approx.cont)
    
    return(fhat)
}

#############################################################################
## Univariate kernel density estimate on a grid
#############################################################################
kde.grid.1d <- function(x, h, gridsize, supp=3.7, positive=FALSE, adj.positive, xmin, xmax, gridtype, w)
{
    if (missing(xmin)) xmin <- min(x) - h*supp
    if (missing(xmax)) xmax <- max(x) + h*supp
    if (missing(gridtype)) gridtype <- "linear"

    if (positive)
    {
        if (missing(adj.positive)) adj.positive <- abs(min(x))
        y <- log(x + adj.positive)  ## transform positive data x to real line

        gridx <- seq(max(0, xmin), xmax, length=gridsize)
        gridy <- log(gridx + adj.positive)
        gridtype.vec <- "linear" 
    }
    else
    {
        y <- x
        gridtype1 <- match.arg(gridtype, c("linear", "sqrt", "quantile", "exp")) 
        if (gridtype1=="linear")
        {
            gridy <- seq(xmin, xmax, length=gridsize)
        }
        else if (gridtype1=="sqrt")
        {
            gridy.temp <- seq(sign(xmin)*sqrt(abs(xmin)), sign(xmax)*sqrt(abs(xmax)), length=gridsize)
            gridy <- sign(gridy.temp) * gridy.temp^2
        }
        else if (gridtype1=="exp")
        {
            gridy.temp <- seq(exp(xmin), exp(xmax), length=gridsize)
            gridy <- log(gridy.temp)
        }
        gridtype.vec <- gridtype1
    }
    n <- length(y)

    est <- dnorm.mixt(x=gridy, mus=y, sigmas=rep(h, n), props=w/n)
    fhat <- list(x=y, eval.points=gridy, estimate=est, h=h, H=h^2, gridtype=gridtype.vec, gridded=TRUE)
  
    if (positive)
    {
        ## compute transformation KDE
        fhat$estimate <- fhat$estimate/(exp(gridy))
        fhat$x <- x
        fhat$eval.points <- exp(gridy) - adj.positive  
    }
    class(fhat) <- "kde"
  
    return(fhat)
}

kde.positive.1d <- function(x, h, adj.positive, binned=FALSE, xmin, xmax, compute.cont=TRUE, approx.cont=TRUE, ...)
{
    if (missing(adj.positive)) adj.positive <- abs(min(x)) 
    y <- log(x + adj.positive) 
    if (missing(h)) h <- hpi(y, binned=binned)

    d <- 1
    tol <- 3.7
    tol.h <-  tol*h
    if (missing(xmin)) xmin <- min(x) - tol.h
    if (missing(xmax)) xmax <- max(x) + tol.h
    xmin[xmin<0] <- 0
    ymin1 <- log(xmin + adj.positive)
    ymax1 <- log(xmax + adj.positive)

    fhaty <- kde(x=y, h=h, xmin=ymin1, xmax=ymax1, gridtype=c("exp"), binned=binned, compute.cont=compute.cont, approx.cont=approx.cont, ...) 
    fhaty$estimate[is.nan(fhaty$estimate)] <- 0

    fhatx <- fhaty
    fhatx$x <- x
    fhatx$eval.points <- exp(fhaty$eval.points) - adj.positive
    jacobian <- abs(exp(fhaty$eval.points))
    jacobian[jacobian<=0] <- min(fhatx$estimate[fhatx$estimate>0]) 
    fhatx$estimate <- fhaty$estimate/jacobian

    if (compute.cont)
        fhatx$cont <- contourLevels(fhatx, cont=1:99, approx=approx.cont)
      
    ## re-sample on regular grid    
    ep <- seq(fhatx$eval.points[1], tail(fhatx$eval.points,n=1), length=length(fhatx$eval.points))
    fhatx$estimate <- predict(fhatx, x=ep)
    fhatx$eval.points <- ep

    return(fhatx)
}

kde.unit.interval.1d <- function(x, h, w, binned=FALSE)
{
    d <- 1
    y <- qnorm(x)
    if (missing(h)) h <- hpi(y)
    xseq <- tail(head(seq(0,1, length=default.gridsize(d)+2),n=-1), n=-1)
    fhaty <- kde(x=y, h=h, w=w, binned=binned)
    fhaty$estimate <- predict(fhaty, x=qnorm(xseq))
    fhatx <- fhaty
    fhatx$eval.points <- xseq
    fhatx$estimate <- fhaty$estimate/dnorm(qnorm(xseq))

    ## apply loess smoothing for unsmooth binned estimates
    if (binned)
    {      
        fhatx.loess <- loess(fhatx$estimate ~ fhatx$eval.points, span=0.1)
        fhatx.smoothed <- fhatx
        fhatx.smoothed$eval.points <- xseq
        fhatx.smoothed$estimate <- predict(fhatx.loess, x=xseq)
        fhatx <- fhatx.smoothed
    }
    fhatx$x <- x

    return(fhatx)
}

#############################################################################
## Bivariate kernel density estimate using normal kernels, evaluated over grid
##
## Parameters
## x - data points
## H - bandwidth matrix
## gridsize - number of interval points in grid
## supp - effective support of kernel
##
## Returns
## list with fields
## x - data points
## eval.points - points that KDE is evaluated at
## estimate - KDE evaluated at eval.points 
## H - bandwidth matrix 
#############################################################################
kde.grid.2d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, verbose=FALSE)
{
    ## initialise grid 
    n <- nrow(x)
    if (is.null(gridx))
    gridx <- make.grid.ks(x, matrix.sqrt(H), tol=supp, gridsize=gridsize, xmin=xmin, xmax=xmax, gridtype=gridtype) 

    suppx <- make.supp(x, matrix.sqrt(H), tol=supp)

    if (is.null(grid.pts)) grid.pts <- find.gridpts(gridx, suppx)    
    fhat.grid <- matrix(0, nrow=length(gridx[[1]]), ncol=length(gridx[[2]]))
    if (verbose) pb <- txtProgressBar()
  
    for (i in 1:n)
    {
        ## compute evaluation points 
        eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
        eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
        eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
        eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
        eval.x.len <- length(eval.x)
        eval.pts <- expand.grid(eval.x, eval.y)
        fhat <- dmvnorm(eval.pts, x[i,], H)

        ## place vector of density estimate values `fhat' onto grid 'fhat.grid' 
        for (j in 1:length(eval.y))
        fhat.grid[eval.x.ind, eval.y.ind[j]] <- 
        fhat.grid[eval.x.ind, eval.y.ind[j]] + 
          w[i]*fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
        if (verbose) setTxtProgressBar(pb, i/n)
    }
    if (verbose) close(pb) 
   
    fhat.grid <- fhat.grid/n
    gridx1 <- list(gridx[[1]], gridx[[2]]) 
    fhat.list <- list(x=x, eval.points=gridx1, estimate=fhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE)

    return(fhat.list)
}

######################################################################
## Bivariate KDE for data in positive quadrant
######################################################################
kde.positive.2d <- function(x, H, adj.positive, binned=FALSE, xmin, xmax, compute.cont=TRUE, approx.cont=TRUE, ...)
{
    if (missing(adj.positive)) adj.positive <- abs(apply(x, 2, min)) 
    y <- log(cbind(x[,1] + adj.positive[1],x[,2] + adj.positive[2])) 
    if (missing(H)) H <- Hpi(y, binned=binned)

    d <- ncol(x)
    tol <- 3.7
    tol.H <-  tol * diag(H)
    if (missing(xmin)) xmin <- apply(x, 2, min) - tol.H
    if (missing(xmax)) xmax <- apply(x, 2, max) + tol.H
    xmin[xmin<0] <- 0
    ymin1 <- log(pmax(xmin + adj.positive, apply(x, 2, min)))
    ymax1 <- log(xmax + adj.positive)

    fhaty <- kde(x=y, H=H, xmin=ymin1, xmax=ymax1, gridtype=c("exp", "exp"), binned=binned, compute.cont=compute.cont, approx.cont=approx.cont, ...) 
    fhaty$estimate[is.nan(fhaty$estimate)] <- 0

    fhatx <- fhaty
    fhatx$x <- x
    fhatx$eval.points[[1]] <- exp(fhaty$eval.points[[1]]) - adj.positive[1]
    fhatx$eval.points[[2]] <- exp(fhaty$eval.points[[2]]) - adj.positive[2]
    jacobian <- abs(exp(fhaty$eval.points[[1]]) %o% exp(fhaty$eval.points[[2]]))
    jacobian[jacobian<=0] <- min(fhatx$estimate[fhatx$estimate>0]) 
    fhatx$estimate <- fhaty$estimate/jacobian

    if (compute.cont)
        fhatx$cont <- contourLevels(fhatx, cont=1:99, approx=approx.cont)

    return(fhatx)
}

#############################################################################
## Trivariate kernel density estimate using normal kernels, evaluated over grid
##
## Parameters
## x - data points
## H - bandwidth matrix
## gridsize - number of interval points in grid
## supp - effective support of kernel
##
## Returns
## list with fields
## x - data points
## eval.points - points that KDE is evaluated at
## estimate - KDE evaluated at eval.points 
## H - bandwidth matrix 
#############################################################################
kde.grid.3d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, verbose=FALSE)
{
    ## initialise grid 
    n <- nrow(x)

    if (is.null(gridx))
    gridx <- make.grid.ks(x, matrix.sqrt(H), tol=supp, gridsize=gridsize, xmin=xmin, xmax=xmax, gridtype=gridtype) 
    suppx <- make.supp(x, matrix.sqrt(H), tol=supp)

    if (is.null(grid.pts))
    grid.pts <- find.gridpts(gridx, suppx)    
    fhat.grid <- array(0, dim=c(length(gridx[[1]]), length(gridx[[2]]), length(gridx[[3]])))
    if (verbose) pb <- txtProgressBar()

    for (i in 1:n)
    {
        ## compute evaluation points
        eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
        eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
        eval.z <- gridx[[3]][grid.pts$xmin[i,3]:grid.pts$xmax[i,3]]
        eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
        eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
        eval.z.ind <- c(grid.pts$xmin[i,3]:grid.pts$xmax[i,3])
        eval.x.len <- length(eval.x)
        eval.pts <- expand.grid(eval.x, eval.y)

        ## place vector of density estimate values `fhat' onto grid 'fhat.grid' 
        for (k in 1:length(eval.z))
        {
            fhat <- w[i]*dmvnorm(cbind(eval.pts, eval.z[k]), x[i,], H)
            for (j in 1:length(eval.y))
                fhat.grid[eval.x.ind,eval.y.ind[j], eval.z.ind[k]] <- 
                    fhat.grid[eval.x.ind, eval.y.ind[j], eval.z.ind[k]] + 
                    fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
        }
        if (verbose) setTxtProgressBar(pb, i/n)
    }
    if (verbose) close(pb)
    fhat.grid <- fhat.grid/n
    gridx1 <- list(gridx[[1]], gridx[[2]], gridx[[3]]) 
    fhat.list <- list(x=x, eval.points=gridx1, estimate=fhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE)

    return(fhat.list)
}

kde.grid.nd <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, verbose=FALSE)
{
    ## initialise grid 
    n <- nrow(x)
    if (is.null(gridx))
        gridx <- make.grid.ks(x, matrix.sqrt(H), tol=supp, gridsize=gridsize, xmin=xmin, xmax=xmax, gridtype=gridtype) 
    
    gridx1 <- gridx
    gridx1$stepsize <- NULL
    gridx1$gridtype <- NULL
    eval.points <- do.call(expand.grid, gridx1)
    est <- kde.points(x=x, H=H, eval.points=eval.points, w=w, verbose=verbose)$estimate 
    fhat.grid <- array(est, dim=gridsize)
        
    fhat.list <- list(x=x, eval.points=gridx1, estimate=fhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE)

    return(fhat.list)
}

#############################################################################
## Multivariate kernel density estimate using normal kernels,
## evaluated at each sample point
##
## Parameters
## x - data points
## H - bandwidth matrix
## eval.points - points where to evaluate density estimate
##
## Returns
## list with fields
## x - data points
## eval.points - points that KDE is evaluated at
## estimate - KDE evaluated at eval.points 
## H - bandwidth matrix 
#############################################################################
kde.points <- function(x, H, eval.points, w, verbose) 
{
    n <- nrow(x)
    d <- ncol(x)
    ne <- nrow(eval.points)
    Hs <- replicate(n, H, simplify=FALSE) 
    Hs <- do.call(rbind, Hs)
    fhat <- dmvnorm.mixt(x=eval.points, mus=x, Sigmas=Hs, props=w/n, verbose=verbose)
    
    return(list(x=x, eval.points=eval.points, estimate=fhat, H=H, gridded=FALSE))
}

kde.points.1d <- function(x, h, eval.points, positive=FALSE, adj.positive, w) 
{
    n <- length(x)

    if (positive)
    {
        if (missing(adj.positive)) adj.positive <- abs(min(x))
        y <- log(x + adj.positive)  ## transform positive data x to real line
        eval.pointsy <- log(eval.points + adj.positive)
    }
    else
    {
        y <- x
        eval.pointsy <- eval.points
    }

    est <- dnorm.mixt(x=eval.pointsy, mus=y, sigmas=rep(h,n), props=w/n)
    if (positive)
        est <- est/(eval.points + adj.positive) 
    fhat <- list(x=x, eval.points=eval.points, estimate=est, h=h, H=h^2, gridded=FALSE)

    return(fhat)
}

#############################################################################
## S3 methods for KDE objects
#############################################################################
## predict method for KDE objects
predict.kde <- function(object, ..., x, zero.flag=TRUE)
{
    fhat <- grid.interp(x=x, gridx=object$eval.points, f=object$estimate)
    if (!zero.flag) warning("zero.flag=FALSE has been deprecated and no longer has any effect")
  
    return(fhat)
}

## plot method
plot.kde <- function(x, ...)
{ 
    fhat <- x
    if (is.vector(fhat$x)) plotkde.1d(fhat, ...)
    else
    {
        d <- ncol(fhat$x)

        if (d==2) 
        {
            opr <- options()$preferRaster
            if (isTRUE(opr)) options("preferRaster"=TRUE)
            plotret <- plotkde.2d(fhat, ...)
            if (!is.null(opr)) options("preferRaster"=opr)
            invisible(plotret)
        }
        else if (d==3)
        {
            plotkde.3d(fhat, ...)
            invisible()
        }
        else 
          stop ("Plot function only available for 1, 2 or 3-d data")
    }
}

plotkde.1d <- function(fhat, xlab, ylab="Density function", add=FALSE,
  drawpoints=FALSE, col=1, col.pt=4, col.cont=1, cont.lwd=1, jitter=FALSE, cont, abs.cont, approx.cont=TRUE, alpha=1, ...) 
{
    if (missing(xlab)) xlab <- fhat$names
    col <- transparency.col(col, alpha=alpha)

    if (add) lines(fhat$eval.points, fhat$estimate, xlab=xlab, ylab=ylab, col=col, ...)
    else plot(fhat$eval.points, fhat$estimate, type="l", xlab=xlab, ylab=ylab, col=col, ...) 

    ## compute contours
    if (!missing(cont) | !missing(abs.cont)) 
    {
        if (missing(abs.cont))
        {
            if (!is.null(fhat$cont))
            {
                cont.ind <- rep(FALSE, length(cont))
                for (j in 1:length(cont))
                {
                    ci <- which(cont[j]==(100-as.numeric(unlist(strsplit(names(fhat$cont),"%")))))
                    if (length(ci)==0) cont.ind[j] <- NA else cont.ind[j] = ci
                }
              
                if (any(is.na(cont.ind)))
                    hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
                else
                    hts <- fhat$cont[cont.ind]
            }
            else
                hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
        }
        else
          hts <- abs.cont 
       
 
        if (is.null(fhat$deriv.order))
        {
            hts <- sort(hts, decreasing=TRUE)
            cont.ind <- 1-as.numeric(fhat$estimate>=hts[1])
            cont.ind[cont.ind==1] <- NA   
            lines(fhat$eval.points, cont.ind, col=col.cont, lwd=cont.lwd)
        }
        else
        {
           for (i in 1:length(hts))
           {     
               cont.ind <- 1-as.numeric(abs(fhat$estimate)>=abs(hts[i]))
               cont.ind[cont.ind==1] <- NA    
               lines(fhat$eval.points, cont.ind, col=col.cont, lwd=cont.lwd)      
           }
        }
    }
  
    if (drawpoints)
        if (jitter) rug(jitter(fhat$x), col=col.pt)
        else rug(fhat$x, col=col.pt)
}

plotkde.2d <- function(fhat, display="slice", cont=c(25,50,75), abs.cont, approx.cont=TRUE, xlab, ylab, zlab="Density function", cex=1, pch=1, labcex=1, add=FALSE, drawpoints=FALSE, drawlabels=TRUE, theta=-30, phi=40, d=4, col.pt=4, col, col.fun, alpha=1, lwd=1, border=1, thin=3, kdde.flag=FALSE, ticktype="detailed", ...) 
{
    disp <- match.arg(display, c("slice", "persp", "image", "filled.contour", "filled.contour2"))
    if (disp=="filled.contour2") disp <- "filled.contour"
    if (!is.list(fhat$eval.points)) stop("Need a grid of density estimates")

    if (missing(xlab)) xlab <- fhat$names[1]
    if (missing(ylab)) ylab <- fhat$names[2]
    if (missing(col.fun)) 
    {
        if (any(fhat$type=="kcurv")) col.fun <- function(n) { hcl.colors(n, palette="Oranges",rev=TRUE, alpha=alpha) }
        else col.fun <- function(n) { hcl.colors(n, palette="heat",rev=TRUE, alpha=alpha) }
    }
    
    ## perspective/wireframe plot
    if (disp=="persp")
    {
        hts <- seq(0, 1.1*max(fhat$estimate,na.rm=TRUE), length=100)
        if (missing(col)) col <- col.fun(length(hts)+1)
        if (length(col)<length(hts)) col <- rep(col, length=length(hts))
        col <- transparency.col(col, alpha=alpha)

        ## thinning indices
        plot.ind <- list(seq(1, length(fhat$eval.points[[1]]), by=thin), seq(1, length(fhat$eval.points[[2]]), by=thin))

        z <- fhat$estimate[plot.ind[[1]], plot.ind[[2]]]
        nrz <- nrow(z)
        ncz <- ncol(z)
        zfacet <- z[-1, -1] + z[-1, -ncz] + z[-nrz, -1] + z[-nrz, -ncz]
        facetcol <- cut(zfacet, length(col)+1)

        plotret <- persp(fhat$eval.points[[1]][plot.ind[[1]]], fhat$eval.points[[2]][plot.ind[[2]]], z, theta=theta, phi=phi, d=d, xlab=xlab, ylab=ylab, zlab=zlab, col=col[facetcol], border=border, ticktype=ticktype, ...)
    }
    else if (disp=="slice") 
    {
        ## compute contours
        if (missing(abs.cont))
        {
            if (!is.null(fhat$cont))
            {
                cont.ind <- rep(FALSE, length(cont))
                for (j in 1:length(cont))
                {
                    ci <- which(cont[j]==(100-as.numeric(unlist(strsplit(names(fhat$cont),"%")))))
                    if (length(ci)==0) cont.ind[j] <- NA else cont.ind[j] = ci
                }
              
                if (any(is.na(cont.ind)))
                    hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
                else
                    hts <- fhat$cont[cont.ind]
            }
            else
                hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
        }
        else
            hts <- abs.cont 

        hts <- sort(hts)
        if (missing(col)) col <- col.fun(length(hts))
        if (length(col)<length(hts)) col <- rep(col, times=length(hts))
        col <- transparency.col(col, alpha=alpha)
        
        ## draw contours
        j <- 0
        for (i in 1:length(hts)) 
        {
            if (missing(abs.cont)) { ni <- length(hts)-i+1; scale <- (100-cont[i])/hts[i]; scale2 <- cont[ni]/hts[ni] }
            else { ni <- i; scale <- 1; scale2 <-1 }

            if (hts[i]>0 | !is.null(fhat$deriv.order))
            {
                j <- j+1
                if (j==1) 
                    contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate*scale, level=hts[i]*scale, label=signif(hts[ni]*scale2), add=add, drawlabels=drawlabels, col=col[i], lwd=lwd, labcex=labcex, xlab=xlab, ylab=ylab, ...)
                else 
                    contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate*scale, level=hts[i]*scale, label=signif(hts[ni]*scale2), add=TRUE, drawlabels=drawlabels, col=col[i], lwd=lwd, labcex=labcex, ...)
            }
        }

        ## add points 
        if (drawpoints) points(fhat$x[,1], fhat$x[,2], col=col.pt, cex=cex, pch=pch)
    }
    ## image plot
    else if (disp=="image")
    {
        if (missing(col)) col <- col.fun(100)
        col <- transparency.col(col, alpha=alpha)
        image(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate, xlab=xlab, ylab=ylab, add=add, col=col, ...)

        ## add points 
        if (drawpoints) points(fhat$x[,1], fhat$x[,2], col=col.pt, cex=cex, pch=pch)
        box()
    }
    else if (disp=="filled.contour")
    {
        ## compute contours
        if (missing(abs.cont))
        {
            if (!is.null(fhat$cont))
            {
                cont.ind <- rep(FALSE, length(cont))
                for (j in 1:length(cont))
                {
                    ci <- which(cont[j]==(100-as.numeric(unlist(strsplit(names(fhat$cont),"%")))))
                    if (length(ci)==0) cont.ind[j] <- NA else cont.ind[j] = ci
                }
              
                if (any(is.na(cont.ind)))
                    hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
                else
                    hts <- fhat$cont[cont.ind]
            }
            else
                hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
        }
        else
            hts <- abs.cont 
        hts <- sort(hts)

        if (missing(col)) col <- c("transparent", col.fun(length(hts)))
        col <- transparency.col(col, alpha=alpha)
        clev <- c(min(c(fhat$estimate, hts)-0.01*max(abs(fhat$estimate))), hts, max(c(fhat$estimate, hts)) + 0.01*max(abs(fhat$estimate)))
        clev <- unique(clev) 

        if (!add) plot(fivenum(fhat$eval.points[[1]]), fivenum(fhat$eval.points[[2]]), type="n", xlab=xlab, ylab=ylab, ...)
        .filled.contour(fhat$eval.points[[1]], fhat$eval.points[[2]], z=fhat$estimate, levels=clev, col=col)
      
        if (!missing(lwd))
        {
            for (i in 1:length(hts)) 
            {
                if (missing(abs.cont)) { ni <- length(hts)-i+1; scale <- (100-cont[i])/hts[i]; scale2 <- cont[ni]/hts[ni] }
                else { ni <- i; scale <- 1; scale2 <-1 }

                if (lwd >=1) contour(fhat$eval.points[[1]], fhat$eval.points[[2]], fhat$estimate*scale, level=hts[i]*scale, label=signif(hts[ni]*scale2,3), add=TRUE, drawlabels=drawlabels, col=1, lwd=lwd, labcex=labcex, ...)
            }
        }

        ## add points 
        if (drawpoints) points(fhat$x[,1], fhat$x[,2], col=col.pt, cex=cex, pch=pch)
        box()
    }

    if (disp=="persp") invisible(plotret)
    else invisible()
}
  
plotkde.3d <- function(fhat, display="plot3D", cont=c(25,50,75), abs.cont, approx.cont=TRUE, colors, col, col.fun, alphavec, size=3, cex=1, pch=1, theta=-30, phi=40, d=4, ticktype="detailed", bty="f", col.pt=4, add=FALSE, xlab, ylab, zlab, drawpoints=FALSE, alpha, box=TRUE, axes=TRUE, ...)
{
    ## compute contours
    if (missing(abs.cont))
    {
        if (!is.null(fhat$cont))
        {
            cont.ind <- rep(FALSE, length(fhat$cont))
            for (j in 1:length(cont))
                cont.ind[which(cont[j] == 100-as.numeric(unlist(strsplit(names(fhat$cont),"%"))))] <- TRUE
           
            if (all(!cont.ind))
                hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
            else
                hts <- fhat$cont[cont.ind]
         }
         else
            hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
    }  
    else
        hts <- abs.cont
        
    nc <- length(hts)
    if (missing(col)) 
    { 
        if (missing(col.fun)) 
        {
            if (any(fhat$type=="kcurv")) col.fun <- function(n) { hcl.colors(n, palette="Oranges",rev=TRUE) }
            else col.fun <- function(n) { hcl.colors(n, palette="heat",rev=TRUE) }
        }
        col <- col.fun(length(hts))
    }    
    colors <- col
    if (missing(xlab)) xlab <- fhat$names[1]
    if (missing(ylab)) ylab <- fhat$names[2]
    if (missing(zlab)) zlab <- fhat$names[3]
    if (missing(alphavec))
    {
        if (is.null(fhat$deriv.order)) alphavec <- seq(0.1,0.5,length=nc)
        else alphavec <- c(rev(seq(0.1,0.4,length=round(nc/2))), seq(0.1,0.4,length=round(nc/2)))
    }
    if (missing(alpha)) alpha <- 0.1 
    #else if (!missing(alpha)) { alphavec <- rep(alpha,nc) }
            
    disp <- match.arg(display, c("plot3D", "rgl")) 
    if (disp %in% "plot3D")
    {
        if (!requireNamespace("plot3D", quietly=TRUE)) stop("Install the plot3D package as it is required.", call.=FALSE) 
        for (i in 1:nc)
            if (hts[nc-i+1] < max(fhat$estimate))
                plot3D::isosurf3D(x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], colvar=fhat$estimate, level=hts[nc-i+1], add=add | (i>1), col=colors[i], alpha=alphavec[i], phi=phi, theta=theta, xlab=xlab, ylab=ylab, zlab=zlab, d=d, ticktype=ticktype, bty=bty, ...)
                
        if (drawpoints) plot3D::points3D(x=fhat$x[,1], y=fhat$x[,2], z=fhat$x[,3], cex=cex, col=col.pt, add=TRUE, pch=pch, d=d)    
    }
    else if (disp %in% "rgl")
    {   
        ## suggestions from Viktor Petukhov 08/03/2018
        if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
        if (!requireNamespace("misc3d", quietly=TRUE)) stop("Install the misc3d package as it is required.", call.=FALSE)
    
        fhat.eval.mean <- sapply(fhat$eval.points, mean)
        if (drawpoints)
            rgl::plot3d(fhat$x[,1],fhat$x[,2],fhat$x[,3], size=size, col=col.pt, alpha=alpha, xlab=xlab, ylab=ylab, zlab=zlab, add=add, box=FALSE, axes=FALSE, ...)
        else
            rgl::plot3d(fhat$x[,1],fhat$x[,2],fhat$x[,3], size=0, col="transparent", alpha=0, xlab=xlab, ylab=ylab, zlab=zlab, add=add, box=FALSE, axes=FALSE, ...)

        for (i in 1:nc)
            if (hts[nc-i+1] < max(fhat$estimate))
                misc3d::contour3d(fhat$estimate, level=hts[nc-i+1], x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], add=TRUE, color=colors[i], alpha=alphavec[i], box=FALSE, axes=FALSE, ...)
    
        if (axes) rgl::axes3d(c("x","y","z"))
        if (box) rgl::box3d()
    }
}

## contourLevels method
## create S3 generic 
contourLevels <- function(x, ...) UseMethod("contourLevels")    

contourLevels.kde <- function(x, prob, cont, approx=TRUE, ...)
{ 
    nlevels <- 5
    approx <- as.integer(approx)
    fhat <- x

    if (is.vector(fhat$x))
    {
        d <- 1; n <- length(fhat$x)
    }
    else
    {
        d <- ncol(fhat$x); n <-nrow(fhat$x)
        if (!is.matrix(fhat$x)) fhat$x <- as.matrix(fhat$x)
    }

    if (is.null(x$w)) w <- rep(1, n)
    else w <- x$w

    if (is.null(fhat$gridded))
    {
        if (d==1) fhat$gridded <- fhat$binned
        else fhat$gridded <- is.list(fhat$eval.points)
    }
  
    if (missing(prob) & missing(cont))
    {    
        hts <- pretty(fhat$estimate, n=nlevels)
        args <- list(...)
        if (any(pmatch(names(args), "nlevels"))) warning("nlevels=5 always so changing nlevels has no effect") 
    }
    else
    {
        if (approx & fhat$gridded)
            dobs <- predict(fhat, x=fhat$x)
        else
            dobs <- kde(x=fhat$x, H=fhat$H, eval.points=fhat$x, w=w)$estimate 
        
        if (approx<=1)
        {
            if (!missing(prob) & missing(cont))
                hts <- quantile(dobs, prob=prob)
          
            if (missing(prob) & !missing(cont))
                hts <- quantile(dobs, prob=(100-cont)/100)
        }
        else 
            hts <- contourLevelskde.grid(x=x, prob=prob, cont=cont)

    }
      
    return(hts)
}

## approximate contour levels using estimation grid only from "kde" object and not 
## original data sample  
contourLevelskde.grid <- function(x, prob, cont)
{ 
    fhat <- x
    if (!is.list(fhat$eval.points)) d <- 1 else d <- length(fhat$eval.points)
    fhat$estimate[fhat$estimate<0] <- 0
    
    ## compute Riemann probability for each bin 
    if (d==1)
    {
        del.mat <- c(0, diff(fhat$eval.points))
    }
    else
    {
        del <- lapply(fhat$eval.points, function(.) c(0, diff(.)))
        del.mat <- as.matrix(expand.grid(del))
        del.mat <- apply(del.mat, 1, prod)
        del.mat <- array(del.mat, dim=dim(fhat$estimate))
    }
    cestimate <- fhat$estimate*del.mat

    ## alpha density contour consists of ordered bins whose prob sum = alpha   
    cestimate.ord <- order(cestimate)
    cestimate[cestimate.ord] <- cumsum(cestimate[cestimate.ord]) 
    cestimate <- (cestimate-min(cestimate))/(max(cestimate)-min(cestimate))

    ## contour levels are min of density values of density contours
    if (missing(prob) & !missing(cont)) prob <- (100-cont)/100
    hts <- sapply(prob, function(.) fhat$estimate[which.min(abs(cestimate-.))]) 
    names(hts) <- paste0(prob*100, "%")  

    return(hts)
}


#############################################################################
## Riemann sums to compute approximate Lebesgue measure of contour set
#############################################################################
contourSizes <- function(x, abs.cont, cont=c(25,50,75), approx=TRUE)
{
    if (missing(abs.cont))
        abs.cont <- contourLevels(x, cont=cont, approx=approx)
    num.int <- rep(0, length(abs.cont))
    if (!is.null(names(abs.cont))) names(num.int) <- names(abs.cont) 
    
    if (!is.list(x$eval.points))
        delta.int <- head(diff(x$eval.points), n=1)
    else
        delta.int <- prod(sapply(lapply(x$eval.points, diff), head, n=1)) 
    
    for (j in 1:length(abs.cont)) 
        num.int[j] <- sum(x$estimate>abs.cont[j])
    
    return(num.int*delta.int)
}

contourBreaks <- function(x, n=3, cont=c(25,50,75), type="density")
{
    is.kdde <- FALSE 
    #if (!inherits(x, c("kde", "kqde"))) message("contourBreaks is defined rigorously for `kde' objects only")
    cb <- .contourBreaks(x=x, n=n, cont=cont, type=type, is.kdde=is.kdde)

    return(cb)
}

.contourBreaks <- function(x, n=3, cont=c(25,50,75), type="density", is.kdde=FALSE)
{
    type <- match.arg(type, c("density", "length", "quantile", "natural"))
    xe <- x
    if (any(names(x) %in% "estimate")) xe <- unlist(x$estimate)
    neg.prop <- abs(mean(xe[xe<=0]))/abs(mean(xe[xe>0]))

    if (type=="density")
        clev <- c(min(xe, na.rm=TRUE), contourLevels(x, cont=cont), max(xe, na.rm=TRUE))
    else if (type=="length")
        clev <- seq(min(xe, na.rm=TRUE), max(xe, na.rm=TRUE), length=n+2) 
    else if (type=="quantile")
    {
        prob <- sort(unique((100-cont)/100)) 
        if (neg.prop > 0.01) 
            clev <- c(min(xe, na.rm=TRUE), quantile(xe[xe<=0], prob=prob), quantile(xe[xe>0], prob=prob), max(xe, na.rm=TRUE))
        else
            clev <- c(min(xe, na.rm=TRUE), quantile(xe, prob=prob), max(xe, na.rm=TRUE))
    }
    else if (type=="natural")
    {
        ## set local seed
        if (!exists(".Random.seed")) rnorm(1)
        current.seed <- .Random.seed
        on.exit( { .Random.seed <<- current.seed } )
        set.seed(8192)
        clev <- c(min(xe, na.rm=TRUE), sort(as.vector(kmeans(as.vector(xe), centers=max(1,n), nstart=10)$centers)), max(xe, na.rm=TRUE))
    }
    
    clev <- sort(clev)
    npos <- sum(clev>=0)
    nneg <- sum(clev<0)
    include.min <- is.kdde | (nneg==1 & neg.prop>0.01)
    include.max <- is.kdde | npos==1
    if (!include.min) clev <- tail(clev, n=-1)
    if (!include.max) clev <- head(clev, n=-1)
    clev <- unique(clev)
    
    return(clev)
}


#############################################################################
## Riemann sums to compute approximate probability of contour set
#############################################################################
contourProbs <- function(x, abs.cont, cont=c(25,50,75), approx=TRUE)
{
    if (missing(abs.cont)) 
        abs.cont <- contourLevels(x, cont=cont, approx=approx)
    num.int <- rep(0, length(abs.cont))
    if (!is.null(names(abs.cont))) 
        names(num.int) <- names(abs.cont)
    if (!is.list(x$eval.points)) 
    {
        delta.int <- head(diff(x$eval.points), n=1)
        eval.points.midpoint <- (head(x$eval.points,n=-1)+tail(x$eval.points,n=-1))/2
    }
    else
    {
        delta.int <- prod(sapply(lapply(x$eval.points, diff), head, n=1)) 
        eval.points.midpoint <- expand.grid(lapply(x$eval.points, function(y) { (head(y,n=-1)+tail(y,n=-1))/2 }))
    }
    
    x.evmp <- predict(x, x=eval.points.midpoint)
    for (i in 1:length(num.int)) num.int[i] <- sum(x.evmp*(x.evmp>=abs.cont[i])*delta.int)
    
    return(num.int)
}
 
#############################################################################
## Generate grid over a set of points
##
## Parameters
## x - data points
## H - bandwidth matrix
## tol - tolerance = extra coverage exceeding the range of x   
## gridsize - number of points for each direction
##
## Returns
## gridx - list of intervals, one for each co-ord direction so that
##         gridx[[1]] x gridx[[2]] x ... x gridx[[d]] is the grid
## stepsize - vector of step sizes 
#############################################################################
make.grid.ks <- function(x, H, tol, gridsize, xmin, xmax, gridtype)
{
    d <- ncol(x)
    tol.H <-  tol * diag(H)
    if (missing(xmin)) xmin <- apply(x, 2, min) - tol.H
    if (missing(xmax)) xmax <- apply(x, 2, max) + tol.H

    stepsize <- rep(0, d)
    gridx <- numeric(0)
    if (length(gridsize)==1)  gridsize <- rep(gridsize, d)
    if (missing(gridtype)) gridtype <- rep("linear", d)
    gridtype.vec <- rep("", d)

    for (i in 1:d)
    {
        gridtype1 <- match.arg(gridtype[i], c("linear", "sqrt", "quantile", "exp")) 
        if (gridtype1=="linear")
        {  
            gridx <- c(gridx, list(seq(xmin[i], xmax[i], length=gridsize[i])))
            stepsize[i] <- abs(gridx[[i]][1] - gridx[[i]][2])
        }
        else if (gridtype1=="sqrt")
        {
            gridx.temp <- seq(sign(xmin[i])*sqrt(abs(xmin[i])), sign(xmax[i])*sqrt(abs(xmax[i])), length=gridsize[i])
            gridx <- c(gridx, list(sign(gridx.temp) * gridx.temp^2))
            stepsize[i] <- NA
            }
        else if (gridtype1=="quantile")
        {
            gridx.temp <- qnorm(seq(1e-2, 1-1e-2, length=gridsize[i]))
            gridx <- c(gridx, list(xmin[i] + (xmax[i]-xmin[i])*(gridx.temp-min(gridx.temp))/(max(gridx.temp)-min(gridx.temp))))
            stepsize[i] <- NA
        }
        else if (gridtype1=="exp")
        {
            gridx.temp <- seq(exp(xmin[i]), exp(xmax[i]), length=gridsize[i])
            gridx <- c(gridx, list(log(gridx.temp)))
            stepsize[i] <- NA
        }
        gridtype.vec[i] <- gridtype1
    }
    gridx <- c(gridx, list(stepsize = stepsize, gridtype=gridtype.vec))

    return(gridx)
}  

#############################################################################
## Generate kernel (rectangular) support at data point
## 
## Parameters
## x - data points
## H - bandwidth matrix
## tol - tolerance = extra coverage exceeding the range of x 
##
## Returns
## list of min and max points of support (here we parameterise rectangles
## by their min = lower left co-ord and max = upper right coord)
#############################################################################
make.supp <- function(x, H, tol)
{
    n <- nrow(x)
    d <- ncol(x)
    tol.H <- tol * diag(H)
    xmin <- matrix(0, nrow=n, ncol=d)
    xmax <- matrix(0, nrow=n, ncol=d)

    for (i in 1:n)
    {
        xmin[i,] <- x[i,] - tol.H
        xmax[i,] <- x[i,] + tol.H 
    }
           
    return(list(xmin=xmin, xmax=xmax))
}

#############################################################################
## Find the grid points contained in kernel support rectangles 
##
## Parameters
## gridx - grid (list of subdivided intervals)
## rectx - rectangles (list of min and max points) 
##
## Returns
## list of min and max points of the grid for each rectangle 
#############################################################################
find.gridpts <- function(gridx, suppx)
{
    xmax <- suppx$xmax
    xmin <- suppx$xmin
    d <- ncol(xmax)
    n <- nrow(xmax)
    gridpts.min <- matrix(0, ncol=d, nrow=n)
    gridpts.max <- gridpts.min
  
    for (i in 1:n)
        for (j in 1:d)    
        {
            ## find index of last element of gridx smaller than min support  
            tsum <- sum(xmin[i,j] >= gridx[[j]])
            if (tsum==0)
                gridpts.min[i,j] <- 1
            else
                gridpts.min[i,j] <- tsum

            ## find index of first element gridx greater than max support 
            gridpts.max[i,j] <- sum(xmax[i,j] >= gridx[[j]])
        }   
        
    return(list(xmin=gridpts.min, xmax=gridpts.max))
} 

#############################################################################
## Interpolate the values of f defined on gridx at new values x 
#############################################################################
grid.interp <- function(x, gridx, f)
{
    if (!is.list(gridx))
    {
        ## uniform grid
        if (isTRUE(all.equal(diff(gridx), rep(diff(gridx)[1], length(gridx)-1))))
            fx <- grid.interp.1d(x=as.vector(x), gridx=gridx, f=f)
        ## non-uniform grid
        else
            fx <- varying.grid.interp.1d(x=as.vector(x), gridx=gridx, f=f)
    
    }
    else
    {  
        if (is.vector(x)) x <- as.matrix(t(x))
        d <- ncol(x)
        n <- nrow(x)

        if (d<2) stop("x should be a vector") 
        gridx.diff <- lapply(lapply(gridx,diff), getElement, 1)
        for (i in 1:length(gridx.diff)) gridx.diff[[i]] <- rep(gridx.diff[[i]], sapply(gridx, length)[i]-1)
        uniform.grid.flag <- isTRUE(all.equal(lapply(gridx,diff), gridx.diff))

        ## uniform grid
        if (d==2 & uniform.grid.flag) fx <- grid.interp.2d(x=x, gridx=gridx, f=f)
        else if (d==3 & uniform.grid.flag) fx <- grid.interp.3d(x=x, gridx=gridx, f=f)
        else ## d >=4 or non-uniform grid
        {    
            gridsize <- sapply(gridx,length)
            gind <- matrix(0, nrow=n, ncol=d)
            
            for (i in 1:n)
                for (j in 1:d)
                {
                    tsum <- sum(x[i,j] >= gridx[[j]])
                    if (tsum==0) gind[i,j] <- 1
                    else gind[i,j] <- tsum
                }
            for (j in 1:d) gind[gind[,j]>=gridsize[j],j] <- gridsize[j]-1

            bperm <- list()
            for (j in 1:d) bperm[[j]] <- elem(1,2)
            binary.perm <- as.matrix(expand.grid(bperm))
            colnames(binary.perm) <- NULL
            
            gind.list <- list()
            fx <- rep(0, length=n)
            for (i in 1:n)
            {
                gind.list[[i]] <- matrix(gind[i,], nrow=2^d, ncol=d, byrow=TRUE) + binary.perm
                w <- matrix(0, nrow=2^d, ncol=d)
                gridw <- matrix(0, nrow=2^d, ncol=d)
                for (j in 1:d)
                {
                    gind.list[[i]][,j][gind.list[[i]][,j]>=gridsize[j]] <- gridsize[j]
                    gridw[,j] <- gridx[[j]][gind.list[[i]][,j]]
                }
                w <- abs(matrix(as.numeric(x[i,]), nrow=2^d, ncol=d, byrow=TRUE) - gridw)
                w <- apply(w, 1, prod)
                w <- w/sum(w)
                fx[i] <- sum(w*f[gind.list[[i]][2^d:1,]])
            }
        }
    }
  
    return(fx)
}

grid.interp.1d <- function(x, gridx, f)
{
   n <- length(x)
   gpoints1 <- gridx
   M1 <- length(gpoints1)
   a1 <- gpoints1[1]
   b1 <- gpoints1[M1]
   
   out <- .C(C_interp1d, x1=as.double(x), n=as.integer(n), a1=as.double(a1), b1=as.double(b1), M1=as.integer(M1), fun=as.double(as.vector(f)), est=double(n))
   return(out$est)
}

grid.interp.2d <- function(x, gridx, f)
{
   n <- nrow(x)
   gpoints1 <- gridx[[1]]
   gpoints2 <- gridx[[2]]
   M1 <- length(gpoints1)
   M2 <- length(gpoints2)
   a1 <- gpoints1[1]
   a2 <- gpoints2[1]
   b1 <- gpoints1[M1]
   b2 <- gpoints2[M2]

   out <- .C(C_interp2d, x1=as.double(x[,1]), x2=as.double(x[,2]), n=as.integer(n), a1=as.double(a1), a2=as.double(a2), b1=as.double(b1), b2=as.double(b2), M1=as.integer(M1), M2=as.integer(M2), fun=as.double(as.vector(f)), est=double(n))
   
   return(out$est)
}

grid.interp.3d <- function(x, gridx, f)
{
   n <- nrow(x)
   gpoints1 <- gridx[[1]]
   gpoints2 <- gridx[[2]]
   gpoints3 <- gridx[[3]]
   M1 <- length(gpoints1)
   M2 <- length(gpoints2)
   M3 <- length(gpoints3)
   a1 <- gpoints1[1]
   a2 <- gpoints2[1]
   a3 <- gpoints3[1]
   b1 <- gpoints1[M1]
   b2 <- gpoints2[M2]
   b3 <- gpoints3[M3]
   
   out <- .C(C_interp3d, x1=as.double(x[,1]), x2=as.double(x[,2]), x3=as.double(x[,3]), n=as.integer(n), a1=as.double(a1), a2=as.double(a2), a3=as.double(a3), b1=as.double(b1), b2=as.double(b2), b3=as.double(b3), M1=as.integer(M1), M2=as.integer(M2), M3=as.integer(M3), fun=as.double(as.vector(f)), est=double(n))
   
   return(out$est)
}

## Linear intepolation based on kernel estimation grid
## alias for predict.kde
kde.approx <- function(fhat, x)
{
    return(grid.interp(x=x, gridx=fhat$eval.points, f=fhat$estimate))
}

#############################################################################
## Find the nearest grid points surrounding point x for non-uniform grids 
#############################################################################
varying.grid.interp <- function(x, gridx, f)
{
    if (!is.list(gridx))
        return(varying.grid.interp.1d(x=x, gridx=gridx, f=f))
    else
    {  
        if (is.vector(x)) x <- as.matrix(t(x))
        d <- ncol(x)
        n <- nrow(x)
        gridsize <- sapply(gridx,length)
        gind <- matrix(0, nrow=n, ncol=d)

        for (i in 1:n)
            for (j in 1:d)
            {
                tsum <- sum(x[i,j] >= gridx[[j]])
                if (tsum==0) gind[i,j] <- 1
                else gind[i,j] <- tsum
            }
    }

    bperm <- list()
    for (j in 1:d) bperm[[j]] <- elem(1,2)
    binary.perm <- as.matrix(expand.grid(bperm))
    colnames(binary.perm) <- NULL

    gind.list <- list()
    fx <- rep(0, length=n)
    for (i in 1:n)
    {
        gind.list[[i]] <- matrix(gind[i,], nrow=2^d, ncol=d, byrow=TRUE) + binary.perm
        w <- matrix(0, nrow=2^d, ncol=d)
        gridw <- matrix(0, nrow=2^d, ncol=d)
        for (j in 1:d)
        {
            gind.list[[i]][,j][gind.list[[i]][,j]>=gridsize[j]] <- gridsize[j]
            gridw[,j] <- gridx[[j]][gind.list[[i]][,j]]
        }
        w <- 1/apply((matrix(as.numeric(x[i,]), nrow=2^d, ncol=d, byrow=TRUE) - gridw)^2, 1, sum)
        w[w>1e5] <- 1e5
        w <- w/sum(w)
        fx[i] <- sum(w*f[gind.list[[i]]])
    }

    return(fx)
}

varying.grid.interp.1d <- function(x, gridx, f)
{
    n <- length(x)
    gind <- rep(0, length=n)

    for (i in 1:length(x))
    {
        tsum <- sum(x[i] >= gridx)
        if (tsum==0) gind[i] <- 1
        else gind[i] <- tsum
    }
 
    gind2 <- gind+1
    gind2[gind2>length(gridx)] <- length(gridx)
    gind2[x<=gridx[1]] <- gind[x<=gridx[1]]
    gind <- cbind(gind, gind2)
    colnames(gind) <- NULL

    fx <- rep(0, n)
    for (i in 1:n)
    {
        w <- 1/(x[i] - gridx[gind[i,]])^2
        w[w>1e5] <- 1e5
        w <- w/sum(w)   
        fx[i] <- sum(w*f[gind[i,]])
    }

  return(fx)
}

#############################################################################
## Convert estimation grid to quasi-kernel density estimator
## Allow estimate to be negative
#############################################################################

as.kde <- function(x, density)
{
    ## eval.points is tidy matrix 
    if (inherits(x, c("data.frame", "matrix","numeric")))
    {
        d <- ncol(x)-1
        eval.points <- x[,1:d]
        ## reduce #s.f. to ensure exactly replicated values in eval.points
        eval.points <- signif(eval.points, 10)
        estimate <- x[,(d+1)]
        
        if (!all(is.na(names(eval.points))) & d>1) xnames <- names(eval.points) 
        else 
        {
            if (d<=3) xnames <- c("x","y","z")[1:d]
            else xnames <- paste0("x[,", 1:d,"]")
        }
        if (d>1)
        {
            eval.points <- lapply(1:d, function(.) sort(unique(unlist(eval.points[,.]))))
            names(eval.points) <- xnames
            estimate <- array(estimate, dim=lengths(eval.points))
        }
        
    }
    else 
    {
        eval.points <- x$eval.points
        estimate <- x$estimate
        if (is.matrix(x$estimate)) d <- length(x$eval.points) else d <- 1
        if (!all(is.na(names(eval.points))) & d>1) xnames <- names(eval.points) 
        else
        {
            if (d<=3) xnames <- c("x","y","z")[1:d]
            else xnames <- paste0("x[,", 1:d,"]")
        }
    }

    if (d==1) x <- 0
    else 
    {
        x <- matrix(0, ncol=d, nrow=1) 
        colnames(x) <- xnames
    }

    if (missing(density)) 
    {
        if (all(estimate>=0)) density <- TRUE
        else if (all(estimate<0)) density <- FALSE
        else density <- (abs(mean(estimate[estimate<0]))/mean(estimate[estimate>=0]) < 1e-6) 
    }

    if (density)
    {
        estimate[estimate<0] <- 0
        fhat <- list(x=x, estimate=estimate, eval.points=eval.points)
        fhat$gridded <- TRUE
        fhat$cont <- contourLevels.kqde(fhat, cont=1:99)
        fhat$names <- xnames
        fhat$type <- "kde" 
        class(fhat) <- "kqde"
    }
    else
    {
        fhat <- list(x=x, estimate=estimate, eval.points=eval.points)
        fhat$gridded <- TRUE
        fhat$cont <- contourLevels.kqdde(fhat, cont=1:99)
        fhat$estimate <- list(estimate)
        fhat$names <- xnames
        fhat$deriv.order <- 1L
        fhat$deriv.ind <- matrix(1, nrow=1) 
        fhat$type <- "kdde" 
        class(fhat) <- "kqdde"
    }
    
    return(fhat)
}

plot.kqde <- plot.kde 
predict.kqde <- predict.kde

contourLevels.kqde <- function(x, prob, cont, ...) 
{ 
    contourLevelskde.grid(x=x, prob=prob, cont=cont) 
}

plot.kqdde <- plot.kdde 
predict.kqdde <- predict.kdde

contourLevels.kqdde <- function(x, prob, cont, ...) 
{ 
    if (is.list(x$estimate)) x$estimate <- x$estimate[[1]]
    xpos <- xneg <- x
    xpos$estimate[xpos$estimate<0] <- 0 
    xneg$estimate[xneg$estimate>=0] <- 0 
    xneg$estimate <- -xneg$estimate 
    xpos.ct <- contourLevelskde.grid(x=xpos, prob=prob, cont=cont)
    xneg.ct <- contourLevelskde.grid(x=xneg, prob=prob, cont=cont)
    names.ct <- names(contourLevelskde.grid(x=xpos, prob=prob, cont=cont))

    xpos.ct[lengths(xpos.ct)==0] <- min(abs(x$estimate[abs(x$estimate)!=0]))
    xneg.ct[lengths(xneg.ct)==0] <- min(abs(x$estimate[abs(x$estimate)!=0]))
    if (is.list(xpos.ct)) xpos.ct <- unlist(xpos.ct)
    if (is.list(xneg.ct)) xneg.ct <- unlist(xneg.ct)
    xct <- rbind(-xneg.ct, xpos.ct)
    colnames(xct) <- names.ct
    rownames(xct) <- NULL

    return(xct)
}

## create S3 generic 
addContourLabels <- function(x, ...) UseMethod("addContourLabels")

addContourLabels.kde <- function(x, cont, abs.cont, ...)
{
    if (missing(abs.cont)) abs.cont <- contourLevels(x, cont=cont)
    else cont <- abs.cont
    abs.cont <- sort(c(min(x$estimate)-0.1*abs(max(x$estimate)), abs.cont, max(x$estimate)+0.1*abs(max(x$estimate))))
    contlabel <- sort(unique(c(0,cont)))
    xcontlabel <- cut(x$estimate, abs.cont, right=TRUE, labels=contlabel)
    xcontlabel <- array(as.numeric(levels(xcontlabel))[xcontlabel], dim=dim(x$estimate))
    x$estimate <- xcontlabel

    return(x)
}

#####################################################################
## Kernel estimators of the multivariate cdf (cumulative distribution function)
#####################################################################
kcde <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, positive=FALSE, adj.positive, w, verbose=FALSE, tail.flag="lower.tail")
{
    ## default values 
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    gridsize <- ksd$gridsize
    bgridsize <- ksd$bgridsize
    tail.flag1 <- match.arg(tail.flag, c("lower.tail", "upper.tail")) 

    ## clip data to xmin, xmax grid for binned estimation
    grid.clip <- binned
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        if (positive & missing(xmin)) { xmin <- rep(0,d) }
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }
    
    ## default bandwidths
    if (d==1 & missing(h)) 
    {
        if (positive) x1 <- log(x) else x1 <- x
        h <- hpi.kcde(x=x1, nstage=2, binned=default.bflag(d=d, n=n))
    }
    if (d>1 & missing(H))
    {
        if (positive) x1 <- log(x) else x1 <- x
        H <- Hpi.kcde(x=x1, binned=default.bflag(d=d, n=n), bgridsize=bgridsize, verbose=FALSE)
    }
    
    ## KCDE is computed as cumulative Riemann sum of KDE on a grid
    if (d==1)
    {
        Fhat <- kde(x=x, h=h, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, binned=binned, bgridsize=bgridsize, positive=positive, adj.positive=adj.positive, w=w)
        if (positive) 
        {
            ep <- seq(Fhat$eval.points[1], tail(Fhat$eval.points,n=1), length=length(Fhat$eval.points))
            Fhat$estimate <- predict(Fhat, x=ep)
            Fhat$eval.points <- ep
        }
        diffe <- abs(diff(Fhat$eval.points))
        if (tail.flag1=="lower.tail") Fhat$estimate <- c(0, diffe) * cumsum(Fhat$estimate)
        else Fhat$estimate <- c(diffe[1], diffe) * (sum(Fhat$estimate) - cumsum(Fhat$estimate))
    }
    else if (d==2)
    {
        Fhat <- kde(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, binned=binned, bgridsize=bgridsize, w=w, verbose=verbose)
        diffe1 <- abs(diff(Fhat$eval.points[[1]]))
        diffe2 <- abs(diff(Fhat$eval.points[[2]]))

        if (tail.flag1=="lower.tail")
        {
            Fhat$estimate <- apply(Fhat$estimate, 1, cumsum)*c(0,diffe1)
            Fhat$estimate <- apply(t(Fhat$estimate), 2, cumsum)*c(0,diffe2)
        }
        else
        {
            Fhatsum <- matrix(apply(Fhat$estimate, 1, sum), ncol=ncol(Fhat$estimate), nrow=nrow(Fhat$estimate), byrow=TRUE)
            Fhat$estimate <- (Fhatsum-apply(Fhat$estimate, 1, cumsum))*c(diffe1[1], diffe1)
            Fhatsum <- matrix(apply(Fhat$estimate, 1, sum), ncol=ncol(Fhat$estimate), nrow=nrow(Fhat$estimate), byrow=TRUE)
            Fhat$estimate <- (Fhatsum-apply(t(Fhat$estimate), 2, cumsum))*c(diffe2[1], diffe2)
        }
    }
    else if (d==3)
    {
        if (missing(H) & !positive) H <- Hpi.kcde(x=x, binned=default.bflag(d=d, n=n), bgridsize=bgridsize, verbose=FALSE)

        Fhat <- kde(x=x, H=H, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, binned=binned, bgridsize=bgridsize, w=w, verbose=verbose)
        Fhat.temp <- Fhat$estimate
        diffe1 <- abs(diff(Fhat$eval.points[[1]]))
        diffe2 <- abs(diff(Fhat$eval.points[[2]]))
        diffe3 <- abs(diff(Fhat$eval.points[[3]]))
        if (tail.flag1=="lower.tail")
        {
            for (i in 1:dim(Fhat$estimate)[3])
            {
                Fhat.temp[,,i] <- apply(Fhat.temp[,,i], 1, cumsum)*c(0,diffe1)
                Fhat.temp[,,i] <- apply(t(Fhat.temp[,,i]), 2, cumsum)*c(0,diffe2)
            }
            for (i in 1:dim(Fhat$estimate)[1])
                for (j in 1:dim(Fhat$estimate)[2])
                    Fhat.temp[i,j,] <- cumsum(Fhat.temp[i,j,])*c(0,diffe3)
            Fhat$estimate <- Fhat.temp
        }
        else
        {
            for (i in 1:dim(Fhat$estimate)[3])
            {
                Fhatsum <- matrix(apply(Fhat.temp[,,i], 1, sum), ncol=ncol(Fhat.temp), nrow=nrow(Fhat.temp), byrow=TRUE)
                Fhat.temp[,,i] <- (Fhatsum-apply(Fhat.temp[,,i], 1, cumsum))*c(diffe1[1], diffe1)
                Fhatsum <- matrix(apply(Fhat.temp[,,i], 1, sum), ncol=ncol(Fhat.temp), nrow=nrow(Fhat.temp), byrow=TRUE)
                Fhat.temp[,,i] <- (Fhatsum-apply(t(Fhat.temp[,,i]), 2, cumsum))*c(diffe2[1],diffe2)
            }

            for (i in 1:dim(Fhat$estimate)[1])
                for (j in 1:dim(Fhat$estimate)[2])
                {
                    Fhatsum <- sum(Fhat.temp[i,j,])
                    Fhat.temp[i,j,] <- (Fhatsum-cumsum(Fhat.temp[i,j,]))*c(diffe3[1],diffe3)
                }
            Fhat$estimate <- Fhat.temp
        }
    }
    ## normalise max CDF estimate equal to 1
    Fhat$estimate <- Fhat$estimate/max(Fhat$estimate)

    if (!missing(eval.points))
    {
        if (d<=3)
        {
            Fhat$estimate <- predict(Fhat, x=eval.points)
            Fhat$eval.points <- eval.points
        }
        else
        {
            Fhat <- kcde.points(x=x, H=H, eval.points=eval.points, w=w, verbose=verbose, tail.flag=tail.flag1)
        }
    }
    Fhat$tail <- tail.flag1
    Fhat$type <- "kcde"
    class(Fhat) <- "kcde"
    
    return(Fhat) 
}

## KCDE is computed at specified estimation points
kcde.points <- function(x, H, eval.points, w, verbose=FALSE, tail.flag="lower.tail") 
{
    n <- nrow(x)
    if (verbose) pb <- txtProgressBar() 
    Fhat <- rep(0, nrow(eval.points))
    pmvnorm.temp <- function(x, ...) { return(pmvnorm(mean=x, ...)) }

    for (i in 1:nrow(eval.points))
    {  
        if (verbose) setTxtProgressBar(pb, i/(nrow(eval.points)-1))
        if (tail.flag=="lower.tail")
            Fhat[i] <- sum(apply(x, 1, pmvnorm.temp, upper=eval.points[i,], sigma=H))
        else
            Fhat[i] <- sum(apply(x, 1, pmvnorm.temp, lower=eval.points[i,], sigma=H))
    }
    Fhat <- Fhat/n
    if (verbose) close(pb)

    return(list(x=x, eval.points=eval.points, estimate=Fhat, H=H, gridded=FALSE, binned=FALSE, names=NULL, w=w))
}

#####################################################################
## Plotting functions for 1-d to 3-d KCDE
#####################################################################
plot.kcde <- function(x, ...)
{ 
    Fhat <- x
    if (is.vector(Fhat$x)) plotkcde.1d(Fhat, ...)
    else
    {
        d <- ncol(Fhat$x)

        if (d==2) 
        {
            opr <- options()$preferRaster; if (!is.null(opr)) if (!opr) options("preferRaster"=TRUE)
            plotret <- plotkcde.2d(Fhat, ...)
            if (!is.null(opr)) options("preferRaster"=opr) 
            invisible(plotret)
        }
        else if (d==3)
        {
            plotkcde.3d(Fhat, ...)
            invisible()
        }
        else stop ("kde.plot function only available for 1, 2 or 3-d data")
    }
}

plotkcde.1d <- function(Fhat, xlab, ylab="Distribution function", add=FALSE, drawpoints=FALSE, col=1, col.pt=4, jitter=FALSE, alpha=1, ...) 
{
    if (missing(xlab)) xlab <- Fhat$names
    if (Fhat$tail=="upper.tail") zlab <- "Survival function"
    col <- transparency.col(col, alpha=alpha)
    if (add) lines(Fhat$eval.points, Fhat$estimate, xlab=xlab, ylab=ylab, col=col, ...)
    else plot(Fhat$eval.points, Fhat$estimate, type="l", xlab=xlab, ylab=ylab, col=col, ...) 

    if (drawpoints)
        if (jitter) rug(jitter(Fhat$x), col=col.pt)
        else rug(Fhat$x, col=col.pt)
}

plotkcde.2d <- function(Fhat, display="persp", cont=seq(10,90, by=10), abs.cont,
    xlab, ylab, zlab="Distribution function", cex=1, pch=1,   
    add=FALSE, drawpoints=FALSE, drawlabels=TRUE, theta=-30, phi=40, d=4,
    col.pt=4, col, col.fun, alpha=1, lwd=1, border=NA, thin=3, labcex=1, ticktype="detailed", ...) 
{
    disp <- match.arg(display, c("slice", "persp", "image", "filled.contour", "filled.contour2"))
    if (disp=="filled.contour2") disp <- "filled.contour"
  
    if (!is.list(Fhat$eval.points)) stop("Needs a grid of density estimates")

    if (missing(xlab)) xlab <- Fhat$names[1]
    if (missing(ylab)) ylab <- Fhat$names[2]
    if (Fhat$tail=="upper.tail") zlab <- "Survival function"
  
    ## perspective/wireframe plot
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n, palette="viridis", alpha=alpha) }
   
    if (disp=="persp")
    {
        hts <- seq(0, 1.1*max(Fhat$estimate, na.rm=TRUE), length=500)
        if (missing(col)) col <- col.fun(n=length(hts)+1)
        if (length(col)<(length(hts)+1)) col <- rep(col, length=length(hts)+1)
        col <- transparency.col(col, alpha=alpha)
        
        ## thinning indices
        plot.ind <- list(seq(1, length(Fhat$eval.points[[1]]), by=thin), seq(1, length(Fhat$eval.points[[2]]), by=thin))

        z <- Fhat$estimate[plot.ind[[1]], plot.ind[[2]]]
        nrz <- nrow(z)
        ncz <- ncol(z)
        zfacet <- z[-1, -1] + z[-1, -ncz] + z[-nrz, -1] + z[-nrz, -ncz]
        facetcol <- cut(zfacet, length(hts)+1, labels=FALSE)
        plotret <- persp(Fhat$eval.points[[1]][plot.ind[[1]]], Fhat$eval.points[[2]][plot.ind[[2]]], z, theta=theta, phi=phi, d=d, xlab=xlab, ylab=ylab, zlab=zlab, col=col[facetcol], border=border, ticktype=ticktype, ...)
    }
    else if (disp=="slice") 
    {
        if (!add) plot(Fhat$x[,1], Fhat$x[,2], type="n", xlab=xlab, ylab=ylab, ...)
 
        if (missing(abs.cont)) hts <- cont/100
        else hts <- abs.cont
    
        if (missing(col)) col <- col.fun(n=length(hts))
        if (length(col)<length(hts)) col <- rep(col, times=length(hts))
        col <- transparency.col(col, alpha=alpha)
         
        ## draw contours         
        for (i in 1:length(hts)) 
        {
            if (missing(abs.cont)) scale <- cont[i]/hts[i]
            else scale <- 1
            if (hts[i]>0)
                contour(Fhat$eval.points[[1]], Fhat$eval.points[[2]], Fhat$estimate*scale, level=hts[i]*scale, add=TRUE, drawlabels=drawlabels, col=col[i], lwd=lwd, labcex=labcex, ...)
        }
        ## add points 
        if (drawpoints) points(Fhat$x[,1], Fhat$x[,2], col=col.pt, cex=cex, pch=pch)
    }
    ## image plot
    else if (disp=="image")
    {
        if (missing(col)) col <- col.fun(100)
        col <- transparency.col(col, alpha=alpha)
        image(Fhat$eval.points[[1]], Fhat$eval.points[[2]], Fhat$estimate, xlab=xlab, ylab=ylab, add=add, col=col, ...)
        box()
    }
    else if (disp=="filled.contour") 
    {
        hts <- cont/100
        clev <- c(-0.01*max(abs(Fhat$estimate)), hts, max(c(Fhat$estimate, hts)) + 0.01*max(abs(Fhat$estimate)))
        if (missing(col)) col <- col.fun(length(hts))
        col <- transparency.col(col, alpha=alpha)
       
        if (!add) plot(Fhat$eval.points[[1]], Fhat$eval.points[[2]], type="n", xlab=xlab, ylab=ylab, ...) 
        if (tail(hts, n=1) < max(Fhat$estimate)) hts2 <- c(hts, max(Fhat$estimate))
       
        .filled.contour(Fhat$eval.points[[1]], Fhat$eval.points[[2]], Fhat$estimate, levels=hts2, col=col)
      
        if (!missing(lwd))
        {
            for (i in 1:length(hts)) 
            {
                if (missing(abs.cont)) scale <- (100-cont[i])/hts[i]
                else scale <- 1
          
                if (lwd >=1) contour(Fhat$eval.points[[1]], Fhat$eval.points[[2]], Fhat$estimate*scale, level=hts[i]*scale, add=TRUE, drawlabels=drawlabels, col=1, lwd=lwd, labcex=labcex, ...)
            }
        }
    }
    if (disp=="persp")  invisible(plotret)
    else invisible()
}
  
plotkcde.3d <- function(Fhat, display="plot3D", cont=c(25,50,75), colors, col, alphavec, size=3, cex=1, pch=1, theta=-30, phi=40, d=4, ticktype="detailed", bty="f", col.pt=4, add=FALSE, xlab, ylab, zlab, drawpoints=FALSE, alpha, box=TRUE, axes=TRUE, ...)
{
    disp <- match.arg(display, c("plot3D", "rgl"))
    hts <- sort(cont/100)
    nc <- length(hts)
    
    if (missing(col)) 
    { 
        col.fun <- function(n) { hcl.colors(n, palette="viridis") }
        col <- col.fun(n=length(hts))
    }
    colors <- col
    if (missing(xlab)) xlab <- Fhat$names[1]
    if (missing(ylab)) ylab <- Fhat$names[2]
    if (missing(zlab)) zlab <- Fhat$names[3]
    if (missing(alphavec)) alphavec <- seq(0.5,0.1,length=nc)
    if (missing(alpha)) alpha <- 0.5 
    if (!missing(alpha)) { alphavec <- rep(alpha,nc) }
    
    disp <- match.arg(display, c("plot3D", "rgl")) 
    if (disp %in% "plot3D")
    {
        if (!requireNamespace("plot3D", quietly=TRUE)) stop("Install the plot3D package as it is required.", call.=FALSE)
        for (i in 1:nc)
            if (hts[nc-i+1] < max(Fhat$estimate))
                plot3D::isosurf3D(x=Fhat$eval.points[[1]], y=Fhat$eval.points[[2]], z=Fhat$eval.points[[3]], colvar=Fhat$estimate, level=hts[nc-i+1], add=add | (i>1), col=colors[nc-i+1], alpha=alphavec[i], phi=phi, theta=theta, xlab=xlab, ylab=ylab, zlab=zlab, d=d, ticktype=ticktype, bty=bty, ...)
                
        if (drawpoints) plot3D::points3D(x=Fhat$x[,1], y=Fhat$x[,2], z=Fhat$x[,3], cex=cex, col=col.pt, add=TRUE, pch=pch, d=d)    
    }
    else if (disp %in% "rgl")
    {
        ## suggestions from Viktor Petukhov 08/03/2018
        if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
        if (!requireNamespace("misc3d", quietly=TRUE)) stop("Install the misc3d package as it is required.", call.=FALSE)
    
        if (drawpoints)
            rgl::plot3d(Fhat$x[,1],Fhat$x[,2],Fhat$x[,3], size=size, col=col.pt, alpha=alpha, xlab=xlab, ylab=ylab, zlab=zlab, add=add, box=FALSE, axes=FALSE, ...)
        else
            rgl::plot3d(Fhat$x[,1],Fhat$x[,2],Fhat$x[,3], type="n", xlab=xlab, ylab=ylab, zlab=zlab, add=add, box=FALSE, axes=FALSE, ...)
        rgl::bg3d(col="white")
    
        for (i in 1:nc)
            if (hts[nc-i+1] < max(Fhat$estimate))
                misc3d::contour3d(Fhat$estimate, level=hts[nc-i+1], x=Fhat$eval.points[[1]], y=Fhat$eval.points[[2]], z=Fhat$eval.points[[3]], add=TRUE, color=colors[nc-i+1], alpha=alphavec[i], box=FALSE, axes=FALSE, ...)
    
        if (box) rgl::box3d()
        if (axes) rgl::axes3d(c("x","y","z"))
    }
}   

#####################################################################
## Bandwidth selectors for KCDE
#####################################################################
## Normal scale bandwidth selectors
hns.kcde <- function(x)
{
    d <- 1
    n <- length(x)
    sigma <- sd(x)
    hns <- 4^(1/3)*sigma*n^(-1/3)

    return(hns)
}

Hns.kcde <- function(x)
{
    if (is.vector(x)) { return(hns.kcde(x)^2) }
    d <- ncol(x)
    n <- nrow(x)
    m1 <- (4*pi)^(-1/2)

    Jd <- matrix(1, ncol=d, nrow=d)
    Sigma <- var(x)
    Hns <- (4*det(Sigma)^(1/2)*tr(matrix.sqrt(Sigma))/tr(Sigma))^(2/3)*Sigma*n^(-2/3)

    return(Hns)
}

## Plug-in bandwidth selector
hpi.kcde <- function(x, nstage=2, binned, amise=FALSE)
{
    n <- length(x)
    d <- 1
    if (missing(binned)) binned <- default.bflag(d,n)

    K2 <- dnorm.deriv(x=0, mu=0, sigma=1, deriv.order=2)  
    K4 <- dnorm.deriv(x=0, mu=0, sigma=1, deriv.order=4) 
    m2 <- 1  
    m1 <- (4*pi)^(-1/2)

    ## formula for bias annihliating bandwidths from Wand & Jones (1995, p.70)
    if (nstage==2)
    {
        psi6.hat <- psins.1d(r=6, sigma=sd(x))
        gamse4 <- (2*K4/(-m2*psi6.hat*n))^(1/(4+3)) 
        psi4.hat <- kfe.1d(x=x, g=gamse4, deriv.order=4, inc=1, binned=binned)
        gamse2 <- (2*K2/(-m2*psi4.hat*n))^(1/(2+3))
        psi2.hat <- kfe.1d(x=x, g=gamse2, deriv.order=2, inc=1, binned=binned)
    }
    else 
    {
        psi4.hat <- psins.1d(r=4, sigma=sd(x))
        gamse2 <- (2*K2/(-m2*psi4.hat*n))^(1/(2+3))
        psi2.hat <- kfe.1d(x=x, g=gamse2, deriv.order=2, inc=1, binned=binned)
    }

    ## formula form Polansky & Baker (2000)
    h <- (2*m1/(-m2^2*psi2.hat*n))^(1/3) 
    if (amise) PI <- -2*n^(-1)*m1*h - 1/4*psi2.hat*h^4

    if (!amise) return(h)
    else return(list(h=h, PI=PI))
}

Hpi.kcde <- function(x, nstage=2, pilot, Hstart, binned, bgridsize, amise=FALSE, verbose=FALSE, optim.fun="optim", pre=TRUE)
{
    n <- nrow(x)
    d <- ncol(x)
    m1 <- (4*pi)^(-1/2)
    Jd <- matrix(1, ncol=d, nrow=d)

    if (missing(binned)) binned <- default.bflag(d,n)
    if (!is.matrix(x)) x <- as.matrix(x)
    if (missing(pilot)) pilot <- "dunconstr"
    pilot1 <- match.arg(pilot, c("dunconstr", "dscalar"))

    if (pre) { S12 <- diag(sqrt(diag(var(x)))); x <- pre.scale(x) }
    D2K0 <- t(dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=2))

    if (nstage==2)
    {  
    ## stage 1
    psi4.ns <- psins(r=4, Sigma=var(x), deriv.vec=TRUE)

    amse2.temp <- function(vechH)
    { 
      H <- invvech(vechH) %*% invvech(vechH)
      Hinv <- chol2inv(chol(H))
      Hinv12 <- matrix.sqrt(Hinv)
      amse2.val <- 1/(det(H)^(1/2)*n)*((Hinv12 %x% Hinv12) %*% D2K0) + 1/2* t(vec(H) %x% diag(d^2)) %*% psi4.ns
      return(sum(amse2.val^2))
    }
      
    Hstart2 <- matrix.sqrt(Gns(r=2, n=n, Sigma=var(x)))
    optim.fun1 <- match.arg(optim.fun, c("nlm", "optim"))

    if (optim.fun1=="nlm")
    {
        result <- nlm(p=vech(Hstart2), f=amse2.temp, print.level=2*as.numeric(verbose))    
        H2 <- invvech(result$estimate) %*% invvech(result$estimate)
    }
    else
    {
        result <- optim(vech(Hstart2), amse2.temp, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
        H2 <- invvech(result$par) %*% invvech(result$par)
    }

    psi2.hat <- kfe(x=x, G=H2, deriv.order=2, add.index=FALSE, binned=binned, bgridsize=bgridsize, verbose=verbose)
    }
    else
    {
        psi2.hat <- psins(r=2, Sigma=var(x), deriv.vec=TRUE)    
        H2 <- Gns(r=2, n=n, Sigma=var(x))
    }
    if (missing(Hstart)) Hstart <- Hns.kcde(x=x)

    ## stage 2
    amise.temp <- function(vechH)
    { 
        H <- invvech(vechH) %*% invvech(vechH)
        H12 <- matrix.sqrt(H)
        amise.val <- -2*n^(-1)*m1*tr(H12) - 1/4*t(vec(H %*% H)) %*% psi2.hat
        return(drop(amise.val)) 
    }

    Hstart <- matrix.sqrt(Hstart)
    optim.fun1 <- match.arg(optim.fun, c("optim", "nlm"))

    if (optim.fun1=="nlm")
    {
        result <- nlm(p=vech(Hstart), f=amise.temp, print.level=2*as.numeric(verbose)) 
        H <- invvech(result$estimate) %*% invvech(result$estimate)
        amise.star <- result$minimum
    }
    else
    {
        result <- optim(vech(Hstart), amise.temp, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
        H <- invvech(result$par) %*% invvech(result$par)
        amise.star <- result$value
    }
    if (pre) H <- S12 %*% H %*% S12
    if (amise) H <- list(H=H, PI=amise.star)

    return(H)
}

Hpi.diag.kcde <- function(x, nstage=2, pilot, Hstart, binned=FALSE, bgridsize, amise=FALSE, verbose=FALSE, optim.fun="optim", pre=TRUE)
{
    n <- nrow(x)
    d <- ncol(x)
    m1 <- (4*pi)^(-1/2)
    Jd <- matrix(1, ncol=d, nrow=d)

    if (missing(binned)) binned <- default.bflag(d,n)
    if(!is.matrix(x)) x <- as.matrix(x)
    if (missing(pilot)) pilot <- "dscalar"
    pilot <- match.arg(pilot, c("dunconstr", "dscalar"))
    if (pre) { S12 <- diag(sqrt(diag(var(x)))); x <- pre.scale(x) }
  
    D2K0 <- t(dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=2))
    if (nstage==2)
    {  
        ## stage 1
        psi4.ns <- psins(r=4, Sigma=var(x), deriv.vec=TRUE)

        amse2.temp <- function(diagH)
        { 
            H <- diag(diagH) %*% diag(diagH)
            Hinv <- chol2inv(chol(H))
            Hinv12 <- matrix.sqrt(Hinv)
            amse2.val <- 1/(det(H)^(1/2)*n)*((Hinv12 %x% Hinv12) %*% D2K0) + 1/2* t(vec(H) %x% diag(d^2)) %*% psi4.ns
            return(sum(amse2.val^2))
        }

        Hstart2 <- matrix.sqrt(Gns(r=2, n=n, Sigma=var(x)))
        optim.fun <- match.arg(optim.fun, c("optim", "nlm")) 

        if (optim.fun=="nlm")
        {
            result <- nlm(p=diag(Hstart2), f=amse2.temp, print.level=2*as.numeric(verbose))    
            H2 <- diag(result$estimate) %*% diag(result$estimate)
        }
        else
        {
            result <- optim(diag(Hstart2), amse2.temp, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
            H2 <- diag(result$par) %*% diag(result$par)
        }

        psi2.hat <- kfe(x=x, G=H2, deriv.order=2, add.index=FALSE, binned=binned, bgridsize=bgridsize, verbose=verbose)
        }
        else
            psi2.hat <- psins(r=2, Sigma=var(x), deriv.vec=TRUE)    

        if (missing(Hstart)) Hstart <- Hns.kcde(x=x)

        ## stage 2
        amise.temp <- function(diagH)
        {    
            H <- diag(diagH) %*% diag(diagH)
            H12 <- matrix.sqrt(H)
            amise.val <- -2*n^(-1)*m1*tr(H12) - 1/4*t(vec(H %*% H)) %*% psi2.hat
            return(drop(amise.val)) 
        }
  
    Hstart <- matrix.sqrt(Hstart)
    optim.fun1 <- match.arg(optim.fun, c("optim", "nlm")) 
  
    if (optim.fun1=="nlm")
    {
        result <- nlm(p=diag(Hstart), f=amise.temp, print.level=2*as.numeric(verbose)) 
        H <- diag(result$estimate) %*% diag(result$estimate)
        amise.star <- result$minimum
    }
    else
    {
        result <- optim(diag(Hstart), amise.temp, method="BFGS", control=list(trace=as.numeric(verbose), REPORT=1))
        H <- diag(result$par) %*% diag(result$par)
        amise.star <- result$value
    }
    if (pre) H <- S12 %*% H %*% S12
    if (amise)  H <- list(H=H, PI=amise.star) 
    
    return(H)
}

#####################################################################
## Multivariate kernel ROC estimators
#####################################################################
kroc <- function(x1, x2, H1, h1, hy, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, positive=FALSE, adj.positive, w, verbose=FALSE)
{
    if (is.vector(x1)) { d <- 1; n1 <- length(x1) } 
    else { d <- ncol(x1); n1 <- nrow(x1); x1 <- as.matrix(x1); x2 <- as.matrix(x2) }
    if (!missing(eval.points)) stop("eval.points in kroc not yet implemented")

    if (d==1)
    {
        if (missing(h1)) h1 <- hpi.kcde(x=x1, binned=default.bflag(d=d, n=n1))
        Fhatx1 <- kcde(x=x1, h=h1, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, binned=binned, bgridsize=bgridsize, positive=positive, adj.positive=adj.positive, w=w, tail.flag="upper.tail")
    }
    else
    {
        if (missing(H1)) H1 <- Hpi.kcde(x=x1, binned=default.bflag(d=d, n=n1), verbose=verbose)
        Fhatx1 <- kcde(x=x1, H=H1, gridsize=gridsize, gridtype=gridtype, xmin=xmin, xmax=xmax, supp=supp, binned=binned, bgridsize=bgridsize, w=w, tail.flag="upper.tail", verbose=verbose)
    }

    ## transform from [0,1] to reals
    y1 <- predict(Fhatx1, x=x1)
    y2 <- predict(Fhatx1, x=x2)
    y1 <- qnorm(y1[y1>0])
    y2 <- qnorm(y2[y2>0])

    if (missing(hy)) hy <- hpi.kcde(y2, binned=default.bflag(d=1, n=n1))
    Fhaty2 <- kcde(x=y2, h=hy, binned=TRUE, xmin=min(y1,y2)-3.7*hy, xmax=max(y1,y2)+3.7*hy)
    Fhaty1 <- kcde(x=y1, h=hy, binned=TRUE, xmin=min(y1,y2)-3.7*hy, xmax=max(y1,y2)+3.7*hy)
    Fhaty1$eval.points <- pnorm(Fhaty1$eval.points)
    Fhaty2$eval.points <- pnorm(Fhaty2$eval.points)

    Rhat <- Fhaty1
    Rhat$eval.points <- Fhaty1$estimate
    Rhat$estimate <- Fhaty2$estimate
    if (d==1) { Rhat$h1 <- h1; Rhat$H1 <- h1^2; Rhat$hy <- hy }
    else { Rhat$H1 <- H1; Rhat$hy <- hy }

    ## Use spline to smooth out transformed ROC curve
    Rhat.smoothed <- smooth.spline(Rhat$eval.points, Rhat$estimate, spar=0.5)
    Rhat.smoothed <- predict(Rhat.smoothed, x=seq(0,1,length=length(Rhat$eval.points))) 
    Rhat$eval.points <- Rhat.smoothed$x
    Rhat$estimate <- Rhat.smoothed$y

    ## add (0,0) and (1,1) as endpoints
    if (head(Rhat$eval.points, n=1)!=0) Rhat$eval.points[1] <- 0
    if (head(Rhat$estimate, n=1)!=0) Rhat$estimate[1] <- 0
    if (tail(Rhat$eval.points, n=1)!=1) Rhat$eval.points[length(Rhat$eval.points)] <- 1
    if (tail(Rhat$estimate, n=1)!=1) Rhat$estimate[length(Rhat$estimate)] <- 1
    Rhat$estimate[Rhat$estimate>1] <- 1
    Rhat$estimate[Rhat$estimate<0] <- 0
    Rhat$indices <- indices.kroc(Rhat)
    Rhat <- Rhat[-c(4,5)] 
    Rhat$type <- "kroc"
    class(Rhat) <- "kroc"
    
    return(Rhat)
}

## summary measure of ROC curves
indices.kroc <- function(Rhat)
{
    auc <- sum(abs((head(Rhat$estimate, n=-1) - tail(Rhat$estimate, n=-1)))*abs(diff(Rhat$eval.points))/2 + head(Rhat$estimate, n=-1)*abs(diff(Rhat$eval.points)))

    youden.val <- Rhat$estimate - Rhat$eval.points
    if (max(youden.val)>0.001)
    {  
        youden.ind <- which.max(youden.val)
        youden <- youden.val[youden.ind]
        LR <- list(minus=(1-Rhat$estimate[youden.ind])/(1-Rhat$eval.points[youden.ind]), plus=Rhat$estimate[youden.ind]/Rhat$eval.points[youden.ind])
    }
    else
        LR <- list(minus=1, plus=1)
      
    return(list(auc=auc, youden=max(youden.val), LR=LR))
}  

#############################################################################
## S3 methods
#############################################################################
## plot method
plot.kroc <- function(x, add=FALSE, add.roc.ref=FALSE, xlab, ylab, alpha=1, col=1, ...)
{
    Rhat <- x
    col <- transparency.col(col, alpha=alpha)
    if (missing(ylab)) ylab <- "True positive rate (sensitivity)"
    if (missing(xlab)) xlab <- expression("False positive rate"~~group("(", list(bar(specificity)), ")"))

    if (add) lines(Rhat$eval.points, Rhat$estimate, ...)
    else plot(Rhat$eval.points, Rhat$estimate, type="l", ylab=ylab, xlab=xlab, col=col, ...)

    if (is.vector(Rhat$x[[1]])) d <- 1 else d <- ncol(Rhat$x[[1]])
    if (add.roc.ref)
    {
        z <- seq(0,1, length=401)
        kind <- 0:(d-1)
        roc.indep <- 0
        for (k in kind) roc.indep <- roc.indep + z*(-log(z))^k/factorial(k)
        lines(z, roc.indep, lty=2, col="grey")
    }
}

## summary method
summary.kroc <- function(object, ...)
{
    cat("Summary measures for ROC curve\nAUC =", signif(object$indices$auc, ...), "\n")
    cat("Youden index =", signif(object$indices$youden, ...), "\n")
    cat(paste("(LR-, LR+) = (",  signif(object$indices$LR$minus, ...), ", ", signif(object$indices$LR$plus, ...),")\n\n",sep=""))
}

## predict methods
predict.kcde <- function(object, ..., x)
{
    return(predict.kde(object=object, ..., x=x))
}

predict.kroc <- function(object, ..., x)
{
    return(predict.kde(object=object, ..., x=x))
}

## contourLevels method
contourLevels.kcde <- function(x, prob, cont, ...)
{
    nlevels <- 5
    fhat <- x 
    if (missing(prob) & missing(cont)) hts <- pretty(fhat$estimate, n=nlevels) 
    if (!missing(prob) & missing(cont)) { hts <- prob/100; names(hts) <- paste0(prob, "%") }     
    if (missing(prob) & !missing(cont)) { prob <- 100-cont; hts <- prob/100; names(hts) <-  paste0(prob, "%") }
    
    return(hts)
}

######################################################################
## Kernel support estimate - contour-based or convex hull 
######################################################################
ksupp <- function(fhat, cont=95, abs.cont, convex.hull=TRUE)
{
    if (missing(abs.cont)) abs.cont <- contourLevels(fhat, cont=cont)
    supp <- expand.grid(fhat$eval.points)[as.vector(fhat$estimate > abs.cont),]
    d <- length(fhat$eval.points)
    if (d==2) { if (convex.hull) supp <- supp[chull(supp),] }
    else 
    { 
        if (convex.hull) 
        { 
            if (!requireNamespace("geometry", quietly=TRUE)) stop("Install the geometry package as it is required.", call.=FALSE)
            supp <- as.matrix(supp[t(geometry::convhulln(supp)),]) 
        }
    }
    class(supp) <- c("ksupp", class(supp))
    
    return(supp)
}

## Devroye-Wise support estimate
dwsupp <- function(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, binned, bgridsize, verbose=FALSE, w)
{
    if (is.vector(x))
    {
        if (missing(H)) { d <- 1; n <- length(x) }
        else
        {
            if (is.vector(H)) { d <- 1; n <- length(x) }
            else { x <- matrix(x, nrow=1); d <- ncol(x); n <- nrow(x) }
        }
    }
    else { d <- ncol(x); n <- nrow(x) }

    if (!missing(w))
    if (!(identical(all.equal(sum(w), n), TRUE)))
    {
        warning("Weights don't sum to sample size - they have been scaled accordingly\n")
        w <- w*n/sum(w)
    }

    if (missing(binned)) binned <- default.bflag(d=d, n=n)
    if (missing(w)) w <- rep(1,n)
    if (d==1)
    {  
        y <- x
        if (missing(h)) h <- hpi(x=y, binned=default.bflag(d=d, n=n), bgridsize=bgridsize)
    }
    if (missing(H) & d>1)
    {
        H <- Hpi(x=x, binned=default.bflag(d=d, n=n), bgridsize=bgridsize, verbose=verbose)
    }

    if (missing(bgridsize)) bgridsize <- default.gridsize(d)
    if (missing(gridsize)) gridsize <- default.gridsize(d)
  
    ## initialise grid 
    n <- nrow(x)
    gridx <- make.grid.ks(x, matrix.sqrt(H), tol=supp, gridsize=gridsize, xmin=xmin, xmax=xmax, gridtype=gridtype) 

    suppx <- make.supp(x, matrix.sqrt(H), tol=supp)

    grid.pts <- find.gridpts(gridx, suppx)    
    fhat.grid <- matrix(0, nrow=length(gridx[[1]]), ncol=length(gridx[[2]]))
    if (verbose) pb <- txtProgressBar()

    for (i in 1:n)
    {
        ## compute evaluation points 
        eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
        eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
        eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
        eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
        eval.x.len <- length(eval.x)
        eval.pts <- expand.grid(eval.x, eval.y)
        fhat <- rep(1, nrow(eval.pts)) 
        ## place vector of density estimate values `fhat' onto grid 'fhat.grid' 
        for (j in 1:length(eval.y))
            fhat.grid[eval.x.ind, eval.y.ind[j]] <- 
                fhat.grid[eval.x.ind, eval.y.ind[j]] + 
                w[i]*fhat[((j-1) * eval.x.len + 1):(j * eval.x.len)]
        if (verbose) setTxtProgressBar(pb, i/n)
    }
    if (verbose) close(pb)
    gridx1 <- list(gridx[[1]], gridx[[2]]) 

    fhat <- list(x=x, eval.points=gridx1, estimate=fhat.grid>=1, H=H, gridtype=gridx$gridtype, gridded=TRUE)

    fhat$binned <- binned
    fhat$names <- parse.name(x)  ## add variable names
    fhat$w <- w
    class(fhat) <- "kde"

    return(fhat)
}

######################################################################
## S3 methods for KSUPP objects
######################################################################
## plot method
plot.ksupp <- function(x, display="plot3D", ...)
{
    d <- ncol(x)
    if (d==2) polygon(x, ...)
    else if (d==3) 
    {
        disp <- match.arg(display, c("plot3D", "rgl"))
        if (disp=="plot3D") 
        {
            if (!requireNamespace("plot3D", quietly=TRUE)) stop("Install the plot3D package as it is required.", call.=FALSE)
            plot3D::triangle3D(x, ...)
        }
        else if (disp=="rgl") 
        {
            if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
            rgl::triangles3d(x=x[,1], y=x[,2], z=x[,3], ...)
        }
    }
}

############################################################################
## Univariate mixture normal densities and derivatives
############################################################################
rnorm.mixt <- function(n=100, mus=0, sigmas=1, props=1, mixt.label=FALSE)
{
    if (!(identical(all.equal(sum(props), 1), TRUE)))
        stop("Proportions don't sum to one")

    ## single component mixture
    if (identical(all.equal(props[1], 1), TRUE))
    {
        if (mixt.label)
            rand <- cbind(rnorm(n=n, mean=mus, sd=sigmas), rep(1, n))
        else
            rand <- rnorm(n=n, mean=mus, sd=sigmas)
    }
    ## multiple component mixture
    else
    {
        k <- length(props)
        n.samp <- sample(1:k, n, replace=TRUE, prob=props) 
        n.prop <- numeric(0)

        ## compute number taken from each mixture
        for (i in 1:k)
            n.prop <- c(n.prop, sum(n.samp == i))

        rand <- numeric(0)

        for (i in 1:k)
        {
            ## compute random sample from normal mixture component
            if (n.prop[i] > 0)
                if (mixt.label)
                    rand <- rbind(rand, cbind(rnorm(n=n.prop[i], mean=mus[i], sd=sigmas[i]), rep(i, n.prop[i])))
                else
                    rand <- c(rand, rnorm(n=n.prop[i], mean=mus[i], sd=sigmas[i]))
        }
    }
    if (mixt.label) return(rand[sample(n),])
    else return(rand[sample(n)])
}

dnorm.mixt <- function(x, mus=0, sigmas=1, props=1)
{
    if (!(identical(all.equal(sum(props), 1), TRUE)))
    stop("Proportions don't sum to one")

    ## single component mixture
    if (identical(all.equal(props[1], 1), TRUE))
    dens <- dnorm(x, mean=mus[1], sd=sigmas[1])

    ## multiple component mixture
    else   
    {   
        k <- length(props)
        dens <- 0

        ## sum of each normal density value from each component at x  
        for (i in 1:k) dens <- dens + props[i]*dnorm(x, mean=mus[i], sd=sigmas[i])
    }
  
    return(dens)
}   

############################################################################
## Derivatives of the univariate normal 
## (code: J.E.Chacon 08/06/2018)
##
## Parameters
## x - points to evaluate at
## sigma - std deviation
## r - derivative index 
#
## Returns
## r-th derivative at x
############################################################################
dnorm.deriv <- function (x, mu=0, sigma=1, deriv.order=0)
{
    r <- deriv.order
    phi <- dnorm(x, mean=mu, sd=sigma)
    x <- (x - mu)
    arg <- x / sigma
    hmold0 <- 1
    hmold1 <- arg
    hmnew <- 1
    if (r == 1)
        hmnew <- hmold1
    if (r >= 2) 
        for (i in (2:r)) {
            hmnew <- arg * hmold1 - (i - 1) * hmold0
            hmold0 <- hmold1
            hmold1 <- hmnew
        }
    derivt <- (-1)^r * phi * hmnew / sigma^r

    return(derivt)
}

############################################################################
## Double sum  of K(X_i - X_j) used in density derivative estimation
#
## Parameters
## x - points to evaluate
## Sigma - variance matrix
## inc - 0 - exclude diagonals
##     - 1 - include diagonals
#
## Returns
## Double sum at x
############################################################################
dnorm.deriv.sum <- function(x, sigma, deriv.order, inc=1, binned=FALSE, bin.par, kfe=FALSE)
{
    r <- deriv.order
    n <- length(x)
    if (binned)
    {
        if (missing(bin.par)) bin.par <- binning(x, h=sigma, supp=4+r) 
        est <- kdde.binned(x=x, H=sigma^2, h=sigma, deriv.order=r, bin.par=bin.par)$estimate
        sumval <- sum(bin.par$counts*est*n)
        if (inc == 0) 
            sumval <- sumval - n*dnorm.deriv(x=0, mu=0, sigma=sigma, deriv.order=r)
    }
    else
    {
        sumval <- 0
        for (i in 1:n)
            sumval <- sumval + sum(dnorm.deriv(x=x[i] - x, mu=0, sigma=sigma, deriv.order=r)) 
        if (inc == 0) 
            sumval <- sumval - n*dnorm.deriv(x=0, mu=0, sigma=sigma, deriv.order=r)
    }

    if (kfe)
        if (inc==1) sumval <- sumval/n^2
    else sumval <- sumval/(n*(n-1))
  
    return(sumval)
}

dnorm.deriv.mixt <- function(x, mus=0, sigmas=1, props=1, deriv.order=0)
{
    if (!(identical(all.equal(sum(props), 1), TRUE)))
    stop("Proportions don't sum to one")

    ## single component mixture
    if (identical(all.equal(props[1], 1), TRUE))
    dens <- dnorm.deriv(x, mu=mus[1], sigma=sigmas[1], deriv.order=deriv.order)

    ## multiple component mixture
    else   
    {   
        k <- length(props)
        dens <- 0

        ## sum of each normal density value from each component at x  
        for (i in 1:k)
            dens <- dens + props[i]*dnorm.deriv(x=x, mu=mus[i], sigma=sigmas[i], deriv.order=deriv.order)
    }

    return(dens)
}   

############################################################################
## Multivariate normal densities and derivatives
############################################################################

############################################################################
## Multivariate normal mixture - random sample
## 
## Parameters
## n - number of samples
## mus - matrix of means (each row is a vector of means from each component
##       density)
## Sigmas - matrix of covariance matrices (every d rows is a covariance matrix 
##          from each component density) 
## props - vector of mixing proportions 
## 
## Returns
## Vector of n observations from the normal mixture 
############################################################################
rmvnorm.mixt <- function(n=100, mus=c(0,0), Sigmas=diag(2), props=1, mixt.label=FALSE)
{
    if (!(identical(all.equal(sum(props), 1), TRUE)))
    stop("Proportions don't sum to one")

    ## single component mixture
    if (identical(all.equal(props[1], 1), TRUE))
    if (mixt.label)
        rand <- cbind(mvtnorm::rmvnorm(n=n, mean=mus, sigma=Sigmas), rep(1, n))
    else
        rand <- cbind(mvtnorm::rmvnorm(n=n, mean=mus, sigma=Sigmas))

    ## multiple component mixture
    else
    {
        k <- length(props)
        d <- ncol(Sigmas)
        n.samp <- sample(1:k, n, replace=TRUE, prob=props) 
        n.prop <- numeric(0)

        ## compute number taken from each mixture
        for (i in 1:k)
            n.prop <- c(n.prop, sum(n.samp == i))
        
        rand <- numeric(0)
        for (i in 1:k)
        {
            ## compute random sample from normal mixture component
            if (n.prop[i] > 0)
            {       
                if (mixt.label)
                    rand <- rbind(rand, cbind(rmvnorm(n=n.prop[i], mean=mus[i,], sigma=Sigmas[((i-1)*d+1) : (i*d),]), rep(i, n.prop[i])))
                else
                    rand <- rbind(rand, rmvnorm(n=n.prop[i], mean=mus[i,], sigma=Sigmas[((i-1)*d+1) : (i*d),]))
            }    
        }
    }

    return(rand[sample(n),])
}

############################################################################
## Multivariate normal mixture - density values
## 
## Parameters
## x - points to compute density at 
## mus - matrix of means
## Sigmas - matrix of covariance matrices 
## props - vector of mixing proportions 
## 
## Returns
## Density values from the normal mixture (at x)
############################################################################
dmvnorm.mixt <- function(x, mus, Sigmas, props=1, verbose=FALSE)
{  
    if (!(identical(all.equal(sum(props), 1), TRUE)))
        stop("Proportions don't sum to one")
    
    if (is.vector(x)) { d <- length(x); n <- 1 } 
    else { d <- ncol(x); n <- nrow(x) }
    
    if (missing(mus)) mus <- rep(0,d)
    if (missing(Sigmas)) Sigmas <- diag(d)
    
    ## single component mixture
    if (identical(all.equal(props[1], 1), TRUE))
    {
        if (is.matrix(mus)) mus <- mus[1,]
        dens <- dmvnorm(x=x, mean=mus, sigma=Sigmas[1:d,])
    }
    ## multiple component mixture
    else   
    {
        if (verbose) pb <- txtProgressBar()
        k <- length(props)
        dens <- 0
        ## sum of each normal density value from each component at x
        for (i in 1:k)
        {
            dens <- dens + props[i]*dmvnorm(x=x, mean=mus[i,], sigma=Sigmas[((i-1)*d+1):(i*d),])
            if (verbose) setTxtProgressBar(pb, i/k)
        }
        if (verbose) close(pb)
    }
  
  return(dens)
}   

##########################################################################
## Computation of the r-th derivative vector of the Gaussian density
##########################################################################
dmvnorm.deriv <- function(x, mu, Sigma, deriv.order=0, deriv.vec=TRUE, add.index=FALSE, only.index=FALSE, type="unique")
{
    type1 <- match.arg(type, c("recursive", "direct", "unique"))

    r <- deriv.order
    if(length(r)>1) stop("deriv.order should be a non-negative integer")
    sumr <- sum(r)

    if (missing(x)) d <- ncol(Sigma)
    else
    {
        if (is.vector(x)) x <- t(as.matrix(x))
        if (is.data.frame(x)) x <- as.matrix(x)
        d <- ncol(x)
        n <- nrow(x)
    }
    if (add.index | only.index | !deriv.vec)
    {    
        ## matrix of derivative indices
        ind.mat <- 0
        sumr.counter <- sumr
        if (sumr>=1) ind.mat <- diag(d)
        {
            while (sumr.counter >1)
            {
                ind.mat <- Ksum(diag(d), ind.mat)
                sumr.counter <- sumr.counter - 1
            }
        }
        ind.mat.minimal <- unique(ind.mat)
        ind.mat.minimal.logical <- !duplicated(ind.mat)

        if (only.index)
            if (deriv.vec) return (ind.mat)
            else  return(ind.mat.minimal)
    }
  
    if (missing(mu)) mu <- rep(0,d)
    if (missing(Sigma)) Sigma <- diag(d)
 
    x.centred <- sweep(x, 2, mu)
    dens <- do.call(paste("dmvnorm.deriv", type1, sep="."), list(x=x.centred, Sigma=Sigma, deriv.order=r))
    if (is.vector(dens) & r>0) dens <- matrix(dens, nrow=1)
  
    if (!deriv.vec & r>0)
    {
        ind.select <- numeric()
        for (i in 1:nrow(ind.mat.minimal))
            ind.select <- c(ind.select, head(which.mat(ind.mat.minimal[i,], ind.mat), n=1))

        dens <- dens[,ind.select]
        ind.mat <- ind.mat.minimal
    }
    
    if (add.index) return(list(deriv=dens, deriv.ind=ind.mat))
    else return(dens)
}

############################################################################
## dmvnorm.deriv.direct computes the vector derivative of the Gaussian
## density phi_Sigma(x) on the basis of Equation (1) and Algotihm 2, as
## described in Chacon and Duong (2014)
############################################################################
dmvnorm.deriv.direct <- function(x,Sigma,deriv.order=0)
{ 
    if (is.vector(x)) { x <- matrix(x,nrow=1) }
    d <- ncol(Sigma)
    n <- nrow(x)
    r <- deriv.order

    Sigmainv <- chol2inv(chol(Sigma))
    Hermx <- matrix(0,nrow=n,ncol=d^r)
    for (i in 1:n)
    {
        for (j in 0:floor(r/2))
            Hermx[i,] <- Hermx[i,] + (-1)^j/(factorial(j)*factorial(r-2*j)*2^j)*
                Kpow(Sigmainv%*%x[i,], r-2*j)%x%Kpow(vec(Sigmainv),j)
            Hermx[i,] <- Sdrv.recursive(d=d,r=r,v=Hermx[i,])
    }
    dens <- drop((-1)^r*factorial(r)*Hermx*dmvnorm(x,mean=rep(0,d), sigma=Sigma))
    
    return(dens)
}

############################################################################
## dmvnorm.deriv.recursive computes the vector derivative of the Gaussian
## density phi_Sigma(x) on the basis of Equation (7) and Algorithm 2 as
## described in Section 5 of Chacon and Duong (2014)
############################################################################
dmvnorm.deriv.recursive <- function(x,Sigma,deriv.order=0)
{ 
    if (is.vector(x)) { x <- matrix(x,nrow=1) }
    d <- ncol(Sigma)
    n <- nrow(x)
    r <- deriv.order
    G <- Sigma
    Ginv <- chol2inv(chol(G))
    vGinv <- vec(Ginv)
    nvGinv <- matrix(rep(vGinv,n),byrow=TRUE,ncol=length(vGinv),nrow=n)
    arg <- matrix(x%*%Ginv, nrow=n)
    
    hmold0 <- matrix(1,nrow=n,ncol=1)
    hmold1 <- arg
    hmnew <- hmold0

    if(r==1) { hmnew <- hmold1 }
    if(r>=2)
    { 
        for(i in 2:r)
        {
            hmnew <- mat.Kprod(hmold1,arg)-(i-1)*mat.Kprod(nvGinv,hmold0)
            hmnew <- matrix(Sdrv.recursive(d=d,r=i,v=hmnew), nrow=n)
            hmold0 <- hmold1
            hmold1 <- hmnew        
        }    
    }
    dens <- dmvnorm(x,mean=rep(0,d),sigma=Sigma)
    result <- drop(matrix(rep(dens,d^r),byrow=FALSE,nrow=n,ncol=d^r)*hmnew*(-1)^r)
    
    return(result)
}

############################################################################
## dmvnorm.deriv.unique computes the whole vector derivative of the Gaussian
## density phi_Sigma(x) from its unique coordinates, based on Algorithm 3 as
## described in Section 5 of Chacon and Duong (2014)
############################################################################
dmvnorm.deriv.unique <- function(x,Sigma,deriv.order=0)
{
    if (is.vector(x)) { x <- matrix(x,nrow=1) }
    d <- ncol(x)
    n <- nrow(x)
    r <- deriv.order 
    G <- Sigma
    Ginv <- chol2inv(chol(G))
    arg <- x%*%Ginv
    
    hmold0 <- matrix(1,nrow=n,ncol=1)
    hmold1 <- arg
    hmnew <- hmold0  
    udind0 <- matrix(rep(0,d),nrow=1,ncol=d)
    udind1 <- diag(d)
     
    if(r==1) { hmnew <- hmold1 }
    if(r>=2) 
    { 
        for (i in 2:r)
        {
            Ndi1 <- ncol(hmold1)
            Ndi0 <- ncol(hmold0)
            hmnew <- numeric()
            
            for (j in 1:d)
            {        
                nrecj <- choose(d-j+i-1,i-1)        
                hmnew.aux <- arg[,j]*hmold1[,Ndi1-(nrecj:1)+1]
                      
                for(k in j:d)
                {        
                    udind0.aux <- matrix(udind1[Ndi1-(nrecj:1)+1,],ncol=d,byrow=FALSE)
                    udind0.aux[,k] <- udind0.aux[,k]-1
                    valid.udind0 <- as.logical(apply(udind0.aux>=0,1,min))
                    enlarged.hmold0 <- matrix(0,ncol=nrow(udind0.aux),nrow=n)
                    for (ell in 1:nrow(udind0.aux))
                    {
                        if(valid.udind0[ell])
                        {
                            pos <- which(rowSums((udind0-matrix(rep(udind0.aux[ell,],nrow(udind0)), nrow=nrow(udind0),byrow=TRUE))^2)==0)
                            enlarged.hmold0[,ell] <- hmold0[,pos] 
                        }
                    }
                    hmnew.aux <- hmnew.aux-Ginv[j,k]*matrix(rep(udind1[Ndi1-(nrecj:1)+1,k],n), nrow=n,byrow=TRUE)*enlarged.hmold0
                }
                hmnew <- cbind(hmnew,hmnew.aux)
            }
            hmold0 <- hmold1
            hmold1 <- hmnew

            ## compute the unique i-th derivative multi-indexes 
            nudind1 <- nrow(udind1)
            udindnew <- numeric()
            for (j in 1:d)
            {
                Ndj1i <- choose(d+i-1-j,i-1)
                udind.aux <- matrix(udind1[nudind1-(Ndj1i:1)+1,],ncol=d,byrow=FALSE)
                udind.aux[,j] <- udind.aux[,j]+1
                udindnew <- rbind(udindnew,udind.aux)
            }
            udind0 <- udind1
            udind1 <- udindnew
        }
    }
        
    if(r==0) result <- dmvnorm(x,mean=rep(0,d),sigma=Sigma)
    if(r==1) result <- (-1)*matrix(rep(dmvnorm(x,mean=rep(0,d),sigma=Sigma),d),
                     nrow=n,byrow=FALSE)*hmnew
    if(r>=2)
    {
        per <- pinv.all(d=d,r=r)
        dind <- numeric()
        udind <- udind1
        dind.base <- rep(0,d^r)        
        udind.base <- rep(0,choose(d+r-1,r))
                
        for (i in 1:d)
        {
            dind <- cbind(dind,rowSums(per==i)) ## Matrix of derivative indices
            dind.base <- dind.base+dind[,i]*(r+1)^(d-i) ## Transform each row to base r+1           
            udind.base <- udind.base+udind[,i]*(r+1)^(d-i) ## Transform each row to base r+1                       
        }        
        
        dlabs <- match(dind.base,udind.base)
        deriv.vector <- hmnew[,dlabs]*matrix(rep((-1)^rowSums(dind),n),nrow=n,byrow=TRUE)
       
        result <- matrix(rep(dmvnorm(x,mean=rep(0,d),sigma=Sigma),ncol(deriv.vector)), nrow=n,byrow=FALSE)*deriv.vector
    }

    return(drop(result))
}

dmvnorm.deriv.mixt <- function(x, mus, Sigmas, props, deriv.order, deriv.vec=TRUE, add.index=FALSE, only.index=FALSE, verbose=FALSE)
{
    if (!(identical(all.equal(sum(props), 1), TRUE)))
        stop("Proportions don't sum to one")

    if (is.vector(x)) d <- length(x)
    else d <- ncol(x)

    if (missing(mus)) mus <- rep(0,d)
    if (missing(Sigmas)) Sigmas <- diag(d)
    r <- deriv.order
    sumr <- sum(r)

    if (only.index | add.index) 
        ind.mat <- dmvnorm.deriv(x=x, mu=mus[1,], Sigma=Sigmas[1:d,], deriv.order=r, only.index=TRUE)
    if (only.index)  
    if (deriv.vec) return (ind.mat)
    else return(unique(ind.mat))
  
    ## derivatives 
    ## single component mixture
    if (identical(all.equal(props[1], 1), TRUE))
    {
    if (is.matrix(mus)) mus <- mus[1,]
    dens <- dmvnorm.deriv(x=x, mu=mus, Sigma=Sigmas[1:d,], deriv.order=sumr)
    }
    ## multiple component mixture
    else   
    {   
        k <- length(props)
        if (verbose) pb <- txtProgressBar()
        dens <- 0
        ## sum of each normal density value from each component at x  
        for (i in 1:k)
        {
            dens <- dens + props[i]*dmvnorm.deriv(x=x, mu=mus[i,], Sigma=Sigmas[((i-1)*d+1):(i*d),], deriv.order=sumr)
            if (verbose) setTxtProgressBar(pb, i/k)
        }
        if (verbose) close(pb)   
    }

    if (!deriv.vec)
    { 
        dens <- dens[,!duplicated(ind.mat)]
        ind.mat <- unique(ind.mat)
    }
 
    if (add.index) return(list(deriv=dens, deriv.ind=ind.mat))
    else return(deriv=dens)
}

############################################################################
## Double sum  of K(X_i - X_j) used in density derivative estimation
##
## Parameters
## x - points to evaluate
## Sigma - variance matrix
## inc - 0 - exclude diagonals
##     - 1 - include diagonals
##
## Returns
## Double sum at x
#############################################################################
dmvnorm.deriv.sum <- function(x, Sigma, deriv.order=0, inc=1, binned=FALSE, bin.par, bgridsize, kfe=FALSE, deriv.vec=TRUE, add.index=FALSE, verbose=FALSE)
{
    r <- deriv.order
    d <- ncol(x)
    n <- nrow(x)
    if (missing(bgridsize)) bgridsize <- default.bgridsize(d)

    if (binned)
    {
        d <- ncol(Sigma)
        n <- nrow(x)

        if (missing(bin.par)) bin.par <- binning(x, H=diag(diag(Sigma)), bgridsize=bgridsize)  
        est <- kdde.binned(x=x, bin.par=bin.par, H=Sigma, deriv.order=r, verbose=verbose)$estimate

        if (r>0)
        {
            sumval <- rep(0, length(est))
            for (j in 1:length(est)) sumval[j] <- sum(bin.par$counts * n * est[[j]])
        }
        else
            sumval <- sum(bin.par$counts * n * est)
        ## transformation approach from Jose E. Chacon 06/12/2010
        if (0)
        {
            Sigmainv12 <- matrix.sqrt(chol2inv(chol(Sigma)))
            y <- x %*% Sigmainv12
            if (missing(bin.par)) bin.par <- binning(x=y, H=diag(d), bgridsize=bgridsize)

            est <- kdde.binned(x=y, bin.par=bin.par, H=diag(d), deriv.order=r, verbose=verbose)$estimate
            if (r>0)
            {
                sumval <- rep(0, length(est))
                for (j in 1:length(est)) sumval[j] <- sum(bin.par$counts *n*est[[j]]) 
            }
            else 
                sumval <- sum(bin.par$counts * n * est)

            sumval <- det(Sigmainv12) * sumval  %*% Kpow(Sigmainv12, pow=r)
        }
    }
    ## exact computation 
    else
    {
        if (verbose) pb <- txtProgressBar() 
        if (r==0)
        {
            n.seq <- block.indices(n, n, d=d, r=r, diff=TRUE)
            sumval <- 0
            for (i in 1:(length(n.seq)-1))
            {  
                difs <- differences(x=x, y=x[n.seq[i]:(n.seq[i+1]-1),])
                sumval <- sumval + sum(dmvnorm.deriv(x=difs, mu=rep(0,d), Sigma=Sigma, deriv.order=r, deriv.vec=deriv.vec))
                if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1)) 
            }
            if (verbose) close(pb)
        }
        else
        {
            ## only with r>0 
            ## original recursive code from Jose E. Chacon 03/2012
            if (2*floor(r/2)!=r) { sumval <- rep(0,d^r) }
            else
            {
                Sigmainv <- chol2inv(chol(Sigma))
                per <- perm.rep(d=d,r=r)
                dind <- numeric()
                for (i in 1:d) dind <- cbind(dind,rowSums(per==i)) ## matrix of derivative indices

                udind <- unique(dind)
                nudind <- nrow(udind)        
                dlabs <- numeric(nrow(dind))
                for (i in 1:nrow(udind))
                    dlabs <- dlabs+i*(rowSums((dind-matrix(rep(udind[i,],nrow(dind)),nrow=nrow(dind),byrow=TRUE))^2)==0)
                      
                result <- rep(0,nudind)

                ndif <- n*(n-1)/2
                dif.ind <- numeric()

                M <- 1e6
                max.loop.size <- ceiling(M/nudind) ## inside the loop we need to store a matrix of order max.loop.size x nudind <= M
                nblocks <- ceiling(ndif/max.loop.size)
                blength <- c(rep(max.loop.size,nblocks-1),ndif-max.loop.size*(nblocks-1)) ## length of each of the blocks, the last one could be smaller
                
                tri.num <- (1:n)*((1:n)-1)/2
                for (kk in 1:nblocks)
                {
                    b <- blength[kk]
                    if (verbose) setTxtProgressBar(pb, kk/nblocks) 
                    kkm <- (kk-1)*max.loop.size
             
                    tri.ind <- findInterval(kkm:(kkm+b-1), tri.num)
                    dif.ind.block <- cbind(kkm:(kkm+b-1) - tri.num[tri.ind]+1, tri.ind+1) 
                    difs.block <- x[dif.ind.block[,1],]-x[dif.ind.block[,2],]
              
                    arg <- difs.block %*% Sigmainv    
                    narg <- nrow(arg) 

                    hmold0 <- matrix(rep(1,narg),ncol=1,nrow=narg)
                    hmold1 <- arg
                    hmnew <- hmold0

                    udind0 <- matrix(rep(0,d),nrow=1,ncol=d)
                    udind1 <- diag(d)
              
                    if (r==1) { hmnew <- hmold1 }
                    if (r >= 2)
                    { 
                        for (i in 2:r)
                        {            
                            Ndi1 <- ncol(hmold1)
                            hmnew <- numeric()
                      
                            for (j in 1:d)
                            {        
                                nrecj <- choose(d-j+i-1,i-1)        
                                hmnew.aux <- arg[,j]*hmold1[,Ndi1-(nrecj:1)+1]
                                
                                for (k in j:d)
                                {        
                                    udind0.aux <- matrix(udind1[Ndi1-(nrecj:1)+1,],ncol=d,byrow=FALSE)
                                    udind0.aux[,k] <- udind0.aux[,k]-1
                                    valid.udind0 <- as.logical(apply(udind0.aux>=0,1,min))
                                    enlarged.hmold0 <- matrix(0,ncol=nrow(udind0.aux),nrow=narg)
                                    for (ell in 1:nrow(udind0.aux))
                                    {
                                        if (valid.udind0[ell])
                                        {
                                            pos <- which(rowSums((udind0-matrix(rep(udind0.aux[ell,],nrow(udind0)),nrow=nrow(udind0),byrow=TRUE))^2)==0)
                                            enlarged.hmold0[,ell] <- hmold0[,pos]
                                        }
                                    }
                                    ## in enlarged.hmold0 we put the vector hmold0 in those positions not having a -1 in any of the derivative order after subtracting e_k
                                    ## the remaining positions are zeroes
                                    ## surely this could be done in a more efficient way

                                    hmnew.aux <- hmnew.aux-Sigmainv[j,k]*matrix(rep(udind1[Ndi1-(nrecj:1)+1,k],narg),nrow=narg,byrow=TRUE)*enlarged.hmold0
                                }
                                hmnew <- cbind(hmnew,hmnew.aux)
                            }
                            hmold0 <- hmold1
                            hmold1 <- hmnew
                      
                            ## compute the unique i-th derivative multi-indexes 
                            nudind1 <- nrow(udind1)
                            udindnew <- numeric()
                      
                            for (j in 1:d)
                            {
                                Ndj1i <- choose(d+i-1-j,i-1)
                                udind.aux <- matrix(udind1[nudind1-(Ndj1i:1)+1,],ncol=d,byrow=FALSE)
                                udind.aux[,j] <- udind.aux[,j]+1
                                udindnew <- rbind(udindnew,udind.aux)
                            }
                            udind0 <- udind1
                            udind1 <- udindnew
                        }
                    }

                    hmnew <- hmnew*matrix(rep((-1)^rowSums(udind),narg),nrow=narg,byrow=TRUE)
                    phi <- dmvnorm(difs.block, mean=rep(0,d), sigma=Sigma)
                    phi <- matrix(rep(phi,nudind),ncol=nudind,byrow=FALSE)
                    result <- result+drop(colSums(phi*hmnew))
                }

                result <- result[dlabs]    
                hm0 <- dmvnorm.deriv(x=rep(0,d),Sigma=Sigma,deriv.order=deriv.order)
                sumval <- 2*result+n*hm0
                if (verbose) close(pb)
            }      
        }        
    }

    if (inc==0)
        sumval <- sumval - n*dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=Sigma, deriv.order=r)

    sumval <- drop(sumval)  
    if (kfe) { if (inc==1) sumval <- sumval/n^2 else sumval <- sumval/(n*(n-1)) }

    if (add.index)
    {
        ind.mat <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=diag(d), deriv.order=r, only.index=TRUE)
        if (deriv.vec) return(list(sum=sumval, deriv.ind=ind.mat))
        else return(list(sum=sumval, deriv.ind=unique(ind.mat)))
    }
    else return(sum=sumval)
}

## Single partial derivative of the multivariate normal with scalar variance matrix sigma^2 I_d  
## Code by Jose  Chacon 04/09/2007
dmvnorm.deriv.scalar <- function(x, mu, sigma, deriv.order, binned=FALSE)
{
    r <- deriv.order
    d <- ncol(x)

    sderiv <- sum(r)
    arg <- x/sigma
    darg <- dmvnorm(arg, mean=mu)/(sigma^(sderiv+d))
    for (j in 1:d)
    {
        hmold0 <- 1
        hmold1 <- arg[,j]
        hmnew <- 1
        if (r[j] == 1) { hmnew <- hmold1 }
        if (r[j] >= 2) ## Multiply by the corresponding Hermite polynomial, coordinate-wise, using Fact C.1.4 in W&J (1995) and Willink (2005, p.273)
            for (i in (2:r[j]))
            {
                hmnew <- arg[,j] * hmold1 - (i - 1) * hmold0
                hmold0 <- hmold1
                hmold1 <- hmnew
            }
        darg <- hmnew * darg
    }
    val <- darg*(-1)^sderiv
    
    return(val)
}

dmvnorm.deriv.scalar.sum <- function(x, sigma, deriv.order=0, inc=1, kfe=FALSE, binned=FALSE, bin.par, verbose=FALSE)
{
    r <- deriv.order
    d <- ncol(x)
    n <- nrow(x)

    if (binned)
    {
        if (missing(bin.par)) bin.par <- binning(x, H=diag(d)*sigma^2)  
        n <- sum(bin.par$counts)

        ind.mat <- dmvnorm.deriv(x=rep(0,d), Sigma=diag(d), deriv.order=sum(r), deriv.vec=TRUE, only.index=TRUE)
        fhatr <- kdde.binned(bin.par=bin.par, H=sigma^2*diag(d), deriv.order=sum(r), deriv.vec=TRUE, w=rep(1,n), deriv.index=which.mat(r=r, ind.mat)[1])
        sumval <- sum(bin.par$counts * n * fhatr$est[[1]])
    }
    else
    {
        if (verbose) pb <- txtProgressBar() 
    
        n.seq <- block.indices(n, n, d=d, r=r, diff=TRUE)
        sumval <- 0
        for (i in 1:(length(n.seq)-1))
        {  
            difs <- differences(x=x, y=x[n.seq[i]:(n.seq[i+1]-1),])
            sumval <- sumval + sum(dmvnorm.deriv.scalar(x=difs, mu=rep(0,d), sigma=sigma, deriv.order=r))
            if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1)) 
        }
    }
    if (verbose) close(pb)
  
    if (inc==0)
        sumval <- sumval - n*dmvnorm.deriv.scalar(x=t(as.matrix(rep(0,d))), mu=rep(0,d), sigma=sigma, deriv.order=r)
  
    if (kfe)
        if (inc==1) sumval <- sumval/n^2
        else sumval <- sumval/(n*(n-1))
  
    return(sumval)
}

##########################################################################
## Normal scale psi functionals
##########################################################################
psins.1d <- function(r, sigma)
{
    if (r %% 2 ==0)
        psins <- (-1)^(r/2)*factorial(r)/((2*sigma)^(r+1)*factorial(r/2)*pi^(1/2))
    else
        psins <- 0
    
    return(psins)  
}

psins <- function(r, Sigma, deriv.vec=length(r)==1)
{
    d <- ncol(Sigma)
    if (deriv.vec)
    {
        dens <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), deriv.order=r, Sigma=2*Sigma, add.index=FALSE)
    }
    else
    {
        dens <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), deriv.order=sum(r), Sigma=2*Sigma, add.index=TRUE)
        if (!is.vector(dens$deriv.ind))
        {
            i <- head(which.mat(r, dens$deriv.ind),n=1)
            dens <- dens$deriv[1,i]
        }
        else
            dens <- dens$deriv      
    }
    dens <- drop(dens)

    return(dens)
}

##########################################################################
## Vector moments of the normal distribution
##########################################################################
mur <- function(r, A, mu, Sigma, type="unique")
{
    type1 <- match.arg(type, c("direct", "recursive", "unique"))
    mur.val <- do.call(paste("mur", type1, sep="."), list(r=r, A=A, mu=mu, Sigma=Sigma))

    return(mur.val)
}

#############################################################################
## mur.direct computes the vector moment E[X^{\otimes r}] for a random
## vector with N(mu,Sigma) distribution, on the basis of Equation (8) in
## Section 6 of Chacon and Duong (2014)
#############################################################################
mur.direct <- function(r, mu, Sigma)
{
    d <- ncol(Sigma)
    result <- as.vector(Kpow(mu,r))
    vS <- vec(Sigma)
    vSj <- 1
    if (r>=2)
    {
        for(j in 1:floor(r/2))
        {
            vSj <- as.vector(vSj%x%vS)
            cj <- prod(r:(r-2*j+1))/(prod(1:j)*2^j)
            mur2j <- as.vector(Kpow(mu,r-2*j))
            result <- result+cj*as.vector(mur2j%x%vSj)
        }
    }

    return(drop(Sdrv.recursive(d=d,r=r,v=result)))
}

#############################################################################
## mur.recursive computes the vector moment E[X^{\otimes r}] for a random
## vector with N(mu,Sigma) distribution, on the basis of Equation (9) in
## Section 6 of Chacon and Duong (2014), using Equation (7) in Section 5
## to obtain the Hermite polynomial
#############################################################################
mur.recursive <- function(r, mu, Sigma)
{ 
    d <- ncol(Sigma)
    G <- -Sigma
    vG <- vec(G)
    arg <- mu
    hmold0 <- 1
    hmold1 <- arg
    hmnew <- hmold0

    if (r==1) { hmnew <- hmold1 }
    if (r>=2) 
    {
        for (i in 2:r)
        {
            hmnew <- as.vector(arg%x%hmold1-(i-1)*(vG%x%hmold0))
            hmold0 <- hmold1
            hmold1 <- hmnew        
        }    
    }

    return(drop(Sdrv.recursive(d=d,r=r,v=hmnew)))   
}

############################################################################
## mur.unique computes the vector moment E[X^{\otimes r}] for a random vector 
## with N(mu,Sigma) distribution, on the basis of Equation (9) in Section 6
## of Chacon and Duong (2014), using Algorithm 3 in Section 5, based on the
## unique partial derivatives, to obtain the Hermite polynomial
############################################################################
mur.unique <- function(r, mu, Sigma)
{
    d <- ncol(Sigma)
    G <- -Sigma
    arg <- mu
    hmold0 <- 1
    hmold1 <- arg
    hmnew <- hmold0
    udind0 <- matrix(rep(0,d),nrow=1,ncol=d)
    udind1 <- diag(d)
    
    if (r==1) { hmnew <- hmold1 }
    if (r>=2)
    {
        for (i in 2:r)
        {
            Ndi1 <- length(hmold1)
            Ndi0 <- length(hmold0)
            hmnew <- numeric()
            
            for (j in 1:d)
            {        
                nrecj <- choose(d-j+i-1,i-1)        
                hmnew.aux <- arg[j]*hmold1[Ndi1-(nrecj:1)+1]
                      
                for(k in j:d)
                {        
                    udind0.aux <- matrix(udind1[Ndi1-(nrecj:1)+1,],ncol=d,byrow=FALSE)
                    udind0.aux[,k] <- udind0.aux[,k]-1
                    valid.udind0 <- as.logical(apply(udind0.aux>=0,1,min))
                    enlarged.hmold0 <- rep(0,nrow(udind0.aux))
                    for (ell in 1:nrow(udind0.aux))
                    {
                        if (valid.udind0[ell])
                        {
                            pos <- which(rowSums((udind0-matrix(rep(udind0.aux[ell,],nrow(udind0)),
                             nrow=nrow(udind0),byrow=TRUE))^2)==0)
                            enlarged.hmold0[ell] <- hmold0[pos]
                        }
                    }
                    hmnew.aux <- hmnew.aux-G[j,k]*udind1[Ndi1-(nrecj:1)+1,k]*enlarged.hmold0
                }
                hmnew <- c(hmnew,hmnew.aux)
            }
            hmold0 <- hmold1
            hmold1 <- hmnew

            ## Compute the unique i-th derivative multi-indexes 
            nudind1 <- nrow(udind1)
            udindnew <- numeric()
            for (j in 1:d)
            {
                Ndj1i <- choose(d+i-1-j,i-1)
                udind.aux <- matrix(udind1[nudind1-(Ndj1i:1)+1,],ncol=d,byrow=FALSE)
                udind.aux[,j] <- udind.aux[,j]+1
                udindnew <- rbind(udindnew,udind.aux)
            }
            udind0 <- udind1
            udind1 <- udindnew
        }
    }
            
    if (r==0) { result <- 1 }
    if (r==1) { result <- (-1)*hmnew } 
    if (r>=2)
    {
        per <- pinv.all(d=d,r=r)    
        dind <- numeric()
        udind <- udind1
        dind.base <- rep(0,d^r)        
        udind.base <- rep(0,choose(d+r-1,r))
                
        for (i in 1:d)
        {
            dind <- cbind(dind,rowSums(per==i)) ## Matrix of derivative indices
            dind.base <- dind.base+dind[,i]*(r+1)^(d-i) ## Transform each row to base r+1           
            udind.base <- udind.base+udind[,i]*(r+1)^(d-i) ## Transform each row to base r+1                       
        }        
        dlabs <- match(dind.base,udind.base)
        result <- hmnew[dlabs]*(-1)^rowSums(dind)
    }

    return(drop(result*(-1)^r))
}

##########################################################################
## Moments of quadratic forms in normal variables
##########################################################################
nur <- function(r, A, mu, Sigma, type="cumulant")
{
    type <- match.arg(type, c("direct", "recursive", "unique", "cumulant"))
    nur.val <- do.call(paste("nur", type, sep="."), list(r=r, A=A, mu=mu, Sigma=Sigma))
  
    return(nur.val)
}

nurs <- function(r, s, A, B, mu, Sigma, type="cumulant")
{
    type <- match.arg(type, c("direct", "recursive", "unique", "cumulant"))
    nur.val <- do.call(paste("nurs", type, sep="."), list(r=r, s=s, A=A, B=B, mu=mu, Sigma=Sigma))

    return(nur.val)
}

#############################################################################
## nur.direct computes the moment E[(X^T AX)^r] of the quadratic form
## X^T AX where X is a random vector with N(mu,Sigma) distribution, using
## Equation (10) in Section 6 of Chacon and Duong (2014), and the direct
## implementation mur.direct of the normal moments
#############################################################################
nur.direct <- function(r,A,mu,Sigma)
{
    vA <- vec(A)
    result <- drop(Kpow(t(vA),r)%*%mur.direct(2*r,mu,Sigma))
    
    return(result)
}

#############################################################################
## nur.recursive computes the moment E[(X^T AX)^r] of the quadratic form
## X^T AX where X is a random vector with N(mu,Sigma) distribution, using
## Equation (10) in Section 6 of Chacon and Duong (2014), and the recursive
## implementation mur.recursive of the normal moments
#############################################################################
nur.recursive <- function(r,A,mu,Sigma)
{
    vA <- vec(A)
    result <- drop(Kpow(t(vA),r)%*%mur.recursive(2*r,mu,Sigma))
    
    return(result)
}

#############################################################################
## nur.unique computes the moment E[(X^T AX)^r] of the quadratic form
## X^T AX where X is a random vector with N(mu,Sigma) distribution, using
## Equation (10) in Section 6 of Chacon and Duong (2014), and the function
## mur.unique to compute the normal moments from its unique coordinates
#############################################################################
nur.unique <- function(r,A,mu,Sigma)
{
    vA <- vec(A)
    result <- sum(Kpow(vA,r)*mur.unique(2*r,mu,Sigma))
    
    return(result)
}

#############################################################################
## nur.cumulant computes the moment E[(X^T AX)^r] of the quadratic form
## X^T AX where X is a random vector with N(mu,Sigma) distribution, using
## the recursive formula relating moments and cumulants
#############################################################################
nur.cumulant <- function(r,A,mu,Sigma)
{
    if(r==0) { result <- 1 }
    if(r==1) { result <- sum(diag(A%*%Sigma))+t(mu)%*%A%*%mu }
    if(r>=2)
    {
        ASigma <- A%*%Sigma
        AS <- ASigma
        Amu <- A%*%mu
        kappas <- sum(diag(ASigma)+mu*Amu)
        nus <- kappas
        for (k in 2:r)
        {
            knew <- k*t(mu)%*%ASigma%*%Amu
            ASigma <- ASigma%*%AS
            knew <- (knew+sum(diag(ASigma)))*factorial(k-1)*2^(k-1)
            nnew <- knew+sum(choose(k-1,1:(k-1))*nus*rev(kappas))
            kappas <- c(kappas,knew)
            nus <- c(nus,nnew)
        }
        result <- nnew
    }

    return(drop(result))
}

#############################################################################
## nurs.direct computes the joint moment E[(X^T AX)^r (X^T BX)^s] of the
## quadratic forms X^T AX and X^T BX, where X is a random vector with
## N(mu,Sigma) distribution, using Equation (10) in Section 6 of Chacon and
## Duong (2014), and the direct implementation mur.direct of the
## normal moments
#############################################################################
nurs.direct <- function(r,s,A,B,mu,Sigma)
{
    vA <- vec(A)
    vB <- vec(B)    
    result <- (Kpow(t(vA),r)%x%Kpow(t(vB),s))%*%mur.direct(2*r+2*s,mu,Sigma)
    
    return(drop(result))
}

#############################################################################
## nurs.recursive computes the joint moment E[(X^T AX)^r (X^T BX)^s] of the
## quadratic forms X^T AX and X^T BX, where X is a random vector with
## N(mu,Sigma) distribution, using Equation (10) in Section 6 of Chacon and
## Duong (2014), and the recursive implementation mur.recursive of the
## normal moments
#############################################################################
nurs.recursive <- function(r,s,A,B,mu,Sigma)
{
    vA <- vec(A)
    vB <- vec(B)    
    result <- (Kpow(t(vA),r)%x%Kpow(t(vB),s))%*%mur.recursive(2*r+2*s,mu,Sigma)
    
    return(drop(result))
}

#############################################################################
## nurs.unique computes the joint moment E[(X^T AX)^r (X^T BX)^s] of the
## quadratic forms X^T AX and X^T BX, where X is a random vector with
## N(mu,Sigma) distribution, using Equation (10) in Section 6 of Chacon and
## Duong (2014), and the function mur.unique to compute the normal moments
## from its unique coordinates
#############################################################################
nurs.unique <- function(r,s,A,B,mu,Sigma)
{
    vA <- vec(A)
    vB <- vec(B)    
    result <- drop((Kpow(t(vA),r)%x%Kpow(t(vB),s))%*%mur.unique(2*r+2*s,mu,Sigma))
    
    return(drop(result))
}

#############################################################################
## nurs.cumulant computes the joint moment E[(X^T AX)^r (X^T BX)^s] of the
## quadratic forms X^T AX and X^T BX, where X is a random vector with
## N(mu,Sigma) distribution, using the recursive formula (11) in Section 6
## of Chacon and Duong (2014), relating moments and cumulants. The cumulants
## are computed using the function kappars, which is based on Theorem 3
#############################################################################
kappars <- function(r,s,A,B,mu,Sigma)
{ 
    d <- ncol(A)
    if (r+s>1 & r>0 & s>0) { ind <- multicool::allPerm(multicool::initMC(c(rep(1,r),rep(2,s)))) }
    if (r+s==1 | r==0 | s==0) { ind <- matrix(c(rep(1,r),rep(2,s)),nrow=1) }
    if (r+s==0) { return(0) }
    nper <- nrow(ind)
   
    result <- 0
    Dmat <- solve(Sigma)%*%mu%*%t(mu)
    ASigma <- A%*%Sigma
    BSigma <- B%*%Sigma
    Id <- diag(d)
    
    for (i in 1:nper)
    {
        product <- Id
        for (j in 1:(r+s))
        {
            if (ind[i,j]==1) { product <- product%*%ASigma }
            else if(ind[i,j]==2) { product <- product%*%BSigma }
        }
        result <- result+sum(diag(product%*%(Id/(r+s)+Dmat))) 
    }
    result <- result*factorial(r)*factorial(s)*2^(r+s-1)

    return(drop(result))
}

nurs.cumulant <- function(r,s,A,B,mu,Sigma)
{ 
    if(r==0  & s>0) { nurs <- nur.cumulant(s,B,mu,Sigma) }
    if(r>0   & s==0) { nurs <- nur.cumulant(r,A,mu,Sigma) }
    if(r==0  & s==0) { nurs <- 1 }
    if((r>0) & (s>0))
    {
        K <- matrix(0,nrow=r+1,ncol=s)
        for (i in 0:r)
        {
            for (j in 1:s)
            {
                K[i+1,j] <- kappars(i,j,A,B,mu,Sigma)
            }
        }

        N <- matrix(0,nrow=r+1,ncol=s)
        for (i in 0:r)
        {
            N[i+1,1] <- nur.cumulant(r=i,A=A,mu=mu,Sigma=Sigma)
        }
        if (s>1)
        {
            for (j in 1:(s-1))
            {
                for (i in 0:r)
                {
                    Choose <- outer(choose(i,0:i), choose(j-1,0:(j-1)))
                    N[i+1,j+1] <- sum(Choose*N[1:(i+1),1:j]*K[(i:0)+1,j:1])
                }
            }
        }
        Choose <- outer(choose(r,0:r),choose(s-1,0:(s-1)))  
        nurs <- sum(Choose*N[1:(r+1),1:s]*K[(r:0)+1,s:1])
    }

    return(nurs)
}

##########################################################################
## V-statistics with multivariate Gaussian derivatives kernel
##########################################################################
Qr <- function(x, y, Sigma, deriv.order=0, inc=1, type="cumulant", verbose=FALSE)
{
    if (missing(y)) y <- x
    type <- match.arg(type, c("direct", "cumulant"))
    if (type=="direct")
        Qr.val <- Qr.direct(x=x, y=y, Sigma=Sigma, r=deriv.order, inc=inc, verbose=verbose)
    else
        Qr.val <- Qr.cumulant(x=x, y=y, Sigma=Sigma, r=deriv.order, inc=inc, verbose=verbose)
    
    return(Qr.val)
}

############################################################################
## Qr.direct computes the V-statistic using the direct approach, as
## described in Section 6 of Chacon and Duong (2014)
############################################################################
Qr.direct <- function(x, y, Sigma, r=0, inc=1, binned=FALSE, bin.par, bgridsize, verbose=FALSE)
{
    if (is.vector(x)) x <- matrix(x, nrow=1)
    d <- ncol(x)

    eta <- drop(Kpow(vec(diag(d)), r/2) %*% kfe(x=x, G=Sigma, deriv.order=r, inc=inc, verbose=verbose, binned=binned, bin.par=bin.par, add.index=FALSE))

    return(eta)
}

############################################################################
## Qr.cumulant computes the V-statistic using the relationship with nur
## shown in Theorem 4 of Chacon and Duong (2014)
############################################################################
Qr.cumulant <- function(x, y, Sigma, r=0, inc=1, verbose=FALSE)
{
    if (is.vector(x)) x <- matrix(x, nrow=1)
    d <- ncol(x)
    r <- r/2
    if (missing(y)) y <- x
    if (is.vector(y)) y <- matrix(y, nrow=1)
    nx <- as.numeric(nrow(x))
    ny <- as.numeric(nrow(y))
    G <- Sigma
    Ginv <- chol2inv(chol(G))
    G2inv <- Ginv%*%Ginv
    G3inv <- G2inv%*%Ginv
    trGinv <- sum(diag(Ginv))
    trG2inv <- sum(diag(G2inv))   
    detG <- det(G)

    ## indices for separating into blocks for double sum calculation
    n.seq <- block.indices(nx, ny, d=d, r=r, diff=FALSE)
    if (verbose) pb <- txtProgressBar()

    if (r==0)
    {
    xG <- x%*%Ginv
    a <- rowSums(xG*x)
    eta <- 0
    for (i in 1:(length(n.seq)-1))
    {
        nytemp <- n.seq[i+1] - n.seq[i]
        ytemp <- matrix(y[n.seq[i]:(n.seq[i+1]-1),], ncol=d)
        aytemp <- rowSums((ytemp %*% Ginv) *ytemp)
        M <- a%*%t(rep(1,nytemp)) + rep(1, nx)%*%t(aytemp) - 2*(xG%*%t(ytemp))
        em2 <- exp(-M/2)
        eta <- eta + (2*pi)^(-d/2)*detG^(-1/2)*sum(em2)
        if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
    }
    } 
    else if (r==1)
    {
        xG <- x%*%Ginv
        xG2 <- x%*%G2inv
        a <- rowSums(xG*x)
        a2 <- rowSums(xG2*x)

        eta <- 0
        for (i in 1:(length(n.seq)-1))
        {
            nytemp <- n.seq[i+1] - n.seq[i]
            ytemp <- matrix(y[n.seq[i]:(n.seq[i+1]-1),], nrow=nytemp)
            aytemp <- rowSums((ytemp %*% Ginv) *ytemp)
            aytemp2 <- rowSums((ytemp %*% G2inv) *ytemp)
            M  <- a%*%t(rep(1,nytemp))+rep(1,nx)%*%t(aytemp)-2*(xG%*%t(ytemp))
            M2 <- a2%*%t(rep(1,nytemp))+rep(1,nx)%*%t(aytemp2)-2*(xG2%*%t(ytemp))
            eta <- eta + (2*pi)^(-d/2)*detG^(-1/2)*sum(exp(-M/2)*(M2-trGinv))
            if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
        }
    }
    else if (r==2)
    {
        xG <- x%*%Ginv
        xG2 <- x%*%G2inv
        xG3 <- x%*%G3inv
        a <- rowSums(xG*x)
        a2 <- rowSums(xG2*x)
        a3 <- rowSums(xG3*x)

        eta <- 0
        for (i in 1:(length(n.seq)-1))
        {
            nytemp <- n.seq[i+1] - n.seq[i]
            ytemp <- matrix(y[n.seq[i]:(n.seq[i+1]-1),], ncol=d)
            aytemp <- rowSums((ytemp %*% Ginv) *ytemp)
            aytemp2 <- rowSums((ytemp %*% G2inv) *ytemp)
            aytemp3 <- rowSums((ytemp %*% G3inv) *ytemp)
            M  <- a%*%t(rep(1,nytemp))+rep(1,nx)%*%t(aytemp)-2*(xG%*%t(ytemp))
            M2 <- a2%*%t(rep(1,nytemp))+rep(1,nx)%*%t(aytemp2)-2*(xG2%*%t(ytemp))
            M3 <- a3%*%t(rep(1,nytemp))+rep(1,nx)%*%t(aytemp3)-2*(xG3%*%t(ytemp))
            eta <- eta + (2*pi)^(-d/2)*detG^(-1/2)*sum(exp(-M/2)*(2*trG2inv-4*M3
                 +(-trGinv+M2)^2))
            if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
        }
    }
    else if (r>2)
    {
        xG <- x%*%Ginv
        a <- rowSums(xG*x)
        eta <- 0
        for (i in 1:(length(n.seq)-1))
        {
            nytemp <- n.seq[i+1] - n.seq[i]
            ytemp <- matrix(y[n.seq[i]:(n.seq[i+1]-1),], ncol=d)
            aytemp <- rowSums((ytemp %*% Ginv) *ytemp)
            M <- a %*% t(rep(1,nytemp)) + rep(1,nx)%*%t(aytemp) - 2*(xG%*%t(ytemp))
            edv2 <- exp(-M/2)

            P0 <- Ginv
            kappas <- matrix(nrow=as.numeric(nx*nytemp), ncol=r)
            for (j in 1:r)
            {
                Gi1inv <- P0%*%Ginv
                trGi0inv <- sum(diag(P0))    
                xGi1inv <- x%*%Gi1inv
                xGi1invx <- rowSums(xGi1inv*x)
                aytemp <- rowSums((ytemp %*% Gi1inv) *ytemp)
                dvi1 <- xGi1invx%*%t(rep(1,nytemp))+rep(1,nx)%*%t(aytemp)-2*(xGi1inv%*%t(ytemp))
                kappas[,j] <- (-2)^(j-1)*factorial(j-1)*(-trGi0inv+j*dvi1)
                P0 <- Gi1inv
            }
          
            nus <- matrix(nrow=as.numeric(nx*nytemp), ncol=r+1)        
            nus[,1] <- 1        
            for (j in 1:r)
            {
                js <- 0:(j-1)
                if (j==1) nus[,2] <- kappas[,1]
                else nus[,j+1] <- rowSums(kappas[,j:1]*nus[,1:j]/matrix(rep(factorial(js)*
                     factorial(rev(js)),nx*nytemp),nrow=nx*nytemp,byrow=TRUE))*factorial(j-1)
            }
            eta <- eta + (2*pi)^(-d/2)*detG^(-1/2)*sum(edv2*nus[,r+1])
            if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
        }
    }
    if (verbose) close(pb)
    if (inc==0) eta <- (eta - (-1)^r*nx*nur.cumulant(r=r, A=Ginv, mu=rep(0,d), Sigma=diag(d))*(2*pi)^(-d/2)*detG^(-1/2))/(nx*(ny-1))
    if (inc==1) eta <- eta/(nx*ny)  

    return(eta)
}

Qr.1d <- function(x, y, sigma, deriv.order=0, inc=1, verbose=FALSE)
{
    d <- 1
    r <- deriv.order/2
    if (missing(y)) y <- x
    nx <- length(x)
    ny <- length(y)
    g <- sigma

    n.seq <- block.indices(nx, ny, d=1, r=0, diff=FALSE)
    eta <- 0
    if (verbose) pb <- txtProgressBar() 
  
    if (r==0)
    {
        a <- x^2
        for (i in 1:(length(n.seq)-1))
        {
            if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
            nytemp <- n.seq[i+1] - n.seq[i]
            ytemp <- y[n.seq[i]:(n.seq[i+1]-1)]
            aytemp <- ytemp^2
            M <- a %*%t(rep(1,nytemp)) + rep(1, nx)%*%t(aytemp) - 2*(x %*% t(ytemp))
            em2 <- exp(-M/(2*g^2))
            eta <- eta + (2*pi)^(-d/2)*g^(-1)*sum(em2)
        }
    }
    else if (r>0)
    {
        a <- x^2
        for (i in 1:(length(n.seq)-1))
        {
            if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
            nytemp <- n.seq[i+1] - n.seq[i]
            ytemp <- y[n.seq[i]:(n.seq[i+1]-1)]
            aytemp <- ytemp^2 
            M <- a %*% t(rep(1,nytemp)) + rep(1,nx)%*%t(aytemp) - 2*(x %*%t(ytemp))
            edv2 <- exp(-M/(2*g^2))

            kappas <- matrix(nrow=as.numeric(nx*nytemp), ncol=r)
            for (i in 1:r)
            {
                aytemp <- ytemp^2
                dvi1 <- (a %*% t(rep(1,nytemp)) + rep(1,nx) %*% t(aytemp) - 2*(x%*%t(ytemp)))/g^(2*(i+1))
                kappas[,i] <- (-2)^(i-1)*factorial(i-1)*(-g^(-2*i)+i*dvi1)
            }
          
            nus <- matrix(nrow=as.numeric(nx*nytemp), ncol=r+1)        
            nus[,1] <- 1        
            for (j in 1:r)
            {
                js <- 0:(j-1)
                if (j==1) nus[,2] <- kappas[,1]
                else nus[,j+1] <- rowSums(kappas[,j:1]*nus[,1:j]/matrix(rep(factorial(js)*factorial(rev(js)),nx*nytemp),nrow=nx*nytemp,byrow=TRUE))*factorial(j-1)
            }
            eta <- eta + (2*pi)^(-d/2)*g^(-1)*sum(edv2*nus[,r+1])
        }
    }
    if (verbose) close(pb)
    if (inc==0) eta <- (eta - nx*dnorm.deriv(x=0, mu=0, sigma=g, deriv.order=deriv.order))/(nx*(ny-1))
    if (inc==1) eta <- eta/(nx*ny) 

    return(eta)
}

############################################################################
## Creates plots of mixture density functions
##
## Parameters
## mus - means
## Sigmas - variances
## props - vector of proportions of each mixture component 
## dfs - degrees of freedom
## dist - "normal" - normal mixture
##      - "t" - t mixture
## ...
############################################################################
plotmixt <- function(mus, sigmas, Sigmas, props, dfs, dist="normal", draw=TRUE, deriv.order=0, which.deriv.ind=1, binned=TRUE, ...)
{
    ## locally set random seed not to interfere with global random number generators
    if (!exists(".Random.seed")) rnorm(1)
    old.seed <- .Random.seed
    on.exit( { .Random.seed <<- old.seed } )
    set.seed(8192)

    if (!missing(sigmas)) plotmixt.1d(mus=mus, sigmas=sigmas, props=props, dfs=dfs, dist=dist, draw=draw, deriv.order=deriv.order, which.deriv.ind=which.deriv.ind, ...)
    else if (ncol(Sigmas)==2)
    plotmixt.2d(mus=mus, Sigmas=Sigmas, props=props, dfs=dfs, dist=dist, draw=draw, deriv.order=deriv.order, which.deriv.ind=which.deriv.ind,binned=binned, ...)
    else if (ncol(Sigmas)==3)
    plotmixt.3d(mus=mus, Sigmas=Sigmas, props=props, dfs=dfs, dist=dist, draw=draw, deriv.order=deriv.order, which.deriv.ind=which.deriv.ind, binned=binned, ...)
}

plotmixt.1d <- function(mus, sigmas, props, dfs, dist="normal", xlim, ylim, gridsize, draw=TRUE, deriv.order, which.deriv.ind, ...)
{
    dist <- match.arg(dist, c("normal", "t")) 
    maxsigmas <- 4*max(sigmas)

    if (missing(xlim)) xlim <- c(min(mus) - maxsigmas, max(mus) + maxsigmas)  
    if (missing(gridsize)) gridsize <- default.gridsize(1)

    x <- seq(xlim[1]-0.1*abs(diff(xlim)), xlim[2]+0.1*abs(diff(xlim)), length=gridsize)
    if (dist=="normal")
    {
        if (deriv.order<=0) dens <- dnorm.mixt(x=x, mus=mus, sigmas=sigmas, props=props)
        else dens <- dnorm.deriv.mixt(x=x, mus=mus, sigmas=sigmas, props=props, deriv.order=deriv.order)
    }
    else if (dist=="t") stop("1-d t mixture not yet implemented")

    fhat <- list()
    fhat$x <- x
    fhat$eval.points <- x
    fhat$estimate <- dens
    fhat$H <- diag(1)
    fhat$h <- 1
    fhat$gridtype <- "linear"
    fhat$gridded <- TRUE
    fhat$binned <- FALSE
    fhat$names <- parse.name(x)
    fhat$w <- rep(1, length(x))
    if (deriv.order>0)
    {
        fhat$deriv.order <- deriv.order
        fhat$deriv.ind <- which.deriv.ind
    }
    class(fhat) <- "kdde"

    if (draw) plot(fhat, xlim=xlim, ...)
    invisible(fhat)
}

plotmixt.2d <- function(mus, Sigmas, props, dfs, dist="normal", xlim, ylim, gridsize, nrand=1e4, draw=TRUE, binned, deriv.order, which.deriv.ind, display="slice", ...)
{
    dist <- match.arg(dist, c("normal", "t"))
    maxSigmas <- 4*max(Sigmas)

    if (is.vector(mus)) mus <- as.matrix(t(mus))
    if (missing(xlim)) xlim <- c(min(mus[,1]) - maxSigmas, max(mus[,1]) + maxSigmas)
    if (missing(ylim)) ylim <- c(min(mus[,2]) - maxSigmas, max(mus[,2]) + maxSigmas)
    if (missing(gridsize)) gridsize <- default.gridsize(2)

    x <- seq(xlim[1], xlim[2], length=gridsize[1])
    y <- seq(ylim[1], ylim[2], length=gridsize[2])
    xy <- expand.grid(x, y)
    d <- ncol(Sigmas)

    if (dist=="normal")
    {
        if (deriv.order<=0) dens <- dmvnorm.mixt(xy, mus=mus, Sigmas=Sigmas, props=props)
        else dens <- dmvnorm.deriv.mixt(xy, mus=mus, Sigmas=Sigmas, props=props, deriv.order=deriv.order)
    }
    else if (dist=="t")
    {
        if (deriv.order>0) stop("deriv.order>0 for t mixture not yet implemented")
        dens <- dmvt.mixt(xy, mus=mus, Sigmas=Sigmas, props=props, dfs=dfs)
    }

    if (deriv.order<=0) dens.mat <- matrix(dens, ncol=length(x), byrow=FALSE)
    else
    {
        dens.mat <- list()
        for (i in 1:ncol(dens))
            dens.mat[[i]] <- matrix(dens[,i], ncol=length(x), byrow=FALSE)
    }

    if (dist=="normal")
        x.rand <- rmvnorm.mixt(n=nrand, mus=mus, Sigmas=Sigmas, props=props)
    else if (dist=="t")
        x.rand <- rmvt.mixt(n=nrand, mus=mus, Sigmas=Sigmas, props=props, dfs=dfs)

    H <- Hns(x=x.rand, deriv.order=deriv.order)
    if (binned) H <- diag(diag(H))
    fhat.rand <- kdde(x=x.rand, H=H, deriv.order=deriv.order, binned=binned)
    fhat <- fhat.rand 
    fhat$x <- x.rand
    fhat$eval.points <- list(x,y)
    fhat$estimate <- dens.mat
    fhat$names <- c("x", "y")

    if (deriv.order>0)
    {
        deriv.ind <- dmvnorm.deriv.mixt(xy, mus=mus, Sigmas=Sigmas, props=props, add.index=TRUE, only.index=TRUE, deriv.order=deriv.order, deriv.vec=TRUE)
        fhat$deriv.order <- deriv.order
        fhat$deriv.ind <- deriv.ind
        class(fhat) <- "kdde"
        if (draw) plot(fhat, which.deriv.ind=which.deriv.ind, xlim=xlim, ylim=ylim, ...)
    }
    else
    {
        if (draw) plot(fhat, xlim=xlim, ylim=ylim, display=display, ...)
    }
    invisible(fhat)
}

plotmixt.3d <- function(mus, Sigmas, props, dfs, dist="normal", xlim, ylim, zlim, gridsize, nrand=1e4, draw=TRUE, binned, deriv.order, which.deriv.ind, ...)
{
    d <- 3
    dist <- match.arg(dist, c("normal", "t")) 
    maxsd <- sqrt(apply(Sigmas, 2, max))

    if (is.vector(mus)) mus <- as.matrix(t(mus))
    if (missing(xlim)) xlim <- c(min(mus[,1]) - 4*maxsd[1], max(mus[,1]) + 4*maxsd[1])
    if (missing(ylim)) ylim <- c(min(mus[,2]) - 4*maxsd[2], max(mus[,2]) + 4*maxsd[2])
    if (missing(zlim)) zlim <- c(min(mus[,3]) - 4*maxsd[3], max(mus[,3]) + 4*maxsd[3])
    if (missing(gridsize)) gridsize <- default.gridsize(3)

    x <- seq(xlim[1], xlim[2], length=gridsize[1])
    y <- seq(ylim[1], ylim[2], length=gridsize[2])
    z <- seq(zlim[1], zlim[2], length=gridsize[3])
    xy <- expand.grid(x,y)

    if (deriv.order>0)
    {
        if (dist=="t")
            stop("deriv.order>0 for t mixture not yet implemented")
        else if (dist=="normal")
            deriv.ind <- dmvnorm.deriv.mixt(cbind(xy,z[1]), mus=mus, Sigmas=Sigmas, props=props, deriv.order=deriv.order, add.index=TRUE, only.index=TRUE)
    }
     
    if (deriv.order<=0) dens.array <- array(0, dim=gridsize)
    else { dens.array <- list(); for (i in 1:nrow(deriv.ind)) dens.array <- c(dens.array, list(array(0, dim=gridsize))) }

    for (i in 1:length(z))
    {
        if (dist=="normal")
        {
            if (deriv.order<=0)
                dens <- dmvnorm.mixt(cbind(xy, z[i]), mus=mus, Sigmas=Sigmas, props=props)
            else
                dens <- dmvnorm.deriv.mixt(cbind(xy, z[i]), mus=mus, Sigmas=Sigmas, props=props, deriv.order=deriv.order)
        }
        else if (dist=="t")
            dens <- dmvt.mixt(cbind(xy, z[i]), mus=mus, Sigmas=Sigmas, dfs=dfs, props=props)

        if (deriv.order<=0)
        {
            dens.mat <- matrix(dens, ncol=length(x), byrow=FALSE)
            dens.array[,,i] <- dens.mat
        }
        else
        {
            for (j in 1:ncol(dens))
            {
                dens.mat <- matrix(dens[,j], ncol=length(x), byrow=FALSE)
                dens.array[[j]][,,i] <- dens.mat
            }
        }
    }

    if (dist=="normal")
    x.rand <- rmvnorm.mixt(n=nrand, mus=mus, Sigmas=Sigmas, props=props)
    else if (dist=="t")
    x.rand <- rmvt.mixt(n=nrand, mus=mus, Sigmas=Sigmas, props=props, dfs=dfs)

    H <- Hns(x=x.rand, deriv.order=deriv.order)
    if (binned) H <- diag(diag(H))
    fhat.rand <- kdde(x=x.rand, H=H, deriv.order=deriv.order, binned=binned)

    fhat <- fhat.rand
    fhat$x <- head(x.rand, n=100)
    fhat$eval.points <- list(x,y,z)
    fhat$estimate <- dens.array
    fhat$names <- c("x", "y", "z")
    fhat$H <- H
    fhat$w <- rep(1,nrow(fhat$x))

    if (deriv.order>0)
    {
        deriv.ind <- dmvnorm.deriv.mixt(xy, mus=mus, Sigmas=Sigmas, props=props, add.index=TRUE, only.index=TRUE, deriv.order=deriv.order, deriv.vec=TRUE)
        fhat$deriv.order <- deriv.order
        fhat$deriv.ind <- deriv.ind
        class(fhat) <- "kdde"
        if (draw) plot(fhat, which.deriv.ind=which.deriv.ind, xlim=xlim, ylim=ylim, zlim=zlim, ...)
    }
    else
    {
        if (draw) plot(fhat, xlim=xlim, ylim=ylim, zlim=zlim, ...)
    }
 
    invisible(fhat)
}

############################################################################
## Multivariate t mixture - density values
##
## Parameters
## x - points to compute density at    
## mus - vector of means 
## Sigmas - dispersion matrices
## dfs - degrees of freedom
## props - vector of mixing proportions
##
## Returns
## Value of multivariate t mixture density at x
############################################################################
dmvt.mixt <- function(x, mus, Sigmas, dfs, props)
{
    if (!(identical(all.equal(sum(props), 1), TRUE)))
        stop("Proportions don't sum to one")
    else if (length(dfs) != length(props))
        stop("Length of df and mixing proportions vectors not equal")

    ## single component mixture
    if (identical(all.equal(props[1], 1), TRUE))
        dens <- dmvt(x, delta=mus, sigma=Sigmas, df=dfs, log=FALSE)
    ## multiple component mixture
    else   
    {   
        if (is.vector(mus)) d <- length(mus)
        else d <- ncol(mus)
        k <- length(props)
        dens <- 0      
        for (i in 1:k)
            dens <- dens+props[i]*dmvt(x,delta=mus[i,],sigma=Sigmas[((i-1)*d+1):(i*d),], df=dfs[i], log=FALSE)
    }
  
    return(dens)
}

############################################################################
## Multivariate t mixture - random sample
## 
## Parameters
## n - number of samples
## mus - means 
## Sigmas - matrix of dispersion matrices
## dfs - vector of degrees of freedom
## props - vector of mixing proportions 
## 
## Returns
## Vector of n observations from the t mixture
############################################################################
rmvt.mixt <- function(n=100, mus=c(0,0), Sigmas=diag(2), dfs=7, props=1)
{
    if (!(identical(all.equal(sum(props), 1), TRUE)))  
        stop("Proportions don't sum to one")
    else if (length(dfs) != length(props))
        stop("Length of df and mixing proportions vectors not equal")  

    ## single component mixture
    if (identical(all.equal(props[1], 1), TRUE))
    {
        rand <- rmvt(n=n, sigma=Sigmas, df=dfs)
        for (i in 1:length(mus)) rand[,i] <- rand[,i] + mus[i]
    }
    ## multiple component mixture
    else
    {
        k <- length(props)
        d <- ncol(Sigmas)
        n.samp <- sample(1:k, n, replace=TRUE, prob=props) 
        n.prop <- numeric(0)

        ## compute number to be drawn from each component 
        for (i in 1:k) n.prop <- c(n.prop, sum(n.samp == i))

        ## generate random samples from each component
        rand <- numeric(0)
        for (i in 1:k)
        {
            if (n.prop[i] > 0)
            {  
                rand.temp <- rmvt(n=n.prop[i],sigma=Sigmas[((i-1)*d+1):(i*d),],df=dfs[i])
                for (j in 1:length(mus[k,]))
                    rand.temp[,j] <- rand.temp[,j] + mus[i,j]

                rand <- rbind(rand, rand.temp)
            }
        }
    }
  
    return(rand[sample(n),])
}

############################################################################
## Moments of multivariate normal mixture
############################################################################
mvnorm.mixt.moment <- function (mus, Sigmas, props)
{
    if (!(identical(all.equal(sum(props), 1), TRUE)))
        stop("Proportions don't sum to one")
    d <- ncol(Sigmas)
    k <- length(props)
    mn <- rep(0, d)
    va <- matrix(0, nrow=d, ncol=d)
    for (i in 1:k)
    {
        mn <- mn + props[i] * mus[i,]
        va <- va + props[i] * (Sigmas[((i-1)*d+1):(i*d),] + mus[i,] %*% t(mus[i,]))
    } 
    va <- va + mn %*% t(mn)
    
    return( list(mean=mn, var=va))
}

######################################################################
## Modes for normal mixture
######################################################################
mvnorm.mixt.mode <- function(mus, Sigmas, props=1, verbose=FALSE)
{
    if (!(identical(all.equal(sum(props), 1), TRUE))) 
        stop("Proportions don't sum to one")
    
    if (identical(all.equal(props[1], 1), TRUE)) 
        mm <- mus
    else
    {
        k <- length(props)
        d <- ncol(Sigmas)
        mm <- matrix(0, nrow=k, ncol=d)
        dmvnorm.mixt.temp <- function(x) {
            return(-1*dmvnorm.mixt(x=x, mus=mus, Sigmas=Sigmas, props=props))
        }
        
        for (i in 1:k)
        {
            result <- nlm(p=mus[i,], f=dmvnorm.mixt.temp, print.level=2*as.numeric(verbose))
            mm[i,] <- result$estimate
        }
    }
    return(mm)
}

######################################################################
## Partition for 2-d normal mixture
######################################################################
mvnorm.mixt.part <- function(mus, Sigmas, props=1, xmin, xmax, gridsize, max.iter=100, verbose=FALSE)
{
    maxSigmas <- 4*max(Sigmas)
    if (is.vector(mus)) mus <- as.matrix(t(mus))
    if (missing(xmin)) xmin <- c(min(mus[,1]) - maxSigmas, min(mus[,2]) - maxSigmas)
    if (missing(xmax)) xmax <- c(max(mus[,1]) + maxSigmas, max(mus[,2]) + maxSigmas)
    if (missing(gridsize)) gridsize <- c(201,201)
    x <- seq(xmin[1], xmax[1], length=gridsize[1])
    y <- seq(xmin[2], xmax[2], length=gridsize[2])
    xy <- expand.grid(x, y)
    xy.orig <- xy
    d <- ncol(Sigmas)
    k <- length(props)
    a <- min(c(xmax[1]-xmin[1], xmax[2]-xmin[2]))/1e3

    ## max.iter mean shift iterations
    for (i in 1:max.iter)
    {
        dens <- dmvnorm.mixt(xy, mus=mus, Sigmas=Sigmas, props=props)
        grad <- dmvnorm.deriv.mixt(xy, mus=mus, Sigmas=Sigmas, props=props,deriv.order=1)
        xy[,1] <- xy[,1] + a*grad[,1]/dens
        xy[,2] <- xy[,2] + a*grad[,2]/dens 
    }
    xy.dist <- matrix(0, ncol=k, nrow=nrow(xy))
    for (i in 1:k) xy.dist[,i] <- apply(sweep(xy, 2, mus[i,])^2, 1, sum)
    xy.lab <- array(apply(xy.dist, 1, which.min), dim=gridsize)

    x.rand <- rmvnorm.mixt(n=1e3, mus=mus, Sigmas=Sigmas, props=props)
    fhat <- kde(x=x.rand, binned=TRUE)
    fhat$eval.points <- list(x, y)
    fhat$estimate <- xy.lab
    fhat$names <- c("x", "y") 
    class(fhat) <- "kde.part"
    
    return(fhat)
}

#############################################################################
## Multivariate kernel density derivative estimate 
##############################################################################
kdde <- function(x, H, h, deriv.order=0, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, positive=FALSE, adj.positive, w, deriv.vec=TRUE, verbose=FALSE)
{
    ## default values 
    r <- deriv.order
    ksd <- ks.defaults(x=x, w=w, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    gridsize <- ksd$gridsize
    bgridsize <- ksd$bgridsize

    ## clip data to xmin,xmax grid for binned estimation
    grid.clip <- binned 
    if (grid.clip) 
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        if (positive & missing(xmin)) { xmin <- rep(0,d) }
        xt <- truncate.grid(x=x, y=w, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- xt$y; n <- length(w)
    }
  
    ## default bandwidths
    if (d==1 & missing(h)) 
    {
        if (positive) x1 <- log(x) else x1 <- x
        h <- hpi(x=x1, nstage=2, binned=default.bflag(d=d, n=n), deriv.order=r)
    }
    if (d>1 & missing(H))
    {
        if (positive) x1 <- log(x) else x1 <- x
        if ((r>0) & (d>2)) nstage <- 1 else nstage <- 2
        H <- Hpi(x=x1, nstage=nstage, binned=default.bflag(d=d, n=n), deriv.order=r, verbose=verbose)
    }
  
    ## compute binned estimator
    if (binned)
    {
        if (positive & is.vector(x))
        {
            y <- log(x)
            fhat <- kdde.binned(x=y, H=H, h=h, deriv.order=r, bgridsize=bgridsize, xmin=xmin, xmax=xmax, w=w)
            fhat$estimate <- fhat$estimate/exp(fhat$eval.points)
            fhat$eval.points <- exp(fhat$eval.points)
            fhat$x <- x
        }
        else
            fhat <- kdde.binned(x=x, H=H, h=h, deriv.order=r, bgridsize=bgridsize, xmin=xmin, xmax=xmax, w=w, deriv.vec=deriv.vec, verbose=verbose)

        if (!missing(eval.points))
        {
            fhat$estimate <- predict(fhat, x=eval.points)
            fhat$eval.points <- eval.points
        }
    }
    else
    {
        ## compute exact (non-binned) estimator
        ## 1-dimensional    
        if (d==1)
        {
            if (!missing(H) & !missing(h)) stop("Both H and h are both specified")
          
            if (missing(h)) h <- sqrt(H)

            if (missing(eval.points))
                fhat <- kdde.grid.1d(x=x, h=h, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, deriv.order=r)
            else
                fhat <- kdde.points.1d(x=x, h=h, eval.points=eval.points, w=w, deriv.order=r)
        }
        ## multi-dimensional
        else
        {   
            if (is.data.frame(x)) x <- as.matrix(x)
          
            if (missing(eval.points))
            {
                if (d==2) 
                    fhat <- kdde.grid.2d(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, deriv.order=r, deriv.vec=deriv.vec, verbose=verbose)
                else if (d==3)
                    fhat <- kdde.grid.3d(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, deriv.order=r, deriv.vec=deriv.vec, verbose=verbose) 
                else 
                    stop("Need to specify eval.points for more than 3 dimensions")
            }
            else
                fhat <- kdde.points(x=x, H=H, eval.points=eval.points, w=w, deriv.order=r, deriv.vec=deriv.vec)
        }
    }

    fhat$binned <- binned
    fhat$names <- parse.name(x)
    ## can remove this patch when eks 1.1.0 replace eks 1.0.7 on cran
    if (check.eks.version() <= "1.0.7") ksd$names <- NULL
    if (inherits(fhat$eval.points, "list")) names(fhat$eval.points) <- ksd$names
    fhat$type <- "kdde"
    class(fhat) <- "kdde"
    
    return(fhat)
}

##############################################################################
## Multivariate binned kernel density derivative estimate
##############################################################################
kdde.binned <- function(x, H, h, deriv.order, bgridsize, xmin, xmax, bin.par, w, deriv.vec=TRUE, deriv.index, verbose=FALSE)
{
    r <- deriv.order
    if (length(r)>1) stop("deriv.order should be a non-negative integer")

    ## linear binning
    if (missing(bin.par))
    {
        if (is.vector(x)) { d  <- 1 }
        else { d <- ncol(x) }

        if (d==1)
            if (missing(H)) { H <- as.matrix(h^2) } 
            else { h <- sqrt(H); H <- as.matrix(H) }

        if (d==1) Hd <- H else Hd <- diag(diag(H)) 
        bin.par <- binning(x=x, H=Hd, h=h, bgridsize=bgridsize, xmin=xmin, xmax=xmax, supp=3.7+max(r), w=w)
    }
    else
    {
        if (!is.list(bin.par$eval.points)) { d <- 1; bgridsize <- length(bin.par$eval.points) }
        else  { d <- length(bin.par$eval.points); bgridsize <- sapply(bin.par$eval.points, length) } 

        w <- bin.par$w
        if (d==1)
            if (missing(H)) H <- as.matrix(h^2)
            else { h <- sqrt(H); H <- as.matrix(H) }
    }
     
    if (d==1)
    {
        fhat <- kdde.binned.1d(h=h, deriv.order=r, bin.par=bin.par)
        eval.points <- fhat$eval.points
        est <- fhat$estimate
    }
    else
    {
        ind.mat <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=H, deriv.order=r, only.index=TRUE, deriv.vec=deriv.vec)
        fhat.grid <- kdde.binned.nd(H=H, deriv.order=r, bin.par=bin.par, verbose=verbose, deriv.vec=deriv.vec)
    }
     
    if (missing(x)) x <- NULL
  
    if (d==1)
    {
        if (r==0) fhat <- list(x=x, eval.points=unlist(eval.points), estimate=est, h=h, H=h^2, gridtype="linear", gridded=TRUE, binned=TRUE, names=NULL, w=w)
        else
        fhat <- list(x=x, eval.points=unlist(eval.points), estimate=est, h=h, H=h^2, gridtype="linear", gridded=TRUE, binned=TRUE, names=NULL, w=w, deriv.order=r, deriv.ind=r)
    }
    else
    {
        if (r==0)
            fhat <- list(x=x, eval.points=fhat.grid$eval.points, estimate=fhat.grid$estimate[[1]], H=H, gridtype="linear", gridded=TRUE, binned=TRUE, names=NULL, w=w)
        else
            fhat <- list(x=x, eval.points=fhat.grid$eval.points, estimate=fhat.grid$estimate, H=H, gridtype="linear", gridded=TRUE, binned=TRUE, names=NULL, w=w, deriv.order=r, deriv.ind=ind.mat)
    }
    class(fhat) <- "kdde"
  
    return(fhat)
}

kdde.binned.1d <- function(h, deriv.order, bin.par)
{
    r <- deriv.order
    n <- sum(bin.par$counts)
    a <- min(bin.par$eval.points)
    b <- max(bin.par$eval.points)
    M <- length(bin.par$eval.points)
    delta <- (b-a)/(M-1)
    L <- min(ceiling((4+r)*h/delta), M-1)
    N <- 2*L-1

    grid1 <- seq(-(L-1), L-1)
    keval <- dnorm.deriv(x=delta*grid1, mu=0, sigma=h, deriv.order=r)/n
    est <- symconv.1d(keval, bin.par$counts)

    return(list(eval.points=bin.par$eval.points, estimate=est))
}

kdde.binned.nd <- function(H, deriv.order, bin.par, verbose=FALSE, deriv.vec=TRUE)
{
    d <- ncol(H)
    r <- deriv.order
    n <- sum(bin.par$counts)
    a <- sapply(bin.par$eval.points, min)
    b <- sapply(bin.par$eval.points, max)
    M <- sapply(bin.par$eval.points, length)
    delta <- (b-a)/(M-1)
    lambda <- max(sqrt(abs(svd(H)$d)))
    L <- pmin(ceiling((4+r)*lambda/delta), M-1)
    N <- 2*L-1
    
    if (min(L)<=1) warning(paste("Binning grid may be too coarse for current (small) bandwidth: consider increasing grid size for dimensions", toString(which(pmin(L)<=1))))
    
    if(d==2)
    {
        grid1 <- seq(-(L[1]-1), L[1]-1)
        grid2 <- seq(-(L[2]-1), L[2]-1)
        xgrid <- expand.grid(delta[1]*grid1, delta[2]*grid2)
    }
    else if (d==3)
    {
        grid1 <- seq(-(L[1]-1), L[1]-1)
        grid2 <- seq(-(L[2]-1), L[2]-1)
        grid3 <- seq(-(L[3]-1), L[3]-1)
        xgrid <- expand.grid(delta[1]*grid1, delta[2]*grid2, delta[3]*grid3)
    }
    else if (d==4)
    {
        grid1 <- seq(-(L[1]-1), L[1]-1)
        grid2 <- seq(-(L[2]-1), L[2]-1)
        grid3 <- seq(-(L[3]-1), L[3]-1)
        grid4 <- seq(-(L[4]-1), L[4]-1)
        xgrid <- expand.grid(delta[1]*grid1, delta[2]*grid2, delta[3]*grid3, delta[4]*grid4)
    }
    deriv.index <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=H, deriv.order=r, add.index=TRUE, only.index=TRUE, deriv.vec=TRUE) 
    deriv.index.minimal <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=H, deriv.order=r, add.index=TRUE, only.index=TRUE, deriv.vec=FALSE)

    if (verbose) pb <- txtProgressBar()
     
    if (r==0)
    {
        n.seq <- block.indices(1, nrow(xgrid), d=d, r=r, diff=FALSE)
        est.list <- vector(1, mode="list")
        est.list[[1]] <- array(0, dim=dim(bin.par$counts))
    }
    else if (r>0)
    {      
          n.deriv <- nrow(deriv.index)
          n.deriv.minimal <- nrow(deriv.index.minimal)
          if (deriv.vec) n.est.list <- n.deriv else n.est.list <- n.deriv.minimal
          est.list <- vector(n.est.list, mode="list")
          
          for (j in 1:n.est.list) est.list[[j]] <- array(0, dim=dim(bin.par$counts))
          if (d^r >= 3^7) n.seq <- block.indices(1, nrow(xgrid), d=d, r=r, diff=FALSE, block.limit=1e4)
          else n.seq <- block.indices(1, nrow(xgrid), d=d, r=r, diff=FALSE, block.limit=1e5)
    }

    for (i in 1:(length(n.seq)-1))
    {
        if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
        keval <- dmvnorm.deriv(x=xgrid[n.seq[i]:(n.seq[i+1]-1),], mu=rep(0,d), Sigma=H, deriv.order=r, add.index=TRUE, deriv.vec=FALSE)$deriv/n

        if (r==0) keval <- as.matrix(keval, ncol=1)
        else if (is.vector(keval)) keval <- as.matrix(t(keval), nrow=1)
        est <- list()

        ## loop over only unique partial derivative indices
        nderiv <- nrow(deriv.index.minimal)
        if (!(is.null(nderiv)))
            for (s in 1:nderiv)
            {
                if (deriv.vec) deriv.rep.index <- which.mat(deriv.index.minimal[s,], deriv.index)
                else deriv.rep.index <- s
                kevals <- array(keval[,s], dim=N)
                est.temp <- symconv.nd(kevals, bin.par$counts, d=d)
                for (s2 in 1:length(deriv.rep.index)) est[[deriv.rep.index[s2]]] <- est.temp
            }
        else
        {
            kevals <- lapply(as.data.frame(keval), function(x) { array(x, dim=N) })
            est <- lapply(kevals, function(x) { symconv.nd(x, bin.par$counts, d=d) })
        }
        if (r==0) est.list[[1]] <- est.list[[1]] + est[[1]]
        else if (r>0) for (j in 1:n.est.list) est.list[[j]] <- est.list[[j]] + est[[j]]  
    }
    if (verbose) close(pb)
    fhatr <- list(eval.points=bin.par$eval.points, estimate=est.list, deriv.order=r)

    return(fhatr)
}

#############################################################################
## Univariate kernel density derivative estimate on a grid
#############################################################################
kdde.grid.1d <- function(x, h, gridsize, supp=3.7, positive=FALSE, adj.positive, xmin, xmax, gridtype, w, deriv.order=0)
{
    r <- deriv.order
    if (r==0)
        fhatr <- kde(x=x, h=h, gridsize=gridsize, supp=supp, positive=positive, adj.positive=adj.positive, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w)
    else
    {  
        if (missing(xmin)) xmin <- min(x) - h*supp
        if (missing(xmax)) xmax <- max(x) + h*supp
        if (missing(gridtype)) gridtype <- "linear"

        y <- x
        gridtype1 <- match.arg(gridtype, c("linear", "sqrt"))
        if (gridtype1=="linear")
            gridy <- seq(xmin, xmax, length=gridsize)
        else if (gridtype1=="sqrt")
        {
            gridy.temp <- seq(sign(xmin)*sqrt(abs(xmin)), sign(xmax)*sqrt(abs(xmax)), length=gridsize)
            gridy <- sign(gridy.temp) * gridy.temp^2
        }
        gridtype.vec <- gridtype1

        n <- length(y)
        est <- dnorm.deriv.mixt(x=gridy, mus=y, sigmas=rep(h, n), props=w/n, deriv.order=r)
        fhatr <- list(x=y, eval.points=gridy, estimate=est, h=h, H=h^2, gridtype=gridtype.vec, gridded=TRUE, binned=FALSE, names=NULL, w=w, deriv.order=r, deriv.ind=deriv.order)
        class(fhatr) <- "kdde"
    }

    return(fhatr)
}

##############################################################################
## Bivariate kernel density derivative estimate on a grid
## Computes all mixed partial derivatives for a given deriv.order
##############################################################################
kdde.grid.2d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, deriv.order=0, deriv.vec=TRUE, verbose=FALSE)
{
    d <- 2
    r <- deriv.order
    if (r==0)
    fhatr <- kde(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, verbose=verbose)
    else
    {  
        ## initialise grid 
        n <- nrow(x)
        if (is.null(gridx))
          gridx <- make.grid.ks(x, matrix.sqrt(H), tol=supp, gridsize=gridsize, xmin=xmin, xmax=xmax, gridtype=gridtype) 
        
        suppx <- make.supp(x, matrix.sqrt(H), tol=supp)
        
        if (is.null(grid.pts))
        grid.pts <- find.gridpts(gridx, suppx)    

        nderiv <- d^r
        fhat.grid <- list()
        for (k in 1:nderiv)
            fhat.grid[[k]] <- matrix(0, nrow=length(gridx[[1]]), ncol=length(gridx[[2]]))
        if (verbose) pb <- txtProgressBar()
        for (i in 1:n)
        {
            ## compute evaluation points 
            eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
            eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
            eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
            eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
            eval.x.len <- length(eval.x)
            eval.pts <- expand.grid(eval.x, eval.y)

            ## Create list of matrices for different partial derivatives
            fhat <- dmvnorm.deriv(x=eval.pts, mu=x[i,], Sigma=H, deriv.order=r)

            ## place vector of density estimate values `fhat' onto grid 'fhat.grid'
            for (k in 1:nderiv)
            for (j in 1:length(eval.y))
                fhat.grid[[k]][eval.x.ind, eval.y.ind[j]] <- fhat.grid[[k]][eval.x.ind, eval.y.ind[j]] + w[i]*fhat[((j-1) * eval.x.len + 1):(j * eval.x.len),k]
            if (verbose) setTxtProgressBar(pb, i/n)
        }
        if (verbose) close(pb)
        
        for (k in 1:nderiv) fhat.grid[[k]] <- fhat.grid[[k]]/n
        gridx1 <- list(gridx[[1]], gridx[[2]]) 

        ind.mat <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=H, deriv.order=r, only.index=TRUE)

        if (!deriv.vec)
        {
            fhat.grid.vech <- list()
            deriv.ind <- unique(ind.mat)

            for (i in 1:nrow(deriv.ind))
            {
                which.deriv <- which.mat(deriv.ind[i,], ind.mat)[1]
                fhat.grid.vech[[i]] <- fhat.grid[[which.deriv]]
            }
            ind.mat <- deriv.ind
            fhat.grid <- fhat.grid.vech
        }

        fhatr <- list(x=x, eval.points=gridx1, estimate=fhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE, binned=FALSE, names=NULL, w=w, deriv.order=deriv.order, deriv.ind=ind.mat)
      }

      return(fhatr)
}

##############################################################################
## Trivariate kernel density derivative estimate on a grid
## Computes all mixed partial derivatives for a given deriv.order
##############################################################################
kdde.grid.3d <- function(x, H, gridsize, supp, gridx=NULL, grid.pts=NULL, xmin, xmax, gridtype, w, deriv.order=0, deriv.vec=TRUE, verbose=FALSE)
{
    d <- 3
    r <- deriv.order
    if (r==0)
        fhatr <- kde(x=x, H=H, gridsize=gridsize, supp=supp, xmin=xmin, xmax=xmax, gridtype=gridtype, w=w, verbose=verbose)
    else
    {
        ## initialise grid 
        n <- nrow(x)
        if (is.null(gridx))
            gridx <- make.grid.ks(x, matrix.sqrt(H), tol=supp, gridsize=gridsize, xmin=xmin, xmax=xmax, gridtype=gridtype) 

        suppx <- make.supp(x, matrix.sqrt(H), tol=supp)

        if (is.null(grid.pts))
        grid.pts <- find.gridpts(gridx, suppx)    

        nderiv <- d^r
        fhat.grid <- list()
        for (k in 1:nderiv)
            fhat.grid[[k]] <- array(0, dim=gridsize)
        if (verbose) pb <- txtProgressBar()
        for (i in 1:n)
        {
            ## compute evaluation points 
            eval.x <- gridx[[1]][grid.pts$xmin[i,1]:grid.pts$xmax[i,1]]
            eval.y <- gridx[[2]][grid.pts$xmin[i,2]:grid.pts$xmax[i,2]]
            eval.z <- gridx[[3]][grid.pts$xmin[i,3]:grid.pts$xmax[i,3]]
            eval.x.ind <- c(grid.pts$xmin[i,1]:grid.pts$xmax[i,1])
            eval.y.ind <- c(grid.pts$xmin[i,2]:grid.pts$xmax[i,2])
            eval.z.ind <- c(grid.pts$xmin[i,3]:grid.pts$xmax[i,3])
            eval.x.len <- length(eval.x)
            eval.pts <- expand.grid(eval.x, eval.y)

            ## create list of matrices for different partial derivatives
            ## place vector of density estimate values `fhat' onto grid 'fhat.grid'
            for (ell in 1:nderiv)
                for (k in 1:length(eval.z))
                {
                    fhat <- w[i]*dmvnorm.deriv(cbind(eval.pts, eval.z[k]), x[i,], H, deriv.order=r)
                    for (j in 1:length(eval.y))
                        fhat.grid[[ell]][eval.x.ind,eval.y.ind[j], eval.z.ind[k]] <- 
                            fhat.grid[[ell]][eval.x.ind, eval.y.ind[j], eval.z.ind[k]] + 
                            fhat[((j-1) * eval.x.len + 1):(j * eval.x.len), ell]
                }
          
            if (verbose) setTxtProgressBar(pb, i/n)
        }
        if (verbose) close(pb)

        for (k in 1:nderiv) fhat.grid[[k]] <- fhat.grid[[k]]/n
        gridx1 <- list(gridx[[1]], gridx[[2]], gridx[[3]])

        ind.mat <- dmvnorm.deriv(x=rep(0,d), mu=rep(0,d), Sigma=H, deriv.order=r, only.index=TRUE)

        if (!deriv.vec)
        {
            fhat.grid.vech <- list()
            deriv.ind <- unique(ind.mat)

            for (i in 1:nrow(deriv.ind))
            {
                which.deriv <- which.mat(deriv.ind[i,], ind.mat)[1]
                fhat.grid.vech[[i]] <- fhat.grid[[which.deriv]]
            }
            ind.mat <- deriv.ind
            fhat.grid <- fhat.grid.vech
        }

        fhatr <- list(x=x, eval.points=gridx1, estimate=fhat.grid, H=H, gridtype=gridx$gridtype, gridded=TRUE, binned=FALSE, names=NULL, w=w, deriv.order=deriv.order, deriv.ind=ind.mat)
    }

    return(fhatr)
}

#############################################################################
## Multivariate kernel density estimate using normal kernels,
## evaluated at each sample point
#############################################################################
kdde.points.1d <- function(x, h, eval.points, w, deriv.order=0) 
{
    r <- deriv.order
    n <- length(x)
    fhat <- dnorm.deriv.mixt(x=eval.points, mus=x, sigmas=rep(h,n), props=w/n, deriv.order=r)

    return(list(x=x, eval.points=eval.points, estimate=fhat, h=h, H=h^2, gridded=FALSE, binned=FALSE, names=NULL, w=w, deriv.order=r, deriv.ind=r))
}

kdde.points <- function(x, H, eval.points, w, deriv.order=0, deriv.vec=TRUE) 
{
    n <- nrow(x)
    Hs <- replicate(n, H, simplify=FALSE) 
    Hs <- do.call(rbind, Hs)
    r <- deriv.order
    fhat <- dmvnorm.deriv.mixt(x=eval.points, mus=x, Sigmas=Hs, props=w/n, deriv.order=r, deriv.vec=deriv.vec, add.index=TRUE)

    return(list(x=x, eval.points=eval.points, estimate=fhat$deriv, H=H, gridded=FALSE, binned=FALSE, names=NULL, w=w, deriv.order=r, deriv.ind=fhat$deriv.ind))
}

#############################################################################
## S3 methods for KDDE objects
#############################################################################
## plot method
plot.kdde <- function(x, ...)
{
    fhat <- x

    if (is.null(fhat$deriv.order))
    {
        class(fhat) <- "kde"
        plot(fhat, ...)
    }
    else
    {  
        if (is.vector(fhat$x))
        {
            plotkdde.1d(fhat, ...)
            invisible()
        }
        else
        {
            d <- ncol(fhat$x)
            if (d==2) 
            {
                opr <- options()$preferRaster; if (!is.null(opr)) if (!opr) options("preferRaster"=TRUE)
                plotret <- plotkdde.2d(fhat, ...)
                if (!is.null(opr)) options("preferRaster"=opr)
                invisible(plotret)
            }
            else if (d==3)
            {
                plotkdde.3d(fhat, ...)
                invisible()
            }
            else 
                stop ("Plot function only available for 1, 2 or 3-d data")
        }
    }
}

plotkdde.1d <- function(fhat, xlab, ylab="Density derivative function", cont=50, abs.cont, ...)
{
    if (missing(xlab)) xlab <- fhat$names[1]
    
    if (missing(abs.cont))
    {
        abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont), ncol = length(cont))
        abs.cont <- c(abs.cont[1, ], rev(abs.cont[2, ]))
    }
    class(fhat) <- "kde"

    plot(fhat, xlab=xlab, ylab=ylab, abs.cont=abs.cont, ...)
}

plotkdde.2d <- function(fhat, which.deriv.ind=1, cont=c(25,50,75), abs.cont, display="slice", xlab, ylab, zlab="Density derivative function", col, col.fun, transp.neutral, alpha=1, kdde.flag=TRUE, thin=3, transf=1, neg.grad=FALSE, ...)
{
    disp <- match.arg(display, c("slice", "persp", "image", "filled.contour", "filled.contour2", "quiver"))
    if (disp=="filled.contour2") disp <- "filled.contour"
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n, palette="Blue-Red", alpha=alpha) }
    if (missing(xlab)) xlab <- fhat$names[1]
    if (missing(ylab)) ylab <- fhat$names[2]
    if (missing(transp.neutral)) transp.neutral <- class(fhat) %in% "kdde"

    if (disp=="slice" | disp=="filled.contour")
    {
        if (missing(abs.cont))
        {
            abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont, which.deriv.ind=which.deriv.ind), ncol=length(cont))
            abs.cont <- c(unlist(abs.cont[1,]), rev(unlist(abs.cont[2,])))
        }
        if (missing(col))
        {
            if (is.matrix(abs.cont)) abs.cont <- c(abs.cont[1,], rev(abs.cont[2,]))
            col <- col.diverging(f=col.fun, levels=abs.cont, display=disp, transp.neutral=transp.neutral)
        }
    } 

    if (disp=="quiver")
    {
        if (fhat$deriv.order==1)
            plotquiver(fhat=fhat, thin=thin, transf=transf, neg.grad=neg.grad, col=col, xlab=xlab, ylab=ylab, alpha=alpha, ...)
        else warning("Quiver plot requires gradient estimate.")
    }
    else
    {
        fhat.temp <- fhat
        fhat.temp$deriv.ind <- fhat.temp$deriv.ind[which.deriv.ind,]
        fhat.temp$estimate <- fhat.temp$estimate[[which.deriv.ind]]
        fhat <- fhat.temp
        class(fhat) <- "kde"
      
        if (disp=="persp") plot(fhat, display=display, abs.cont=abs.cont, xlab=xlab, ylab=ylab, zlab=zlab, col.fun=col.fun, kdde.flag=kdde.flag, col=col, thin=thin, alpha=alpha, ...) 
        else plot(fhat, display=display, abs.cont=abs.cont, xlab=xlab, ylab=ylab, zlab=zlab, col.fun=col.fun, kdde.flag=kdde.flag, col=col, alpha=alpha, ...) 
    }
}

plotkdde.3d <- function(fhat, which.deriv.ind=1, display="plot3D", cont=c(25,50,75), abs.cont, colors, col, col.fun, xlab, ylab, zlab, ...)
{
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n=n, palette="Blue-Red") }
    if (missing(xlab)) xlab <- fhat$names[1]
    if (missing(ylab)) ylab <- fhat$names[2]
    if (missing(zlab)) ylab <- fhat$names[3]

    if (missing(abs.cont))
    {
        abs.cont <- as.matrix(contourLevels(fhat, approx=TRUE, cont=cont, which.deriv.ind=which.deriv.ind), ncol=length(cont))
        abs.cont <- c(abs.cont[1,], rev(abs.cont[2,]))
    }
    fhat.temp <- fhat 
    fhat.temp$deriv.ind <- fhat.temp$deriv.ind[which.deriv.ind,]
    fhat.temp$estimate <- fhat.temp$estimate[[which.deriv.ind]]
    fhat <- fhat.temp
    class(fhat) <- "kde"

    if (missing(col))
    {
        col <- col.fun(length(abs.cont)+1)
        nc <- length(col)
        col <- col[-(max(which(abs.cont<0))+1)]
    }

    colors <- col
    plot(fhat, display=display, abs.cont=abs.cont, colors=col, col=col, xlab=xlab, ylab=ylab, zlab=zlab, ...) 
}

## diverging hcl.colors with transparent mid point (primarily for display="filled.contour")
col.diverging <- function(levels, f, alpha=1, transp.neutral=FALSE, ind.neutral, insert.neutral=TRUE, display="filled.contour")
{
    disp <- match.arg(display, c("filled.contour", "slice"))
    if (missing(f)) f <- function(.) { hcl.colors(n=., palette="Blue-Red", alpha=alpha) } 
    if (is.matrix(levels)) levels <- c(levels[1,], rev(levels[2,]))

    levels.rle <- rle(sign(levels))
    levels.len <- max(levels.rle$lengths)
           
    ## for unblanced number of negative + positive contour levels
    col.neg <- head(f(2*length(which(levels<=0))), n=length(which(levels<=0)))
    col.pos <- tail(f(2*length(which(levels>0))), n=length(which(levels>0))) 
     
    if (disp=="slice") cols <- c(col.neg, col.pos) 
    else if (disp=="filled.contour") 
    {
        cols <- c(col.neg, col.pos)
        n <- length(cols)
        if (missing(ind.neutral)) ind.neutral <- length(col.neg)+1
        #if (ind.neutral<1 | ind.neutral>(n-1)) ind.neutral <- 1
        if (ind.neutral<1) ind.neutral <- 1
        else if (ind.neutral>n) ind.neutral <- n
        if (insert.neutral) cols <- cols[c(1:ind.neutral, ind.neutral:n)] 
        if (transp.neutral) cols[ind.neutral] <- "transparent"
        else cols[ind.neutral] <- f(3)[2]
    }

    return(cols)
}

######################################################################
## Quiver plot
######################################################################
plotquiver <- function(fhat, thin=5, transf=1, neg.grad=FALSE, xlab, ylab, col, add=FALSE, scale, length=0.1, alpha=1, ...)
{
    if (!requireNamespace("pracma", quietly=TRUE)) stop("Install the pracma package as it is required.", call.=FALSE)
    if (missing(col)) col <- 1
    col <- transparency.col(col, alpha=alpha)
    
    ev <- fhat$eval.points
    est <- fhat$estimate

    if (transf!=0)
    {
        est[[1]] <- sign(est[[1]])*abs(est[[1]])^(transf)
        est[[2]] <- sign(est[[2]])*abs(est[[2]])^(transf)
    }
    
    thin1.ind <- seq(1, length(ev[[1]]), by=thin)
    thin2.ind <- seq(1, length(ev[[2]]), by=thin)
    evx <- ev[[1]][thin1.ind]
    evy <- ev[[2]][thin2.ind]
    fx <- est[[1]][thin1.ind, thin2.ind]
    fy <- est[[2]][thin1.ind, thin2.ind]
    if (neg.grad) { fx <- -fx; fy <- -fy }
    if (missing(xlab)) xlab <- fhat$names[1]
    if (missing(ylab)) ylab <- fhat$names[2]
   
    if (!add) plot(fhat, abs.cont=c(0,0), col="transparent", xlab=xlab, ylab=ylab)
    grid.xy <- pracma::meshgrid(evy, evx)
    x0 <- grid.xy$Y
    y0 <- grid.xy$X
    
    ## default scale factor - arrows should exceed a bin
    arrow.len <- sqrt(fx^2 + fy^2) 
    bin.diag <- min(diff(evx), diff(evy)) 
    if (missing(scale)) scale <- bin.diag/max(arrow.len)
    
    ## remove `zero-length' arrows i.e. length < 1e-3 inches
    ## adapted from https://stackoverflow.com/questions/52689959/how-small-is-a-zero-length-arrow/52690054
    units <- par(c('usr', 'pin'))
    x2in <- with(units, pin[1L]/diff(usr[1:2]))
    y2in <- with(units, pin[2L]/diff(usr[3:4]))
    arrow2in <- sqrt((x2in*scale*fx)^2 + (y2in*scale*fy)^2)
    
    arrowint <- sort(c(min(arrow2in)-0.1*abs(max(arrow2in)), seq(1e-3, length, length=5), max(arrow2in)+0.1*abs(max(arrow2in))))
    nsmall.f <- cut(arrow2in, arrowint, labels=FALSE)
    nsf.lab <- unique(nsmall.f[nsmall.f>1])
    
    for (i in nsf.lab)
    {
        nsf <- which(nsmall.f==i)
        if (length(nsf)>0) pracma::quiver(x=x0[nsf], y=y0[nsf], u=fx[nsf], v=fy[nsf], col=col, scale=scale, length=arrowint[i], ...)
    }
}

######################################################################
## Summary kernel curvature 
######################################################################
kcurv <- function(fhat, compute.cont=TRUE)
{
    fhat.curv <- fhat
    if (is.vector(fhat$H)) d <- 1 else d <- ncol(fhat$H)
    if (fhat$deriv.order!=2) stop("Requires output from kdde(, deriv.order=2).")

    if (d==1)
    {
        Hessian.det <- fhat$estimate 
        local.mode <- fhat$estimate <0
        fhat.curv$estimate <- local.mode*abs(Hessian.det)
    }
    else if (d>1)
    {
        fhat.est <- sapply(fhat$estimate, as.vector)
        Hessian.det <- sapply(seq(1,nrow(fhat.est)), function(i) { det(invvec(fhat.est[i,])) })
        Hessian.eigen <- lapply(lapply(seq(1,nrow(fhat.est)), function(i) { invvec(fhat.est[i,]) }), eigen, only.values=TRUE)
        Hessian.eigen <- t(sapply(Hessian.eigen, getElement, "values"))
        local.mode <- apply(Hessian.eigen <= 0, 1, all)
        fhat.curv$estimate <- local.mode*array(abs(Hessian.det), dim=dim(fhat$estimate[[1]]))
    }
    
    fhat.curv$deriv.order <- NULL
    fhat.curv$deriv.ind <- NULL
    if (compute.cont)
    {
        fhat.temp <- fhat.curv
        fhat.temp$x <- fhat.curv$x[predict(fhat.curv, x=fhat.curv$x)>0,]
        fhat.temp$estimate <- fhat.temp$estimate
        fhat.curv$cont <- contourLevels(fhat.temp, cont=1:99)
    }
    class(fhat.curv) <- "kde"
    fhat.curv$type <- "kcurv"
    
    return(fhat.curv)
}

## contourLevels method 
contourLevels.kdde <- function(x, prob, cont, approx=TRUE, which.deriv.ind=1, ...)
{ 
    nlevels <- 5
    approx <- as.integer(approx)
    fhat <- x
    if (is.vector(fhat$x))
    {
        d <- 1; n <- length(fhat$x)
        if (!is.null(fhat$deriv.order))  
        {
          fhat.temp <- fhat 
          fhat.temp$deriv.ind <-fhat.temp$deriv.ind[which.deriv.ind]
          fhat <- fhat.temp
        }    
    }
    else
    {
        d <- ncol(fhat$x); n <-nrow(fhat$x)
        if (!is.matrix(fhat$x)) fhat$x <- as.matrix(fhat$x)

        if (!is.null(fhat$deriv.order))  
        {
            fhat.temp <- fhat 
            fhat.temp$estimate <- fhat.temp$estimate[[which.deriv.ind]]
            fhat.temp$deriv.ind <-fhat.temp$deriv.ind[which.deriv.ind,]
            fhat <- fhat.temp
        }
    } 

    if (is.null(x$w)) w <- rep(1, n)
    else w <- x$w

    if (is.null(fhat$gridded))
    {
        if (d==1) fhat$gridded <- fhat$binned
        else fhat$gridded <- is.list(fhat$eval.points)
    }

    if (missing(prob) & missing(cont))
    {    
        hts <- pretty(fhat$estimate, n=nlevels)
        args <- list(...)
        if (any(pmatch(names(args), "nlevels"))) warning("nlevels=5 always so changing nlevels has no effect") 
    } 
    else
    {
        if (approx & fhat$gridded)
            dobs <- predict.kde(fhat, x=fhat$x)
        else
            dobs <- kdde(x=fhat$x, H=fhat$H, eval.points=fhat$x, w=w, deriv.order=fhat$deriv.order)$estimate[,which.deriv.ind] 
            
        if (is.null(fhat$deriv.order))
        {
            if (!missing(prob) & missing(cont)) hts <- quantile(dobs[dobs>=0], prob=prob)
            if (missing(prob) & !missing(cont)) hts <- quantile(dobs[dobs>=0], prob=(100-cont)/100)
        }
        else
        {
            if (!missing(prob) & missing(cont)) hts <- rbind(-quantile(abs(dobs[dobs<0]), prob=prob), quantile(dobs[dobs>=0], prob=prob))
            if (missing(prob) & !missing(cont)) hts <- rbind(-quantile(abs(dobs[dobs<0]), prob=(100-cont)/100), quantile(dobs[dobs>=0], prob=(100-cont)/100))
        }
    }
  
    return(hts)
}

## predict method for KDDE 
predict.kdde <- function(object, ..., x)
{
    fhat <- object
    #if (is.vector(fhat$H)) d <- 1 else d <- ncol(fhat$H)
    if (!is.list(fhat$eval.points)) d <- 1 else d <- length(fhat$eval.points)
    if (d==1) n <- length(x)
    else
    {
        if (is.vector(x)) x <- matrix(x, nrow=1)
        else x <- as.matrix(x)
        n <- nrow(x)
    }

    if (!is.null(fhat$deriv.ind))
    {
        if (is.vector(fhat$deriv.ind)) pk.mat <- predict.kde(fhat, x=x, ...)
        else
        {
            nd <- nrow(fhat$deriv.ind)
            pk.mat <- matrix(0, ncol=nd, nrow=n)
            for (i in 1:nd)
            {
                fhat.temp <- fhat
                fhat.temp$estimate <- fhat$estimate[[i]]
                pk.mat[,i] <- predict.kde(fhat.temp, x=x, ...)
            }
        }
    }
    else
        pk.mat <- predict.kde(fhat, x=x, ...)
    
    return(drop(pk.mat))
}

##############################################################################
## Convert estimation grid to quasi-kernel density derivative estimator
##############################################################################
kqdde <- function(estimate, eval.points, xnames)
{
    if (missing(eval.points))
    {
        if (is.vector(estimate)) { d <- 1; eval.points <- 1:length(estimate) }
        else { d <- length(dim(estimate)); eval.points <- lapply(dim(estimate), function(.) 1:.) }
    } 
    else 
    {
        if (!is.list(eval.points)) d <- 1 else d <- length(eval.points)
    }
   
    if (d==1) { x <- 0; deriv.ind <- 1 }
    else { x <- matrix(0, ncol=d, nrow=1); deriv.ind <- matrix(1, nrow=1) }
    if (missing(xnames))
    {
        if (!all(is.na(names(eval.points)))) colnames(x) <- names(eval.points) 
    }
    if (!missing(xnames)) colnames(x) <- xnames
    
    fhat <- list(x=x, estimate=list(estimate, estimate), eval.points=eval.points, deriv.order=1, deriv.ind=deriv.ind)
    fhat$gridded <- TRUE
    fhat$cont <- contourLevels.kqdde(fhat, cont=1:99)
    fhat$names <- parse.name(x)
    fhat$type <- "kdde" 
    class(fhat) <- "kqdde"
    
    return(fhat)
}

plot.kqdde <- plot.kdde 
predict.kqdde <- predict.kdde

#############################################################################
## Kernel discriminant analysis
#############################################################################

#############################################################################
## KDEs of individual densities for KDA - 1- to 3-dim
##
## Parameters
## x - data values
## group - group variable
## Hs - bandwidth matrices
##
## Returns
## List with components (class dade)
## x - list of data values
## eval.points - evaluation points of dnesity estimate
## estimate - list of density estimate
## H - list of bandwidth matrices
#############################################################################
kda <- function(x, x.group, Hs, hs, prior.prob=NULL, gridsize, xmin, xmax, supp=3.7, eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, kde.flag=TRUE)
{
    if (missing(eval.points)) eval.points <- x
    if (!is.factor(x.group)) x.group <- factor(x.group) 
    gr <- levels(x.group)
    m <- length(gr)
    
    ## default values 
    ksd <- ks.defaults(x=x, w=w, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    bgridsize <- ksd$bgridsize
    gridsize <- ksd$gridsize  

    if (!is.null(prior.prob))
        if (!(identical(all.equal(sum(prior.prob), 1), TRUE))) stop("Sum of weights not equal to 1")
    
    ## clip data to xmin,xmax grid for binned estimation
    grid.clip <- binned 
    if (grid.clip)
    {
        if (!missing(xmax)) xmax <- xmax[1:d]
        if (!missing(xmin)) xmin <- xmin[1:d]
        xt <- truncate.grid(x=x, y=1:n, xmin=xmin, xmax=xmax)
        x <- xt$x; w <- w[xt$y]; x.group <- x.group[xt$y]; n <- length(w)
    }

    if (d==1)
    {
        if (missing(hs)) hs <- hkda(x=x, x.group=x.group, bw="plugin", nstage=2, binned=default.bflag(d=d,n=n))
        
        ## compute KDA on grid
        if (kde.flag)
            fhat.list <- kda.1d(x=x, x.group=x.group, hs=hs, prior.prob=prior.prob, gridsize=gridsize, supp=supp, binned=binned, bgridsize=bgridsize, xmin=xmin, xmax=xmax, compute.cont=compute.cont, approx.cont=approx.cont, w=w)
        
        ## compute KDA at eval.points
        fhat <- kda.1d(x=x, x.group=x.group, hs=hs, prior.prob=prior.prob, gridsize=gridsize, supp=supp, binned=binned, bgridsize=bgridsize, xmin=xmin, xmax=xmax, eval.points=eval.points, compute.cont=compute.cont, approx.cont=approx.cont, w=w)
        fhat.wt <- matrix(0, ncol=m, nrow=length(eval.points)) 
    }
    else
    {
        if (missing(Hs)) Hs <- Hkda(x=x, x.group=x.group, bw="plugin", binned=default.bflag(d=d, n=n))
        
        ## Compute KDA on grid
        if (d>3) kde.flag <- FALSE
        if (kde.flag)
            fhat.list <- kda.nd(x=x, x.group=x.group, Hs=Hs, prior.prob=prior.prob, gridsize=gridsize, supp=supp, binned=binned, bgridsize=bgridsize, xmin=xmin, xmax=xmax, compute.cont=compute.cont, approx.cont=approx.cont)
        
        ## compute KDA at eval.points
        fhat <- kda.nd(x=x, x.group=x.group, Hs=Hs, prior.prob=prior.prob, gridsize=gridsize, supp=supp, binned=FALSE, bgridsize=bgridsize, xmin=xmin, xmax=xmax, eval.points=eval.points, compute.cont=compute.cont, approx.cont=approx.cont)
        fhat.wt <- matrix(0, ncol=m, nrow=nrow(eval.points))  
    }

    for (j in 1:m)
        fhat.wt[,j] <- fhat$estimate[[j]]* fhat$prior.prob[j]
    
    ## assign y according largest weighted density value 
    disc.gr.temp <- apply(fhat.wt, 1, which.max)
    disc.gr <- as.factor(gr[disc.gr.temp])
    if (is.numeric(gr)) disc.gr <- as.numeric(levels(disc.gr))[disc.gr]
    
    if (kde.flag) fhat.list$x.group.estimate <- disc.gr
    else fhat.list <- disc.gr
    fhat.list$type <- "kda"
    
    return(fhat.list)
}

kda.1d <- function(x, x.group, hs, prior.prob, gridsize, supp, eval.points, binned, bgridsize, xmin, xmax, w, compute.cont, approx.cont)
{
    gr <- levels(x.group)
    m <- length(gr)
    d <- 1
    hmax <- max(hs)
    if (missing(xmin)) xmin <- min(x) - supp*hmax
    if (missing(xmax)) xmax <- max(x) + supp*hmax
    fhat.list <- list()
    
    for (j in 1:m)
    {
        xx <- x[x.group==gr[j]]
        ww <- w[x.group==gr[j]]
        h <- hs[j]
        
        ## compute individual density estimate
        if (missing(eval.points))
            fhat.temp <- kde(x=xx, h=h, supp=supp, xmin=xmin, xmax=xmax, bgridsize=bgridsize, gridsize=gridsize, w=ww, binned=binned)
        else
            fhat.temp <- kde(x=xx, h=h, w=ww, binned=binned, bgridsize=bgridsize, gridsize=gridsize, eval.points=eval.points)
    
        fhat.list$estimate <- c(fhat.list$estimate, list(fhat.temp$estimate))
        fhat.list$eval.points <- fhat.temp$eval.points
        fhat.list$x <- c(fhat.list$x, list(xx))
        fhat.list$h <- c(fhat.list$h, h)
        fhat.list$H <- c(fhat.list$H, h^2)
        fhat.list$w <- c(fhat.list$w, list(ww))
        
        ## compute prob contour levels
        if (compute.cont & missing(eval.points))
        {
            contlev <- contourLevels(fhat.temp, cont=1:99, approx.cont=approx.cont)
            fhat.list$cont <- c(fhat.list$cont, list(contlev))
        }
    }
    fhat.list$names <- parse.name(x)  ## add variable names
    fhat.list$binned <- binned
    fhat.list$gridded <- fhat.temp$gridded
    
    if (is.null(prior.prob))
    {
        pr <- rep(0, length(gr))
        for (j in 1:length(gr)) pr[j] <- length(which(x.group==gr[j]))
        pr <- pr/length(x)
        fhat.list$prior.prob <- pr
    }
    else
        fhat.list$prior.prob <- prior.prob
    
    fhat.list$x.group <- x.group
    class(fhat.list) <- "kda"
    
    return(fhat.list) 
}

kda.nd <- function(x, x.group, Hs, prior.prob, gridsize, supp, eval.points, binned, bgridsize, xmin, xmax, w, compute.cont, approx.cont)
{
    if (is.data.frame(x)) x <- as.matrix(x)
    gr <- levels(x.group)
    m <- length(gr)
    d <- ncol(x)
  
    ## find largest bandwidth matrix to initialise grid
    detH <- vector() 
    for (j in 1:m)
        detH[j] <- det(Hs[((j-1)*d+1) : (j*d),])  
    Hmax.ind <- which.max(detH)
    Hmax <- Hs[((Hmax.ind-1)*d+1) : (Hmax.ind*d),]
    if (missing(xmin)) xmin <- apply(x, 2, min) - supp*max(sqrt(diag(Hmax)))
    if (missing(xmax)) xmax <- apply(x, 2, max) + supp*max(sqrt(diag(Hmax)))
    if (missing(w)) w <- rep(1, nrow(x))

    if (binned & d > 4) stop("Binning only available for 1- to 4-d data")
    if (missing(bgridsize)) bgridsize <- default.gridsize(d)
    if (missing(gridsize)) gridsize <- default.gridsize(d)

    fhat.list <- list()
    for (j in 1:m)
    {
        xx <- x[x.group==gr[j],]
        ww <- w[x.group==gr[j]]   
        H <- Hs[((j-1)*d+1) : (j*d),]     
        
        ## compute individual density estimate
        if (binned)
            fhat.temp <- kdde(x=xx, bgridsize=bgridsize, H=H, xmin=xmin, xmax=xmax, w=ww, deriv.order=0, binned=TRUE)
        else if (missing(eval.points))
            fhat.temp <- kde(x=xx, H=H, supp=supp, xmin=xmin, xmax=xmax, gridsize=gridsize, w=ww)
        else
            fhat.temp <- kde(x=xx, H=H, eval.points=eval.points, w=ww)
        
        fhat.list$estimate <- c(fhat.list$estimate, list(fhat.temp$estimate))
        fhat.list$eval.points <- fhat.temp$eval.points
        fhat.list$x <- c(fhat.list$x, list(xx))
        fhat.list$H <- c(fhat.list$H, list(H))
        fhat.list$w <- c(fhat.list$w, list(ww))
        
        ## compute prob contour levels
        if (compute.cont & missing(eval.points))
        {
            contlev <- contourLevels(fhat.temp, cont=1:99, approx.cont=approx.cont)
            fhat.list$cont <- c(fhat.list$cont, list(contlev))
        }
    }

    fhat.list$names <- parse.name(x)  ## add variable names
    fhat.list$binned <- binned
    fhat.list$gridded <- fhat.temp$gridded

    if (is.null(prior.prob))
    {
        pr <- rep(0, length(gr))
        for (j in 1:length(gr)) pr[j] <- length(which(x.group==gr[j]))
        pr <- pr/nrow(x)
        fhat.list$prior.prob <- pr
    }
    else
        fhat.list$prior.prob <- prior.prob

    fhat.list$x.group <- x.group
    class(fhat.list) <- "kda"

    return (fhat.list)
}

  
##############################################################################
## S3 methods for kda objects
##############################################################################
## contourLevels method
contourLevels.kda <- function(x, prob, cont, approx=TRUE,...) 
{
    nlevels <- 5
    approx <- as.integer(approx)
    fhat <- x
    m <- length(fhat$x)
    hts <- list()

    for (j in 1:m)
    {
        fhatj <- list(x=fhat$x[[j]], eval.points=fhat$eval.points, estimate=fhat$estimate[[j]], H=fhat$H[[j]], binned=fhat$binned, gridded=fhat$gridded)
        class(fhatj) <- "kde"
        hts[[j]] <- contourLevels(x=fhatj, prob=prob, cont=cont, approx=approx, ...)
    }
   
    return(hts) 
}

## predict method
predict.kda <- function(object, ..., x)
{
    fhat <- object
    m <- length(fhat$prior.prob)
    if (is.vector(fhat$x[[1]])) n <- length(x) else { if (is.vector(x)) n <- 1 else n <- nrow(x) }
    fhat.temp <- matrix(0, ncol=m, nrow=n)
    for (j in 1:m)    
        fhat.temp[,j] <- fhat$prior.prob[j]*grid.interp(x=x, gridx=fhat$eval.points, f=fhat$estimate[[j]])
    
    est.group <- apply(fhat.temp, 1, which.max)
    est.group <- as.factor(sort(unique(fhat$x.group))[est.group])
    
    return(est.group)
}

## plot method
## fhat - output from `kda.kde'
## y - data points (separate from training data inside fhat)
## y.group - data group labels
plot.kda <- function(x, y, y.group, ...) 
{
    if (is.vector(x$x[[1]]))
        plotkda.1d(x=x, y=y, y.group=y.group,  ...)
    else
    {  
        d <- ncol(x$x[[1]])      
        if (d==2)
        {
            opr <- options()$preferRaster; if (!is.null(opr)) if (!opr) options("preferRaster"=TRUE)
            plotkda.2d(x=x, y=y, y.group=y.group, ...)
            if (!is.null(opr)) options("preferRaster"=opr) 
        }
        else if (d==3)  
            plotkda.3d(x=x, y=y, y.group=y.group, ...)
    }
}

plotkda.1d <- function(x, y, y.group, prior.prob=NULL, xlim, ylim, xlab, ylab="Weighted density function", drawpoints=FALSE, col, col.fun, col.part, col.pt, lty, jitter=TRUE, rugsize, add=FALSE, alpha=1, ...)
{ 
    fhat <- x
    m <- length(fhat$x)
    if (missing(xlab)) xlab <- fhat$names
    if (is.null(prior.prob))
    prior.prob <- fhat$prior.prob

    if (m != length(prior.prob))
        stop("prior.prob not same length as number of components in fhat")
    if (!(identical(all.equal(sum(prior.prob), 1), TRUE)))  
        stop("Sum of weights not equal to 1")

    weighted.fhat <- matrix(0, nrow=length(fhat$eval.points), ncol=m) 
    for (j in 1:m)  weighted.fhat[,j] <- fhat$estimate[[j]]*fhat$prior.prob[j]

    if (missing(xlim)) xlim <- range(fhat$eval.points)
    if (missing(ylim)) ylim <- range(weighted.fhat)
    if (missing(lty)) lty <- rep(1, m)
    if (length(lty) < m) lty <- rep(lty, m)
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n, palette="Dark2") }
    if (missing(col)) col <- col.fun(m)
    if (length(col) < m) col <- rep(col, m)
    if (missing(col.part)) col.part <- plot3D.alpha.col(col, alpha=alpha) 
    if (missing(col.pt)) col.pt <- col.fun(m)
    if (length(col.pt)==1) col.pt <- rep(col.pt, m)
  
    ## plot each training group's KDE in separate colour and line type 
    if (!add) plot(fhat$eval.points, weighted.fhat[,1], type="l", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, lty=lty[1], col=col[1], ...)
  
    if (m > 1)
        for (j in 2:m) lines(fhat$eval.points, weighted.fhat[,j], lty=lty[j], col=col[j], ...)

    ydata <- seq(min(fhat$eval.points), max(fhat$eval.points), length=401)
    x.gr <- 1:m
    ydata.gr <- x.gr[apply(weighted.fhat,1, which.max)] 

    ## draw partition class as rug-like plot
    plot.lim <- par()$usr
    if (missing(rugsize)) rugsize <- abs(plot.lim[4]-plot.lim[3])/50

    image(ydata, c(plot.lim[3], plot.lim[3]+rugsize), cbind(as.numeric(ydata.gr), as.numeric(ydata.gr)), breaks=0.5+(0:m), add=TRUE, col=col.part, ...)  

    for (j in 1:m)
    {     
        ## draw data points
        if (drawpoints)
        {
            if (missing(y)) yy <- fhat$x[[j]]
            else 
            {
                if (missing(y.group)) yy <- fhat$x[[j]]
                else yy <- y[y.group==levels(y.group)[j]]
            }
            if (jitter) yy <- jitter(yy)
            rug(yy, col=col.pt[j], ticksize=-0.03)
        }
    }   
}

plotkda.2d <- function(x, y, y.group, prior.prob=NULL, display.part="filled.contour",
    cont=c(25,50,75), abs.cont, approx.cont=TRUE, xlim, ylim, xlab, ylab,
    drawpoints=FALSE, drawlabels=TRUE, cex=1, pch, lty, part=TRUE, col, col.fun, col.part, col.pt, alpha=1, lwd=1, lwd.part=0, labcex=1, add=FALSE, ...)
{ 
    fhat <- x
    m <- length(fhat$x)
    xtemp <- numeric()
    for (j in 1:m) xtemp <- rbind(xtemp, fhat$x[[j]]) 
    if (missing(xlim)) xlim <- range(xtemp[,1])
    if (missing(ylim)) ylim <- range(xtemp[,2])
    if (missing(pch)) pch <- 1:m
    if (missing(lty)) lty <- rep(1, m)
    if (length(lty) < m) lty <- rep(lty, m)
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n, palette="Dark2", alpha=1) }
    if (missing(col)) col <- col.fun(m)
    if (length(col) < m) col <- rep(col, m)
    if (missing(col.part)) col.part <- plot3D.alpha.col(col, alpha=alpha) 
    if (missing(col.pt)) col.pt <- col.fun(m)
    if (length(col.pt)==1) col.pt <- rep(col.pt, m)

    x.names <- fhat$names
    if (!is.null(x.names))
    {
        if (missing(xlab)) xlab <- x.names[1]
        if (missing(ylab)) ylab <- x.names[2]
    }
    else
    {
        xlab="x"
        ylab="y"
    }
    if (is.null(prior.prob)) prior.prob <- fhat$prior.prob

    if (m != length(prior.prob))
        stop("prior.prob not same length as number of components in fhat")
    if (!(identical(all.equal(sum(prior.prob), 1), TRUE)))  
        stop("Sum of weights not equal to 1")

    ## set up plot
    if (!add)
    {
        if (missing(y)) 
            plot(fhat$x[[1]], type="n", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, ...)
        else
        plot(y, type="n", xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, ...)
    }

    ## set up common grid for all densities 
    class.grid <- array(0, dim=dim(fhat$estimate[[1]]))
    temp <- matrix(0, ncol=length(fhat$est), nrow=nrow(fhat$est[[1]]))
    for (j in 1:ncol(fhat$estimate[[1]]))
    {
        for (k in 1:length(fhat$estimate))
            temp[,k] <- fhat$estimate[[k]][,j]* prior.prob[k]
        class.grid[,j] <- max.col(temp)
    }

    ## draw partition
    fhat.part <- fhat
    fhat.part$estimate <- class.grid  
    fhat.part$H <- fhat$H[[1]]
    fhat.part$x <- fhat$x[[1]]
    fhat.part$w <- fhat$w[[1]]
    fhat.part$cont <- fhat$cont[[1]]
    class(fhat.part) <- "kde.part"

    if (part) plot(fhat.part, col=col.part, add=TRUE, display=display.part, lwd=lwd.part, alpha=alpha, ...)

    ## common contour levels removed from >= v1.5.3 
    if (missing(abs.cont))
    {
        hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
        nhts <- length(hts[[1]])
    }
    else
    {
        hts <- abs.cont
        nhts <- length(hts)
    }
 
        ## draw contours
        for (j in 1:m)
        {
            for (i in 1:nhts) 
            {
                if (missing(abs.cont))
                {
                    scale <- cont[i]/hts[[j]][i]
                    contour(fhat$eval.points[[1]], fhat$eval.points[[2]], 
                        fhat$estimate[[j]]*scale, level=hts[[j]][i]*scale, add=TRUE, 
                        drawlabels=drawlabels, lty=lty[j], col=col[j], lwd=lwd, labcex=labcex, 
                        ...)
                }
                else
                {
                    contour(fhat$eval.points[[1]], fhat$eval.points[[2]], 
                        fhat$estimate[[j]], level=hts[i], add=TRUE, 
                        drawlabels=drawlabels, lty=lty[j], col=col[j], lwd=lwd, labcex=labcex, 
                        ...)
                }
            }
        }
  
    for (j in 1:m)
    {  
        ## draw data points
        if (drawpoints)
        {
            if (missing(y))
                points(fhat$x[[j]], pch=pch[j], col=col.pt[j], cex=cex)
            else 
            {
                if (missing(y.group))
                {  
                    points(y, col=col.pt[1], cex=cex)
                    j <- m+1
                }
                else
                    points(y[y.group==levels(y.group)[j],], pch=pch[j], col=col.pt[j], cex=cex) 
            }
        }
    }   
}

plotkda.3d <- function(x, y, y.group, prior.prob=NULL, display="plot3D", cont=c(25,50,75), abs.cont, approx.cont=TRUE, colors, col, col.fun, col.pt, alpha=0.5, alphavec, xlab, ylab, zlab, drawpoints=FALSE, size=3, cex=1, pch, theta=-30, phi=40, d=4, ticktype="detailed", bty="f", add=FALSE, ...)
{
    fhat <- x
    
    m <- length(fhat$x) 
    if (is.null(prior.prob)) prior.prob <- fhat$prior.prob
    if (m != length(prior.prob))
        stop("prior.prob not same length as number of components in fhat")
    if (!(identical(all.equal(sum(prior.prob), 1), TRUE)))  
        stop("Sum of prior weights not equal to 1")
    
    if (missing(col.fun)) col.fun <- function(n) { hcl.colors(n, palette="Dark2") }
    if (missing(col)) col <- col.fun(m)
    if (length(col) < m) col <- rep(col, m)
    if (missing(col.pt)) col.pt <- col.fun(m)
    if (length(col.pt)==1) col.pt <- rep(col.pt, m)
    if (missing(pch)) pch <- 1:m
    if (length(pch)==1) pch <- rep(pch, m)
   
    x.names <- colnames(fhat$x[[1]])
    if (missing(xlab)) if (is.null(x.names)) xlab <- "x" else xlab <- x.names[1]
    if (missing(ylab)) if (is.null(x.names)) ylab <- "y" else ylab <- x.names[2]
    if (missing(zlab)) if (is.null(x.names)) zlab <- "z" else zlab <- x.names[3]
    
    xx <- numeric(0)
    for (j in 1:m) xx <- rbind(xx, fhat$x[[j]])
    
    ## common contour levels removed from >= v1.5.3 
    if (missing(abs.cont))
    {
        hts <- contourLevels(fhat, prob=(100-cont)/100, approx=approx.cont)
        nhts <- length(hts[[1]])
    }
    else
    {
        hts <- abs.cont
        nhts <- length(hts)
    }
    
    if (missing(alphavec)) alphavec <- seq(0.05,0.2,length=nhts)
    
    disp <- match.arg(display, c("plot3D", "rgl")) 
    if (disp %in% "plot3D")
    {
        if (!requireNamespace("plot3D", quietly=TRUE)) stop("Install the plot3D package as it is required.", call.=FALSE)
        for (j in 1:m)
        {
            for (i in 1:nhts)
            { 
                cti <- hts[[j]][nhts-i+1]
                if (cti <= max(fhat$estimate[[j]]))
                    plot3D::isosurf3D(x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], colvar=fhat$estimate[[j]], level=cti, add=add | (j>1 | i>1), alpha=alphavec[i], col=col[j], phi=phi, theta=theta, xlab=xlab, ylab=ylab, zlab=zlab, d=d, ticktype=ticktype, bty=bty, ...)
            }
        }
        
        ## plot points
        if (drawpoints)   
        {
            for (j in 1:m)
            {
                if (missing(y))
                    plot3D::points3D(fhat$x[[j]][,1], fhat$x[[j]][,2], fhat$x[[j]][,3], col=col.pt[j], cex=cex, alpha=1, add=TRUE, pch=pch[j])
                else
                {
                    if (missing(y.group))
                        plot3D::points3D(y[,1], y[,2], y[,3], col=col.pt, cex=cex, alpha=alpha, add=TRUE, pch=pch[j])
                    else
                    {
                        y.temp <- y[y.group==levels(y.group)[j],]
                        if (nrow(y.temp)>0)
                            plot3D::points3D(y.temp[,1], y.temp[,2], y.temp[,3], col=col.pt[j], cex=cex, alpha=alpha, add=TRUE, pch=pch[j])
                    }
                }
            }
        }
    
    }
    else if (disp %in% "rgl")
    {
        ## suggestions from Viktor Petukhov 08/03/2018
        if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
        if (!requireNamespace("misc3d", quietly=TRUE)) stop("Install the misc3d package as it is required.", call.=FALSE)

        if (missing(colors)) colors <- col
        
        xtemp <- numeric(); for (i in 1:length(fhat$x)) xtemp <- rbind(xtemp, fhat$x[[i]])
        rgl::plot3d(x=xtemp[,1], y=xtemp[,2], z=xtemp[,3], type="n", xlab=xlab, ylab=ylab, zlab=zlab, ...)

        for (j in 1:m)
        {
            for (i in 1:nhts)
            { 
                cti <- hts[[j]][nhts-i+1]
                if (cti <= max(fhat$estimate[[j]]))
                    misc3d::contour3d(x=fhat$eval.points[[1]], y=fhat$eval.points[[2]], z=fhat$eval.points[[3]], f=fhat$estimate[[j]], level=cti, add=TRUE, alpha=alphavec[i], color=colors[j], ...)
            }
            if (drawpoints)   ## plot points
            {
                if (missing(y))
                    rgl::points3d(fhat$x[[j]][,1], fhat$x[[j]][,2], fhat$x[[j]][,3], color=col.pt[j], size=size, alpha=alpha)
                else
                {
                    if (missing(y.group))
                        rgl::points3d(y[,1], y[,2], y[,3], color=col.pt, size=size, alpha=1)
                    else
                    {
                        y.temp <- y[y.group==levels(y.group)[j],]
                        if (nrow(y.temp)>0)
                            rgl::points3d(y.temp[,1], y.temp[,2], y.temp[,3], color=col.pt[j], size=size, alpha=alpha)
                    }
                }
            }
        }
    }   
}

#############################################################################
## Find bandwidths for each class in training set, for 2- to 6-dim 
##
## Parameters
## x - data values
## group - group variable
## bw - type of bandwidth selector
## nstage, pilot, pre - parameters for plugin bandwidths
## diag - FALSE - use full b/w matrices
##      - TRUE - use diag b/w matrices
##
## Returns
## Matrix of bandwidths for each group in training set
#############################################################################
hkda <- function(x, x.group, bw="plugin", ...)
{
    gr <- sort(unique(x.group))
    m <- length(gr)
    bw <- match.arg(bw, c("lscv", "plugin", "scv"))
    hs <- numeric(0) 
    for (i in 1:m)
    {
        y <- x[x.group==gr[i]]
        if (bw=="plugin") h <- hpi(y, ...)
        else if ((bw=="lscv") | (bw=="ucv")) h <- hlscv(y, ...)
        else if (bw=="scv") h <- hscv(y, ...)
        hs <- c(hs, h)
    }

    return(hs)
}
   
Hkda <- function(x, x.group, Hstart, bw="plugin", ...)
{
    d <- ncol(x)
    gr <- sort(unique(x.group))
    m <- length(gr)
    bw <- match.arg(bw, c("lscv", "plugin", "scv", "ucv")) 
    Hs <- numeric(0)

    for (i in 1:m)
    {
        y <- x[x.group==gr[i],]
        if (!missing(Hstart)) 
        {
            Hstarty <- Hstart[((i-1)*d+1) : (i*d),]
            if ((bw=="lscv") | (bw=="ucv")) H <- Hlscv(y, Hstart=Hstarty, ...)
            else if (bw=="scv") H <- Hscv(y, Hstart=Hstarty, ...)
            else if (bw=="plugin") H <- Hpi(y, Hstart=Hstarty, ...)
        }
        else
        {
            if ((bw=="lscv") | (bw=="ucv")) H <- Hlscv(y, ...)
            else if (bw=="scv") H <- Hscv(y, ...)
            else if (bw=="plugin") H <- Hpi(y, ...)
        }
        Hs <- rbind(Hs, H)
    }

    return(Hs)   
}

Hkda.diag <- function(x, x.group, bw="plugin", ...)
{
    d <- ncol(x)
    gr <- sort(unique(x.group))
    m <- length(gr)
    bw <- match.arg(bw, c("lscv", "plugin", "scv", "ucv"))
    Hs <- numeric(0)

    for (i in 1:m)
    {
        y <- x[x.group==gr[i],]
        if ((bw=="lscv") | (bw=="ucv")) H <- Hlscv.diag(y, ...)
        else if (bw=="plugin") H <- Hpi.diag(y, ...)
        else if (bw=="scv") H <- Hscv.diag(y, ...)
        Hs <- rbind(Hs, H)
    }

    return(Hs)   
}

#############################################################################
## Compares true group classification with an estimated one
##
## Parameters
## group - true group variable
## est.group - estimated group variable
##
## Returns
## List with components
## comp - cross-classification table of groupings - true groups are the rows,
##        estimated groups are the columns
## error - total mis-classification rate
#############################################################################
compare <- function(x.group, est.group, by.group=FALSE)
{
    if (length(x.group)!=length(est.group))
    stop("Group label vectors not the same length")

    if (!is.factor(x.group)) x.group <- factor(x.group)
    grlab <- levels(x.group)
    m <- length(grlab)
    comp <- table(x.group, est.group)

    if (by.group)
    {
        er <- vector()
        for (i in 1:m)
          er[i] <- 1-comp[i,i]/rowSums(comp)[i]
        er <- matrix(er, ncol=1)
        er <- rbind(er, 1 - sum(diag(comp))/sum(comp)) 
        rownames(er) <- c(as.character(paste(grlab, "(true)")), "Total")
        colnames(er) <- "error"    
    }
    else 
        er <- 1 - sum(diag(comp))/sum(comp)

    comp <- cbind(comp, rowSums(comp))
    comp <- rbind(comp, colSums(comp))

    colnames(comp) <- c(as.character(paste(grlab, "(est.)")), "Total")
    rownames(comp) <- c(as.character(paste(grlab, "(true)")), "Total")

    if (nrow(comp)==2 & nrow(comp)==2)
    {
        TN <- comp[1,1]; FP <- comp[1,2]; FN <- comp[2,1]; TP <- comp[2,2]
        spec <- 1-(FP/(FP+TN))
        sens <- 1-(FN/(TP+FN))
        comp <- list(cross=comp, error=er, TP=TP, FP=FP, FN=FN, TN=TN, spec=spec, sens=sens) 
    }
    else comp <- list(cross=comp, error=er)

    return(comp)
}

#############################################################################
## Computes cross-validated misclassification rates (for use when test data is
## not independent of training data) for KDA
##
## Parameters
## x - training data
## x.group - group variable for x
## y - data values to be classified
## Hs - bandwidth matrices
## prior.prob - prior probabilities
##
## Returns
## List with components
## comp - cross-classification table of groupings - true groups are the rows,
##        estimated groups are the columns
## error - total mis-classification rate
#############################################################################
compare.kda.cv <- function(x, x.group, bw="plugin", prior.prob=NULL, Hstart, by.group=FALSE, verbose=FALSE, recompute=FALSE, ...)
{
    if (verbose) pb <- txtProgressBar()
    bw <- match.arg(bw, c("lscv", "plugin", "scv", "ucv"))

    ## 1-d
    if (is.vector(x))
    {
        n <- length(x)
        h <- hkda(x, x.group, bw=bw, ...)
        gr <- sort(unique(x.group)) 
        kda.cv.gr <- x.group

        for (i in 1:n)
        {
            h.mod <- h
            ## find group that x[i] belongs to 
            ind <- which(x.group[i]==gr)
            indx <- x.group==gr[ind]
            indx[i] <- FALSE

            if ((bw=="lscv") | (bw=="ucv")) h.temp <- hlscv(x[indx], , ...)
            else if (bw=="plugin") h.temp <- hpi(x[indx], , ...)
            else if (bw=="scv") h.temp <- hscv(x[indx], , ...)
            h.mod[ind] <- h.temp

            ## recompute KDA estimate of groups with x[i] excluded
            if (verbose) setTxtProgressBar(pb, i/n)
            kda.cv.gr[i] <- kda(x[-i], x.group[-i], hs=h.mod, eval.points=x, prior.prob=prior.prob, kde.flag=FALSE)[i]
        }
        if (verbose) close(pb)
        
        return(compare(x.group, kda.cv.gr, by.group=by.group)) 
    }

    ## multi-dimensional   
    n <- nrow(x)
    d <- ncol(x)

    if (!missing(Hstart))
        H <- Hkda(x, x.group, bw=bw, Hstart=Hstart, ...)
    else
        H <- Hkda(x, x.group, bw=bw, ...)

    ## classify data x using KDA rules based on x itself
    ## kda.group <- kda(x, x.group, Hs=H, y=x, prior.prob=prior.prob)
    ## comp <- compare(x.group, kda.group)

    gr <- sort(unique(x.group)) 
    kda.cv.gr <- x.group

    for (i in 1:n)
    {
        H.mod <- H
        ## find group that x[i] belongs to 
        ind <- which(x.group[i]==gr)
        indx <- x.group==gr[ind]
        indx[i] <- FALSE

        if (recompute)
        {
            ## compute b/w matrix for that group with x[i] excluded
            if (!missing(Hstart))
            {  
                Hstart.temp <- Hstart[((ind-1)*d+1):(ind*d),]

                if (bw=="plugin") H.temp <- Hpi(x[indx,], Hstart=Hstart.temp,  ...)
                else if (bw=="scv") H.temp <- Hscv(x[indx,],  Hstart=Hstart.temp, ...)
                else if ((bw=="lscv") | (bw=="ucv")) H.temp <- Hlscv(x[indx,],  Hstart=Hstart.temp, ...)
            }
            else
            {
                if (bw=="plugin") H.temp <- Hpi(x[indx,], ...)
                else if (bw=="scv") H.temp <- Hscv(x[indx,], ...)
                else if ((bw=="lscv") | (bw=="ucv")) H.temp <- Hlscv(x[indx,], ...) 
            }

            H.mod[((ind-1)*d+1):(ind*d),] <- H.temp
        }
        ## recompute KDA estimate of groups with x[i] excluded
        if (verbose) setTxtProgressBar(pb, i/n)  

        kda.cv.gr[i] <- kda(x[-i,], x.group[-i], Hs=H.mod, eval.points=x, prior.prob=prior.prob, kde.flag=FALSE)[i]
    }
    if (verbose) close(pb)

    return(compare(x.group, kda.cv.gr, by.group=by.group)) 
}

#############################################################################
## Same as compare.kda.cv except uses diagonal b/w matrices
#############################################################################
compare.kda.diag.cv <- function(x, x.group, bw="plugin", prior.prob=NULL, by.group=FALSE, verbose=FALSE, recompute=FALSE, ...)
{
    if (is.vector(x))  return(compare.kda.cv(x=x, x.group=x.group, by.group=by.group, verbose=verbose, prior.prob=prior.prob, recompute=recompute, ...))
    n <- nrow(x)
    d <- ncol(x)

    H <- Hkda.diag(x, x.group, bw=bw, ...)

    if (verbose) pb <- txtProgressBar()
    bw <- match.arg(bw, c("lscv", "plugin", "scv", "ucv"))
    gr <- sort(unique(x.group)) 
    kda.cv.gr <- x.group
  
    for (i in 1:n)
    {
        H.mod <- H

        if (recompute)
        {
            ind <- which(x.group[i]==gr)
            indx <- x.group==gr[ind]
            indx[i] <- FALSE
            if (bw=="plugin")
                H.temp <- Hpi.diag(x[indx,], ...)
            else if ((bw=="lscv") | (bw=="ucv"))
                H.temp <- Hlscv.diag(x[indx,], ...)
            else if (bw=="scv")
                H.temp <- Hscv.diag(x[indx,], ...)
            H.mod[((ind-1)*d+1):(ind*d),] <- H.temp
        }
    
        if (verbose) setTxtProgressBar(pb, i/n) 
        kda.cv.gr[i] <- kda(x[-i,], x.group[-i], Hs=H.mod, eval.points=x, prior.prob=prior.prob, kde.flag=FALSE)[i]  
    }
    if (verbose) close(pb)
  
    return(compare(x.group, kda.cv.gr, by.group=by.group)) 
}

######################################################################
## Kernel density ridge estimation for 2D/3D data
#####################################################################
kdr <- function(x, y, H, p=1, max.iter=400, tol.iter, segment=TRUE, k, kmax, min.seg.size=4, keep.path=FALSE, gridsize, xmin, xmax, binned, bgridsize, w, fhat, density.cutoff, pre=TRUE, verbose=FALSE)
{
    ## default values 
    xnames <- parse.name(x)
    x <- as.matrix(x)
    x.orig <- x
    if (pre) 
    { 
        S12 <- diag(apply(x.orig, 2, sd))  
        Sinv12 <- matrix.pow(S12,-1)
        x <- pre.scale(x)
        if (!missing(xmin)) xmin <- xmin %*% Sinv12
        if (!missing(xmax)) xmax <- xmax %*% Sinv12
        rescale <- function(x) { as.matrix(x) %*% S12 }   
    }
    
    ksd <- ks.defaults(x=x, binned=binned, bgridsize=bgridsize, gridsize=gridsize)
    d <- ksd$d; n <- ksd$n; w <- ksd$w
    binned <- ksd$binned
    bgridsize <- ksd$bgridsize
    gridsize <- ksd$gridsize

    ## default bandwidth
    if (missing(H)) H <- Hpi(x=x, nstage=2-(d>2), binned=binned, deriv.order=2, verbose=verbose)
    Hinv <- chol2inv(chol(H))
    tol <- 3.7
    tol.H <-  tol * diag(H)

    if (missing(xmin)) xmin <- apply(x, 2, min) - tol.H 
    if (missing(xmax)) xmax <- apply(x, 2, max) + tol.H
    if (missing(tol.iter)) tol.iter <- 1e-3*min(apply(x, 2, IQR))
    if (missing(y))
    {    
        xx <- seq(xmin[1], xmax[1], length = gridsize[1])
        yy <- seq(xmin[2], xmax[2], length = gridsize[2])
       
        if (d==2) y <- expand.grid(xx, yy)
        else if (d==3)
        {
            zz <- seq(xmin[3], xmax[3], length = gridsize[3])
            y <- expand.grid(xx, yy, zz)
        }
    }
    else { y <- as.matrix(y); if (pre) y <- y %*% Sinv12 }    
    if (is.vector(y)) y <- matrix(y, nrow=1)
    #if (missing(min.seg.size)) min.seg.size <- min(round(1e-3*nrow(y)),1)

    ## exclude low density regions from ridge search 
    if (missing(fhat)) fhat <- kde(x=x, w=w, binned=binned)
    if (missing(density.cutoff)) density.cutoff <- contourLevels(fhat, cont=99)
    y.ind <- predict(fhat, x=y)>density.cutoff
    y <- y[y.ind,]
   
    fhat2 <- kdde(x=x, H=H, deriv.order=2, xmin=xmin, xmax=xmax, binned=binned, bgridsize=bgridsize, gridsize=gridsize, w=w, verbose=verbose) 
    
    ## projected gradient mean shift iterations
    n.seq <- block.indices(n, nrow(y), d=d, r=0, diff=FALSE)
    if (verbose) pb <- txtProgressBar() 
    pc <- list()
    
    i <- 1
    if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
    pc <- kdr.base(x=x, fhat2=fhat2, y=y[n.seq[i]:(n.seq[i+1]-1),], H=H, tol.iter=tol.iter, Hinv=Hinv, verbose=verbose, max.iter=max.iter, p=p)
    
    if (pre)
    {
        pc[c("x","y","end.points")] <- lapply(pc[c("x","y","end.points")], rescale)
        pc[["path"]] <- lapply(pc[["path"]], rescale)
    }
    
    if (length(n.seq)>2)
    {
        for (i in 2:(length(n.seq)-1))
        {  
            if (verbose) setTxtProgressBar(pb, i/(length(n.seq)-1))
            pc.temp <- kdr.base(x=x, fhat2=fhat2, y=y[n.seq[i]:(n.seq[i+1]-1),], H=H, tol.iter=tol.iter, Hinv=Hinv, verbose=verbose, max.iter=max.iter, p=p)

            if (pre)
            {
                pc.temp[c("y","end.points")] <- lapply(pc.temp[c("y","end.points")], rescale)
                pc.temp[["path"]] <- lapply(pc.temp[["path"]], rescale)
            }
            pc$y <- rbind(pc$y, pc.temp$y)
            pc$end.points <- rbind(pc$end.points, pc.temp$end.points)
            pc$path <- c(pc$path, pc.temp$path)
        }
    }
    if (verbose) close(pb)
    ## remove short segments for p=1
    # if (p==1)
    # {     
    #     tol.seg <- 1e-2*max(apply(x.orig, 2, IQR))
    #     pc.dendo <- hclust(dist(pc$end.points), method="single")
    #     pc.label <- cutree(pc.dendo, h=tol.seg)
    #     pc.label.ind <- pc.label %in% which(table(pc.label)>min.seg.size)
    #     pc$y <- pc$y[pc.label.ind,]
    #     pc$end.points <- pc$end.points[pc.label.ind,]
    #     pc$path <- pc$path[pc.label.ind]
    # }
    pc$H <- H
    pc$names <- xnames
    if (pre) pc$H <- S12 %*% pc$H %*% S12
    
    ## keep segments with at least min.seg.size points
    if (segment) pc <- kdr.segment(x=pc, k=k, kmax=kmax, min.seg.size=min.seg.size, verbose=verbose)
    else pc$end.points <- data.frame(pc$end.points, segment=factor(1L))
     
    ## force segment from factor to nuemric as this causes error in eks <= 1.0.7
    pc$end.points$segment <- as.numeric(levels(pc$end.points$segment)[pc$end.points$segment])

    ## put paths as last element in list
    path.temp <- pc$path
    pc$path <- NULL
    pc$tol.iter <- tol.iter
    pc$min.seg.size <- min.seg.size
    pc$binned <- binned
    pc$names <- xnames
    pc$w <- w
    if (keep.path) pc$path <- path.temp
    
    return(pc)
}

kdr.base <-function(x, fhat2, H, y, max.iter, tol.iter, p=1, verbose=FALSE, Hinv, ...)
{
    if (!is.matrix(x)) x <- as.matrix(x)
    if (!is.matrix(y)) y <- as.matrix(y)
    if (missing(Hinv)) Hinv <- chol2inv(chol(H))
    nx <- nrow(x)
    ny <- nrow(y)
    d <- ncol(y)
    y.path <- split(y, row(y), drop=FALSE)
    names(y.path) <- NULL
 
    xHinv <- x %*% Hinv
    xHinvx <- rowSums(xHinv*x)
    y.update <- y
    i <- 1
    eps <- max(sqrt(rowSums(y.update^2)))
    disp.ind <- head(sample(1:nrow(y)), n=min(1000,nrow(y)))

    while (eps > tol.iter & i < max.iter)
    {
        y.curr <- y.update
        yHinvy <- t(rowSums(y.curr%*%Hinv *y.curr))
        Mah <- apply(yHinvy, 2, "+", xHinvx) - 2*xHinv %*% t(y.curr)
        w <- exp(-Mah/2)
        denom <- colSums(w)
        num <- t(w)%*%x
        mean.shift.H <- num/denom - y.curr

        fhat2.y.curr <- predict(fhat2, x=y.curr)
        for (j in 1:ny)
        {
            Hessian <- invvec(fhat2.y.curr[j,])
            Hessian.svd <- eigen(Hessian, symmetric=TRUE)
            Up <- Hessian.svd$vectors[,tail(1:d,n=d-p)]
            mean.shift.H[j,] <- drop(Up %*% t(Up) %*% mean.shift.H[j,])
        }
        y.update <- y.curr + mean.shift.H                             
        y.update.list <- split(y.update, row(y.update), drop=FALSE)
        y.path <- mapply(rbind, y.path, y.update.list, SIMPLIFY=FALSE)
        eps <- max(sqrt(rowSums((y.curr-y.update)^2)))
    
        if (verbose>1)
        {
            if (d==2) plot(y.update[disp.ind,], col=1, xlab="x", ylab="y")
            else pairs(y.update[disp.ind,], col=1)
        }
        i <- i+1
    }
    
    pc.endpt <- t(sapply(y.path, tail, n=1, SIMPLIFY=FALSE))
    
    pc <- list(x=x, y=y, end.points=pc.endpt, path=y.path, type="kdr")
    class(pc) <- "kdr"
    
    return(pc)
}

## create segment of KDR filaments
## x = output from kdr
kdr.segment <- function(x, k, kmax, min.seg.size, verbose=FALSE)
{
    ep <- x$end.points
    if (any(names(ep) %in% "segment")) ep <- x$end.points[,-which(names(ep)=="segment")]
    if (missing(min.seg.size)) min.seg.size <- x$min.seg.size
    hc <- hclust(dist(ep), method="single")
    
    if (missing(kmax)) kmax <- 30
    kmax <- min(kmax, nrow(ep))
    if (missing(k))
    {
        if (verbose) pb <- txtProgressBar() 
         
        clust.ind <- rep(0, kmax)
        for (i in 1:kmax)
        {
            if (verbose) setTxtProgressBar(pb, i/kmax)
            clust.ind[i] <- clust.crit(hc=hc, x=as.matrix(ep), k=i, min.seg.size=min.seg.size)
        }
        if (verbose) close(pb)
        clust.ind[is.na(clust.ind)] <- 0
        kopt <- which.max(clust.ind) 
    }
    else kopt <- k
   
    ## remove singleton segments
    ep <- data.frame(ep, segment=as.integer(cutree(hc, k=kopt)))
    label <- ep$segment
    tlabel <- as.integer(names(table(label))[table(label) > max(min.seg.size,1)])
    ep <- ep[label %in% tlabel,]
    ep$segment <- factor(ep$segment)
    levels(ep$segment) <- 1:length(levels(ep$segment))
    ep$segment <- as.integer(levels(ep$segment))[ep$segment]
    
    ## re-order KDR segments into 'reasonable' linestring order
    ## experimental 
    j <- 1
    if (nrow(ep)>0)
    {
        for (i in unique(ep$segment))
        {
            ep.temp <- as.matrix(ep[ep$segment==i,-ncol(ep)])
            ep.temp <- data.frame(chain.knnx(ep.temp, k1=1, k2=1), segment=i) 
            if (j==1) ep.ord <- ep.temp else ep.ord <- rbind(ep.ord, ep.temp) 
            j <- j+1
        }

        names(ep.ord) <- c(x$names, "segment")
        rownames(ep.ord) <- NULL
        
        ## order in decreasing #points in segment 
        ep.ord <- split(ep.ord, ep.ord$segment)
        ep.ord <- ep.ord[order(sapply(ep.ord, nrow), decreasing=TRUE)]
        ep.ord <- lapply(1:length(ep.ord), function(.) { ep.ord[[.]]$segment <- .; ep.ord[[.]]}) 
        ep.ord <- do.call(rbind, ep.ord)
        ep.ord$segment <- factor(ep.ord$segment)
        levels(ep.ord$segment) <- 1:length(levels(ep.ord$segment))
        x$end.points <- ep.ord
        x$min.seg.size <- min.seg.size
        x$k <- length(levels(ep.ord$segment))
        if (exists("clust.ind")) x$clust.ind <- clust.ind
    }
    else 
    {
        x$end.points <- data.frame(x$end.points, segment=factor(1L))
        names(x$end.points) <- c(x$names, "segment")
    }
    
    return(x)
}

## rbind nearest neighbour of y from x to y
add.knnx <- function(x, y, k=1)
{
    xynn <- FNN::get.knnx(x, y, k=k)
    y <- rbind(y, x[xynn$nn.index,])
    d <- ncol(x)
    xy <- list(x=matrix(x[-xynn$nn.index,],ncol=d), y=y)
    
    return(xy)      
}

## arrange points in KDR to form a "reasonable" linestring 
chain.knnx <- function(x, k1=1, k2=5)
{
    ## concatenate the nearest neighbours in a chain
    ## start with first point in x
    if (!is.matrix(x)) x <- as.matrix(x)
    d <- ncol(x)
    if (nrow(x)>1)
    {
        x.ord.list <- add.knnx(x=matrix(x[-1,], ncol=d), y=matrix(x[1,], ncol=d), k=k1)
        x.ord <- x.ord.list$y

        while (nrow(x.ord.list$x)>0)
        {
            y.temp <- matrix(apply(as.matrix(tail(x.ord.list$y, n=k2)), 2, mean), ncol=d)
            colnames(y.temp) <- names(x.ord.list$x)
            x.ord.list.temp <- add.knnx(x=x.ord.list$x, y=y.temp, k=k1)  
            x.ord <- rbind(x.ord, matrix(tail(x.ord.list.temp$y,n=1),ncol=d))
            x.ord.list <- x.ord.list.temp
        }
        
        ## decide which permutation is "best" linestring
        ## break at max discontinuity
        ind <- which.max(rowSums((head(x.ord, n=-1)-tail(x.ord,n=-1))^2))
        ind1 <- c(1:ind, (ind+1):nrow(x.ord))
        ind2 <- c(1:ind, rev((ind+1):nrow(x.ord)))
        ind3 <- c(rev(1:ind), (ind+1):nrow(x.ord))
        ind4 <- c(rev(1:ind), rev((ind+1):nrow(x.ord)))
        x.ord1 <- x.ord[ind1,] 
        x.ord2 <- x.ord[ind2,]
        x.ord3 <- x.ord[ind3,] 
        x.ord4 <- x.ord[ind4,]
        x.ord.dist <- rep(0,4)
        x.ord.dist[1] <- sum(rowSums((head(x.ord1, n=-1)-tail(x.ord1,n=-1))^2))
        x.ord.dist[2] <- sum(rowSums((head(x.ord2, n=-1)-tail(x.ord2,n=-1))^2))
        x.ord.dist[3] <- sum(rowSums((head(x.ord3, n=-1)-tail(x.ord3,n=-1))^2))
        x.ord.dist[4] <- sum(rowSums((head(x.ord4, n=-1)-tail(x.ord4,n=-1))^2))
        x.ord <- get(paste0("x.ord", which.min(x.ord.dist)))
    }
    else x.ord <- x
    
    return(x.ord)
}

## Calinski-Harabasz clustering criterion for hierarchical clustering object
clust.crit <- function(hc, x, k, min.seg.size=1)
{
    label <- cutree(hc, k=k)
    tlabel <- as.integer(names(table(label))[table(label) > min.seg.size])
    tlabel.ind <- label %in% tlabel
    cc <- fpc.calinhara(x=x, clustering=label)
    
    return(cc)
}

## copied from fpc::calinhara 2020-09-18
fpc.calinhara <- function(x, clustering, cn = max(clustering)) 
{
    x <- as.matrix(x)
    p <- ncol(x)
    n <- nrow(x)
    cln <- rep(0, cn)
    W <- matrix(0, p, p)
    for (i in 1:cn) cln[i] <- sum(clustering == i)
    for (i in 1:cn) {
        clx <- x[clustering == i, ]
        cclx <- cov(as.matrix(clx))
        if (cln[i] < 2) 
            cclx <- 0
        W <- W + ((cln[i] - 1) * cclx)
    }
    S <- (n - 1) * cov(x)
    B <- S - W
    out <- (n - cn) * sum(diag(B))/((cn - 1) * sum(diag(W)))
    
    return(out)
}

#############################################################################
## S3 methods for KDR objects
#############################################################################
## plot method 
plot.kdr <- function(x, ...)
{ 
    fhat <- x
    d <- ncol(fhat$x)

    if (d==2) 
    {
        plotret <- plotkdr.2d(fhat, ...)
        invisible(plotret)
    }
    else if (d==3)
    {
        plotkdr.3d(fhat, ...)
        invisible()
    }
    else 
      stop ("Plot function only available for 2 or 3-d data")
}

plotkdr.2d <- function(x, add=FALSE, col, type="p", alpha=1, ...)
{
    xp <- x$end.points
    if (!any(names(xp) %in% "segment"))
    {
        if (missing(col)) col <- 6
        col <- transparency.col(col, alpha=alpha)
        if (!add) plot(xp, col=col, type=type, ...)
        else points(xp, col=col, ...)  
    }
    else
    {
        xps <- unique(xp$segment)
        if (missing(col)) col <- hcl.colors(n=length(xps), palette="Set2")
        if (length(col) < length(xps)) col <- rep(col, length(xps))
        col <- transparency.col(col, alpha=alpha)
        if (!add) plot(xp[,-ncol(xp)], col="transparent", ...)
        for (i in 1:length(xps)) lines(xp[xp$segment==xps[i],-ncol(xp)], col=col[i], type=type, ...)  
    }
}

plotkdr.3d <- function(x, display="plot3D", colors, col, col.fun, alphavec, size=3, cex=1, pch=1, theta=-30, phi=40, d=4, ticktype="detailed", add=FALSE, xlab, ylab, zlab, alpha=1, box=TRUE, axes=TRUE, type="p", ...)
{
    fhat <- x
    xp <- x$end.points
  
    if (missing(xlab)) xlab <- fhat$names[1]
    if (missing(ylab)) ylab <- fhat$names[2]
    if (missing(zlab)) zlab <- fhat$names[3]
    if (!any(names(xp) %in% "segment"))
    {
        if (missing(col)) col <- 6
    }
    else 
    {
        xps <- unique(xp$segment)
        if (missing(col)) col <- hcl.colors(n=length(xps), palette="Set2")
        if (length(col) < length(xps)) col <- rep(col, length(xps))
    }
    colors <- col
    disp <- match.arg(display, c("plot3D", "rgl")) 
	if (disp %in% "plot3D") 
	{
        if (!requireNamespace("plot3D", quietly=TRUE)) stop("Install the plot3D package as it is required.", call.=FALSE)
        if (!add) plot3D::scatter3D(x=xp[,1], y=xp[,2], z=xp[,3], add=add, theta=theta, phi=phi, d=d, type=type, xlab=xlab, ylab=ylab, zlab=zlab, ticktype=ticktype, type="n", col=NA, ...) 
        for (i in 1:length(xps))
            plot3D::scatter3D(x=xp[xp$segment==xps[i],1], y=xp[xp$segment==xps[i],2], z=xp[xp$segment==xps[i],3], cex=cex, col=col[i], add=TRUE, pch=pch, type=type, alpha=alpha, ...) 
    }
    else if (disp %in% "rgl")
    {
        if (!requireNamespace("rgl", quietly=TRUE)) stop("Install the rgl package as it is required.", call.=FALSE)
        for (i in 1:length(xps))
            rgl::plot3d(x=xp[xp$segment==xps[i],1], y=xp[xp$segment==xps[i],2], z=xp[xp$segment==xps[i],3], col=col[i], alpha=alpha, xlab=xlab, ylab=ylab, zlab=zlab, add=add | (i>1), box=box, axes=axes, type=type, size=size, ...)
    }
}

---
title: "ks: Kernel density estimation for bivariate data"
output: rmarkdown::html_vignette
description: ks Kernel density estimation for bivariate data
author: Tarn Duong
vignette: >
  %\VignetteIndexEntry{ks: Kernel density estimation for bivariate data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


```{r, echo=FALSE, message=FALSE}
knitr::opts_chunk$set(global.par=TRUE, collapse=TRUE, comment="#>", fig.width=5, fig.height=5, fig.align="center", dpi=96)
options(tibble.print_min=3L, tibble.print_max=3L)
```

## Introduction 

Kernel density estimation is a popular tool for visualising the distribution of data. See Chacon \& Duong (2018), for example, for an overview. When multivariate kernel density estimation is considered it is usually in the constrained context with diagonal bandwidth matrices, e.g. in the `R` packages `sm` and `KernSmooth`. In contrast, the `ks` package implements both diagonal and unconstrained data-driven bandwidth matrices for kernel density estimation. 

## Bivariate data example

For our target density, we use the `dumbbell' density, given by the normal mixture
$$ 
\frac{4}{11} N \bigg( \begin{bmatrix}-2 \\ 2\end{bmatrix}, 
\begin{bmatrix}1 & 0 \\ 0 & 1 \end{bmatrix} \bigg)+ 
\frac{3}{11} N \bigg( \begin{bmatrix}0 \\ 0\end{bmatrix},
\begin{bmatrix}0.8 & -0.72 \\ -0.72 & 0.8\end{bmatrix} \bigg)+
\frac{4}{11} N \bigg( \begin{bmatrix}2 \\ -2\end{bmatrix}, 
\begin{bmatrix}1 & 0 \\ 0 & 1 \end{bmatrix} \bigg).
$$
This density is unimodal and we draw a sample of 2000 data points.
 
```{r, message=FALSE}
library(ks)
mus <- rbind(c(-2,2), c(0,0), c(2,-2))
Sigmas <- rbind(diag(2), matrix(c(0.8, -0.72, -0.72, 0.8), nrow=2), diag(2))
cwt <- 3/11
props <- c((1-cwt)/2, cwt, (1-cwt)/2)
plotmixt(mus=mus, Sigmas=Sigmas, props=props, display="filled.contour", xlim=c(-4,4), ylim=c(-4,4))
set.seed(88192)
samp <- 2000
x <- rmvnorm.mixt(n=samp, mus=mus, Sigmas=Sigmas, props=props)
colnames(x) <- c("x","y")
plot(x, pch=16, cex=0.5, xlim=c(-4,4), ylim=c(-4,4))
```

We use `Hpi` for unconstrained plug-in selectors and `Hpi.diag` for diagonal plug-in selectors.

```{r}
Hpi1 <- Hpi(x=x)
Hpi2 <- Hpi.diag(x=x)
```

To compute a kernel density estimate, the command is `kde`, which creates a `kde` class object

```{r}
fhat.pi1 <- kde(x=x)         ## same as Hpi(x=x, H=Hpi1)
fhat.pi2 <- kde(x=x, H=Hpi2)
``` 

We use the `plot` method for `kde` objects to display these kernel density estimates. The default is a contour plot with the upper 25%, 50% and 75% contours of the 
(sample) highest density regions. The diagonal bandwidth matrix constrains the smoothing to be performed in directions parallel to the co-ordinate axes, so it is not able to apply accurate levels of smoothing to the obliquely oriented central portion. The result is a multimodal density estimate. The unconstrained bandwidth matrix correctly produces a unimodal density estimate. 

```{r}
plot(fhat.pi1, main="Plug-in", display="filled.contour", xlim=c(-4,4), ylim=c(-4,4))
plot(fhat.pi2, main="Plug-in diagonal", display="filled.contour", xlim=c(-4,4), ylim=c(-4,4)) 
```

The unconstrained SCV (Smoothed Cross Validation) selector is `Hscv` and 
its diagonal version is `Hscv.diag`. The most reasonable density estimate below is from the unconstrained SCV selector. 

```{r}
Hscv1 <- Hscv(x=x)
Hscv2 <- Hscv.diag(x=x)
fhat.cv1 <- kde(x=x, H=Hscv1)
fhat.cv2 <- kde(x=x, H=Hscv2)
plot(fhat.cv1, main="SCV", display="filled.contour", xlim=c(-4,4), ylim=c(-4,4))
plot(fhat.cv2, main="SCV diagonal", display="filled.contour", xlim=c(-4,4), ylim=c(-4,4))
```

The unconstrained bandwidth selectors will be better than their diagonal counterparts when the data have large mass oriented obliquely to the co-ordinate axes, like for the dumbbell data. 
The unconstrained plug-in and the SCV selectors can be viewed as generally recommended selectors.

## References

Chacon, J. E. and Duong, T. (2018). [_Multivariate Kernel Smoothing and Its Applications_](https://mvstat.net/mvksa/) Chapman & Hall/CRC Press, Boca Raton.

Duong, T. 2007). [ks: Kernel density estimation and kernel discriminant analysis for
  multivariate data in R](https://www.jstatsoft.org/v21/i07/) _Journal of Statistical Software_ **21 (7)**.

&nbsp; 

-- Generated `r format(Sys.time(), '%d %B %Y')`. 

\documentclass[a4paper,11pt]{article}

%\usepackage{amsmath,amssymb,amsthm,amsopn,natbib}
\usepackage{amsmath,amssymb,amsopn,natbib}
\usepackage[left=2.5cm,top=2.5cm,right=2.5cm,bottom=2.5cm]{geometry}

\renewcommand{\today}{\begingroup
\number \day\space  \ifcase \month \or January\or February\or March\or 
April\or May\or June\or July\or August\or September\or October\or 
November\or December\fi 
\space  \number \year \endgroup}

\renewcommand{\vec}[1]{\boldsymbol{#1}}
\newcommand{\mat}[1]{\mathbf{#1}}
\def\bH{\mat{H}}
\def\vecx{\vec{x}}
\def\vecX{\vec{X}}


\DeclareMathOperator{\E}{\boldsymbol{\mathbb{E}}}
%\DeclareMathOperator{\Var}{Var}
\let\code=\texttt
\let\proglang=\texttt
\let\pkg=\texttt

%\VignetteIndexEntry{kde} 
%\SweaveOpts{eps=FALSE}

\title{ks: Kernel density estimation for bivariate data}
\author{Tarn Duong}


\begin{document}

\maketitle


\noindent Kernel density estimation is a popular tool for visualising 
the distribution of data. See \citet*{simonoff1996}, for example, for
an overview.
When multivariate kernel density estimation is considered it is usually
in the constrained context with diagonal bandwidth matrices, e.g. 
in the \proglang{R} packages \pkg{sm} \citep*{sm} and \pkg{KernSmooth} 
\citep*{KernSmooth}.  
We introduce a new \proglang{R} package \pkg{ks} 
which implements diagonal and unconstrained data-driven bandwidth matrices
for kernel density estimation,
which can also be used for multivariate kernel
discriminant analysis.  
The \pkg{ks} package implements selectors for 1- to 6-dimensional
data. 

This vignette contains only a brief introduction 
to using \pkg{ks} for kernel density estimation
for 2-dimensional data. 
See \citet*{duong2007c} for a more detailed account. 
For a bivariate random sample $\vecX_1, \vecX_2, \ldots, \vecX_n$ 
drawn from a density $f$, 
the kernel density estimate is defined by
$$
\hat{f} (\vecx; \bH) = n^{-1}\sum_{i=1}^n K_{\bH} ( \vecx - \vec{X}_i)
$$
where $\vecx = (x_1, x_2)^T$ and $\vec{X}_i = (X_{i1}, X_{i2})^T, i = 1, 2,  
\ldots, n$.  Here 
$K(\vecx)$ is the kernel which is a symmetric probability density function, 
$\bH$ 
is the bandwidth matrix which is symmetric and positive-definite,  
and $K_{\bH}(\vecx) = |\bH|^{-1/2} 
K( \bH^{-1/2} \vecx)$. 
The choice of $K$ is not crucial: we take 
$K(\vecx) = (2\pi)^{-1} \exp(-\tfrac{1}{2} \vecx^T \vecx)$ the standard normal
throughout.  
In contrast, the choice of $\bH$ is crucial in determining 
the performance of $\hat f$. 
The most common parameterisations of the bandwidth matrix
are the diagonal and the 
general or unconstrained which has no restrictions on $\bH$
provided that $\bH$
remains positive definite and symmetric, that is 
$$
\bH = \begin{bmatrix}h_1^2 & 0 \\0 & h_2^2 \end{bmatrix}
\ \mathrm{or} \ 
\bH = \begin{bmatrix}h_1^2 & h_{12} \\ h_{12}  & h_2^2 \end{bmatrix}.
$$
This latter
parameterisation allows kernels to have an arbitrary orientation
whereas the former only allows kernels which are oriented to the
co-ordinate axes.

For our target density, we use the 
 `dumbbell' density, given by the normal mixture
$$ \frac{4}{11} N \bigg( \begin{bmatrix}-2 \\ 2\end{bmatrix}, 
\begin{bmatrix}1 & 0 \\ 0 & 1 \end{bmatrix} \bigg)+ 
\frac{3}{11} N \bigg( \begin{bmatrix}0 \\ 0\end{bmatrix},
\begin{bmatrix}0.8 & -0.72 \\ -0.72 & 0.8\end{bmatrix} \bigg)+
\frac{4}{11} N \bigg( \begin{bmatrix}2 \\ -2\end{bmatrix}, 
\begin{bmatrix}1 & 0 \\ 0 & 1 \end{bmatrix} \bigg),
$$
displayed on the left in Figure \ref{fig:dens-db}. This 
density is unimodal. On the right is a
sample of 200 data points.
 
<<results=hide,print=FALSE>>=
library(ks)
set.seed(8192)
samp <- 200
mus <- rbind(c(-2,2), c(0,0), c(2,-2))
Sigmas <- rbind(diag(2), matrix(c(0.8, -0.72, -0.72, 0.8), nrow=2), diag(2))
cwt <- 3/11
props <- c((1-cwt)/2, cwt, (1-cwt)/2)
x <- rmvnorm.mixt(n=samp, mus=mus, Sigmas=Sigmas, props=props)
@ 

\setkeys{Gin}{width=0.45\textwidth}
\begin{figure}[!ht]
\begin{center}
<<fig=TRUE,print=FALSE,echo=FALSE>>=
plotmixt(mus=mus, Sigmas=Sigmas, props=props, xlim=c(-4,4), ylim=c(-4,4))
@ 
<<fig=TRUE,print=FALSE,echo=FALSE>>=
plot(x, xlim=c(-4,4), ylim=c(-4,4), xlab="x", ylab="y")
@
\end{center}
\caption{Target `dumbbell' density. (Left) contour plot. (Right) Scatter plot.}
\label{fig:dens-db}
\end{figure}



We use \code{Hpi} for 
unconstrained plug-in selectors and \code{Hpi.diag} for diagonal plug-in selectors.
<<print=TRUE>>=
Hpi1 <- Hpi(x=x)
Hpi2 <- Hpi.diag(x=x)
@ 
To compute a kernel density estimate, the 
command is \code{kde}, which creates a \code{kde} class object
<<>>=
fhat.pi1 <- kde(x=x, H=Hpi1)
fhat.pi2 <- kde(x=x, H=Hpi2)
@ 
We use the \code{plot} method for \code{kde} objects to display these
kernel density estimates. The default is a contour plot with 
the upper 25\%, 50\% and 75\% contours of the 
(sample) highest density regions. %, as
%defined in \citet*{bowman1993} and \citet*{hyndman1996}.
These regions are also plotted by the \pkg{sm} library.
<<eval=FALSE>>=
plot(fhat.pi1)
plot(fhat.pi2)
@
The respective kernel density estimates are produced in Figure \ref{fig:pi}.
The diagonal bandwidth matrix constrains the smoothing to be performed in directions
parallel to the co-ordinate axes, so it is not able to apply accurate levels
of smoothing to the obliquely oriented central portion. The result is a 
multimodal
density estimate. The unconstrained bandwidth matrix correctly produces 
a unimodal density estimate. 


\begin{figure}[!ht]
\centering
<<fig=TRUE,echo=FALSE,print=FALSE>>=
plot(fhat.pi1, main="Plug-in", cex.main=1.4)
@ 
<<fig=TRUE,echo=FALSE,print=FALSE>>=
plot(fhat.pi2, main="Plug-in diagonal", cex.main=1.4) 
@  
\caption{Kernel density estimates with plug-in selectors}
\label{fig:pi}
\end{figure}


%The commands
%\code{Hlscv} and \code{Hlscv.diag} are the unconstrained and diagonal
%LSCV (Least Squares Cross Validation) selectors. 
The unconstrained SCV (Smoothed Cross Validation) selector is \code{Hscv} and 
its diagonal version is \code{Hscv.diag}.
In Figure \ref{fig:cv}, the most
reasonable density estimate is from the unconstrained SCV selector. 
 
<<print=TRUE>>=
Hscv1 <- Hscv(x=x)
Hscv2 <- Hscv.diag(x=x)
@ 

\begin{figure}[!ht]
\centering
<<echo=FALSE,print=FALSE>>=
fhat.cv1 <- kde(x=x, H=Hscv1)
fhat.cv2 <- kde(x=x, H=Hscv2)
@ 
<<fig=TRUE,echo=FALSE,print=FALSE>>=
plot(fhat.cv1, main="SCV", cex.main=1.4)
@ 
<<fig=TRUE,echo=FALSE,print=FALSE>>=
plot(fhat.cv2, main="SCV diagonal", cex.main=1.4)
@ 
\caption{Kernel density estimates with cross validation selectors}
\label{fig:cv}
\end{figure}

The unconstrained bandwidth selectors will be better than their diagonal counterparts
when the data have large mass oriented obliquely to the co-ordinate axes,
like for the dumbbell data. 
The unconstrained plug-in and the SCV selectors
can be viewed as generally recommended selectors.



\bibliographystyle{apalike}

\begin{thebibliography}{}


\bibitem[Bowman and Azzalini, 2007]{sm}
Bowman, A. W. and Azzalini, A. (2007).
\newblock {\em sm: kernel smoothing methods: Bowman and Azzalini (1997)}.
\newblock R package version 2.2.

\bibitem[Duong, 2007]{duong2007c}
Duong, T. (2007).
\newblock ks: {K}ernel density estimation and kernel discriminant analysis for
  multivariate data in {R}.
\newblock {\em Journal of Statistical Software}. \textbf{21 (7)}, URL \texttt{http://www.jstatsoft.org/v21/i07}.


\bibitem[Simonoff, 1996]{simonoff1996}
Simonoff, J. S. (1996).
\newblock {\em Smoothing Methods in Statistics}.
\newblock Springer-Verlag, New York.


\bibitem[Wand, 2006]{KernSmooth}
Wand, M. P. (2006).
\newblock {\em KernSmooth: Functions for kernel smoothing for Wand \& Jones
  (1995)}.
\newblock R package version 2.22-19. R port by Brian Ripley.

\end{thebibliography}

\end{document}

\name{unicef}
\docType{data}
\alias{unicef}
\title{Unicef child mortality - life expectancy data}

\description{
This data set contains the number of deaths of children under 5 years of
age per 1000 live births and the average life expectancy (in years) at birth
for 73 countries with GNI (Gross National Income) less
than 1000 US dollars per annum per capita.
}

\usage{data(unicef)}

\format{A matrix with 2 columns and 73 rows. Each row corresponds to a country.
The first column is the under 5 mortality rate and the second is the average
life expectancy.}

\source{
Unicef (2003). \emph{State of the World's Children Report 2003}, Oxford University Press, for Unicef.
}

\keyword{datasets}

\name{histde}
\alias{histde}
\alias{predict.histde}
\title{Histogram density estimate}

\description{
  Histogram density estimate for 1- and 2-dimensional data.
}

\usage{
histde(x, binw, xmin, xmax, adj=0)

\method{predict}{histde}(object, ..., x)
}

\arguments{
  \item{x}{matrix of data values}
  \item{binw}{(vector) of binwidths}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{adj}{displacement of default anchor point, in percentage of 1
    bin}
  \item{object}{object of class \code{histde}}
  \item{...}{other parameters}
}
  
\value{
  A histogram density estimate is an object of class \code{histde} which is a
  list with fields: 
  \item{x}{data points - same as input}
  \item{eval.points}{vector or list of points at which the estimate is evaluated}
  \item{estimate}{density estimate at \code{eval.points}}
  \item{binw}{(vector of) bandwidths}
  \item{nbin}{(vector of) number of bins}
  \item{names}{variable names}   
}

\details{
  If \code{binw} is missing, the default binwidth is \eqn{\hat{b}_i = 2 \cdot 3^{1/(d+2)} \pi^{d/(2d+4)} S_i
  n^{-1/(d+2)}}{b_i = 2*3^(1/(d+2))*pi^(d/(2d+4))*S_i*n^(-1/(d+2))}, the
  normal scale selector.  

  If \code{xmin} is missing then it defaults to the data minimum.  If
  \code{xmax} is missing then it defaults to the data maximum.  
}

\seealso{\code{\link{plot.histde}}}

\examples{
## positive data example
set.seed(8192)
x <- 2^rnorm(100)
fhat <- histde(x=x)
plot(fhat, border=6)
points(c(0.5, 1), predict(fhat, x=c(0.5, 1)))

## large data example on a non-default grid
set.seed(8192)
x <- rmvnorm.mixt(10000, mus=c(0,0), Sigmas=invvech(c(1,0.8,1)))
fhat <- histde(x=x, xmin=c(-5,-5), xmax=c(5,5))
plot(fhat)

## See other examples in ? plot.histde
}

\keyword{smooth}

\name{plot.kcde}
\alias{plot.kcde}
\title{Plot for kernel cumulative distribution estimate}

\description{
  Plot for kernel cumulative distribution estimate 1- to 3-dimensional data.
}

\usage{
\method{plot}{kcde}(x, ...)
}

\arguments{
  \item{x}{object of class \code{kcde} (output from \code{\link{kcde}})}
  \item{...}{other graphics parameters used in \code{\link{plot.kde}}}
}
  
\value{
  Plots for 1-d and 2-d are sent to graphics window. Plot for 3-d is sent to
  graphics/RGL window.
}

\details{
  For \code{kcde} objects, the function headers for the different dimensional data are  
  \preformatted{
  ## univariate
  plot(Fhat, xlab, ylab="Distribution function", add=FALSE, drawpoints=FALSE, 
       col.pt=4, jitter=FALSE, alpha=1, ...) 

  ## bivariate
  plot(Fhat, display="persp", cont=seq(10,90, by=10), abs.cont, xlab, ylab,    
       zlab="Distribution function", cex=1, pch=1, add=FALSE, drawpoints=FALSE, 
       drawlabels=TRUE, theta=-30, phi=40, d=4, col.pt=4, col, col.fun, alpha=1, 
       lwd=1, border=NA, thin=3, lwd.fc=5, ...) 
  
  ## trivariate     
  plot(Fhat, display="plot3D", cont=c(25,50,75), colors, col, alphavec, 
       size=3, cex=1, pch=1, theta=-30, phi=40, d=4, ticktype="detailed", 
       bty="f", col.pt=4, add=FALSE, xlab, ylab, zlab, drawpoints=FALSE, 
       alpha, box=TRUE, axes=TRUE, ...)}
}
  
\seealso{\code{\link{plot.kde}}}

\examples{
data(iris)
Fhat <- kcde(x=iris[,1])
plot(Fhat, xlab="Sepal.Length")
Fhat <- kcde(x=iris[,1:2])
plot(Fhat)
\donttest{Fhat <- kcde(x=iris[,1:3])
plot(Fhat, alpha=0.3)}
}

\keyword{hplot}

\name{worldbank}
\docType{data}
\alias{worldbank}
\title{Development indicators from the World Bank Group}

\description{
This data set contains six development indicators for national
entities for the year 2011, which is the latest year for which they are
consistently available. 
}

\usage{data(worldbank)}

\format{A matrix with 7 columns and 218 rows.
  Each row corresponds to a country.
  The first column is the country,
  the second is the per capita carbon dioxide emissions (thousands Kg),
  the third is the per capita GDP (thousands of current USD),
  the fourth is the annual GDP growth rate (\%),
  the fifth is the annual inflation rate (\%),
  the sixth is the percentage of internet users in the population (\%),
  the seventh is the added value agricultural production as a ratio of the total GDP (\%). }

\source{
World Bank Group (2016) 
\href{https://databank.worldbank.org/source/world-development-indicators/}{World development indicators}. Accessed 2016-10-03.
}

\keyword{datasets}

\name{as.kde}
\alias{as.kde}
\title{Quasi kernel density estimate}

\description{
  Quasi kernel density estimate for gridded data. 
}

\usage{
as.kde(x, density)
}

\arguments{
  \item{x}{tidy matrix with first d columns are evaluation points of an estimation grid, and last column is estimate value; or list with elements \code{eval.points}, \code{estimate}}
  \item{density}{flag to compute bona fide density. Default is TRUE.} 
}

\value{
  A quasi-kernel density estimate is an object of class \code{kqde} (for \code{density=TRUE}) or \code{kqdde} (for \code{density=FALSE}) which is a
  list with fields: 
  \item{x}{zero vector or matrix}
  \item{eval.points}{vector or list of points at which the estimate is evaluated - copied from input \code{x}} 
  \item{estimate}{density estimate - copied from input \code{x}}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names} 
  \item{cont}{vector of probability contour levels}
  This mimics the output from \code{\link{kde}} (for \code{density=TRUE}) or \code{\link{kdde}}(for \code{density=FALSE}).   
}

\details{
  This function converts the input \code{x} into a quasi-kernel density estimate which is treated similarly to the output from \code{\link{kde}} or \code{\link{kdde}}. If \code{density=TRUE} then all the estimate values are non-negative. If \code{density=FALSE} then the estimate values can be positive or negative. A \code{contourLevels} method computes an approximation of the density contour levels, and a \code{plot} method computes analogous plots. 
}

\seealso{\code{\link{kde}}, \code{\link{kdde}}}

\examples{
## univariate 
ep1 <- seq(0,1, length=401)
est1 <- dbeta(ep1, shape1=2, shape2=5)
f1 <- data.frame(ep1, estimate=est1)
f1.kde <- as.kde(f1)
plot(f1.kde)

## bivariate
ep2 <- expand.grid(x=seq(-pi,pi, length=151), y=seq(-pi,pi, length=151))
est2 <- cos(ep2[,1]*pi/2) + sin(ep2[,2]*pi/2)  
f2 <- data.frame(ep2, estimate=est2)
f2.kde <- as.kde(f2, density=FALSE)
plot(f2.kde, display="filled.contour")
plot(f2.kde, display="persp", phi=10)

\donttest{## trivariate
mus3 <- c(0,0,0)
Sigmas3 <- 1/2*diag(c(1,1/3,2))
props3 <- 1
ep3 <- expand.grid(x=seq(-3,3, length=51), y=seq(-3,3, length=51), z=seq(-3,3, length=51))
est3 <- dmvnorm.mixt(ep3, mus=mus3, Sigmas=Sigmas3, props=props3)
f3 <- data.frame(ep3, estimate=est3)
f3.kde <- as.kde(f3)
plot(f3.kde)}
}

\keyword{smooth}

\name{kde.boundary}
\alias{kde.boundary}
\title{Kernel density estimate for bounded data}

\description{
  Kernel density estimate for bounded 1- to 3-dimensional data.
}

\usage{
kde.boundary(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, 
   binned=FALSE, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE,
   boundary.supp, boundary.kernel="beta", verbose=FALSE)
}

\arguments{
  \item{x}{matrix of data values}
  \item{H,h}{bandwidth matrix/scalar bandwidth. If these are missing, \code{Hpi} or \code{hpi} is called by default.}
  \item{gridsize}{vector of number of grid points}
  \item{gridtype}{not yet implemented}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{vector or matrix of points at which estimate is evaluated}
  \item{binned}{flag for binned estimation.}
  \item{bgridsize}{vector of binning grid sizes}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{compute.cont}{flag for computing 1\% to 99\% probability contour levels. Default is TRUE.}
  \item{approx.cont}{flag for computing approximate probability contour
    levels. Default is TRUE.}
  \item{boundary.supp}{effective support for boundary region}
  \item{boundary.kernel}{"beta" = beta boundary kernel, "linear" = linear boundary kernel}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
}
  
\value{
  A kernel density estimate for bounded data is an object of class \code{kde}.    
}

\details{
  There are two forms of density estimates which are suitable for
  bounded data, based on the modifying the kernel function.
  For \code{boundary.kernel="beta"}, the 2nd form of the Beta boundary kernel of Chen (1999) is employed. It is suited for rectangular data  boundaries.  

  For \code{boundary.kernel="linear"}, the linear boundary kernel of
  Hazelton & Marshall (2009) is employed. It is suited for arbitrarily
  shaped data boundaries, though it is currently only implemented for rectangular boundaries.   
}

\references{
  Chen, S. X. (1999) Beta kernel estimators for density functions. 
  \emph{Computational Statistics and Data Analysis}, \bold{31}, 131-145.

  Hazelton, M. L. & Marshall, J. C. (2009) Linear boundary kernels for
  bivariate density estimation. \emph{Statistics and Probability
    Letters}, \bold{79}, 999-1003.
}

\seealso{\code{\link{kde}}}

\examples{
data(worldbank)
wb <- as.matrix(na.omit(worldbank[,c("internet", "ag.value")]))
fhat <- kde(x=wb)
fhat.beta <- kde.boundary(x=wb, xmin=c(0,0), xmax=c(100,100), boundary.kernel="beta")  
plot(fhat, col=1, xlim=c(0,100), ylim=c(0,100))
plot(fhat.beta, add=TRUE, col=2)
rect(0,0,100,100, lty=2)

\donttest{fhat.LB <- kde.boundary(x=wb, xmin=c(0,0), xmax=c(100,100), boundary.kernel="linear")
plot(fhat, col=1, xlim=c(0,100), ylim=c(0,100))
plot(fhat.LB, add=TRUE, col=3)
rect(0,0,100,100, lty=2)} 
}  

\keyword{smooth}

\name{Hns}
\alias{Hns}
\alias{Hns.diag}
\alias{hns}
\alias{Hns.kcde}
\alias{hns.kcde}
\title{Normal scale bandwidth}

\description{
Normal scale bandwidth.  
}

\usage{
Hns(x, deriv.order=0)
Hns.diag(x)
hns(x, deriv.order=0)
Hns.kcde(x)
hns.kcde(x)
}

\arguments{
  \item{x}{vector/matrix of data values}
  \item{deriv.order}{derivative order}
}

\value{
  Normal scale bandwidth. 
}

\details{
  \code{Hns} is equal to \code{(4/(n*(d+2*r+2)))^(2/(d+2*r+4))*var(x)},
  n = sample size, d = dimension of data, r = derivative
  order. \code{hns} is the analogue  of \code{Hns} for 1-d data. These
  can be used for density (derivative) estimators
  \code{\link{kde}}, \code{\link{kdde}}. 
  The equivalents for distribution estimators \code{\link{kcde}} are
  \code{Hns.kcde} and \code{hns.cde}. 
}

\references{Chacon J.E., Duong, T. & Wand, M.P. (2011). Asymptotics for
  general multivariate kernel density derivative
  estimators. \emph{Statistica Sinica}, \bold{21}, 807-840.   
}

\examples{
data(forbes, package="MASS")
Hns(forbes, deriv.order=2)
hns(forbes$bp, deriv.order=2)
}

\keyword{smooth}

\name{contour}
\alias{contourLevels}
\alias{contourLevels.kde}
\alias{contourLevels.kda}
\alias{contourLevels.kdde}
\alias{contourSizes}
\alias{contourProbs}
\alias{contourBreaks}
\alias{.contourBreaks}
\title{Contour functions}

\description{
  Contour levels and sizes.
} 

\usage{
contourLevels(x, ...)
\method{contourLevels}{kde} (x, prob, cont, approx=TRUE, ...)
\method{contourLevels}{kda} (x, prob, cont, approx=TRUE, ...)
\method{contourLevels}{kdde}(x, prob, cont, approx=TRUE, which.deriv.ind=1, ...) 

contourSizes(x, abs.cont, cont=c(25,50,75), approx=TRUE)
contourProbs(x, abs.cont, cont=c(25,50,75), approx=TRUE)
contourBreaks(x, n=3, cont=c(25,50,75), type="density")
}

\arguments{
  \item{x}{object of class \code{kde}, \code{kdde} or \code{kda}}
  \item{prob}{vector of probabilities corresponding to highest density regions}
  \item{cont}{vector of percentages which correspond to the complement
	of \code{prob}}
  \item{abs.cont}{vector of absolute contour levels}
  \item{approx}{flag to compute approximate contour levels. Default is
    TRUE.} 
  \item{which.deriv.ind}{partial derivative index. Default is 1.} 
  \item{...}{other parameters}
  \item{type}{type of contour levels: one of \code{"density", "length", "quantile", "natural"}. Default is "density".}
  \item{n}{number of contour levels for \code{type="length", "natural"}. Default is 3.}
}
  
\value{
  --For \code{contourLevels}/\code{contourBreaks}, for \code{kde} objects, returns a vector of heights. For \code{kda} objects, returns a list of vectors, one for each training group. For \code{kdde} objects, returns a matrix, one row for each partial derivative. 

  --For \code{contourSizes}, returns an approximation of the Lebesgue measure of the level set, i.e. length (d=1), area (d=2), volume (d=3), hyper-volume (d>4).
  
  --For \code{contourProbs}, returns an approximation of the probability measure of the level set. 
}

\details{
  --For \code{contourLevels}, the most straightforward is to specify \code{prob}.
  The heights of the corresponding highest density region with probability \code{prob} are
  computed. The \command{cont} parameter here is consistent with
  \command{cont} parameter from \command{plot.kde}, \command{plot.kdde}, and \command{plot.kda}
  i.e. \code{cont=(1-prob)*100\%}. 
  If both \code{prob} and \code{cont} are missing then a pretty set of
  \code{nlevels} contours are computed. From \pkg{ks} \eqn{\geq}{>=} 1.15.0, \code{nlevels=5} always.
 
  --For \code{contourSizes}, the length, area, volume etc. and for \code{contourProbs}, 
  the probability, are approximated by Riemann sums. These are rough approximations and
  depend highly on resolution of the estimation grid, and so should be interpreted carefully. 

  If \code{approx=FALSE}, then the exact KDE is computed. Otherwise
  it is interpolated from an existing KDE grid: this can dramatically
  reduce computation time for large data sets. 

  --For \code{contourBreaks}, different contour level selection methods are available, including equal length ("length"), quantile ("quantile") and natural breaks/Jenks ("natural") for density estimates only. The default \code{contourBreaks(type="density")} is equivalent to \code{contourLevels}.
}
 
\seealso{\code{\link{contour}}, \code{\link{contourLines}}}

\examples{
set.seed(8192)
x <- rmvnorm.mixt(n=1000, mus=c(0,0), Sigmas=diag(2), props=1)
fhat <- kde(x=x, binned=TRUE)
contourLevels(fhat, cont=c(25,50,75))
contourProbs(fhat, abs.cont=contourLevels(fhat, cont=50))
## compare approx prob with target prob=0.5
contourSizes(fhat, cont=25, approx=TRUE) 
## compare to approx circle of radius=0.75 with area=1.77
contourBreaks(fhat, n=3, type="natural")
## natural or Jenks contour levels 
}

\keyword{hplot}

\name{Hbcv}
\alias{Hbcv}
\alias{Hbcv.diag}
\title{Biased cross-validation (BCV) bandwidth matrix selector
  for bivariate data}

\description{
BCV bandwidth matrix for bivariate data.}

\usage{
Hbcv(x, whichbcv=1, Hstart, binned=FALSE, amise=FALSE, verbose=FALSE)
Hbcv.diag(x, whichbcv=1, Hstart, binned=FALSE, amise=FALSE, verbose=FALSE)
}

\arguments{
  \item{x}{matrix of data values}
  \item{whichbcv}{1 = BCV1, 2 = BCV2.  See details below.}
  \item{Hstart}{initial bandwidth matrix, used in numerical optimisation}
  \item{binned}{flag for binned kernel estimation. Default is FALSE.}
  \item{amise}{flag to return the minimal BCV value. Default is FALSE.}
  \item{verbose}{flag to print out progress information. Default is FALSE.}  
}

\value{
BCV bandwidth matrix. If \code{amise=TRUE} then the minimal BCV value is returned too. 
}

\references{Sain, S.R, Baggerly, K.A. & Scott, D.W. (1994)
  Cross-validation of multivariate densities. \emph{Journal of the
  American Statistical Association}, \bold{82}, 1131-1146.
}

\details{
  Use \code{Hbcv} for unconstrained bandwidth matrices and \code{Hbcv.diag}
  for diagonal bandwidth matrices. These selectors are only
  available for bivariate data. Two types of BCV criteria are
  considered here.  They are known as BCV1 and BCV2, from Sain, Baggerly
  & Scott (1994) and only differ slightly. These BCV
  surfaces can have multiple minima and so it can be quite difficult to
  locate the most appropriate minimum. Some times, there can be no local minimum at all so there
  may be no finite BCV selector.

  For details about the advanced options for \code{binned}, \code{Hstart}, see \code{\link{Hpi}}.
}

\seealso{\code{\link{Hlscv}}, \code{\link{Hpi}}, \code{\link{Hscv}}}

\examples{
data(unicef)
Hbcv(unicef)
Hbcv.diag(unicef)
}

\keyword{smooth}

\name{kdr}
\alias{kdr}
\alias{kdr.segment}
\alias{plot.kdr}
\title{Kernel density ridge estimation}

\description{
  Kernel density ridge estimation for 2- to 3-dimensional data.
}

\usage{
kdr(x, y, H, p=1, max.iter=400, tol.iter, segment=TRUE, k, kmax, min.seg.size=4,
    keep.path=FALSE, gridsize, xmin, xmax, binned, bgridsize, w, fhat,
    density.cutoff, pre=TRUE, verbose=FALSE) 
kdr.segment(x, k, kmax, min.seg.size, verbose=FALSE) 

\method{plot}{kdr}(x, ...)
}

\arguments{
  \item{x}{matrix of data values or an object of class \code{kdr}}
  \item{y}{matrix of initial values}
  \item{p}{dimension of density ridge}
  \item{H}{bandwidth matrix/scalar bandwidth. If missing,
  \code{Hpi(x,deriv,order=2)} is called by default.}
  \item{max.iter}{maximum number of iterations. Default is 400.}
  \item{tol.iter}{distance under which two successive iterations are
    considered convergent. Default is 0.001*min marginal IQR of \code{x}.}
  \item{segment}{flag to compute segments of density ridge. Default is TRUE.}
  \item{k}{number of segments to partition density ridge}
  \item{kmax}{maximum number of segments to partition density ridge. Default is 30.}
  \item{min.seg.size}{minimum number of points in a segment of a density
  ridge. Default is 4.}
  \item{keep.path}{flag to store the density gradient ascent
    paths. Default is FALSE.}
  \item{gridsize}{vector of number of grid points}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{binned}{flag for binned estimation.}
  \item{bgridsize}{vector of binning grid sizes}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{fhat}{kde of \code{x}. If missing \code{kde(x=x,w=w)} is
    executed.}
  \item{density.cutoff}{density threshold under which the \code{y} are
    excluded from the density ridge estimation. Default is
  \code{contourLevels(fhat, cont=99)}.}
  \item{pre}{flag for pre-scaling data. Default is TRUE.}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
  \item{...}{other graphics parameters}
}
  
\value{
  A kernel density ridge set is an object of class \code{kdr}
  which is a list with fields: 
  \item{x,y}{data points - same as input}
  \item{end.points}{matrix of final iterates starting from \code{y}}
  \item{H}{bandwidth matrix}
  \item{names}{variable names}
  \item{tol.iter,tol.clust,min.seg.size}{tuning parameter values -
    same as input}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names} 
  \item{w}{vector of weights}
  \item{path}{list of density gradient ascent paths where \code{path[[i]]} is
    the path of \code{y[i,]} (if \code{keep.path=TRUE})}
}

\details{ 
  Kernel density ridge estimation is based on reduced dimension kernel
  mean shift. See Ozertem & Erdogmus (2011). 

  If \code{y} is missing, then it defaults to the grid of size
  \code{gridsize} spanning from \code{xmin} to \code{xmax}.
  
  If the bandwidth \code{H} is missing, then
  the default bandwidth is the plug-in selector for the density gradient
  \code{Hpi(x,deriv.order=2)}. Any bandwidth that is suitable for the
  density Hessian is also suitable for the kernel density ridge. 
  
  \code{kdr(segment=TRUE)} or \code{kdr.segment()} carries out the segmentation of the density ridge points in \code{end.points}. If \code{k} is set, then \code{k} segments are created. If \code{k} is not set, then the optimal number of segments is chosen from 1:\code{kmax}, with \code{kmax=30} by default. The segments are created via a hierarchical clustering with single linkage. *Experimental: following the segmentation, the points within each segment are ordered to facilitate a line plot in \code{plot(type="l")}. The optimal ordering is not always achieved in this experimental implementation, though a scatterplot \code{plot(type="p")} always suffices, regardless of this ordering.*    
}

\references{
  Ozertem, U. & Erdogmus, D. (2011) Locally defined principal curves and
  surfaces, \emph{Journal of Machine Learning Research}, \bold{12}, 
  1249-1286. 
}  

\examples{
\donttest{data(cardio)
set.seed(8192)
cardio.train.ind <- sample(1:nrow(cardio), round(nrow(cardio)/4,0))
cardio2 <- cardio[cardio.train.ind,c(8,18)]
cardio.dr2 <- kdr(x=cardio2, gridsize=c(21,21))
## gridsize=c(21,21) is for illustrative purposes only
plot(cardio2, pch=16, col=3)
plot(cardio.dr2, cex=0.5, pch=16, col=6, add=TRUE)}

\dontrun{
cardio3 <- cardio[cardio.train.ind,c(8,18,11)]
cardio.dr3 <- kdr(x=cardio3)
plot(cardio.dr3, pch=16, col=6, xlim=c(10,90), ylim=c(70,180), zlim=c(0,40))
plot3D::points3D(cardio3[,1], cardio3[,2], cardio3[,3], pch=16, col=3, add=TRUE)

library(maps)
data(quake) 
quake <- quake[quake$prof==1,]  ## Pacific Ring of Fire 
quake$long[quake$long<0] <- quake$long[quake$long<0] + 360
quake <- quake[, c("long", "lat")]
data(plate)                     ## tectonic plate boundaries
plate <- plate[plate$long < -20 | plate$long > 20,]
plate$long[plate$long<0 & !is.na(plate$long)] <- plate$long[plate$long<0
& !is.na(plate$long)] + 360

quake.dr <- kdr(x=quake, xmin=c(70,-70), xmax=c(310, 80))
map(wrap=c(0,360), lty=2)
lines(plate[,1:2], col=4, lwd=2)
plot(quake.dr, type="p", cex=0.5, pch=16, col=6, add=TRUE)}
}

\keyword{cluster}

\name{vector}
\alias{vec}
\alias{vech}
\alias{invvec}
\alias{invvech}
\title{Vector and vector half operators}

\description{
The vec (vector) operator takes a \eqn{d \times d}{d x d} matrix and stacks the
columns into a single vector of length \eqn{d^2}{d^2}. The vech (vector
half) operator
takes a symmetric \eqn{d \times d}{d x d} matrix and stacks the lower
triangular half into a single vector of length \eqn{d(d+1)/2}{d(d+1)/2}.
The functions invvec and invvech are the inverses of vec and
vech i.e. they form matrices from vectors. 
}

\usage{
vec(x, byrow=FALSE)
vech(x)
invvec(x, ncol, nrow, byrow=FALSE)
invvech(x)
}

\arguments{
  \item{x}{vector or matrix}
  \item{ncol,nrow}{number of columns and rows for inverse of vech}
  \item{byrow}{flag for stacking row-wise or column-wise. Default is FALSE.}  
  }
  
\references{ Magnus, J.R. & Neudecker H.M. (2007) \emph{Matrix
  Differential Calculus with Applications in Statistics and
  Econometrics (3rd edition)}, Wiley & Sons. Chichester.
}

\examples{
x <- matrix(1:9, nrow=3, ncol=3)
vec(x)
invvec(vec(x))
}

\keyword{algebra}

\name{mixt}
\alias{rnorm.mixt}
\alias{dnorm.mixt}
\alias{rmvnorm.mixt}
\alias{dmvnorm.mixt}
\alias{rmvt.mixt}
\alias{dmvt.mixt}
\alias{mvnorm.mixt.mode}
\title{Normal and t-mixture distributions}

\description{
Random generation and density values from normal and t-mixture distributions.}

\usage{
dnorm.mixt(x, mus=0, sigmas=1, props=1)
rnorm.mixt(n=100, mus=0, sigmas=1, props=1, mixt.label=FALSE)
dmvnorm.mixt(x, mus, Sigmas, props=1, verbose=FALSE)
rmvnorm.mixt(n=100, mus=c(0,0), Sigmas=diag(2), props=1, mixt.label=FALSE)
rmvt.mixt(n=100, mus=c(0,0), Sigmas=diag(2), dfs=7, props=1)
dmvt.mixt(x, mus, Sigmas, dfs, props)
mvnorm.mixt.mode(mus, Sigmas, props=1, verbose=FALSE)
}

\arguments{
  \item{n}{number of random variates}
  \item{x}{matrix of quantiles}
  \item{mus}{(stacked) matrix of mean vectors (>1-d) or vector of means (1-d)}
  \item{Sigmas}{(stacked) matrix of variance matrices (>1-d)}
  \item{sigmas}{vector of standard deviations (1-d)}
  \item{props}{vector of mixing proportions}
  \item{mixt.label}{flag to output numeric label indicating mixture
    component. Default is FALSE.}
  \item{verbose}{flag to print out progress information. Default is FALSE.}
  \item{dfs}{vector of degrees of freedom}
}
  
\value{Normal and t-mixture random vectors and density values.}

\details{
  \code{rmvnorm.mixt} and \code{dmvnorm.mixt} are based on the
  \code{rmvnorm} and \code{dmvnorm} functions from the \pkg{mvtnorm}
  package. Likewise for \code{rmvt.mixt} and \code{dmvt.mixt}.

  For the normal mixture densities, \code{mvnorm.mixt.mode} computes the
  local modes: these are usually very close but not exactly equal to the
  component means.
}

\examples{
## univariate normal mixture
x <- rnorm.mixt(1000, mus=c(-1,1), sigmas=c(0.5, 0.5), props=c(1/2, 1/2))

## bivariate mixtures 
mus <- rbind(c(-1,0), c(1, 2/sqrt(3)), c(1,-2/sqrt(3)))
Sigmas <- 1/25*rbind(invvech(c(9, 63/10, 49/4)), invvech(c(9,0,49/4)), invvech(c(9,0,49/4)))
props <- c(3,3,1)/7
dfs <- c(7,3,2)
x <- rmvnorm.mixt(1000, mus=mus, Sigmas=Sigmas, props=props)
y <- rmvt.mixt(1000, mus=mus, Sigmas=Sigmas, dfs=dfs, props=props)

mvnorm.mixt.mode(mus=mus, Sigmas=Sigmas, props=props)
}

\keyword{distribution}

\name{kroc}
\alias{kroc}
\alias{predict.kroc}
\alias{summary.kroc}
\title{Kernel receiver operating characteristic (ROC) curve}

\description{ Kernel receiver operating characteristic (ROC) curve for 1- to 3-dimensional data.}

\usage{
kroc(x1, x2, H1, h1, hy, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points,
   binned, bgridsize, positive=FALSE, adj.positive, w, verbose=FALSE)

\method{predict}{kroc}(object, ..., x)
\method{summary}{kroc}(object, ...)
}

\arguments{
  \item{x,x1,x2}{vector/matrix of data values}
  \item{H1,h1,hy}{bandwidth matrix/scalar bandwidths. If these are
    missing, \code{Hpi.kcde}, \code{hpi.kcde} is called by default.} 
  \item{gridsize}{vector of number of grid points}
  \item{gridtype}{not yet implemented}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{not yet implemented}
  \item{binned}{flag for binned estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{positive}{flag if 1-d data are positive. Default is FALSE.}
  \item{adj.positive}{adjustment applied to positive 1-d data}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
  \item{object}{object of class \code{kroc}, output from \code{kroc}}
  \item{...}{other parameters} 
}

\value{
  A kernel ROC curve is an object of class \code{kroc} which is a list
  with fields:
  \item{x}{list of data values \code{x1, x2} - same as input}
  \item{eval.points}{vector or list of points at which the estimate is evaluated}
  \item{estimate}{ROC curve estimate at \code{eval.points}}
  \item{gridtype}{"linear"}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names}
  \item{w}{vector of weights}
  \item{tail}{"lower.tail"}
  \item{h1}{scalar bandwidth for first sample (1-d only)}
  \item{H1}{bandwidth matrix for first sample}
  \item{hy}{scalar bandwidth for ROC curve}
  \item{indices}{summary indices of ROC curve.}
}

\details{
  In this set-up, the values in the first sample \code{x1} should
  be larger in general that those in the second sample \code{x2}. The
  usual method for computing 1-d ROC curves is not valid for
  multivariate data. Duong (2014), 
  based on Lloyd (1998), develops an alternative formulation
  \eqn{(F_{Y_1}(z), F_{Y_2}(z))}{(F_Y1(z), F_Y2(z))} based on the
  cumulative distribution functions of \eqn{Y_j = \bar{F}_1(\bold{X}_j), j=1,2}{Yj=bar(F)_1(Xj), j=1,2}.

  If the bandwidth \code{H1} is missing from \code{kroc}, then
  the default bandwidth is the  plug-in selector
  \code{Hpi.kcde}. Likewise for missing \code{h1,hy}. A bandwidth matrix
  \code{H1} is required for \code{x1} for d>1, but the second bandwidth \code{hy} is always a scalar since \eqn{Y_j}{Yj} are 1-d variables.

  The effective support, binning, grid size, grid range, positive
  parameters are the same as \code{\link{kde}}. 
   
  --The \code{summary} method for \code{kroc} objects prints out the
  summary indices of the ROC curve, as contained in the \code{indices}
  field, namely the AUC (area under the curve) and Youden index.

}

\references{
  Duong, T. (2016) Non-parametric smoothed estimation of multivariate
  cumulative distribution and survival functions, and receiver operating
  characteristic curves. \emph{Journal of the Korean Statistical
  Society}, \bold{45}, 33-50.

  Lloyd, C. (1998) Using smoothed receiver operating curves to summarize
  and compare diagnostic systems. \emph{Journal of the American
  Statistical Association}, \bold{93}, 1356-1364.
}

\seealso{\code{\link{kcde}}, \code{\link{plot.kroc}}}
 
\examples{
samp <- 1000
x <- rnorm.mixt(n=samp, mus=0, sigmas=1, props=1)
y <- rnorm.mixt(n=samp, mus=0.5, sigmas=1, props=1)
Rhat <- kroc(x1=x, x2=y)
summary(Rhat)
predict(Rhat, x=0.5) 
}

\keyword{smooth}

\name{plot.kde.part}
\alias{plot.kde.part}
\alias{kms.part}
\alias{mvnorm.mixt.part}
\title{Partition plot for kernel density clustering}

\description{
  Plot of partition for kernel density clustering for 2-dimensional data.
}

\usage{
mvnorm.mixt.part(mus, Sigmas, props=1, xmin, xmax, gridsize, max.iter=100,
   verbose=FALSE)
kms.part(x, H, xmin, xmax, gridsize, verbose=FALSE, ...)

\method{plot}{kde.part}(x, display="filled.contour", col, col.fun, alpha=1, add=FALSE, ...)
}

\arguments{
  \item{mus}{(stacked) matrix of mean vectors}
  \item{Sigmas}{(stacked) matrix of variance matrices}
  \item{props}{vector of mixing proportions}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{gridsize}{vector of number of grid points}
   \item{max.iter}{maximum number of iterations}
  \item{verbose}{flag to print out progress information. Default is FALSE.}
  \item{x}{matrix of data values or an object of class \code{kde.part}}
  \item{H}{bandwidth matrix. If missing,
    \code{Hpi(x,deriv,order=1)} is called by default.}
  \item{display}{type of display, "filled.contour" for filled contour plot} 
  \item{col,col.fun}{vector of plotting colours or colour function}
  \item{alpha}{colour transparency. Default is 1.}
  \item{add}{flag to add to current plot. Default is FALSE.}
  \item{...}{other parameters}
}
  
\value{
  A kernel partition is an object of class \code{kde.part} which is a
  list with fields:
  \item{x}{data points - same as input}
  \item{eval.points}{vector or list of points at which the estimate is evaluated}
  \item{estimate}{density estimate at \code{eval.points}}
  \item{H}{bandwidth matrix}
  \item{gridtype}{"linear"}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names} 
  \item{w}{vector of weights}
  \item{cont}{vector of probability contour levels}   
  \item{end.points}{matrix of final iterates starting from \code{x}}
  \item{label}{vector of cluster labels}
  \item{mode}{matrix of cluster modes}
  \item{nclust}{number of clusters}
  \item{nclust.table}{frequency table of cluster labels}
  \item{tol.iter,tol.clust,min.clust.size}{tuning parameter values -
    same as input}
  
  Plot is sent to graphics window. 
}

\details{
  For 2-d data, \code{kms.part} and \code{mvnorm.mixt.part} produce a
  \code{kde.part} object whose 
  values are the class labels, rather than probability density values.   
}

\seealso{\code{\link{plot.kde}}, \code{\link{kms}}}

\examples{
## normal mixture partition
mus <- rbind(c(-1,0), c(1, 2/sqrt(3)), c(1,-2/sqrt(3)))
Sigmas <- 1/25*rbind(invvech(c(9, 63/10, 49/4)), invvech(c(9,0,49/4)), invvech(c(9,0,49/4)))
props <- c(3,3,1)/7
gridsize <- c(11,11) ## small gridsize illustrative purposes only 
nmixt.part <- mvnorm.mixt.part(mus=mus, Sigmas=Sigmas, props=props, gridsize=gridsize)
plot(nmixt.part, asp=1, xlim=c(-3,3), ylim=c(-3,3), alpha=0.5)

## kernel mean shift partition
\donttest{set.seed(81928192)
x <- rmvnorm.mixt(n=10000, mus=mus, Sigmas=Sigmas, props=props)
msize <- round(prod(gridsize)*0.1)
kms.nmixt.part <- kms.part(x=x, min.clust.size=msize, gridsize=gridsize)
plot(kms.nmixt.part, asp=1, xlim=c(-3,3), ylim=c(-3,3), alpha=0.5)}}

\keyword{hplot}

\name{kms}
\alias{kms}
\alias{summary.kms}
\alias{plot.kms}
\title{Kernel mean shift clustering}

\description{
  Kernel mean shift clustering for 2- to 6-dimensional data.
}

\usage{
kms(x, y, H, max.iter=400, tol.iter, tol.clust, min.clust.size, merge=TRUE,
    keep.path=FALSE, verbose=FALSE)

\method{plot}{kms}(x, display="splom", col, col.fun, alpha=1, xlab, ylab, zlab, theta=-30, 
    phi=40, add=FALSE, ...)
\method{summary}{kms}(object, ...)
}

\arguments{
  \item{x}{matrix of data values or object of class \code{kms}}
  \item{y}{matrix of candidate data values for which the mean shift will
    estimate their cluster labels. If missing, \code{y=x}.}
  \item{H}{bandwidth matrix/scalar bandwidth. If missing,
  \code{Hpi(x,deriv.order=1,nstage=2-(d>2))} is called by default.}
  \item{max.iter}{maximum number of iterations. Default is 400.}
  \item{tol.iter}{distance under which two successive iterations are
    considered convergent. Default is 0.001*min marginal IQR of \code{x}.}
  \item{tol.clust}{distance under which two cluster modes are considered
    to form one cluster. Default is 0.01*max marginal IQR of \code{x}.}
  \item{min.clust.size}{minimum cluster size (cardinality). Default is \code{0.01*nrow(y)}.}
  \item{merge}{flag to merge clusters which are smaller than
    \code{min.clust.size}. Default is TRUE.}
  \item{keep.path}{flag to store the density gradient ascent paths. Default is FALSE.}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
  \item{object}{object of class \code{kms}}
  \item{display}{type of display, "splom" (>=2-d) or "plot3D" (3-d)}
  \item{col,col.fun}{vector or colours (one for each group) or colour function}
  \item{alpha}{colour transparency. Default is 1.}
  \item{xlab,ylab,zlab}{axes labels}
  \item{theta,phi}{graphics parameters for perspective plots (3-d)}
  \item{add}{flag to add to current plot. Default is FALSE.}
  \item{...}{other (graphics) parameters}
}
  
\value{
  A kernel mean shift clusters set is an object of class \code{kms}
  which is a list with fields: 
  \item{x,y}{data points - same as input}
  \item{end.points}{matrix of final iterates starting from \code{y}}
  \item{H}{bandwidth matrix}
  \item{label}{vector of cluster labels}
  \item{nclust}{number of clusters}
  \item{nclust.table}{frequency table of cluster labels}
  \item{mode}{matrix of cluster modes}
  \item{names}{variable names}
  \item{tol.iter,tol.clust,min.clust.size}{tuning parameter values -
  same as input}
  \item{path}{list of density gradient ascent paths where \code{path[[i]]} is
    the path of \code{y[i,]} (if \code{keep.path=TRUE})}
}

\details{ 
  Mean shift clustering belongs to the class of modal or density-based
  clustering methods.
  The mean shift recurrence of the candidate point \eqn{{\bold x}}{x} is
  \eqn{{\bold x}_{j+1} = {\bold x}_j + \bold{{\rm H}} {\sf D} \hat{f}({\bold
      x}_j)/\hat{f}({\bold x}_j)}{x_j+1 = x_j + H D hat(f)(x_j)/hat(f)(x_j)}
  where \eqn{j\geq 0}{j>=0} and \eqn{{\bold x}_0 = {\bold x}}{x_0 = x}.
  The sequence \eqn{\{{\bold x}_0, {\bold
  x}_1, \dots \}}{x_0, x_1, ...} follows the density gradient ascent
  paths to converge to a local mode of the
  density estimate \eqn{\hat{f}}{hat(f)}. Hence \eqn{{\bold x}}{x} is
  iterated until it converges to its local mode, and this determines its
  cluster label.  

  The mean shift recurrence is terminated if successive iterations are
  less than \code{tol.iter} or the maximum number of iterations
  \code{max.iter} is reached. Final iterates which are less than
  \code{tol.clust} distance apart are considered to form a single
  cluster. If \code{merge=TRUE} then the clusters whose cardinality is less
  than \code{min.clust.size} are iteratively merged with their nearest cluster.   
  
  If the bandwidth \code{H} is missing, then
  the default bandwidth is the plug-in selector for the density gradient
  \code{Hpi(x,deriv.order=1)}. Any bandwidth that is suitable for the
  density gradient is also suitable for the mean shift. 
}

\references{
  Chacon, J.E. & Duong, T. (2013) Data-driven density estimation, with
  applications to nonparametric clustering and bump hunting. \emph{Electronic
    Journal of Statistics}, \bold{7}, 499-532.

  Comaniciu, D. & Meer, P. (2002). Mean shift: a robust approach
  toward feature space analysis. \emph{ IEEE Transactions on Pattern
  Analysis and Machine Intelligence}, \bold{24}, 603-619.
}  

\seealso{\code{\link{kde}}}

\examples{
data(crabs, package="MASS")
kms.crabs <- kms(x=crabs[,c("FL","CW")])
plot(kms.crabs, pch=16)
summary(kms.crabs)

kms.crabs <- kms(x=crabs[,c("FL","CW","RW")])
plot(kms.crabs, pch=16)
\donttest{plot(kms.crabs, display="plot3D", pch=16)}
}

\keyword{cluster}

\name{plot.histde}
\alias{plot.histde}
\title{Plot for histogram density estimate}

\description{
  Plot for histogram density estimate for 1- and 2-dimensional data.
}

\usage{
\method{plot}{histde}(x, ...)
}

\arguments{
  \item{x}{object of class \code{histde} (output from \code{\link{histde}})}
  \item{...}{other graphics parameters:
    \describe{
      \item{\code{col}}{plotting colour for density estimate}
      \item{\code{col.fun}}{plotting colour function for levels}
      \item{\code{col.pt}}{plotting colour for data points}
      \item{\code{jitter}}{flag to jitter rug plot (1-d). Default is TRUE.}
      \item{\code{xlim,ylim}}{axes limits}
      \item{\code{xlab,ylab}}{axes labels}
      \item{\code{add}}{flag to add to current plot. Default is FALSE.}
      \item{\code{drawpoints}}{flag to draw data points on density estimate. Default is FALSE.}
      \item{\code{breaks}}{vector of break values of density estimate. Default is an \code{nbreaks} equilinear sequence over the data range.}
      \item{\code{nbreaks}}{number of breaks in \code{breaks} sequence}
      \item{\code{lty.rect},\code{lwd.rect}}{line type/width for histogram box lines (2-d)}
      \item{\code{border}}{colour of histogram box lines (2-d)}
      \item{\code{col.rect}}{colour of histogram bars (1-d)}
      \item{\code{add.grid}}{flag to add histogram grid (2-d). Default is TRUE.}
    }
  }
}
  
\value{
  Plots for 1-d and 2-d are sent to graphics window. 
}

\details{
  For \code{histde} objects, the function headers for the different dimensional data are  
  \preformatted{
  ## univariate
  plot(fhat, xlab, ylab="Density function", add=FALSE, drawpoints=FALSE,
     col.pt=4, jitter=FALSE, border=1, alpha=1, ...) 
  
  ## bivariate
  plot(fhat, breaks, nbreaks=11, xlab, ylab, zlab="Density function", cex=1, 
     pch=1, add=FALSE, drawpoints=FALSE, col, col.fun, alpha=1, col.pt=4,
     lty.rect=2, cex.text=1, border, lwd.rect=1, col.rect="transparent",
     add.grid=TRUE, ...)}

  The 1-d plot is a standard plot of a histogram generated by \code{hist}. If
  \code{drawpoints=TRUE} then a rug plot is added.

  The 2-d plot is similar to the \code{display="filled.contour"} option from
  \code{\link{plot.kde}} with the default \code{nbreaks=11} contour
  levels. 
}

\seealso{\code{\link{plot.kde}}}
 
\examples{
## univariate example
data(iris)
fhat <- histde(x=iris[,2])
plot(fhat, xlab="Sepal length")

## bivariate example
fhat <- histde(x=iris[,2:3])
plot(fhat, drawpoints=TRUE)
box()
}

\keyword{hplot}

\name{binning}
\alias{binning}
\title{Linear binning for multivariate data}

\description{
  Linear binning for 1- to 4-dimensional data.
}

\usage{
binning(x, H, h, bgridsize, xmin, xmax, supp=3.7, w, gridtype="linear")
}

\arguments{
  \item{x}{matrix of data values}
  \item{H,h}{bandwidth matrix, scalar bandwidth}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal is [-supp,supp]}
  \item{bgridsize}{vector of binning grid sizes}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{gridtype}{not yet implemented}
}
  
\value{
  Returns a list with 2 fields
  \item{counts}{linear binning counts}
  \item{eval.points}{vector (d=1) or list (d>=2) of grid points in each
	dimension }
}

\details{
  For \pkg{ks} \eqn{\geq}{>=} 1.10.0, binning is available for unconstrained
  (non-diagonal) bandwidth matrices. Code is used courtesy of A. &
  J. Gramacki, and M.P. Wand. Default \code{bgridsize} are 
  d=1: 401; d=2: rep(151, 2); d=3: rep(51, 3); d=4: rep(21, 4). 
}

\references{
  Gramacki, A. & Gramacki, J. (2016) FFT-based fast computation of
  multivariate kernel estimators with unconstrained bandwidth
  matrices. \emph{Journal of Computational & Graphical Statistics},
  \bold{26}, 459-462.

  Wand, M.P. & Jones, M.C. (1995) \emph{Kernel Smoothing}.
  Chapman & Hall. London.
} 

\examples{
data(unicef)
ubinned <- binning(x=unicef)
}

\keyword{algebra}

\name{vkde}
\alias{kde.balloon}
\alias{kde.sp}
\title{Variable kernel density estimate.}

\description{
  Variable kernel density estimate for 2-dimensional data.
}

\usage{
kde.balloon(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, 
   binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE)
kde.sp(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, 
   binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE, verbose=FALSE)
}

\arguments{
  \item{x}{matrix of data values}
  \item{H}{bandwidth matrix. If this missing, \code{Hns}
  is called by default.}
  \item{h}{not yet implemented}
  \item{gridsize}{vector of number of grid points}
  \item{gridtype}{not yet implemented}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{vector or matrix of points at which estimate is evaluated}
  \item{binned}{flag for binned estimation.}
  \item{bgridsize}{vector of binning grid sizes}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{compute.cont}{flag for computing 1\% to 99\% probability contour levels. Default is TRUE.}
  \item{approx.cont}{flag for computing approximate probability contour
    levels. Default is TRUE.}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
}
  
\value{
  A variable kernel density estimate for bounded data is an object of class \code{kde}.    
}

\details{
  The balloon density estimate \code{kde.balloon} employs bandwidths
  which vary at each 
  estimation point (Loftsgaarden & Quesenberry, 1965). There are as many bandwidths as there are estimation
  grid points. The default bandwidth is \code{Hns(deriv.order=2)} and
  the subsequent bandwidths are derived via a minimal MSE formula.  

  The sample point density estimate \code{kde.sp} employs bandwidths
  which vary for each data point (Abramson, 1982).
  There are as many bandwidths as there are data
  points. The default bandwidth is \code{Hns(deriv.order=4)} and the 
  subsequent bandwidths are derived via the Abramson formula.
}

\references{
  Abramson, I. S. (1982) On bandwidth variation in kernel estimates - a
  square root law. \emph{Annals of Statistics}, \bold{10}, 1217-1223.

  Loftsgaarden, D. O. & Quesenberry, C. P. (1965) A nonparametric
  estimate of a multivariate density function. \emph{Annals of
  Mathematical Statistics}, \bold{36}, 1049-1051.  
}

\seealso{\code{\link{kde}}, \code{\link{plot.kde}}}

\examples{
\donttest{data(worldbank)
wb <- as.matrix(na.omit(worldbank[,4:5]))
xmin <- c(-70,-35); xmax <- c(35,70)
fhat <- kde(x=wb, xmin=xmin, xmax=xmax)
fhat.sp <- kde.sp(x=wb, xmin=xmin, xmax=xmax)
zmax <- max(fhat.sp$estimate)
plot(fhat, display="persp", box=TRUE, phi=20, thin=1, border=grey(0,0.2), zlim=c(0,zmax))
plot(fhat.sp, display="persp", box=TRUE, phi=20, thin=1, border=grey(0,0.2), zlim=c(0,zmax))}
\dontrun{
fhat.ball <- kde.balloon(x=wb, xmin=xmin, xmax=xmax)
plot(fhat.ball, display="persp", box=TRUE, phi=20, zlim=c(0,zmax))}
}

\keyword{smooth}

\name{kcopula}
\alias{kcopula}
\alias{kcopula.de}
\title{Kernel copula (density) estimate}

\description{
  Kernel copula and copula density estimator for 2-dimensional data.
}

\usage{
kcopula(x, H, hs, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points,
  binned, bgridsize, w, marginal="kernel", verbose=FALSE)
kcopula.de(x, H, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, 
  binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE,
  marginal="kernel", boundary.supp, boundary.kernel="beta", verbose=FALSE)
}

\arguments{
  \item{x}{matrix of data values}
  \item{H,hs}{bandwidth matrix. If these are missing, \code{Hpi.kcde}/\code{Hpi} or \code{hpi.kcde}/\code{hpi} is called by default.}
  \item{gridsize}{vector of number of grid points}
  \item{gridtype}{not yet implemented}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{matrix of points at which estimate is evaluated}
  \item{binned}{flag for binned estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{marginal}{"kernel" = kernel cdf or "empirical" = empirical cdf
    to calculate pseudo-uniform values. Default is "kernel".}
  \item{compute.cont}{flag for computing 1\% to 99\% probability contour levels. Default is TRUE.}
  \item{approx.cont}{flag for computing approximate probability contour
    levels. Default is TRUE.}
  \item{boundary.supp}{effective support for boundary region}
  \item{boundary.kernel}{"beta" = beta boundary kernel, "linear" = linear boundary kernel}
 \item{verbose}{flag to print out progress information. Default is FALSE.}  
}
  
\value{
  A kernel copula estimate, output from \code{kcopula}, is an object of
  class \code{kcopula}. A kernel copula density estimate, output from
  \code{kcopula.de}, is an object of class \code{kde}. These two classes
  of objects have the same fields as \code{kcde} and \code{kde} objects
  respectively, except for
  \item{x}{pseudo-uniform data points}
  \item{x.orig}{data points - same as input}
  \item{marginal}{marginal function used to compute pseudo-uniform data}
  \item{boundary}{flag for data points in the boundary region
  (\code{kcopula.de} only)}
}

\details{
  For kernel copula estimates, a transformation approach is used to
  account for the boundary effects. If \code{H} is missing, the default
  is \code{Hpi.kcde}; if \code{hs} are missing, the default is
  \code{hpi.kcde}.  
  
  For kernel copula density estimates, for those points which are in
  the interior region, the usual kernel density estimator
  (\code{\link{kde}}) is used. For those points in the boundary region,
  a product beta kernel based on the boundary corrected univariate beta
  kernel of Chen (1999) is used (\code{\link{kde.boundary}}). If \code{H}
  is missing, the default is \code{Hpi.kcde}; if \code{hs} are missing,
  the default is \code{hpi}.
  
  The effective support, binning, grid size, grid range parameters are
  the same  as for \code{\link{kde}}.  
  
}
 
\references{
  Duong, T. (2014) Optimal data-based smoothing for non-parametric
  estimation of copula functions and their densities. Submitted.

  Chen, S.X. (1999). Beta kernel estimator for density
  functions. \emph{Computational Statistics & Data Analysis},
  \bold{31}, 131--145.
}

\seealso{\code{\link{kcde}}, \code{\link{kde}}}

\examples{
data(fgl, package="MASS")
x <- fgl[,c("RI", "Na")]
Chat <- kcopula(x=x)
plot(Chat, display="filled.contour", lwd=1)
plot(Chat, display="persp", border=1)
}

\keyword{smooth}

\name{kde.truncate}
\alias{kde.truncate}
\alias{kdde.truncate}
\title{Truncated kernel density derivative estimate}

\description{
  Truncated kernel density derivative estimate for 2-dimensional data.
}

\usage{
kde.truncate(fhat, boundary) 
kdde.truncate(fhat, boundary) 
}

\arguments{
  \item{fhat}{object of class \code{kde} or \code{kdde}}
  \item{boundary}{two column matrix delimiting the boundary for truncation}
}
  
\value{
  A truncated kernel density (derivative) estimate inherits the same
  object class as the input estimate. 
}

\details{
  A simple truncation is performed on the kernel estimator. All the
  points in the estimation grid which are outside of the regions
  delimited by \code{boundary} are set to 0, and their probability
  mass is distributed proportionally to the remaining density (derivative) values.
}

\seealso{\code{\link{kde}}, \code{\link{kdde}}}

\examples{
data(worldbank)
wb <- as.matrix(na.omit(worldbank[,c("internet", "ag.value")]))
fhat <- kde(x=wb)
rectb <- cbind(x=c(0,100,100,0,0), y=c(0,0,100,100,0))
fhat.b <- kde.truncate(fhat, boundary=rectb)
plot(fhat, col=1, xlim=c(0,100), ylim=c(0,100))
plot(fhat.b, add=TRUE, col=4)
rect(0,0,100,100, lty=2)

library(oz)
data(grevillea)
wa.coast <- ozRegion(section=1)
wa.polygon <- cbind(wa.coast$lines[[1]]$x, wa.coast$lines[[1]]$y)
fhat1 <- kdde(x=grevillea, deriv.order=1)
fhat1 <- kdde.truncate(fhat1, wa.polygon)
oz(section=1, xlim=c(113,122), ylim=c(-36,-29))
plot(fhat1, add=TRUE, display="filled.contour")
}

\keyword{smooth}

\name{quake}
\docType{data}
\alias{quake}
\alias{quakesf}
\alias{plate}
\alias{platesf}
\title{Geographical locations of earthquakes and tectonic plates}

\description{
The \code{quake} data set contains the geographical locations 
of severe earthquakes in the years 100 and 2016 inclusive. The
\code{plate} data set contains the geographical locations of the tectonic
plate boundaries.    
}

\usage{data(quake)
data(plate)
data(quakesf)
data(platesf)
}

\format{--For \code{quake}, a matrix with 5871 rows and 5 columns.
  Each row corresponds to an earthquake.
  The first column is the year (negative years indicate B.C.E.),
  the second is the longitude (decimal degrees),
  the third is the latitude (decimal degrees),
  the fourth is the depth beneath the surface of the Earth (km),
  the fifth is a flag for the location inside the circum-Pacific belt
  (aka Pacific Ring of Fire). \code{quakesf} is a WGS84 \code{sf} version with a point geometry.

  --For \code{plate}, a matrix with 6276 rows and 3 columns.
  Each row corresponds to an location of the tectonic plate boundaries.
  The first is the longitude,
  the second is the latitude,
  the third is the label of the tectonic plate. 
  \code{platesf} is a WGS84 \code{sf} spatial version with a multipolygon geometry, where the individual plate line segments have been merged into a single multipolygon.} 

\source{
Alhenius, H., Nordpil and Bird, P. (2014). \href{https://github.com/fraxen/tectonicplates}{World Tectonic Plates and Boundaries}. Accessed 2021-03-11.

Bird, P. (2003) An updated digital model of plate boundaries,
\emph{Geochemistry, Geophysics, Geosystems} \bold{4(3)}, 1-52. 1027.
\href{http://peterbird.name/publications/2003_PB2002/2003_PB2002.htm}{Data set} accessed 2016-03-24.
  
NGDC/WDS (2017) Global significant earthquake database, National
  Geophysical Data Center, NOAA, \doi{10.7289/V5TD9V7K}.
National Geophysical Data Center/World Data Service. Accessed 2017-03-30.
}

\keyword{datasets}

\name{kde}
\alias{kde}
\alias{predict.kde}
\title{Kernel density estimate}

\description{
  Kernel density estimate for 1- to 6-dimensional data.
}

\usage{
kde(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, 
    bgridsize, positive=FALSE, adj.positive, w, compute.cont=TRUE, 
    approx.cont=TRUE, unit.interval=FALSE, density=FALSE, verbose=FALSE)

\method{predict}{kde}(object, ..., x, zero.flag=TRUE)
}

\arguments{
  \item{x}{matrix of data values}
  \item{H,h}{bandwidth matrix/scalar bandwidth. If these are missing, \code{Hpi} or \code{hpi} is called by default.}
  \item{gridsize}{vector of number of grid points}
  \item{gridtype}{not yet implemented}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{vector or matrix of points at which estimate is evaluated}
  \item{binned}{flag for binned estimation.}
  \item{bgridsize}{vector of binning grid sizes}
  \item{positive}{flag if data are positive (1-d, 2-d). Default is FALSE.}
  \item{adj.positive}{adjustment applied to positive 1-d data}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{compute.cont}{flag for computing 1\% to 99\% probability contour levels. Default is TRUE.}
  \item{approx.cont}{flag for computing approximate probability contour
    levels. Default is TRUE.}
  \item{unit.interval}{flag for computing log transformation KDE on 1-d data bounded on unit interval [0,1]. Default is FALSE.} 
  \item{density}{flag if density estimate values are forced to be non-negative function. Default is FALSE.} 
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
  \item{object}{object of class \code{kde}}
  \item{zero.flag}{deprecated (retained for backwards compatibilty)}
  \item{...}{other parameters}
}

\value{
  A kernel density estimate is an object of class \code{kde} which is a
  list with fields: 
  \item{x}{data points - same as input}
  \item{eval.points}{vector or list of points at which the estimate is evaluated}
  \item{estimate}{density estimate at \code{eval.points}}
  \item{h}{scalar bandwidth (1-d only)}
  \item{H}{bandwidth matrix}
  \item{gridtype}{"linear"}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names} 
  \item{w}{vector of weights}
  \item{cont}{vector of probability contour levels}    
}

\details{
  For d=1, if \code{h} is missing, the default bandwidth is \code{hpi}.
  For d>1, if \code{H} is missing, the default is \code{Hpi}.

  For d=1, if \code{positive=TRUE} then \code{x} is transformed to
  \code{log(x+adj.positive)} where the default \code{adj.positive} is
  the minimum of \code{x}. This is known as a log transformation density
  estimate. If \code{unit.interval=TRUE} then \code{x} is transformed to 
  \code{qnorm(x)}. See \code{\link{kde.boundary}} for boundary kernel density estimates, as these tend to be more robust than transformation density estimates. 
  
  For d=1, 2, 3, and if \code{eval.points} is not specified, then the
  density estimate is computed over a grid 
  defined by \code{gridsize} (if \code{binned=FALSE}) or
  by \code{bgridsize} (if \code{binned=TRUE}). This form is suitable for
  visualisation in conjunction with the \code{plot} method.

  For d=4, 5, 6,  and if \code{eval.points} is not specified, then the
  density estimate is computed over a grid defined by \code{gridsize}. 
  
  If \code{eval.points} is specified, as a vector (d=1) or 
  as a matrix (d=2, 3, 4), then the density estimate is computed at
  \code{eval.points}. This form is suitable for numerical summaries
  (e.g. maximum likelihood), and is not compatible with the \code{plot}
  method. Despite that the density estimate is returned only at
  \code{eval.points}, by default, a binned gridded estimate is
  calculated first and then the density estimate at \code{eval.points}
  is computed using the \code{predict} method. If this default intermediate
  binned grid estimate is not required, then set \code{binned=FALSE} to
  compute directly the exact density estimate at \code{eval.points}.   

  Binned kernel estimation is an approximation to the exact kernel
  estimation and is available for d=1, 2, 3, 4. This makes
  kernel estimators feasible for large samples. The default value of the
  binning flag \code{binned} is n>1 (d=1), n>500 (d=2), n>1000 (d>=3). 
  Some times binned estimation leads to negative density values: if non-negative
  values are required, then set \code{density=TRUE}. 
  
  The default \code{bgridsize,gridsize} are d=1: 401; d=2: rep(151, 2); d=3: 
  rep(51, 3); d=4: rep(21, 4). 
  
  The effective support for a normal kernel is where  
  all values outside \code{[-supp,supp]^d} are zero. 

  The default \code{xmin} is \code{min(x)-Hmax*supp} and \code{xmax}
  is \code{max(x)+Hmax*supp}  where \code{Hmax} is the maximum of the
  diagonal elements of \code{H}. The grid produced is the outer
  product of \code{c(xmin[1], xmax[1])}, ..., \code{c(xmin[d], xmax[d])}. 
  For \pkg{ks} \eqn{\geq}{>=} 1.14.0, when \code{binned=TRUE} and \code{xmin,xmax} 
  are not missing, the data values \code{x} are clipped to the estimation grid 
  delimited by \code{xmin,xmax} to prevent potential memory leaks. 
}

\seealso{\code{\link{plot.kde}}, \code{\link{kde.boundary}}}

\examples{
## unit interval data 
set.seed(8192)             
fhat <- kde(runif(10000,0,1), unit.interval=TRUE)
plot(fhat, ylim=c(0,1.2))

## positive data 
data(worldbank)
wb <- as.matrix(na.omit(worldbank[,2:3]))
wb[,2] <- wb[,2]/1000
fhat <- kde(x=wb)
fhat.trans <- kde(x=wb, adj.positive=c(0,0), positive=TRUE)
plot(fhat, col=1, xlim=c(0,20), ylim=c(0,80))
plot(fhat.trans, add=TRUE, col=2)
rect(0,0,100,100, lty=2)

## large data on non-default grid
## 151 x 151 grid = [-5,-4.933,..,5] x [-5,-4.933,..,5]
set.seed(8192)
x <- rmvnorm.mixt(10000, mus=c(0,0), Sigmas=invvech(c(1,0.8,1)))
fhat <- kde(x=x, compute.cont=TRUE, xmin=c(-5,-5), xmax=c(5,5), bgridsize=c(151,151))
plot(fhat)

## See other examples in ? plot.kde
}

\keyword{smooth}

\name{tempb}
\docType{data}
\alias{tempb}
\title{Daily temperature}

\description{
  This data set contains the daily minimum and maximum temperatures from
  the weather station in Badajoz, Spain, from 1 January 1955 to 31 December
  2015.  
}

\usage{data(tempb)}

\format{A matrix with  21908 rows and 5 columns.
  Each row corresponds to a daily measurement. 
  The first column is the year (yyyy), 
  the second is the month (mm),
  the third is the day (dd), 
  the fourth is the minimum temperature (degrees Celsius),
  the fifth is the maximum temperature (degrees Celsius).
}

\source{
Menne, M. J., Durre, I., Vose, R. S., Gleason, B. E. & Houston,
T. (2012) An overview of the global historical climatology network-daily
database, \emph{Journal of Atmospheric and Oceanic Technology}
\bold{429}, 897 - 910. \href{https://climexp.knmi.nl/selectdailyseries.cgi}{KNMI climate explorer}. Accessed 2016-10-20.
}

\keyword{datasets}

\name{plot.kde.loctest}
\alias{plot.kde.loctest}
\title{Plot for kernel local significant difference regions}

\description{
  Plot for kernel local significant difference regions for 1- to 3-dimensional data. 
}

\usage{
\method{plot}{kde.loctest}(x, ...)
}

\arguments{
  \item{x}{object of class \code{kde.loctest} (output from \code{\link{kde.local.test}})}
  \item{...}{other graphics parameters:
    \describe{
      \item{\code{lcol}}{colour for KDE curve (1-d)}
      \item{\code{col}}{vector of 2 colours. First colour: sample 1>sample 2, second colour: sample 1<sample2.}
      \item{\code{add}}{flag to add to current plot. Default is FALSE.}
      \item{\code{rugsize}}{height of rug-like plot (1-d)}
      \item{\code{add.legend}}{flag to add legend. Default is TRUE.}
      \item{\code{pos.legend}}{position label for legend (1-d, 2-d)}
      \item{\code{alphavec}}{vector of transparency values for contour (3-d)} 
    }
    
    and those used in \code{\link{plot.kde}}
  }
} 

\value{
  Plots for 1-d and 2-d are sent to graphics window. Plot for 3-d is
  sent to graphics/RGL window. 
}

\details{
  For \code{kde.loctest} objects, the function headers are  
  \preformatted{
   ## univariate
   plot(x, lcol, col, add=FALSE, xlab="x", ylab, rugsize, add.legend=TRUE, 
        pos.legend="topright", alpha=1, ...)
   
   ## bivariate
   plot(x, col, add=FALSE, add.legend=TRUE, pos.legend="topright", alpha=1, 
        ...)

   ## trivariate 
   plot(x, col, color, add=FALSE, box=TRUE, axes=TRUE, alphavec=c(0.5, 0.5), 
        add.legend=TRUE, ...)}
}
 
\seealso{\code{\link{kde.local.test}}}

\examples{
## bivariate
data(air)
air.var <- c("co2","pm10","no")
air <- air[, c("date","time",air.var)]
air2 <- reshape(air, idvar="date", timevar="time", direction="wide")
a1 <- as.matrix(na.omit(air2[, paste0(air.var, ".08:00")]))
a2 <- as.matrix(na.omit(air2[, paste0(air.var, ".20:00")]))
colnames(a1) <- air.var
colnames(a2) <- air.var
air08 <- a1[,c("co2","pm10")]
air20 <- a2[,c("co2","pm10")]

loct <- kde.local.test(x1=air08, x2=air20)
plot(loct, lwd=1)

\donttest{## significant curvature regions
air20.fs <- kfs(air20)
plot(air20.fs, add=TRUE)

## trivariate
air08 <- a1; air20 <- a2
loct <- kde.local.test(x1=air08, x2=air20)
plot(loct, xlim=c(0,800), ylim=c(0,300), zlim=c(0,300))}
}

\keyword{hplot}

\name{ks-package}
\alias{ks}
\alias{ks-package}
\docType{package}
\title{ks}

\description{
  Kernel smoothing for data from 1- to 6-dimensions.
}

\details{
  There are three main types of functions in this package:
  \itemize{
    \item computing kernel estimators - these function names begin with `k'
    \item computing bandwidth selectors - these begin with `h' (1-d) or
    `H' (>1-d)
    \item displaying kernel estimators - these begin with `plot'.
  }
  The kernel used throughout is the normal (Gaussian) kernel \eqn{K}{K}.
  For 1-d data, the bandwidth \eqn{h}{h} is the standard deviation of
  the normal kernel, whereas for multivariate data, the bandwidth matrix 
  \eqn{\bold{{\rm H}}}{H} is the variance matrix.
 
  --For kernel density estimation, \code{\link{kde}} computes  
  \deqn{\hat{f}(\bold{x}) = n^{-1} \sum_{i=1}^n K_{\bold{{\rm H}}} (\bold{x} - \bold{X}_i).}{hat(f)(x) = n^(-1) sum_i K_H (x - X_i).} 

  The bandwidth matrix \eqn{\bold{{\rm H}}}{H} is a matrix of smoothing
  parameters and its choice is crucial for the performance of kernel
  estimators. For display, its \code{plot} method calls \code{\link{plot.kde}}.
    
  --For kernel density estimation, there are several varieties of bandwidth selectors 
   \itemize{
  \item plug-in \code{\link{hpi}} (1-d); 
  \code{\link{Hpi}}, \code{\link{Hpi.diag}} (2- to 6-d) 
  \item least squares (or unbiased) cross validation (LSCV or UCV) \code{\link{hlscv}} (1-d);
  \code{\link{Hlscv}}, \code{\link{Hlscv.diag}} (2- to 6-d) 
  \item biased cross validation (BCV) 
  \code{\link{Hbcv}}, \code{\link{Hbcv.diag}} (2- to 6-d) 
  \item smoothed cross validation (SCV) \code{\link{hscv}} (1-d);
  \code{\link{Hscv}}, \code{\link{Hscv.diag}} (2- to 6-d) 
  \item normal scale \code{\link{hns}} (1-d); \code{\link{Hns}} (2- to 6-d).
  }

  --For kernel density support estimation, the main function is
  \code{\link{ksupp}} which is (the convex hull of) 
  \deqn{\{\bold{x}: \hat{f}(\bold{x}) >
    \tau\}}{\{x: hat(f) > tau\}}
  for a suitable level \eqn{\tau}{tau}. This is closely related to the \eqn{\tau}{tau}-level set of
  \eqn{\hat{f}}{hat(f)}. 
  
  --For truncated kernel density estimation, the main function is
  \code{\link{kde.truncate}}
  \deqn{\hat{f} (\bold{x}) \bold{1}\{\bold{x} \in \Omega\} /
  \int_{\Omega}\hat{f} (\bold{x}) \, d\bold{x}}{hat(f)(x) 1\{x in Omega\}/int hat(f) 1\{x in Omega\}}
  for a bounded data support \eqn{\Omega}{Omega}. The standard density
  estimate \eqn{\hat{f}}{hat(f)} is truncated and rescaled to give
  unit integral over \eqn{\Omega}{Omega}. Its \code{plot} method calls \code{\link{plot.kde}}.

  --For boundary kernel density estimation where the kernel function is
  modified explicitly in the boundary region, the main function is
  \code{\link{kde.boundary}}
  \deqn{ n^{-1} \sum_{i=1}^n K^*_{\bold{{\rm H}}} (\bold{x} - \bold{X}_i)}{hat(f)(x) = n^(-1) sum_i K*_H (x - X_i)} 
  for a boundary kernel \eqn{K^*}{K*}. Its \code{plot} method calls \code{\link{plot.kde}}.

  --For variable kernel density estimation where the bandwidth is not a
  constant matrix, the main functions are \code{\link{kde.balloon}}
  \deqn{\hat{f}_{\rm ball}(\bold{x}) = n^{-1} \sum_{i=1}^n K_{\bold{{\rm H}}(\bold{x})} (\bold{x} - \bold{X}_i)}{hat(f)_ball(x) = n^(-1) sum_i K_H(x) (x - X_i)}
  and
  \code{\link{kde.sp}} 
  \deqn{\hat{f}_{\rm SP}(\bold{x}) = n^{-1} \sum_{i=1}^n K_{\bold{{\rm H}}(\bold{X}_i)} (\bold{x} - \bold{X}_i).}{hat(f)_SP(x) = n^(-1) sum_i K_H(X_i) (x - X_i).}
  For the balloon estimation \eqn{\hat{f}_{\rm ball}}{hat(f)_ball} the
  bandwidth varies with the estimation point \eqn{\bold{x}}{x}, whereas
  for the sample point estimation  \eqn{\hat{f}_{\rm SP}}{hat(f)_SP}
  the bandwidth varies with the data point
  \eqn{\bold{X}_i, i=1,\dots,n}{X_i, i=1, ..., n}. 
  Their \code{plot} methods call \code{\link{plot.kde}}. The bandwidth
  selectors for \code{kde.balloon} are based on the normal scale bandwidth
  \code{Hns(deriv.order=2)} via the MSE minimal formula, and for
  \code{kde.SP} on \code{Hns(deriv.order=4)} via the Abramson formula.
  
  --For kernel density derivative estimation, the main function is \code{\link{kdde}}
  \deqn{{\sf D}^{\otimes r}\hat{f}(\bold{x}) = n^{-1} \sum_{i=1}^n {\sf
  D}^{\otimes r}K_{\bold{{\rm H}}} (\bold{x} -
\bold{X}_i).}{hat(f)^(r)(x) = n^(-1) sum_i D^r K_H (x - X_i).} 
  The bandwidth selectors are a modified subset of those for
  \code{\link{kde}}, i.e. \code{\link{Hlscv}}, \code{\link{Hns}}, \code{\link{Hpi}}, \code{\link{Hscv}} with \code{deriv.order>0}. 
   Its \code{plot} method is \code{\link{plot.kdde}} for plotting each
   partial derivative singly.

  --For kernel summary curvature estimation, the main function is
  \code{\link{kcurv}}
  \deqn{\hat{s}(\bold{x})= - \bold{1}\{{\sf D}^2 \hat{f}(\bold{x}) <
    0\} \mathrm{abs}(|{\sf D}^2 \hat{f}(\bold{x})|)}{hat(s)(x) = -1\{D^2 hat(f)(x) <0)*abs(det(D^2 hat(f)(x)))\}} where \eqn{{\sf D}^2
    \hat{f}(\bold{x})}{D^2 hat(f)(x)} is the kernel Hessian matrix estimate.
  It has the same structure as a kernel density estimate so its \code{plot}
  method calls \code{\link{plot.kde}}. 
   
  --For kernel discriminant analysis, the main function is
  \code{\link{kda}} which computes density estimates for each the 
  groups in the training data, and the discriminant surface. 
  Its \code{plot} method is \code{\link{plot.kda}}. The wrapper function
  \code{\link{hkda}}, \code{\link{Hkda}} computes 
  bandwidths for each group in the training data for \code{kde},
  e.g. \code{hpi}, \code{Hpi}.
  
  --For kernel functional estimation, the main function is
  \code{kfe} which computes the \eqn{r}{r}-th order integrated density functional  
  \deqn{\hat{{\bold \psi}}_r  = n^{-2} \sum_{i=1}^n \sum_{j=1}^n {\sf D}^{\otimes r}K_{\bold{{\rm H}}}(\bold{X}_i-\bold{X}_j).}{hat(psi)_r = n^(-2) sum_i sum_j D^r K_H (X_i - X_j).} The plug-in selectors are \code{\link{hpi.kfe}} (1-d), \code{\link{Hpi.kfe}} (2- to 6-d).
  Kernel functional estimates are usually not required to computed
  directly by the user, but only within other functions in the package.
    
  --For kernel-based 2-sample testing, the main function is
  \code{\link{kde.test}} which computes the integrated 
  \eqn{L_2}{L2} distance between the two density estimates as the test
  statistic, comprising a linear combination of 0-th order kernel
  functional estimates:
  \deqn{\hat{T} = \hat{\psi}_{0,1} + \hat{\psi}_{0,2} - (\hat{\psi}_{0,12} +
  \hat{\psi}_{0,21}),}{hat(T) = hat(psi)_0,1 + hat(psi)_0,2 - (hat(psi)_0,12 +
  hat(psi)_0,21),} and the corresponding p-value. The \eqn{\psi}{psi} are
  zero order kernel functional estimates with the subscripts indicating
  that 1 = sample 1 only, 2 = sample 2 only, and 12, 21 =
  samples 1 and 2.  The bandwidth selectors are \code{\link{hpi.kfe}},
  \code{\link{Hpi.kfe}} with \code{deriv.order=0}. 

  --For kernel-based local 2-sample testing, the main function is
   \code{\link{kde.local.test}} which computes the squared distance
   between the two density estimates as the test  
   statistic \deqn{\hat{U}(\bold{x}) = [\hat{f}_1(\bold{x}) -
     \hat{f}_2(\bold{x})]^2}{hat(U)(x) = [hat(f)_1(x) - hat(f)_2(x)]^2} and the corresponding local
   p-values.  The bandwidth selectors are those used with \code{\link{kde}},
   e.g. \code{\link{hpi}, \link{Hpi}}.  

  --For kernel cumulative distribution function estimation, the main
  function is \code{\link{kcde}}
  \deqn{\hat{F}(\bold{x}) = n^{-1} \sum_{i=1}^n
    \mathcal{K}_{\bold{{\rm H}}} (\bold{x} - \bold{X}_i)}{hat(F)(x) = n^(-1) sum_i intK_H (x - X_i)}
  where \eqn{\mathcal{K}}{intK} is the integrated kernel.  
  The bandwidth selectors are \code{\link{hpi.kcde}},
  \code{\link{Hpi.kcde}}.  Its \code{plot} method is
  \code{\link{plot.kcde}}.
  There exist analogous functions for the survival function \eqn{\hat{\bar{F}}}{hat(bar(F))}.

  --For kernel estimation of a ROC (receiver operating characteristic)
  curve to compare two samples from \eqn{\hat{F}_1,
  \hat{F}_2}{hat(F)_1, hat(F)_2}, the main function is \code{\link{kroc}}
  \deqn{\{\hat{F}_{\hat{Y}_1}(z),
  \hat{F}_{\hat{Y}_2}(z)\}}{\{hat(F)_hat(Y1))(z), hat(F_hat(Y2))(z)\}} based on the cumulative distribution functions of
  \eqn{\hat{Y}_j = \hat{\bar{F}}_1(\bold{X}_j), j=1,2}{hat(Yj)=hat(bar(F))_1(X_j), j=1,2}.
    
  The bandwidth selectors are those used with \code{\link{kcde}},
  e.g. \code{\link{hpi.kcde}, \link{Hpi.kcde}}  for
  \eqn{\hat{F}_{\hat{Y}_j}, \hat{\bar{F}}_1}{hat(F)_hat(Yj), hat(bar(F))_1}. Its \code{plot} method
  is \code{\link{plot.kroc}}.   

  --For kernel estimation of a copula, the
  main function is \code{\link{kcopula}}
  \deqn{\hat{C}(\bold{z}) = \hat{F}(\hat{F}_1^{-1}(z_1), \dots,
  \hat{F}_d^{-1}(z_d))}{hat(C)(z) = hat(F)(hat(F)_1^(-1)(z_1),..., hat(F)_d^(-1)(z_d))}
  where \eqn{\hat{F}_j^{-1}(z_j)}{hat(F)_j^(-1)(z_j)} is
  the \eqn{z_j}{z_j}-th quantile of of the \eqn{j}{j}-th marginal
  distribution \eqn{\hat{F}_j}{hat(F_j)}.
  The bandwidth selectors are those used with \code{\link{kcde}} for
  \eqn{\hat{F}, \hat{F}_j}{hat(F), hat(F)_j}.
  Its \code{plot} method is \code{\link{plot.kcde}}.

  %--For kernel estimation of a copula density, the
  %main function is \code{\link{kcopula.de}}
  %\deqn{\hat{c}(\bold{z}) = n^{-1} \sum_{i=1}^n
  %K_{\bold{{\rm H}}} (\bold{z} - \hat{\bold{Z}}_i)}{hat(c)(z) =
  %hat(f)(z) = n^(-1) sum_i K_H (z - hat(Z)_i)}
  %where \eqn{\hat{\bold{Z}}_i = (\hat{F}_1(X_{i1}), \dots,
  %  \hat{F}_d(X_{id}))}{hat(Z)_i = (hat(F)_1(X_i1), \dots, hat(F)_d(X_id))}.
  %The bandwidth selectors are those used with \code{\link{kde}} for
  %\eqn{\hat{c}}{hat(c)} and \code{\link{kcde}} for \eqn{\hat{F}_j}{hat(F)_j}.
  %Its \code{plot} method is \code{\link{plot.kde}}.

  --For kernel mean shift clustering, the main function is
  \code{\link{kms}}. The mean shift recurrence relation of the candidate
  point \eqn{{\bold x}}{x}   
  \deqn{{\bold x}_{j+1} = {\bold x}_j + \bold{{\rm H}} {\sf D} \hat{f}({\bold
      x}_j)/\hat{f}({\bold x}_j),}{x_j+1 = x_j + H D hat(f)(x_j)/hat(f)(x_j),}
  where \eqn{j\geq 0}{j>=0} and \eqn{{\bold x}_0 = {\bold x}}{x_0 = x},
  is iterated until \eqn{{\bold x}}{x} converges to its
  local mode in the density estimate \eqn{\hat{f}}{hat(f)} by following
  the density gradient ascent paths. This mode determines the cluster
  label for \eqn{\bold{x}}{x}.  The bandwidth selectors are those used with
  \code{\link{kdde}(deriv.order=1)}.  

  --For kernel density ridge estimation, the main function is
    \code{\link{kdr}}. The kernel density ridge recurrence relation of
    the candidate point \eqn{{\bold x}}{x} 
    \deqn{{\bold x}_{j+1} = {\bold x}_j + \bold{{\rm U}}_{(d-1)}({\bold
    x}_j)\bold{{\rm U}}_{(d-1)}({\bold x}_j)^T \bold{{\rm H}} {\sf D}
    \hat{f}({\bold x}_j)/\hat{f}({\bold x}_j),}{x_j+1 = x_j +
    U_(d-1)(x_j) U_(d-1)(x_j)^T H D hat(f)(x_j)/hat(f)(x_j),}
    where \eqn{j\geq 0}{j>=0}, \eqn{{\bold x}_0 = {\bold x}}{x_0 =
    x} and \eqn{\bold{{\rm U}}_{(d-1)}}{U_(d-1)} is the 1-dimensional projected
    density gradient, 
    is iterated until \eqn{{\bold x}}{x} converges to the ridge in the
    density estimate. The bandwidth selectors are those used with 
    \code{\link{kdde}(deriv.order=2)}. 
  
  -- For kernel feature significance, the main function
     \code{\link{kfs}}.  The hypothesis test at a point \eqn{\bold{x}}{x} is
  \eqn{H_0(\bold{x}): \mathsf{H} f(\bold{x}) < 0}{H0(x): H f(x) < 0},
     i.e. the density Hessian matrix \eqn{\mathsf{H} f(\bold{x})}{H f(x)} is negative definite.
  The test statistic is  
  \deqn{W(\bold{x}) = \Vert 
    \mathbf{S}(\bold{x})^{-1/2} \mathrm{vech} \ \mathsf{H} \hat{f} (\bold{x})\Vert ^2}{%
    W(x) = ||S(x)^(-1/2) vech H hat{f}(x)||^2}
  where \eqn{{\sf H}\hat{f}}{H hat{f}} is the
  Hessian estimate, vech is the vector-half operator, and
  \eqn{\mathbf{S}}{S} is an estimate of the null variance.
  \eqn{W(\bold{x})}{W(x)} is
  approximately \eqn{\chi^2}{chi-squared} distributed with
     \eqn{d(d+1)/2}{d(d+1)/2} degrees of freedom.
  If \eqn{H_0(\bold{x})}{H0(x)} is rejected, then \eqn{\bold{x}}{x}
  belongs to a significant modal region.
  The bandwidth selectors are those used with
  \code{\link{kdde}(deriv.order=2)}. Its \code{plot} method is
  \code{\link{plot.kfs}}.
     
  --For deconvolution density estimation, the main function is
  \code{\link{kdcde}}. A weighted kernel density
  estimation with the contaminated data \eqn{{\bold W}_1, \dots, {\bold
  W}_n}{W_1, ..., W_n}, 
  \deqn{\hat{f}_w({\bold x}) = n^{-1} \sum_{i=1}^n
    \alpha_i K_{\bold{{\rm H}}}({\bold x} - {\bold W}_i),}{hat(f)(x) = n^(-1) sum_i alpha_i K_H (x - W_i),}
  is utilised, where the weights \eqn{\alpha_1, \dots,
  \alpha_n}{alpha_1, ..., alpha_n} are chosen via a
  quadratic optimisation involving the error variance and the
  regularisation parameter. The bandwidth selectors are those used with
  \code{\link{kde}}.
  
  --For quasi kernel density estimation, the main function is \code{\link{as.kde}}. The gridded data is converted to a quasi kernel density estimate which mimics the output from \code{\link{kde}}.  

  --Binned kernel estimation is an approximation to the exact kernel
  estimation and is available for d=1, 2, 3, 4. This makes
  kernel estimators feasible for large samples. 
  
  --For an overview of this package with 2-d density estimation, see 
  \code{vignette("kde")}.

  --For \pkg{ks} \eqn{\geq}{>=} 1.11.1, the \pkg{misc3d} and
  \pkg{rgl} (3-d plot), \pkg{oz} (Australian map) packages, and for \pkg{ks} \eqn{\geq}{>=} 1.14.2, the \pkg{plot3D} (3-d plot) package, have been moved from
  Depends to Suggests. This was done to allow \pkg{ks} to be installed
  on systems where these latter graphical-based packages can't be
  installed. For \pkg{ks} \eqn{\geq}{>=}1.12.0, \pkg{plot3D} becomes the 
  default for 3D plotting, since the future of OpenGL in R is not certain. 
  RGL plots are still supported though these may be deprecated in the future.  

  --This package forms the basis for the practical data analysis in the book \href{https://mvstat.net/mvksa/}{Multivariate Kernel Smoothing and Its Applications}.
}

\author{
  Tarn Duong for most of the package. 
  M. P. Wand for the binned estimation, univariate plug-in selector and 
  univariate density derivative estimator code.
  J. E. Chacon for the unconstrained pilot functional estimation and fast implementation of derivative-based estimation code.
  A. and J. Gramacki for the binned estimation for unconstrained bandwidth 
  matrices. 
}

\references{
  Bowman, A. & Azzalini, A. (1997) \emph{Applied Smoothing Techniques
    for Data Analysis}. Oxford University Press, Oxford.

  Chacon, J.E. & Duong, T. (2018) \emph{Multivariate Kernel Smoothing
  and Its Applications}. Chapman & Hall/CRC, Boca Raton. 
  
  Duong, T. (2004) \emph{Bandwidth Matrices for Multivariate Kernel Density 
    Estimation.} Ph.D. Thesis, University of Western Australia. 

  Scott, D.W. (2015) \emph{Multivariate Density Estimation: Theory,
    Practice, and Visualization (2nd edn)}. John Wiley & Sons, New York.

  Silverman, B. (1986) \emph{Density Estimation for Statistics and
  Data Analysis}. Chapman & Hall/CRC, London.

  Simonoff, J. S. (1996) \emph{Smoothing Methods in Statistics}.
  Springer-Verlag, New York.

  Wand, M.P. & Jones, M.C. (1995) \emph{Kernel Smoothing}. Chapman &
  Hall/CRC, London.
}

\keyword{package}

\seealso{\pkg{feature}, \pkg{sm}, \pkg{KernSmooth}}

\name{plot.kda}
\alias{plot.kda}
\title{Plot for kernel discriminant analysis}

\description{
  Plot for kernel discriminant analysis for 1- to 3-dimensional data.
}

\usage{
\method{plot}{kda}(x, y, y.group, ...)
}

\arguments{
  \item{x}{ object of class \code{kda} (output from \code{\link{kda}})}
  \item{y}{matrix of test data points}
  \item{y.group}{vector of group labels for test data points}
  \item{...}{other graphics parameters:
    \describe{
      \item{\code{rugsize}}{height of rug-like plot for partition classes (1-d)}
      \item{\code{prior.prob}}{vector of prior probabilities}
      \item{\code{col.part}}{vector of colours for partition classes (1-d, 2-d)}
    }
    
    and those used in \code{\link{plot.kde}}
  }
}
  
\value{
  Plots for 1-d and 2-d are sent to graphics window. Plot for 3-d is sent to
  graphics/RGL window. 
}

\details{
  For \code{kda} objects, the function headers for the different dimensional data are 
  \preformatted{
  ## univariate
  plot(x, y, y.group, prior.prob=NULL, xlim, ylim, xlab, 
       ylab="Weighted density function", drawpoints=FALSE, col, col.fun, 
       col.part, col.pt, lty, jitter=TRUE, rugsize, add=FALSE, alpha=1, ...)

  ## bivariate
  plot(x, y, y.group, prior.prob=NULL, display.part="filled.contour",
       cont=c(25,50,75), abs.cont, approx.cont=TRUE, xlim, ylim, xlab, ylab,
       drawpoints=FALSE, drawlabels=TRUE, cex=1, pch, lty, part=TRUE, col, 
       col.fun, col.part, col.pt, alpha=1, lwd=1, lwd.part=0, add=FALSE, ...)

  ## trivariate
  plot(x, y, y.group, prior.prob=NULL, display="plot3D", cont=c(25,50,75), 
       abs.cont, approx.cont=TRUE, colors, col, col.fun, col.pt, alpha=0.5, 
       alphavec, xlab, ylab, zlab, drawpoints=FALSE, size=3, cex=1, pch, 
       theta=-30, phi=40, d=4, ticktype="detailed", bty="f", add=FALSE, ...)}
}
 
\seealso{\code{\link{kda}}, \code{\link{kde}}}

\examples{
## univariate example
data(iris)
ir <- iris[,1]
ir.gr <- iris[,5]
kda.fhat <- kda(x=ir, x.group=ir.gr, xmin=3, xmax=9)
plot(kda.fhat, xlab="Sepal length")

## bivariate example
ir <- iris[,1:2]
ir.gr <- iris[,5]
kda.fhat <- kda(x=ir, x.group=ir.gr)
plot(kda.fhat, alpha=0.2, drawlabels=FALSE)

\donttest{## trivariate example
ir <- iris[,1:3]
ir.gr <- iris[,5]
kda.fhat <- kda(x=ir, x.group=ir.gr)
plot(kda.fhat) ## colour=species, transparency=density heights}
}

\keyword{hplot}

\name{rkde}
\alias{dkde}
\alias{pkde}
\alias{qkde}
\alias{rkde}
\title{Derived quantities from kernel density estimates}

\description{
   Derived quantities from kernel density estimates.
}

\usage{
 dkde(x, fhat)
 pkde(q, fhat)
 qkde(p, fhat)
 rkde(n, fhat, positive=FALSE) 
}

\arguments{
   \item{x,q}{vector of quantiles}
   \item{p}{vector of probabilities}
   \item{n}{number of observations}
   \item{positive}{flag to compute KDE on the positive real
   line. Default is FALSE.}  
   \item{fhat}{kernel density estimate, object of class \code{kde}}
}

\value{
   For the 1-d kernel density estimate \code{fhat},
   \code{pkde} computes the cumulative probability for the quantile
   \code{q}, \code{qkde} computes the quantile corresponding to the probability
   \code{p}.

   For any kernel density estimate, \code{dkde} computes the density value at
   \code{x} (it is an alias for \code{predict.kde}), \code{rkde}
   computes a random sample of size \code{n}.   
}

\details{ 
  \code{pkde} uses the trapezoidal rule for the numerical
  integration. \code{rkde} uses
  Silverman (1986)'s method to generate a random sample from a KDE. 
}

\references{ Silverman, B. (1986) \emph{Density Estimation for Statistics and
  Data Analysis}. Chapman & Hall/CRC. London.}

\examples{
set.seed(8192)
x <- rnorm.mixt(n=10000, mus=0, sigmas=1, props=1)
fhat <- kde(x=x)
p1 <- pkde(fhat=fhat, q=c(-1, 0, 0.5))
qkde(fhat=fhat, p=p1)    
y <- rkde(fhat=fhat, n=100)

x <- rmvnorm.mixt(n=10000, mus=c(0,0), Sigmas=invvech(c(1,0.8,1)))
fhat <- kde(x=x)
y <- rkde(fhat=fhat, n=1000)
fhaty <- kde(x=y)
plot(fhat, col=1)
plot(fhaty, add=TRUE, col=2)
}

\keyword{smooth}

\name{hsct}
\docType{data}
\alias{hsct}
\title{Haematopoietic stem cell transplant}

\description{
  This data set contains the haematopoietic stem cell transplant (HSCT)
  measurements obtained by a flow cytometer from mouse subjects. A flow
  cytometer measures the spectra of fluorescent signals from biological
  cell samples to study their properties.    
}

\usage{data(hsct)}

\format{A matrix with 39128 rows and 6 columns.
  The first column is the FITC-CD45.1 fluorescence (0-1023), 
  the second is the PE-Ly65/Mac1 fluorescence (0-1023),
  the third is the PI-LiveDead fluorescence (0-1023),
  the fourth is the APC-CD45.2 fluorescence (0-1023),
  the fifth is the class label of the cell type (1, 2, 3, 4, 5),
  the sixth the mouse subject number (5, 6, 9, 12).
}

\source{
Aghaeepour, N., Finak, G., The FlowCAP Consortium, The DREAM Consortium,
Hoos, H., Mosmann, T. R., Brinkman, R., Gottardo, R. & Scheuermann, R. H. (2013) Critical assessment of automated flow cytometry data analysis techniques, \emph{Nature Methods} \bold{10}, 228-238.
}

\keyword{datasets}

\name{kde.local.test}
\alias{kde.local.test}
\title{Kernel density based local two-sample comparison test}

\description{
  Kernel density based local two-sample comparison test for 1- to 6-dimensional data.}

\usage{
kde.local.test(x1, x2, H1, H2, h1, h2, fhat1, fhat2, gridsize, binned, 
   bgridsize, verbose=FALSE, supp=3.7, mean.adj=FALSE, signif.level=0.05,
   min.ESS, xmin, xmax)
}

\arguments{
  \item{x1,x2}{vector/matrix of data values}
  \item{H1,H2,h1,h2}{bandwidth matrices/scalar bandwidths.  If these are missing, \code{Hpi} or \code{hpi} is called by default.}
  \item{fhat1,fhat2}{objects of class \code{kde}}
  \item{binned}{flag for binned estimation}
  \item{gridsize}{vector of grid sizes}
  \item{bgridsize}{vector of binning grid sizes}
  \item{verbose}{flag to print out progress information. Default is FALSE.}
  \item{supp}{effective support for normal kernel}
  \item{mean.adj}{flag to compute second order correction for mean value of critical sampling distribution. Default is FALSE. Currently implemented for d<=2 only.}
  \item{signif.level}{significance level. Default is 0.05.}
  \item{min.ESS}{minimum effective sample size. See below for details.}
  \item{xmin,xmax}{vector of minimum/maximum values for grid} 
}

\value{
  A kernel two-sample local significance is an object of class
  \code{kde.loctest} which is a list with fields:
  \item{fhat1,fhat2}{kernel density estimates, objects of class \code{kde}}
  \item{chisq}{chi squared test statistic}
  \item{pvalue}{matrix of local \eqn{p}{p}-values at each grid point}
  \item{fhat.diff}{difference of KDEs}
  \item{mean.fhat.diff}{mean of the test statistic}  
  \item{var.fhat.diff}{variance of the test statistic}
  \item{fhat.diff.pos}{binary matrix to indicate locally significant fhat1 > fhat2}
  \item{fhat.diff.neg}{binary matrix to indicate locally significant fhat1 < fhat2}
  \item{n1,n2}{sample sizes}
  \item{H1,H2,h1,h2}{bandwidth matrices/scalar bandwidths}
}

\details{The null hypothesis is \eqn{H_0(\bold{x}): f_1(\bold{x}) = f_2(\bold{x})}{H_0(x): f_1(x) = f_2(x)} where \eqn{f_1, f_2}{f_1, f_2} 
  are the respective density functions. The measure of discrepancy is  
  \eqn{U(\bold{x}) = [f_1(\bold{x}) - f_2(\bold{x})]^2}{U(x) = [f_1(x) - f_2(x)]^2}. 
  Duong (2013) shows that the test statistic obtained, by substituting the
  KDEs for the true densities, has a null 
  distribution which is asymptotically chi-squared with 1 d.f.

  The required input is either \code{x1,x2} and \code{H1,H2}, or
  \code{fhat1,fhat2}, i.e. the data values and bandwidths or objects of class
  \code{kde}. In the former case, the \code{kde} objects are created.
  If the \code{H1,H2} are missing then the default are the plug-in
  selectors \code{Hpi}.  Likewise for missing \code{h1,h2}. 

  The \code{mean.adj} flag determines whether the
second order correction to the mean value of the test statistic should be computed. 
  \code{min.ESS} is borrowed from Godtliebsen et al. (2002)
  to reduce spurious significant results in the tails, though by it is usually
  not required for small to moderate sample sizes. 
}

\references{
  Duong, T. (2013) Local significant differences from non-parametric
  two-sample tests. \emph{Journal of Nonparametric Statistics},
  \bold{25}, 635-645.
    
  Godtliebsen, F., Marron, J.S. & Chaudhuri, P. (2002) 
  Significance in scale space for bivariate density estimation.
  \emph{Journal of Computational and Graphical Statistics},
  \bold{11}, 1-22.

}

\seealso{\code{\link{kde.test}}, \code{\link{plot.kde.loctest}}}
  
\examples{
data(crabs, package="MASS")
x1 <- crabs[crabs$sp=="B", 4]
x2 <- crabs[crabs$sp=="O", 4]
loct <- kde.local.test(x1=x1, x2=x2)
plot(loct)

## see examples in ? plot.kde.loctest
}

\keyword{test}

\name{plot.kfs}
\alias{plot.kfs}
\title{Plot for kernel feature significance}

\description{
  Plot for kernel significant regions for 1- to 3-dimensional data.
}

\usage{
\method{plot}{kfs}(x, display="filled.contour", col=7, colors, abs.cont,
   alpha=1, alphavec=0.4, add=FALSE, ...)
}

\arguments{
  \item{x}{object of class \code{kfs} (output from
    \code{\link{kfs}})}
  \item{display}{type of display, "slice" for contour plot,
    "persp" for perspective plot, "image" for image plot, "filled.contour"
    for filled contour plot (2-d); "plot3D", "rgl" (3-d)}
  \item{col,colors}{colour for contour region}
  \item{abs.cont}{absolute contour height. Default is 0.5.}
  \item{alpha}{transparency value for contour (2-d)}
  \item{alphavec}{vector of transparency values for contour (3-d)}
  \item{add}{flag to add to current plot. Default is FALSE.}
  \item{...}{other graphics parameters used in \code{\link{plot.kde}}}
}
  
\value{
  Plots for 1-d and 2-d are sent to graphics window. Plot for 3-d is sent to
  graphics/RGL window.
}

\seealso{\code{\link{plot.kde}}}

\examples{
\donttest{data(geyser, package="MASS")
geyser.kde <- kde(geyser)
geyser.fs <- kfs(geyser, binned=TRUE)
plot(geyser.kde, col=1)
plot(geyser.fs, add=TRUE, alpha=0.6)
}}

\keyword{hplot}

\name{kdcde}
\alias{kdcde}
\alias{dckde}
\title{Deconvolution kernel density derivative estimate}

\description{
  Deconvolution kernel density derivative estimate for 1- to 6-dimensional data.
}

\usage{
kdcde(x, H, h, Sigma, sigma, reg, bgridsize, gridsize, binned, 
      verbose=FALSE, ...)
dckde(...) 
}

\arguments{
  \item{x}{matrix of data values}
  \item{H,h}{bandwidth matrix/scalar bandwidth. If these are missing, \code{Hpi} or \code{hpi} is called by default.}
  \item{Sigma,sigma}{error variance matrix}
  \item{reg}{regularisation parameter}
  \item{gridsize}{vector of number of grid points}
  \item{binned}{flag for binned estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
  \item{...}{other parameters to \code{\link{kde}}} 
}
  
\value{
  A deconvolution kernel density derivative estimate is an object of class
  \code{kde} which is a list with fields:
  \item{x}{data points - same as input}
  \item{eval.points}{vector or list of points at which the estimate is evaluated}
  \item{estimate}{density estimate at \code{eval.points}}
  \item{h}{scalar bandwidth (1-d only)}
  \item{H}{bandwidth matrix}
  \item{gridtype}{"linear"}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names} 
  \item{w}{vector of weights}
  \item{cont}{vector of probability contour levels}   
}

\details{ 
  A weighted kernel density estimate is utilised to perform the
  deconvolution. The weights \code{w} are the solution to a
  quadratic programming problem, and then input into \code{kde(w=w)}.
  This weighted estimate also requires an estimate of the error
  variance matrix from repeated observations, and of the regularisation
  parameter. If the latter is missing, it is calculated internally using
  a 5-fold cross validation method. See Hazelton & Turlach (2009).
  \code{dckde} is an alias for \code{kdcde}.
  
  If the bandwidth \code{H} is missing from \code{kde}, then
  the default bandwidth is the plug-in selector
  \code{Hpi}. Likewise for missing \code{h}.
  
  The effective support, binning, grid size, grid range, positive
  parameters are the same as \code{\link{kde}}.
}

\references{
Hazelton, M. L. & Turlach, B. A. (2009), Nonparametric density
deconvolution by weighted kernel density estimators, \emph{Statistics
  and Computing}, \bold{19}, 217-228.
}

\seealso{\code{\link{kde}}}

\examples{
\donttest{
data(air)
air <- air[, c("date", "time", "co2", "pm10")]
air2 <- reshape(air, idvar="date", timevar="time", direction="wide")
air <- as.matrix(na.omit(air2[,c("co2.20:00", "pm10.20:00")]))
Sigma.air <- diag(c(var(air2[,"co2.19:00"] - air2["co2.21:00"], na.rm=TRUE),
   var(air2[,"pm10.19:00"] - air2[,"pm10.21:00"], na.rm=TRUE)))
fhat.air.dec <- kdcde(x=air, Sigma=Sigma.air, reg=0.00021, verbose=TRUE)
plot(fhat.air.dec, drawlabels=FALSE, display="filled.contour", lwd=1)}
}

\keyword{smooth}

\name{air}
\docType{data}
\alias{air}
\title{Air quality measurements in an underground train station}

\description{
This data set contains the hourly mean air
quality measurements from 01 January 2013 to 31 December 2016 in the
Chatelet underground train station in the Paris metro. 
}

\usage{data(air)}

\format{A matrix with 35039 rows and 8 columns.
Each row corresponds to an hourly measurement. 
The first column is the date (yyyy-mm-dd),
the second is the time (hh:mm),
the third is the nitric oxide NO concentration (g/m3),
the fourth is the nitrogen dioxide NO\eqn{_2}{2} concentration (g/m3),
the fifth is the concentration of particulate matter less than 10
microns PM10 (ppm),
the sixth is the carbon dioxide concentration CO\eqn{_2}{2} (g/m3),
the seventh is the temperature (degrees Celsius),
the eighth is the relative humidity (percentage). }

\source{
RATP (2016) 
\href{https://data.iledefrance.fr/explore/dataset/qualite-de-l-air-mesuree-dans-la-station-chatelet-rer-a/}{Qualite de l'air mesuree dans la station Chatelet}. Regie autonome des transports parisiens - Departement Developpement, Innovation et Territoires. Accessed 2017-09-27.
}

\keyword{datasets}

\name{Hscv}
\alias{Hscv}
\alias{Hscv.diag}
\alias{hscv}
\title{Smoothed cross-validation (SCV) bandwidth selector}

\description{
  SCV bandwidth for 1- to 6-dimensional data.}

\usage{
Hscv(x, nstage=2, pre="sphere", pilot, Hstart, binned, 
     bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
Hscv.diag(x, nstage=2, pre="scale", pilot, Hstart, binned, 
     bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
hscv(x, nstage=2, binned=TRUE, bgridsize, plot=FALSE)
}

\arguments{
  \item{x}{vector or matrix of data values}
  \item{pre}{"scale" = \code{\link{pre.scale}}, "sphere" = \code{\link{pre.sphere}}}
  \item{pilot}{"amse" = AMSE pilot bandwidths \cr
    "samse" = single SAMSE pilot bandwidth \cr 
    "unconstr" = single unconstrained pilot bandwidth \cr
    "dscalar" = single pilot bandwidth for deriv.order>0 \cr
    "dunconstr" = single unconstrained pilot bandwidth for deriv.order>0}
  \item{Hstart}{initial bandwidth matrix, used in numerical
    optimisation}
  \item{binned}{flag for binned kernel estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{amise}{flag to return the minimal scaled SCV value. Default is FALSE.}
  \item{deriv.order}{derivative order}
  \item{verbose}{flag to print out progress information. Default is FALSE.}
  \item{optim.fun}{optimiser function: one of \code{nlm} or \code{optim}}
  \item{nstage}{number of stages in the SCV bandwidth selector (1 or 2)}
  \item{plot}{flag to display plot of SCV(h) vs h (1-d only). Default is FALSE.}
}

\value{
SCV bandwidth. If \code{amise=TRUE} then the minimal scaled SCV value is returned too. 
}

\details{ \code{hscv} is the univariate SCV
  selector of Jones, Marron & Park (1991). \code{Hscv} is a
  multivariate generalisation of this, see Duong & Hazelton (2005).
  Use \code{Hscv} for unconstrained bandwidth matrices and \code{Hscv.diag}
  for diagonal bandwidth matrices. 
 
  The default pilot is \code{"samse"} for d=2, r=0, and
  \code{"dscalar"} otherwise. For SAMSE pilot bandwidths, see Duong &
  Hazelton (2005). Unconstrained and higher order derivative pilot
  bandwidths are from Chacon & Duong (2011).  

  For d=1, the selector \code{hscv} is not always stable for large
  sample sizes with binning.
  Examine the plot from \code{hscv(plot=TRUE)} to
  determine the appropriate smoothness of the SCV function. Any
  non-smoothness is due to the discretised nature of binned estimation.

  For details about the advanced options for \code{binned, Hstart, optim.fun}, 
  see \code{\link{Hpi}}.
}

\references{
  Chacon, J.E. & Duong, T. (2011) Unconstrained pilot selectors for
  smoothed cross validation. \emph{Australian & New Zealand Journal of
  Statistics}, \bold{53}, 331-351. 

  Duong, T. & Hazelton, M.L. (2005) Cross-validation bandwidth
  matrices for multivariate kernel density
  estimation. \emph{Scandinavian Journal of Statistics}, \bold{32}, 485-506.
   
  Jones, M.C., Marron, J.S. & Park, B.U. (1991) A simple root \eqn{n}{n}
  bandwidth selector. \emph{Annals of Statistics}, \bold{19}, 1919-1932.
} 

\seealso{\code{\link{Hbcv}}, \code{\link{Hlscv}}, \code{\link{Hpi}}}

\examples{
data(unicef)
Hscv(unicef)
hscv(unicef[,1])
}

\keyword{smooth}

\name{plot.kde}
\alias{plot.kde}
\title{Plot for kernel density estimate}

\description{
  Plot for kernel density estimate for 1- to 3-dimensional data.
}

\usage{
\method{plot}{kde}(x, ...)
}

\arguments{
  \item{x}{object of class \code{kde} (output from \code{\link{kde}})}
  \item{...}{other graphics parameters:
    \describe{
      \item{\code{display}}{type of display, "slice" for contour plot,
	"persp" for perspective plot, "image" for image plot, "filled.contour"
	for filled contour plot (2-d); "plot3D", "rgl" (3-d)}
      \item{\code{cont}}{vector of percentages for contour level curves}
      \item{\code{abs.cont}}{vector of absolute density estimate heights for contour level curves}
      \item{\code{approx.cont}}{flag to compute approximate contour levels. Default is FALSE.}
      \item{\code{col}}{plotting colour for density estimate (1-d, 2-d)}
      \item{\code{col.cont}}{plotting colour for contours}
      \item{\code{col.fun}}{plotting colour function for contours}
      \item{\code{col.pt}}{plotting colour for data points}
      \item{\code{colors}}{vector of colours for each contour (3-d)}
      \item{\code{jitter}}{flag to jitter rug plot (1-d). Default is TRUE.}
      \item{\code{lwd.fc}}{line width for filled contours (2-d)}
      \item{\code{xlim,ylim,zlim}}{axes limits}
      \item{\code{xlab,ylab,zlab}}{axes labels}
      \item{\code{add}}{flag to add to current plot. Default is FALSE.}
      \item{\code{theta,phi,d,border}}{graphics parameters for perspective plots (2-d)}
      \item{\code{drawpoints}}{flag to draw data points on density estimate. Default is FALSE.}
      \item{\code{drawlabels}}{flag to draw contour labels (2-d). Default is TRUE.} 
      \item{\code{alpha}}{transparency value of plotting symbol}
      \item{\code{alphavec}}{vector of transparency values for contours (3-d)}
      \item{\code{size}}{size of plotting symbol (3-d).}
    }
  }
}

\value{
  Plots for 1-d and 2-d are sent to graphics window. Plot for 3-d is sent to
  graphics/RGL window. 
}

\details{
  For \code{kde} objects, the function headers for the different dimensional data are  
  \preformatted{
  ## univariate
  plot(fhat, xlab, ylab="Density function", add=FALSE, drawpoints=FALSE, col=1,
       col.pt=4, col.cont=1, cont.lwd=1, jitter=FALSE, cont, abs.cont, 
       approx.cont=TRUE, alpha=1, ...)
  
  ## bivariate
  plot(fhat, display="slice", cont=c(25,50,75), abs.cont, approx.cont=TRUE, 
       xlab, ylab, zlab="Density function", cex=1, pch=1, add=FALSE, 
       drawpoints=FALSE, drawlabels=TRUE, theta=-30, phi=40, d=4, col.pt=4, 
       col, col.fun, alpha=1, lwd=1, border=1, thin=3, kdde.flag=FALSE, 
       ticktype="detailed", ...) 

  ## trivariate
  plot(fhat, display="plot3D", cont=c(25,50,75), abs.cont, approx.cont=TRUE, 
       colors, col, col.fun, alphavec, size=3, cex=1, pch=1, theta=-30, phi=40, 
       d=4, ticktype="detailed", bty="f", col.pt=4, add=FALSE, xlab, ylab, 
       zlab, drawpoints=FALSE, alpha, box=TRUE, axes=TRUE, ...)}

  For 1-dimensional data, the plot is a standard plot of a 1-d curve. If
  \code{drawpoints=TRUE} then a rug plot is added. If \code{cont} is specified,
  the horizontal line on the x-axis indicates the \code{cont}\% highest 
  density level set.  
  
  For 2-dimensional data, the different types of plotting displays are 
  controlled by the \code{display} parameter.
  (a) If \code{display="slice"} then a slice/contour plot
  is generated using \code{contour}.  
  (b) If \code{display} is \code{"filled.contour"} 
  then a filled contour plot is generated.
  The default contours are at 25\%, 50\%, 75\% or
  \code{cont=c(25,50,75)} which are upper percentages of
  highest density regions. 
  (c) If \code{display="persp"} then a perspective/wire-frame plot
  is generated.  The default z-axis limits \code{zlim} are the default
  from the usual \code{persp} command.
  (d) If \code{display="image"} then an image plot
  is generated. 

  For 3-dimensional data, the plot is a series of nested
  3-d contours. The default contours are \code{cont=c(25,50,75)}. The
  default opacity \code{alphavec} ranges from 0.1 to 0.5. 
  For \pkg{ks} \eqn{\geq}{>=} 1.12.0, base R graphics becomes the default plotting engine: 
  to create an \pkg{rgl} plot like in previous versions, set \code{display="rgl"}. 

  To specify contours, either one of \code{cont} or \code{abs.cont}
  is required. \code{cont} specifies upper percentages which
  correspond to probability contour regions. If \code{abs.cont} is set
  to particular values, then contours at these levels are drawn.
  This second option is useful for plotting
  multiple density estimates with common contour levels. See
  \code{\link{contourLevels}} for details on computing contour levels.   
  If \code{approx=FALSE}, then the exact KDE is computed. Otherwise
  it is interpolated from an existing KDE grid, which can dramatically
  reduce computation time for large data sets. 
  
  If a colour function is specified in \code{col.fun}, it should have the number of colours as a single argument, e.g. \code{function(n)\{hcl.colors(n, ...)\}}. The transparent background colour is automatically concatenated before this colour function. If \code{col} is specified, it overrides \code{col.fun}. There should be one more colour than the number of contours, i.e. background colour plus one for each contour. 
}
 
\examples{
## univariate example
data(iris)
fhat <- kde(x=iris[,2])
plot(fhat, cont=50, col.cont=4, cont.lwd=2, xlab="Sepal length")

## bivariate example
fhat <- kde(x=iris[,2:3])
plot(fhat, display="filled.contour", cont=seq(10,90,by=10), lwd=1, alpha=0.5)
plot(fhat, display="persp", border=1, alpha=0.5)

\donttest{## trivariate example
fhat <- kde(x=iris[,2:4])
plot(fhat)
if (interactive()) plot(fhat, display="rgl")}
}

\keyword{hplot}

\name{ksupp}
\alias{ksupp}
\alias{plot.ksupp}
\title{Kernel support estimate}

\description{
  Kernel support estimate for 2 and 3-dimensional data.
}
\usage{
ksupp(fhat, cont=95, abs.cont, convex.hull=TRUE)

\method{plot}{ksupp}(x, display="plot3D", ...)
}

\arguments{
  \item{fhat}{object of class \code{kde}}
  \item{cont}{percentage for contour level curve. Default
  is 95.}
  \item{abs.cont}{absolute density estimate height for contour level curve}
  \item{convex.hull}{flag to compute convex hull of contour level
  curve. Default is TRUE.}
  \item{x}{object of class \code{ksupp}}
  \item{display}{one of "plot3D", "rgl" (required for 3-d only)}
  \item{...}{other graphics parameters}
}
  
\value{
  A kernel support estimate is an object of class \code{ksupp}, i.e. a 2- or 3-column matrix which delimits the (convex hull of the) level set of the density estimate \code{fhat}.  
}

\details{
  The kernel support estimate is the level set of the density estimate
  that exceeds the \code{cont} percent contour level. If this level set
  is a simply connected region, then this can suffice to be a
  conservative estimate of the density support. Otherwise, the convex
  hull of the level set is advised. For 2-d data, the convex hull is computed by \code{chull}; for 3-d data, it is computed by \code{geometry::convhulln}.
}

\seealso{\code{\link{kde}}}

\examples{
data(grevillea)
fhat <- kde(x=grevillea)
fhat.supp <- ksupp(fhat)
plot(fhat, display="filled.contour", cont=seq(10,90,by=10))
plot(fhat, cont=95, add=TRUE, col=1)
plot(fhat.supp, lty=2)

\donttest{data(iris)
fhat <- kde(x=iris[,1:3])
fhat.supp <- ksupp(fhat)
plot(fhat)
plot(fhat.supp, add=TRUE, col=3, alpha=0.1)}
}

\keyword{smooth}

\name{Hnm}
\alias{Hnm}
\alias{hnm}
\alias{Hnm.diag}
\title{Normal mixture bandwidth}

\description{
Normal mixture bandwidth.  
}

\usage{
Hnm(x, deriv.order=0, G=1:9, subset.ind, mise.flag=FALSE, verbose, ...)
Hnm.diag(x, deriv.order=0, G=1:9, subset.ind, mise.flag=FALSE, verbose, ...)
hnm(x, deriv.order=0, G=1:9, subset.ind, mise.flag=FALSE, verbose, ... )
}

\arguments{
  \item{x}{vector/matrix of data values}
  \item{deriv.order}{derivative order}
  \item{G}{range of number of mixture components}
  \item{subset.ind}{index vector of subset of \code{x} for fitting}
  \item{mise.flag}{flag to use MISE or AMISE minimisation. Default is FALSE.}
  \item{verbose}{flag to print out progress information. Default is FALSE.}
  \item{...}{other parameters for \code{Mclust}}
}

\value{
  Normal mixture bandwidth. If \code{mise=TRUE} then the minimal MISE value is returned too. 
}

\details{
  The normal mixture fit is provided by the \code{Mclust} function in
  the \pkg{mclust} package. \code{Hnm} is then \code{Hmise.mixt} (if
  \code{mise.flag=TRUE}) or \code{Hamise.mixt} (if \code{mise.flag=FALSE}) with
  these fitted normal mixture parameters. Likewise for 
  \code{Hnm.diag}, \code{hnm}.   
}

\seealso{ \link{Hmise.mixt}, \link{Hamise.mixt} }

\references{Cwik, J. & Koronacki, J.  (1997). A combined
  adaptive-mixtures/plug-in estimator of multivariate probability
  densities. \emph{Computational Statistics and Data Analysis},
  \bold{26}, 199-218.   
}

\examples{
data(unicef)
Hnm(unicef) 
}

\keyword{smooth}

\name{Hlscv}
\alias{Hlscv}
\alias{Hlscv.diag}
\alias{Hucv}
\alias{Hucv.diag}
\alias{hlscv}
\alias{hucv}
\title{Least-squares cross-validation (LSCV) bandwidth matrix selector
  for multivariate data}

\description{
LSCV bandwidth for 1- to 6-dimensional data}

\usage{
Hlscv(x, Hstart, binned, bgridsize, amise=FALSE, deriv.order=0, 
      verbose=FALSE, optim.fun="optim", trunc)
Hlscv.diag(x, Hstart, binned, bgridsize, amise=FALSE, deriv.order=0, 
      verbose=FALSE, optim.fun="optim", trunc)
hlscv(x, binned=TRUE, bgridsize, amise=FALSE, deriv.order=0, bw.ucv=TRUE)
Hucv(...)
Hucv.diag(...)
hucv(...)
}

\arguments{
  \item{x}{vector or matrix of data values}
  \item{Hstart}{initial bandwidth matrix, used in numerical
    optimisation}
  \item{binned}{flag for binned kernel estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{amise}{flag to return the minimal LSCV value. Default is FALSE.}	
  \item{deriv.order}{derivative order} 
  \item{verbose}{flag to print out progress information. Default is FALSE.}
  \item{optim.fun}{optimiser function: one of \code{nlm} or \code{optim}}
  \item{trunc}{parameter to control truncation for numerical
  optimisation. Default is 4 for density.deriv>0, otherwise no
  truncation. For details see below.}
  \item{bw.ucv}{flag to use \code{stats::bw.ucv} as minimiser function. Default is TRUE.}
  \item{...}{parameters as above}
}

\value{
LSCV bandwidth. If \code{amise=TRUE} then the minimal LSCV value is returned too. 
}

\references{
  Bowman, A. (1984) An alternative method of cross-validation for the
  smoothing of kernel density estimates. \emph{Biometrika}, \bold{71},
  353-360.

  Rudemo, M. (1982) Empirical choice of histograms and kernel density
  estimators. \emph{Scandinavian Journal of Statistics}, \bold{9},
  65-78.
}

\details{\code{hlscv} is the univariate LSCV
  selector of Bowman (1984) and Rudemo (1982). \code{Hlscv} is a
  multivariate generalisation of this. Use \code{Hlscv} for unconstrained
  bandwidth matrices and \code{Hlscv.diag} for diagonal bandwidth matrices. 
  \code{Hucv}, \code{Hucv.diag} and \code{hucv} are aliases with UCV
  (unbiased cross validation) instead of LSCV. 
  
  For \pkg{ks} \eqn{\geq}{>=} 1.13.0, the default minimiser in \code{hlscv} is based on the UCV minimiser 
  \code{stats::bw.ucv}. To reproduce prior behaviour, set \code{bw.ucv=FALSE}. 
  
  Truncation of the parameter space is usually required for the LSCV selector,
  for r > 0, to find a reasonable solution to the numerical optimisation. 
  If a candidate matrix \code{H} is
  such that \code{det(H)} is not in \code{[1/trunc, trunc]*det(H0)} or  
  \code{abs(LSCV(H)) > trunc*abs(LSCV0)} then the \code{LSCV(H)} is reset to \code{LSCV0} where 
  \code{H0=Hns(x)} and \code{LSCV0=LSCV(H0)}.

  For details about the advanced options for \code{binned,Hstart,optim.fun}, 
  see \code{\link{Hpi}}. 
}

\seealso{\code{\link{Hbcv}}, \code{\link{Hpi}}, \code{\link{Hscv}}}

\examples{
data(forbes, package="MASS")
Hlscv(forbes)
hlscv(forbes$bp)
}

\keyword{smooth}

\name{Hpi}
\alias{Hpi}
\alias{Hpi.diag}
\alias{hpi}
\title{Plug-in bandwidth selector}

\description{
  Plug-in bandwidth for for 1- to 6-dimensional data.}

\usage{
Hpi(x, nstage=2, pilot, pre="sphere", Hstart, binned, bgridsize,
   amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
Hpi.diag(x, nstage=2, pilot, pre="scale", Hstart, binned, bgridsize,
   amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
hpi(x, nstage=2, binned=TRUE, bgridsize, deriv.order=0)
}

\arguments{
  \item{x}{vector or matrix of data values}
  \item{nstage}{number of stages in the plug-in bandwidth selector (1 or 2)}
  \item{pilot}{"amse" = AMSE pilot bandwidths \cr
    "samse" = single SAMSE pilot bandwidth \cr 
    "unconstr" = single unconstrained pilot bandwidth \cr
    "dscalar" = single pilot bandwidth for deriv.order >= 0 \cr
    "dunconstr" = single unconstrained pilot bandwidth for deriv.order >= 0}
  \item{pre}{"scale" = \code{\link{pre.scale}}, "sphere" = \code{\link{pre.sphere}}}
  \item{Hstart}{initial bandwidth matrix, used in numerical
    optimisation}
  \item{binned}{flag for binned kernel estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{amise}{flag to return the minimal scaled PI value}
  \item{deriv.order}{derivative order}
  \item{verbose}{flag to print out progress information. Default is FALSE.}
  \item{optim.fun}{optimiser function: one of \code{nlm} or \code{optim}}
}

\value{
Plug-in bandwidth. 
If \code{amise=TRUE} then the minimal scaled PI value is returned too.}

\details{\code{hpi(deriv.order=0)} is the univariate plug-in 
  selector of Wand & Jones (1994), i.e. it is exactly the same as
  \pkg{KernSmooth}'s \code{dpik}. For deriv.order>0, the formula is
  taken from Wand & Jones (1995). \code{Hpi} is a multivariate
  generalisation of this. Use \code{Hpi} for unconstrained bandwidth matrices and
  \code{Hpi.diag} for diagonal bandwidth matrices.

  The default pilot is \code{"samse"} for d=2, r=0, and
  \code{"dscalar"} otherwise.
  For AMSE pilot bandwidths, see Wand & Jones (1994). For
  SAMSE pilot bandwidths, see Duong & Hazelton (2003).  The latter is a
  modification of the former, in order to remove any possible problems
  with non-positive definiteness. Unconstrained and higher order
  derivative pilot bandwidths are from Chacon & Duong (2010). 

  For d=1, 2, 3, 4 and \code{binned=TRUE}, 
  estimates are computed over a binning grid defined 
  by \code{bgridsize}. Otherwise it's computed exactly.  
  If \code{Hstart} is not given then it defaults to \code{Hns(x)}.

  For \pkg{ks} \eqn{\geq}{>=} 1.11.1, the default optimisation function is
  \code{optim.fun="optim"}. To reinstate the previous functionality, use
  \code{optim.fun="nlm"}.  
}

\references{
  Chacon, J.E. & Duong, T. (2010) Multivariate plug-in bandwidth
  selection with unconstrained pilot matrices. \emph{Test}, \bold{19}, 375-398.
  
  Duong, T. & Hazelton, M.L. (2003) Plug-in bandwidth matrices for
  bivariate kernel density estimation. \emph{Journal of Nonparametric
    Statistics}, \bold{15}, 17-30.
  
  Sheather, S.J. & Jones, M.C. (1991) A reliable data-based bandwidth selection
  method for kernel density estimation. \emph{Journal of the Royal
    Statistical Society Series B}, \bold{53}, 683-690.
  
  Wand, M.P. & Jones, M.C. (1994) Multivariate plug-in bandwidth
  selection. \emph{Computational Statistics}, \bold{9}, 97-116.
}

\seealso{\code{\link{Hbcv}}, \code{\link{Hlscv}}, \code{\link{Hscv}}}
  
\examples{
data(unicef)
Hpi(unicef, pilot="dscalar")
hpi(unicef[,1])
}

\keyword{smooth}

\name{ks-internal}
\alias{block.indices}
\alias{col.diverging}
\alias{dwsupp}
\alias{getRow}
\alias{kr}
\alias{Lpdiff}
\alias{matrix.sqrt}
\alias{mur}
\alias{now.time}
\alias{now_time}
\alias{nur}
\alias{nurs}
\alias{Qr}
\alias{rowKpow}
\alias{Sdr}
\alias{Sdrv}
\alias{symconv.1d}
\alias{symconv.nd}
\alias{reg.ucv}

\title{Internal functions in the ks library}

\description{
These functions are user-level but which the user is not required to use directly.
}

\value{The user is not required to use directly these outputs.}

\keyword{internal}

\name{plotmixt}
\alias{plotmixt}
\title{Plot for 1- to 3-dimensional normal and t-mixture density functions}

\description{
  Plot for 1- to 3-dimensional normal and t-mixture density functions.
}

\usage{
plotmixt(mus, sigmas, Sigmas, props, dfs, dist="normal", draw=TRUE,
   deriv.order=0, which.deriv.ind=1, binned=TRUE, ...)
}

\arguments{
  \item{mus}{(stacked) matrix of mean vectors}
  \item{sigmas}{vector of standard deviations (1-d)}
  \item{Sigmas}{(stacked) matrix of variance matrices (2-d, 3-d)}
  \item{props}{vector of mixing proportions}
  \item{dfs}{vector of degrees of freedom}
  \item{dist}{"normal" - normal mixture, "t" - t-mixture}
  \item{draw}{flag to draw plot. Default is TRUE.}
  \item{deriv.order}{derivative order}
  \item{which.deriv.ind}{index of which partial derivative to plot}
  \item{binned}{flag for binned estimation of contour levels. Default is TRUE.} 
  \item{...}{other graphics parameters, see \code{\link{plot.kde}}}
  }

\value{
  If \code{draw=TRUE}, the 1-d, 2-d plot is sent to graphics window, 3-d plot to 
  graphics/RGL window. If \code{draw=FALSE}, then a \code{kdde}-like object is returned.
}

\examples{
## bivariate 
mus <- rbind(c(0,0), c(-1,1))
Sigma <- matrix(c(1, 0.7, 0.7, 1), nr=2, nc=2) 
Sigmas <- rbind(Sigma, Sigma)
props <- c(1/2, 1/2)
plotmixt(mus=mus, Sigmas=Sigmas, props=props, display="filled.contour", lwd=1)

\donttest{## trivariate 
mus <- rbind(c(0,0,0), c(-1,0.5,1.5))
Sigma <- matrix(c(1, 0.7, 0.7, 0.7, 1, 0.7, 0.7, 0.7, 1), nr=3, nc=3) 
Sigmas <- rbind(Sigma, Sigma)
props <- c(1/2, 1/2)
plotmixt(mus=mus, Sigmas=Sigmas, props=props, dfs=c(11,8), dist="t")}
}

\keyword{hplot}

\name{kcde}
\alias{kcde}
\alias{Hpi.kcde}
\alias{Hpi.diag.kcde}
\alias{hpi.kcde}
\alias{predict.kcde}
\title{Kernel cumulative distribution/survival function estimate}

\description{
  Kernel cumulative distribution/survival function estimate for 1- to 3-dimensional data.
}

\usage{
kcde(x, H, h, gridsize, gridtype, xmin, xmax, supp=3.7, eval.points, binned, 
  bgridsize, positive=FALSE, adj.positive, w, verbose=FALSE, 
  tail.flag="lower.tail")
Hpi.kcde(x, nstage=2, pilot, Hstart, binned, bgridsize, amise=FALSE, 
  verbose=FALSE, optim.fun="optim", pre=TRUE)
Hpi.diag.kcde(x, nstage=2, pilot, Hstart, binned, bgridsize, amise=FALSE,
  verbose=FALSE, optim.fun="optim", pre=TRUE)
hpi.kcde(x, nstage=2, binned, amise=FALSE)

\method{predict}{kcde}(object, ..., x)
}

\arguments{
  \item{x}{matrix of data values}
  \item{H,h}{bandwidth matrix/scalar bandwidth. If these are missing, then
  \code{Hpi.kcde} or \code{hpi.kcde} is called by default.}
  \item{gridsize}{vector of number of grid points}
  \item{gridtype}{not yet implemented}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{vector or matrix of points at which estimate is evaluated}
  \item{binned}{flag for binned estimation. Default is FALSE.}
  \item{bgridsize}{vector of binning grid sizes}
  \item{positive}{flag if 1-d data are positive. Default is FALSE.}
  \item{adj.positive}{adjustment applied to positive 1-d data}
  \item{w}{not yet implemented}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
  \item{tail.flag}{"lower.tail" = cumulative distribution, "upper.tail" =
  survival function}
  \item{nstage}{number of stages in the plug-in bandwidth selector (1 or 2)}
  \item{pilot}{"dscalar" = single pilot bandwidth (default for
  \code{Hpi.diag.kcde}) \cr
    "dunconstr" = single unconstrained pilot bandwidth (default for
  \code{Hpi.kcde})} 
  \item{Hstart}{initial bandwidth matrix, used in numerical
    optimisation}
  \item{amise}{flag to return the minimal scaled PI value}
  \item{optim.fun}{optimiser function: one of \code{nlm} or
    \code{optim}}
  \item{pre}{flag for pre-scaling data. Default is TRUE.}
  \item{object}{object of class \code{kcde}}
  \item{...}{other parameters} 
}
  
\value{
  A kernel cumulative distribution estimate is an object of class
  \code{kcde} which is a list with fields:
  \item{x}{data points - same as input}
  \item{eval.points}{vector or list of points at which the estimate is evaluated}
  \item{estimate}{cumulative distribution/survival function estimate at
  \code{eval.points}} 
  \item{h}{scalar bandwidth (1-d only)}
  \item{H}{bandwidth matrix}
  \item{gridtype}{"linear"}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names}
  \item{w}{vector of weights}
  \item{tail}{"lower.tail"=cumulative distribution, "upper.tail"=survival function}
}

\details{
  If \code{tail.flag="lower.tail"} then the cumulative distribution
  function \eqn{\mathrm{Pr}(\bold{X}\leq\bold{x})}{Pr(X<=x)} is estimated, otherwise
  if \code{tail.flag="upper.tail"}, it is the survival function
  \eqn{\mathrm{Pr}(\bold{X}>\bold{x})}{P(X>x)}. For \eqn{d>1}{d>1},
  \eqn{\mathrm{Pr}(\bold{X}\leq\bold{x}) \neq 1 - \mathrm{Pr}(\bold{X}>\bold{x})}{Pr(X<=x) != 1-Pr(X>x)}.
  
  If the bandwidth \code{H} is missing in \code{kcde}, then
  the default bandwidth is the plug-in selector
  \code{Hpi.kcde}. Likewise for missing \code{h}.
  No pre-scaling/pre-sphering is used since the \code{Hpi.kcde} is not
  invariant to translation/dilation.

  The effective support, binning, grid size, grid range, positive, optimisation function
  parameters are the same as \code{\link{kde}}.
}

\references{
  Duong, T. (2016) Non-parametric smoothed estimation of multivariate
  cumulative distribution and survival functions, and receiver operating
  characteristic curves. \emph{Journal of the Korean Statistical
  Society}, \bold{45}, 33-50.
}

\seealso{\code{\link{kde}}, \code{\link{plot.kcde}}}

\examples{
data(iris)
Fhat <- kcde(iris[,1:2])  
predict(Fhat, x=as.matrix(iris[,1:2]))

## See other examples in ? plot.kcde
}

\keyword{smooth}

\name{plot.kdde}
\alias{plot.kdde}
\title{Plot for kernel density derivative estimate}

\description{
  Plot for kernel density derivative estimate for 1- to 3-dimensional data.
}

\usage{
\method{plot}{kdde}(x, ...)
}

\arguments{
  \item{x}{object of class \code{kdde} (output from \code{\link{kdde}})}
  \item{...}{other graphics parameters:
    \describe{
      \item{\code{which.deriv.ind}}{index of the partial derivative to
	be plotted (>1-d)}
    }

    and those used in \code{\link{plot.kde}}}
}
  
\value{
  Plots for 1-d and 2-d are sent to graphics window. Plot for 3-d is sent to
  graphics/RGL window.

  In addition to the display options inherited from \code{plot.kde}, the
  first derivative has \code{display="quiver"}. This is a quiver plot
  where the size and direction of the arrow indicates the
  magnitude/direction of the density gradient. See \code{quiver} from
  the \pkg{pracma} package for more details. 
}

\details{
  For \code{kdde} objects, the function headers for the different dimensional data are  
  \preformatted{
  ## univariate
  plot(fhat, ylab="Density derivative function", cont=50, abs.cont, alpha=1, ...)

  ## bivariate
  plot(fhat, which.deriv.ind=1, cont=c(25,50,75), abs.cont, display="slice", 
       zlab="Density derivative function", col, col.fun, alpha=1, kdde.flag=TRUE, 
       thin=3, transf=1, neg.grad=FALSE, ...)
  
  ## trivariate 
  plot(fhat, which.deriv.ind=1, display="plot3D", cont=c(25,50,75), abs.cont, 
       colors, col, col.fun, ...)}
}
 
\seealso{\code{\link{plot.kde}}}

\examples{
## univariate example
data(tempb)
fhat1 <- kdde(x=tempb[,"tmin"], deriv.order=1)   ## gradient [df/dx, df/dy]
plot(fhat1, xlab="Min. temp.", col.cont=4)       ## df/dx
points(20,predict(fhat1, x=20))

## bivariate example
fhat1 <- kdde(x=tempb[,c("tmin", "tmax")], deriv.order=1)
## gradient [df/dx, df/dy]
plot(fhat1, display="quiver")
  

\donttest{fhat2 <- kdde(x=tempb[,c("tmin", "tmax")], deriv.order=2)
plot(fhat2, which.deriv.ind=2, display="persp", phi=10)
## d^2 f/(dx dy): blue=-ve, red=+ve
plot(fhat2, which.deriv.ind=2, display="filled.contour", lwd=1)
## summary curvature 
s2 <- kcurv(fhat2)
plot(s2, display="filled.contour", lwd=1)

## trivariate example  
data(iris)
fhat1 <- kdde(iris[,2:4], deriv.order=1)
plot(fhat1)}}

\keyword{hplot}

\name{kda}
\alias{Hkda}
\alias{Hkda.diag}
\alias{kda}
\alias{hkda}
\alias{predict.kda}
\alias{compare}
\alias{compare.kda.diag.cv}
\alias{compare.kda.cv}
\title{Kernel discriminant analysis (kernel classification)}

\description{
  Kernel discriminant analysis (kernel classification) for 1- to d-dimensional data.
}

\usage{
kda(x, x.group, Hs, hs, prior.prob=NULL, gridsize, xmin, xmax, supp=3.7,
  eval.points, binned, bgridsize, w, compute.cont=TRUE, approx.cont=TRUE,
  kde.flag=TRUE)
Hkda(x, x.group, Hstart, bw="plugin", ...)
Hkda.diag(x, x.group, bw="plugin", ...)
hkda(x, x.group, bw="plugin", ...)

\method{predict}{kda}(object, ..., x)

compare(x.group, est.group, by.group=FALSE)
compare.kda.cv(x, x.group, bw="plugin", prior.prob=NULL, Hstart, by.group=FALSE,
   verbose=FALSE, recompute=FALSE, ...)
compare.kda.diag.cv(x, x.group, bw="plugin", prior.prob=NULL, by.group=FALSE, 
   verbose=FALSE, recompute=FALSE, ...)
}

\arguments{
  \item{x}{matrix of training data values}
  \item{x.group}{vector of group labels for training data}
  \item{Hs,hs}{(stacked) matrix of bandwidth matrices/vector of scalar
  bandwidths. If these are missing, \code{Hkda} or \code{hkda} is called by default.}
  \item{prior.prob}{vector of prior probabilities}
  \item{gridsize}{vector of grid sizes}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{vector or matrix of points at which estimate is evaluated}
  \item{binned}{flag for binned estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{w}{vector of weights. Not yet implemented.}
  \item{compute.cont}{flag for computing 1\% to 99\% probability contour levels. Default is TRUE.}
  \item{approx.cont}{flag for computing approximate probability contour
    levels. Default is TRUE.}
  \item{kde.flag}{flag for computing KDE on grid. Default is TRUE.}
  \item{object}{object of class \code{kda}}
  \item{bw}{bandwidth: "plugin" = plug-in, "lscv" = LSCV, 
    "scv" = SCV}
  \item{Hstart}{(stacked) matrix of initial bandwidth matrices, used in
    numerical optimisation}
  \item{est.group}{vector of estimated group labels}
  \item{by.group}{flag to give results also within each group}
  \item{verbose}{flag for printing progress information. Default is FALSE.}
  \item{recompute}{flag for recomputing the bandwidth matrix after
    excluding the i-th data item}
  \item{...}{other optional parameters for bandwidth selection, see
    \code{\link{Hpi}}, \code{\link{Hlscv}}, \code{\link{Hscv}}} 
}

\value{
  --For \code{kde.flag=TRUE}, a kernel discriminant analysis is an object of class \code{kda} which is a list with fields
  \item{x}{list of data points, one for each group label}
  \item{estimate}{list of density estimates at \code{eval.points}, one for each group label}  
  \item{eval.points}{vector or list of points that the estimate is evaluated at, one
    for each group label}
  \item{h}{vector of bandwidths (1-d only)}
  \item{H}{stacked matrix of bandwidth matrices or vector of bandwidths}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{w}{vector of weights}
  \item{prior.prob}{vector of prior probabilities}
  \item{x.group}{vector of group labels - same as input}
  \item{x.group.estimate}{vector of estimated group labels. If the test data
    \code{eval.points} are given then these are classified. Otherwise
    the training data \code{x} are classified.}
  For \code{kde.flag=FALSE}, which is always the case for d > 3,
  then only the vector of estimated group labels is returned. 
  
  --The result from \code{Hkda} and \code{Hkda.diag} is a stacked matrix
  of bandwidth matrices, one for each training data group. The result
  from \code{hkda} is a vector of bandwidths, one for each training group.  

  --The \code{compare} functions create a comparison between the true
  group labels \code{x.group} and the estimated ones. 
  It returns a list with fields
  \item{cross}{cross-classification table with the rows
    indicating the true group and the columns the estimated group}
  \item{error}{misclassification rate (MR)}
    
  In the case where the test data are independent of the
  training data, \code{compare} computes MR = (number of points wrongly
  classified)/(total number of points). In the case where the test data
  are not independent e.g. 
  we are classifying the training data set itself, then the cross
  validated estimate of MR is more appropriate. These
  are implemented as \code{compare.kda.cv} (unconstrained bandwidth
  selectors) and \code{compare.kda.diag.cv} (for diagonal bandwidth
  selectors). These functions are only available for d > 1.

  If \code{by.group=FALSE} then only the total MR rate is given. If it
  is set to TRUE, then the MR rates for each class are also given
  (estimated number in group divided by true number).
}

\references{
  Simonoff, J. S. (1996) \emph{Smoothing Methods in Statistics}.
  Springer-Verlag. New York
}

\details{
  If the bandwidths \code{Hs} are missing from \code{kda}, then the
  default bandwidths are the plug-in selectors \code{Hkda(bw="plugin")}.
  Likewise for missing \code{hs}. Valid options for \code{bw}
  are \code{"plugin"}, \code{"lscv"} and \code{"scv"} which in turn call
  \code{\link{Hpi}}, \code{\link{Hlscv}} and \code{\link{Hscv}}.

  The effective support, binning, grid size, grid range, positive
  parameters are the same as \code{\link{kde}}.
  
  If prior probabilities are known then set \code{prior.prob} to these.
  Otherwise \code{prior.prob=NULL} uses the sample
  proportions as estimates of the prior probabilities.
  
  For \pkg{ks} \eqn{\geq}{>=} 1.8.11, \code{kda.kde} has been subsumed
  into \code{kda}, so all prior calls to \code{kda.kde} can be replaced
  by \code{kda}. To reproduce the previous behaviour of \code{kda}, the
  command is \code{kda(kde.flag=FALSE)}.
}

\seealso{\code{\link{plot.kda}}}

\examples{
set.seed(8192)
x <- c(rnorm.mixt(n=100, mus=1), rnorm.mixt(n=100, mus=-1))
x.gr <- rep(c(1,2), times=c(100,100))
y <- c(rnorm.mixt(n=100, mus=1), rnorm.mixt(n=100, mus=-1))
y.gr <- rep(c(1,2), times=c(100,100))
kda.gr <- kda(x, x.gr)
y.gr.est <- predict(kda.gr, x=y)
compare(y.gr, y.gr.est)

## See other examples in ? plot.kda
}

\keyword{smooth}

\name{cardio}
\docType{data}
\alias{cardio}
\title{Foetal cardiotocograms}

\description{
  This data set contains the cardiotocographic measurements from healthy,
  suspect and pathological foetuses. 
}

\usage{data(cardio)}

\format{A matrix with 2126 rows and 8 columns.
  Each row corresponds to a foetal cardiotocogram. The
  class label for the foetal state is the last column: N = normal, S =
  suspect, P = pathological. Details for all variables are found in the
  link below.  
}

\source{
Lichman, M. (2013) UCI Machine learning repository: cardiotocography data set.
\doi{10.24432/C51S4N}.
University of California, Irvine, School of Information and Computer
Sciences. Accessed 2017-05-18.
}

\keyword{datasets}

\name{kfs}
\alias{kfs}
\title{Kernel feature significance }

\description{
  Kernel feature significance for 1- to 6-dimensional data.
}

\usage{
kfs(x, H, h, deriv.order=2, gridsize, gridtype, xmin, xmax, supp=3.7,
    eval.points, binned, bgridsize, positive=FALSE, adj.positive, w, 
    verbose=FALSE, signif.level=0.05)
}

\arguments{
  \item{x}{matrix of data values}
  \item{H,h}{bandwidth matrix/scalar bandwidth. If these are missing, \code{Hpi} or \code{hpi} is called by default.}
  \item{deriv.order}{derivative order (scalar)}
  \item{gridsize}{vector of number of grid points}
  \item{gridtype}{not yet implemented}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{vector or matrix of points at which estimate is evaluated}
  \item{binned}{flag for binned estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{positive}{flag if 1-d data are positive. Default is FALSE.}
  \item{adj.positive}{adjustment applied to positive 1-d data}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
  \item{signif.level}{overall level of significance for hypothesis
  tests. Default is 0.05.} 
}
  
\value{
  A kernel feature significance estimate is an object of class
  \code{kfs} which is a list with fields 
  \item{x}{data points - same as input}
  \item{eval.points}{vector or list of points at which the estimate is evaluated}
  \item{estimate}{binary matrix for significant feature at
  \code{eval.points}: 0 = not signif.,  1 = signif.}
  \item{h}{scalar bandwidth (1-d only)}
  \item{H}{bandwidth matrix}
  \item{gridtype}{"linear"}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names}
  \item{w}{vector of weights}
  \item{deriv.order}{derivative order (scalar)}
  \item{deriv.ind}{martix where each row is a vector of partial derivative indices.}

  This is the same structure as a \code{kdde} object, except that
  \code{estimate} is a binary matrix rather than real-valued. 
}

\details{
  Feature significance is based on significance testing of the gradient
  (first derivative) and curvature (second derivative) of a kernel
  density estimate. Only the latter is currently implemented, and is
  also known as significant modal regions.

  The hypothesis test at a grid point \eqn{\bold{x}}{x} is
  \eqn{H_0(\bold{x}): \mathsf{H} f(\bold{x}) < 0}{H0(x): H f(x) < 0},
  i.e.  the density Hessian matrix \eqn{\mathsf{H} f(\bold{x})}{H f(x)} is negative definite.
  The \eqn{p}{p}-values are computed for each \eqn{\bold{x}}{x} using that
  the test statistic is
  approximately chi-squared distributed with \eqn{d(d+1)/2}{d(d+1)/2} d.f.
  We then use a Hochberg-type simultaneous testing procedure, based on the
  ordered \eqn{p}{p}-values, to control the
  overall level of significance to be \code{signif.level}. If
  \eqn{H_0(\bold{x})}{H0(x)} is rejected then \eqn{\bold{x}}{x}
  belongs to a significant modal region. 

  The computations are based on \code{kdde(x, deriv.order=2)} so
  \code{kfs} inherits its behaviour from \code{\link{kdde}}.
  If the bandwidth \code{H} is missing, then
  the default bandwidth is the plug-in selector
  \code{Hpi(deriv.order=2)}. Likewise for missing \code{h}.
  The effective support, binning, grid size, grid range, positive
  parameters are the same as \code{\link{kde}}.

  This function is similar to the \code{featureSignif} function in the
  \pkg{feature} package, except that it accepts unconstrained bandwidth
  matrices. 
}

\references{
  Chaudhuri, P. & Marron, J.S. (1999) 
  SiZer for exploration of structures in curves.
  \emph{Journal of the American Statistical Association},
  \bold{94},  807-823.

  Duong, T., Cowling, A., Koch, I. & Wand, M.P. (2008)
  Feature significance for multivariate kernel density estimation.
  \emph{Computational Statistics and Data Analysis}, \bold{52},
  4225-4242. 
  
  Godtliebsen, F., Marron, J.S. & Chaudhuri, P. (2002) 
  Significance in scale space for bivariate density estimation.
  \emph{Journal of Computational and Graphical Statistics},
  \bold{11}, 1-22.
}

\seealso{\code{\link{kdde}}, \code{\link{plot.kfs}}}

\examples{
data(geyser, package="MASS")
geyser.fs <- kfs(geyser$duration, binned=TRUE)
plot(geyser.fs, xlab="duration")

## see example in ? plot.kfs
}

\keyword{smooth}

\name{kfe}
\alias{kfe}
\alias{Hpi.kfe}
\alias{Hpi.diag.kfe}
\alias{hpi.kfe}
\title{Kernel functional estimate}

\description{
  Kernel functional estimate for 1- to 6-dimensional data.}

\usage{
kfe(x, G, deriv.order, inc=1, binned, bin.par, bgridsize, deriv.vec=TRUE,
    add.index=TRUE, verbose=FALSE)
Hpi.kfe(x, nstage=2, pilot, pre="sphere", Hstart, binned=FALSE, 
    bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
Hpi.diag.kfe(x, nstage=2, pilot, pre="scale", Hstart, binned=FALSE,
    bgridsize, amise=FALSE, deriv.order=0, verbose=FALSE, optim.fun="optim")
hpi.kfe(x, nstage=2, binned=FALSE, bgridsize, amise=FALSE, deriv.order=0)
}

\arguments{
  \item{x}{vector/matrix of data values}
  \item{nstage}{number of stages in the plug-in bandwidth selector (1 or 2)}
  \item{pilot}{"dscalar" = single pilot bandwidth (default) \cr
    "dunconstr" = single unconstrained pilot bandwidth}
  \item{pre}{"scale" = \code{\link{pre.scale}}, "sphere" = \code{\link{pre.sphere}}}
  \item{Hstart}{initial bandwidth matrix, used in numerical optimisation}
  \item{binned}{flag for binned estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{amise}{flag to return the minimal scaled PI value}
  \item{deriv.order}{derivative order}
  \item{verbose}{flag to print out progress information. Default is FALSE.}
  \item{optim.fun}{optimiser function: one of \code{nlm} or \code{optim}}
  \item{G}{pilot bandwidth matrix}
  \item{inc}{0=exclude diagonal, 1=include diagonal terms in kfe
    calculation}
  \item{bin.par}{binning parameters - output from \code{\link{binning}}}
  \item{deriv.vec}{flag to compute duplicated partial derivatives in the 
    vectorised form. Default is FALSE.}
  \item{add.index}{flag to output derivative indices matrix. Default is true.}  
}

\value{
   Plug-in bandwidth matrix for \eqn{r}{r}-th order kernel functional estimator.
}

\details{
  \code{Hpi.kfe} is the optimal plug-in bandwidth for \eqn{r}{r}-th order kernel functional estimator
  based on the unconstrained pilot selectors of Chacon & Duong (2010).
  \code{hpi.kfe} is the 1-d equivalent, using the formulas from
  Wand & Jones (1995, p.70).

  \code{kfe} does not usually need to be called explicitly by the user.     
}

\references{
  Chacon, J.E. & Duong, T. (2010) Multivariate plug-in bandwidth
  selection with unconstrained pilot matrices. \emph{Test}, \bold{19},
  375-398.
  
  Wand, M.P. & Jones, M.C. (1995) \emph{Kernel Smoothing}. Chapman &
  Hall/CRC, London. 
}

\seealso{\code{\link{kde.test}}}
 
\keyword{smooth}

\name{kdde}
\alias{kdde}
\alias{predict.kdde}
\alias{kcurv}
\title{Kernel density derivative estimate}

\description{
  Kernel density derivative estimate for 1- to 6-dimensional data.
}

\usage{
kdde(x, H, h, deriv.order=0, gridsize, gridtype, xmin, xmax, supp=3.7, 
    eval.points, binned, bgridsize, positive=FALSE, adj.positive, w,
    deriv.vec=TRUE, verbose=FALSE)
kcurv(fhat, compute.cont=TRUE)

\method{predict}{kdde}(object, ..., x)
}

\arguments{
  \item{x}{matrix of data values}
  \item{H,h}{bandwidth matrix/scalar bandwidth. If these are missing, \code{Hpi} or \code{hpi} is called by default.}
  \item{deriv.order}{derivative order (scalar)}
  \item{gridsize}{vector of number of grid points}
  \item{gridtype}{not yet implemented}
  \item{xmin,xmax}{vector of minimum/maximum values for grid}
  \item{supp}{effective support for standard normal}
  \item{eval.points}{vector or matrix of points at which estimate is evaluated}
  \item{binned}{flag for binned estimation}
  \item{bgridsize}{vector of binning grid sizes}
  \item{positive}{flag if data are positive (1-d, 2-d). Default is FALSE.}
  \item{adj.positive}{adjustment applied to positive 1-d data}
  \item{w}{vector of weights. Default is a vector of all ones.}
  \item{deriv.vec}{flag to compute all derivatives in vectorised
  derivative. Default is TRUE. If FALSE then only the unique derivatives are computed.}
  \item{verbose}{flag to print out progress information. Default is
    FALSE.}
  \item{compute.cont}{flag for computing 1\% to 99\% probability contour
    levels. Default is TRUE.}
  \item{fhat}{object of class \code{kdde} with \code{deriv.order=2}} 
  \item{object}{object of class \code{kdde}}
  \item{...}{other parameters} 
}
  
\value{
  A kernel density derivative estimate is an object of class
  \code{kdde} which is a list with fields:
  \item{x}{data points - same as input}
  \item{eval.points}{vector or list of points at which the estimate is evaluated}
  \item{estimate}{density derivative estimate at \code{eval.points}}
  \item{h}{scalar bandwidth (1-d only)}
  \item{H}{bandwidth matrix}
  \item{gridtype}{"linear"}
  \item{gridded}{flag for estimation on a grid}
  \item{binned}{flag for binned estimation}
  \item{names}{variable names}
  \item{w}{vector of weights}
  \item{deriv.order}{derivative order (scalar)}
  \item{deriv.ind}{martix where each row is a vector of partial derivative indices}
}

\details{ 
  For each partial derivative, for grid estimation, the estimate is a
  list whose elements  
  correspond to the partial derivative indices in the rows of \code{deriv.ind}. 
  For points estimation, the estimate is a matrix whose columns correspond to 
  the rows of \code{deriv.ind}.

  If the bandwidth \code{H} is missing from \code{kdde}, then
  the default bandwidth is the  plug-in selector
  \code{Hpi}. Likewise for missing \code{h}.
  
  The effective support, binning, grid size, grid range, positive
  parameters are the same as \code{\link{kde}}.

  The summary curvature is computed by \code{kcurv}, i.e.
  \deqn{\hat{s}(\bold{x})= - \bold{1}\{\mathsf{D}^2 \hat{f}(\bold{x}) <
  0\} \mathrm{abs}(|\mathsf{D}^2 \hat{f}(\bold{x})|)}{hat(s)(x) = -1(D^2
  hat(f)(x) <0)*abs(det(D^2 hat(f)(x)))} where \eqn{\mathsf{D}^2
  \hat{f}(\bold{x})}{D^2 hat(f)(x)} is the kernel Hessian matrix
  estimate. So \eqn{\hat{s}}{hat{s}} calculates the absolute value of
  the determinant of the Hessian matrix and whose sign is the opposite of
  the negative definiteness indicator.
 
}

\seealso{\code{\link{kde}}}

\examples{
set.seed(8192)
x <- rmvnorm.mixt(1000, mus=c(0,0), Sigmas=invvech(c(1,0.8,1)))
fhat <- kdde(x=x, deriv.order=1) ## gradient [df/dx, df/dy]
predict(fhat, x=x[1:5,])

## See other examples in ? plot.kdde
}

\keyword{smooth}

\name{plot.kroc}
\alias{plot.kroc}
\title{Plot for kernel receiver operating characteristic curve (ROC) estimate}

\description{
  Plot for kernel receiver operating characteristic curve (ROC) estimate 1- to 3-dimensional data.
}

\usage{
\method{plot}{kroc}(x, add=FALSE, add.roc.ref=FALSE, xlab, ylab, 
   alpha=1, col=1, ...)
}

\arguments{
  \item{x}{object of class \code{kroc} (output from \code{\link{kroc}})}
  \item{add}{flag to add to current plot. Default is FALSE.}
  \item{add.roc.ref}{flag to add reference ROC curve. Default is FALSE.} 
  \item{xlab}{x-axis label. Default is "False positive rate
  (bar(specificity))".}
  \item{ylab}{y-axis label. Default is "True positive rate (sensitivity)".} 
  \item{alpha,col}{transparency value and colour of line}
  \item{...}{other graphics parameters used in \code{\link{plot.kde}}.} 
}
  
\value{
  Plots for 1-d and 2-d are sent to graphics window. Plot for 3-d is sent to
  graphics/RGL window. 
}

\seealso{\code{\link{plot.kde}}}

\examples{
data(fgl, package="MASS")
x1 <- fgl[fgl[,"type"]=="WinF",c("RI", "Na")]
x2 <- fgl[fgl[,"type"]=="Head",c("RI", "Na")]
Rhat <- kroc(x1=x1, x2=x2) 
plot(Rhat, add.roc.ref=TRUE)
}

\keyword{hplot}

\name{grevillea}
\docType{data}
\alias{grevillea}
\title{Geographical locations of grevillea plants}

\description{
This data set contains the geographical locations 
of the specimens of \emph{Grevillea uncinulata},
more commonly known as the Hook leaf grevillea, which is an endemic floral
species to south Western Australia. This region is
one of the 25 `biodiversity hotspots' which are `areas featuring exceptional
concentrations of endemic species and experiencing exceptional loss of habitat'.
}

\usage{data(grevillea)}

\format{A matrix with 222 rows and 2 columns. Each row corresponds to an
  observed plant. The first column is the longitude (decimal degrees),
  the second is the latitude (decimal degrees). }

\source{
CSIRO (2016) Atlas of Living Australia: \href{https://ala.org.au/}{\emph{Grevillea uncinulata Diels}}. Commonwealth Scientific and Industrial Research Organisation. Accessed 2016-03-11. %https://bie.ala.org.au/species/https://id.biodiversity.org.au/node/apni/2895039
}

\keyword{datasets}

\name{ise.mixt}
\alias{Hmise.mixt}
\alias{Hamise.mixt}
\alias{Hmise.mixt.diag}
\alias{Hamise.mixt.diag}
\alias{hmise.mixt}
\alias{hamise.mixt}
\alias{ise.mixt}
\alias{amise.mixt}
\alias{mise.mixt}
\title{Squared error bandwidth matrix selectors for normal
  mixture densities}

\description{
  The global errors
  ISE (Integrated Squared Error), MISE (Mean Integrated Squared Error) and the 
  AMISE (Asymptotic Mean Integrated Squared Error) for 1- to 6-dimensional
  data. Normal mixture densities have closed form expressions for the MISE and
  AMISE. So in these cases, we can numerically minimise these criteria
  to find MISE- and AMISE-optimal matrices.
}

\usage{
Hamise.mixt(mus, Sigmas, props, samp, Hstart, deriv.order=0)
Hmise.mixt(mus, Sigmas, props, samp, Hstart, deriv.order=0)
Hamise.mixt.diag(mus, Sigmas, props, samp, Hstart, deriv.order=0)
Hmise.mixt.diag(mus, Sigmas, props, samp, Hstart, deriv.order=0)
hamise.mixt(mus, sigmas, props, samp, hstart, deriv.order=0)
hmise.mixt(mus, sigmas, props, samp, hstart, deriv.order=0)
amise.mixt(H, mus, Sigmas, props, samp, h, sigmas, deriv.order=0)
ise.mixt(x, H, mus, Sigmas, props, h, sigmas, deriv.order=0, binned=FALSE, 
         bgridsize)  
mise.mixt(H, mus, Sigmas, props, samp, h, sigmas, deriv.order=0)
}

\arguments{
  \item{mus}{(stacked) matrix of mean vectors (>1-d), vector of means (1-d)}
  \item{Sigmas,sigmas}{(stacked) matrix of variance matrices (>1-d), vector of standard deviations (1-d)}
  \item{props}{vector of mixing proportions}
  \item{samp}{sample size}
  \item{Hstart,hstart}{initial bandwidth (matrix), used in numerical optimisation}
  \item{deriv.order}{derivative order}
  \item{x}{matrix of data values}
  \item{H,h}{bandwidth (matrix)}
  \item{binned}{flag for binned kernel estimation. Default is FALSE.}
  \item{bgridsize}{vector of binning grid sizes}
}

\value{
  MISE- or AMISE-optimal bandwidth matrix. ISE, MISE or AMISE value. 
}

\details{
  ISE is a random variable that depends on the data
  \code{x}. MISE and AMISE are non-random and don't
  depend on the data. For normal mixture densities, ISE, MISE and AMISE 
  have exact formulas for all dimensions.
}

\references{Chacon J.E., Duong, T. & Wand, M.P. (2011). Asymptotics for
  general multivariate kernel density derivative
  estimators. \emph{Statistica Sinica}, \bold{21}, 807-840.  
}

\examples{
x <- rmvnorm.mixt(100)
Hamise.mixt(samp=nrow(x), mus=rep(0,2), Sigmas=var(x), props=1, deriv.order=1)
}

\keyword{smooth}

\name{kde.test}
\alias{kde.test}
\title{Kernel density based global two-sample comparison test}

\description{
  Kernel density based global two-sample comparison test for 1- to 6-dimensional data.}

\usage{
kde.test(x1, x2, H1, H2, h1, h2, psi1, psi2, var.fhat1, var.fhat2, 
    binned=FALSE, bgridsize, verbose=FALSE)
}

\arguments{
  \item{x1,x2}{vector/matrix of data values}
  \item{H1,H2,h1,h2}{bandwidth matrices/scalar bandwidths. If these are
  missing, \code{Hpi.kfe}, \code{hpi.kfe} is called by default.}
  \item{psi1,psi2}{zero-th order kernel functional estimates}
  \item{var.fhat1,var.fhat2}{sample variance of KDE estimates evaluated at x1, x2}
  \item{binned}{flag for binned estimation. Default is FALSE.}
  \item{bgridsize}{vector of binning grid sizes}
  \item{verbose}{flag to print out progress information. Default is FALSE.}
}

\value{
A kernel two-sample global significance test is a list with fields:
  \item{Tstat}{T statistic}
  \item{zstat}{z statistic - normalised version of Tstat}
  \item{pvalue}{\eqn{p}{p}-value of the double sided test}
  \item{mean,var}{mean and variance of null distribution}
  \item{var.fhat1,var.fhat2}{sample variances of KDE values evaluated at data points}
  \item{n1,n2}{sample sizes}
  \item{H1,H2}{bandwidth matrices}
  \item{psi1,psi12,psi21,psi2}{kernel functional estimates}
}

\details{The null hypothesis is \eqn{H_0: f_1 \equiv f_2}{H_0: f_1 = f_2} where \eqn{f_1, f_2}{f_1, f_2} 
  are the respective density functions. The measure of discrepancy is
  the integrated squared error (ISE)
  \eqn{T = \int [f_1(\bold{x}) - f_2(\bold{x})]^2 \, d \bold{x}}{int [ f_1(x) - f_2(x)]^2 dx}. If 
  we rewrite this as \eqn{T = \psi_{0,1} - \psi_{0,12} - \psi_{0,21} + \psi_{0,2}}{T = psi_0,1 - psi_0,12 - psi_0,21 + psi_0,2} 
  where \eqn{\psi_{0,uv} = \int f_u (\bold{x}) f_v (\bold{x})  \, d \bold{x}}{psi_0,uv = int f_u(x) f_v(x) dx},
  then we can use kernel functional estimators. This test statistic has a null 
  distribution which is asymptotically normal, so no bootstrap
  resampling is required to compute an approximate \eqn{p}{p}-value.     
  
  If \code{H1,H2} are missing then the  plug-in selector \code{\link{Hpi.kfe}}
  is automatically called by \code{kde.test} to estimate the
  functionals with \code{kfe(deriv.order=0)}. Likewise for missing
  \code{h1,h2}. 

  For \pkg{ks} \eqn{\geq}{>=} 1.8.8, \code{kde.test(binned=TRUE)} invokes binned
  estimation for the computation of the bandwidth selectors, and not the
  test statistic and \eqn{p}{p}-value.  
}

\references{
  Duong, T., Goud, B. & Schauer, K. (2012) Closed-form density-based framework for automatic detection of cellular morphology changes. \emph{PNAS}, \bold{109}, 8382-8387. 
}

\seealso{\code{\link{kde.local.test}}}
 
\examples{
set.seed(8192)
samp <- 1000
x <- rnorm.mixt(n=samp, mus=0, sigmas=1, props=1)
y <- rnorm.mixt(n=samp, mus=0, sigmas=1, props=1)
kde.test(x1=x, x2=y)$pvalue   ## accept H0: f1=f2

data(crabs, package="MASS")
x1 <- crabs[crabs$sp=="B", c(4,6)]
x2 <- crabs[crabs$sp=="O", c(4,6)]
kde.test(x1=x1, x2=x2)$pvalue  ## reject H0: f1=f2
}

\keyword{test}

\name{pre.transform}
\alias{pre.sphere}
\alias{pre.scale}
\title{Pre-sphering and pre-scaling}

\description{
Pre-sphered or pre-scaled version of data.}

\usage{
pre.sphere(x, mean.centred=FALSE)
pre.scale(x, mean.centred=FALSE)
}

\arguments{
  \item{x}{matrix of data values}
  \item{mean.centred}{flag to centre the data values to have zero mean. Default is FALSE.}
}

\value{Pre-sphered or pre-scaled version of data. These
  pre-transformations are required for implementing the plug-in
  \code{\link{Hpi}} selectors and the smoothed cross validation
  \code{\link{Hscv}} selectors. 
}

\details{ For pre-scaling, the data values are pre-multiplied by
  \eqn{\mathbf{S}^{-1/2}}{S^(-1/2)} and for pre-scaling, by
  \eqn{\mathbf{S}_D^{-1/2}}{S_D^(-1/2)} where
  \eqn{\mathbf{S}}{S} is the sample variance and \eqn{\mathbf{S}_D}{S_D}
  is \eqn{\mathrm{diag} \, (S_1^2, S_2^2, \dots, S_d^2)}{diag (S_1^2,
    S_2^2, ..., S_d^2)} where
  \eqn{S_i^2}{S_i^2} is the i-th marginal sample variance. 
}

\examples{
data(unicef)
unicef.sp <- pre.sphere(as.matrix(unicef))
}

\keyword{algebra}
